/*****************************************************************************

  FileName:	 Q931.c

  Contents:	 Implementation of Q.931 stack main interface functions.
				See q931.h for description.

  License/Copyright:

  Copyright (c) 2007, Jan Vidar Berger, Case Labs, Ltd. All rights reserved.
  email:janvb@caselaboratories.com

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are
  met:

	* Redistributions of source code must retain the above copyright notice,
	  this list of conditions and the following disclaimer.
	* Redistributions in binary form must reproduce the above copyright notice,
	  this list of conditions and the following disclaimer in the documentation
	  and/or other materials provided with the distribution.
	* Neither the name of the Case Labs, Ltd nor the names of its contributors
	  may be used to endorse or promote products derived from this software
	  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "Q921.h"
#include "Q931.h"
#include "Q931priv.h"
#include "Q931call.h"

#if defined(HAVE_DIALECT_5ESS)
#include "5ESS.h"
#endif

/* Global variables */

L3INT Q931L4HeaderSpace = {0};	/* header space to be ignoder/inserted */
				/* at head of each message. */
L3INT Q931L2HeaderSpace = {4};	/* Q921 header space, sapi, tei etc */

L3ULONG (*Q931GetTimeProc) (void) = NULL;	/* callback for func reading time in ms */

/*****************************************************************************

  Function:	 Q931SetL4HeaderSpace

  Description:  Set the # of bytes to be inserted/ignored at the head of
		each message. Q931 will issue a message with space for header
		and the user will use this to fill in whatever header info
		is required to support the architecture used.

*****************************************************************************/
void Q931SetL4HeaderSpace(L3INT space)
{
	Q931L4HeaderSpace = space;
}

/*****************************************************************************

  Function:	 Q931SetL2HeaderSpace

  Description:  Set the # of bytes to be inserted/ignored at the head of
		each message. Q931 will issue a message with space for header
		and the user will use this to fill in whatever header info
		is required to support the architecture used.

*****************************************************************************/
void Q931SetL2HeaderSpace(L3INT space)
{
	Q931L2HeaderSpace = space;
}


/**
 * Dummy function for message processing.
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call handle
 * @param[in]	msg	Message
 * @param[in]	from	Message coming from Layer 2 or Layer 4?
 * @return	always returns Q931E_INTERNAL (Unhandled message)
 * @internal
 * @todo	move
 */
L3INT Q931ProcDummy(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	return Q931E_INTERNAL;
}

/**
 * Dummy function for message decoding.
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	IBuf	Message input buffer
 * @param[out]	OBuf	Message output buffer
 * @param[in]	IOff	Input buffer offset
 * @param[in]	Size	Size of input
 * @param[out]	errs	Message parser errors
 * @return	always returns Q931E_UNKNOWN_MESSAGE (unhandled message)
 * @internal
 * @todo	move
 */
L3INT Q931UmesDummy(Q931_TrunkInfo_t *trunk, L3UCHAR *IBuf, Q931mes_Generic *OBuf, L3INT IOff, L3INT Size, struct Q931MesgErrors *errs)
{
	return Q931E_UNKNOWN_MESSAGE;
}

/**
 * Dummy function for IE decoding.
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	msg	Message
 * @param[in]	IBuf	Input buffer
 * @param[out]	OBuf	Output buffer
 * @param[in,out]	IOff	Input buffer offset
 * @param[in,out]	OOff	Output buffer offset
 * @return	always returns Q931E_UNKNOWN_IE (unhandled IE)
 * @internal
 * @todo	move
 */
L3INT Q931UieDummy(Q931_TrunkInfo_t *trunk, Q931mes_Generic *msg, L3UCHAR *IBuf, L3UCHAR *OBuf, L3INT *IOff, L3INT *OOff)
{
	return Q931E_UNKNOWN_IE;
}

/**
 * Dummy function for message encoding.
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	IBuf	Message input buffer
 * @param[in]	ISize	Input message size
 * @param[out]	OBuf	Message output buffer
 * @param[out]	OSize	Output message size
 * @return	always returns Q931E_UNKNOWN_MESSAGE (unhandled message)
 * @internal
 * @todo	move
 */
L3INT Q931PmesDummy(Q931_TrunkInfo_t *trunk, Q931mes_Generic *IBuf, L3INT ISize, L3UCHAR *OBuf, L3INT *OSize)
{
	return Q931E_UNKNOWN_MESSAGE;
}

/**
 * Dummy function for IE encoding.
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	IBuf	Message input buffer
 * @param[out]	OBuf	Message output buffer (encoded IE)
 * @param[out]	Octet	Output buffer offset
 * @return	always returns Q931E_UNKNOWN_IE (unhandled IE)
 * @internal
 * @todo	move
 */
L3INT Q931PieDummy(Q931_TrunkInfo_t *trunk, L3UCHAR *IBuf, L3UCHAR *OBuf, L3INT *Octet)
{
	return Q931E_UNKNOWN_IE;
}



/**
 * Initialize Q.931 stack
 * @note	Call once before using any other function!
 */
Q931_API void Q931Initialize(void)
{
	/*
	 * Initialize Q.931 dialect registry and register dialects
	 */
	Q931DialectRegistryInit();

	/* Register Q.931 */
	Q931DialectRegister(Q931_Dialect_Q931, Q931_TE, Q931CreateTE);
	Q931DialectRegister(Q931_Dialect_Q931, Q931_NT, Q931CreateNT);

#if defined(HAVE_DIALECT_5ESS)
	/* Register 5ESS */
	Q931DialectRegister(Q931_Dialect_5ESS, Q931_TE, ATT5ESSCreateTE);
/*	Q931DialectRegister(Q931_Dialect_5ESS, Q931_NT, ATT5ESSCreateNT); */
#endif
}

/**
 * Periodically called to update and check for expired timers
 * @param	trunk	Q.931 trunk
 * @todo	move (q931_timer.c)
 */
Q931_API void Q931TimerTick(Q931_TrunkInfo_t *trunk)
{
	struct Q931_Call *call = NULL;
	L3ULONG now = 0;
	L3INT x;

	/* TODO: Loop through all active calls, check timers and call timout procs
	 * if timers are expired.
	 * Implement a function array so each dialect can deal with their own
	 * timeouts.
	 */
	now = Q931GetTime();

	for (x = 0; x < Q931MAXCALLPERTRUNK; x++) {
		call = &trunk->call[x];

		if (!call->InUse || !call->Timer || !call->TimerID)
			continue;

		if (call->Timer <= now) {
			int id = call->TimerID;	/* Q931CallStopTimer resets the TimerID */

			/* Stop Timer */
			Q931CallStopTimer(call, id);

			/* Increment expire counter */
			Q931CallIncrementTimerExpireCount(call);

			/* Invoke dialect timeout callback */
			Q931Timeout(trunk, call, id);

			/*
			 * Release CRV if call has gone back to Null state after timer event
			 * and send an event to L4 before doing so
			 */
			if (!Q931CallIsGlobal(call)) {
				if ((Q931CallGetState(call) & 0xff) == Q931_U0) {
					/* Send event, so L4 can clean up */
					struct Q931_CallEvent event;

					event.id    = Q931_EVENT_RELEASE_CRV;
					event.type  = Q931_EVENT_TYPE_CRV;
					event.error = 0;

					Q931CallSendEvent(call, &event);

					Q931Log(trunk, Q931_LOG_DEBUG, "Garbage-collecting call %d\n", Q931CallGetCRV(call));

					/* Release call */
					Q931CallRelease(call);
				}
			}
		}
	}
}


/**
 * Receive message from Q.921 (Layer 2) stack
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	ind	Type of Q.921 message: Q921_DL_UNIT_DATA, Q921_DL_DATA
 * @param[in]	tei	TEI this message is coming in on
 * @param[in]	buf	Message buffer
 * @param[in]	Size	Size of message
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 */
Q931_API L3INT Q931Rx23(Q931_TrunkInfo_t *trunk, L3INT ind, L3UCHAR tei, L3UCHAR *buf, L3INT Size)
{
	L3UCHAR *Mes = NULL;
	L3INT RetCode = Q931E_NO_ERROR;
	Q931mes_Generic *m = (Q931mes_Generic *) trunk->L3Buf;
	struct Q931_Call *call = NULL;
	struct Q931MesgErrors Errs;
	L3INT ISize;
	L3INT IOff = 0;
	L3INT L2HSize = Q931L2HeaderSpace;

	switch (ind) {
	case Q921_DL_UNIT_DATA:		/* DL-UNITDATA indication (UI frame, 3 byte header) */
		L2HSize = 3;

	case Q921_DL_DATA:		/* DL-DATA indication (I frame, 4 byte header) */
		/* Reset our decode buffer */
		memset(trunk->L3Buf, 0, sizeof(trunk->L3Buf));

		/* L2 Header Offset */
		Mes = &buf[L2HSize];

		/* Protocol Discriminator */
		m->ProtDisc = Mes[IOff++];

		/* CRV */
		m->CRVFlag = (Mes[IOff + 1] >> 7) & 0x01;
		m->CRV = Q931Uie_CRV(trunk, Mes, m->buf, &IOff, &ISize);

		/* Message Type */
		m->MesType = Mes[IOff++];

		/* Store tei */
		m->Tei = tei;

		/* vvv new code starts here */
		if (m->CRV) {
			/* Lookup call reference */
			call = Q931GetCallByCRV(trunk, m->CRV);
			if (!call) {
				/* oops, no call found */
				if (m->MesType == Q931mes_SETUP) {
					/* Setup message, create new call handle */
					call = Q931CallNewIncoming(trunk, m->CRV);
				} else {
					/* Not a setup message -> INVALID CRV */
					/* TODO: respond with proper RELEASE COMPLETE + cause combination */
					return Q931E_INVALID_CRV;
				}

				/* Still no call handle? */
				if (!call) {
					/* blah, failed to allocate new call -> Disconnect */
					Q931Disconnect(trunk, MSGF_FROM_L2, m->CRV, 42);
					return Q931E_INTERNAL;
				}
			}

			/* d'oh a little ugly but this saves us from:
			 *	a) doing Q.921 work in the lower levels (extracting the TEI ourselves)
			 *	b) adding a tei parameter to _all_ Proc functions
			 */
			if (tei && !call->Tei) {
				call->Tei = tei;
			}

			/* Diagnostic message */
			Q931Log(trunk, Q931_LOG_DEBUG, "Call is in state %d [%s]\n", Q931CallGetState(call), Q931CallGetStateName(call));
		} else {
			/* Global CRV ?? */
			call = Q931GetGlobalCall(trunk);

			Q931Log(trunk, Q931_LOG_DEBUG, "Using global call for message\n");
		}
		/* ^^^ new code ends here */

		Q931Log(trunk, Q931_LOG_DEBUG, "Received message from Q.921 (ind %d, tei %d, size %d)\nMesType: %d, CRVFlag %d (%s), CRV %d (Dialect: %s)\n", ind, m->Tei, Size,
						 m->MesType, m->CRVFlag, m->CRVFlag ? "Terminator" : "Originator", m->CRV, Q931DialectGetName(trunk->Dialect));

		Q931MesgErrorsInit(&Errs, m->MesType);

		RetCode = Q931Umes(trunk, m->MesType, Mes, (Q931mes_Generic *)trunk->L3Buf, IOff, Size - L2HSize, &Errs);
		if (RetCode >= Q931E_NO_ERROR) {
			RetCode = Q931Proc(trunk, call, (Q931mes_Generic *)trunk->L3Buf, Q931_MSG_FROM_L2);
			Q931Log(trunk, Q931_LOG_DEBUG, "Q931Proc[%s][0x%02x]() returned %d\n", Q931DialectGetName(trunk->Dialect), m->MesType, RetCode);
		} else {
			Q931Log(trunk, Q931_LOG_DEBUG, "Q931Umes[%s][0x%02x]() returned %d\n", Q931DialectGetName(trunk->Dialect), m->MesType, RetCode);

			/* Output error report */
			Q931MesgErrorsPrint(trunk, &Errs, Mes, Size - L2HSize);
		}

		/* Error handling for Umes() and Proc() */
		if (RetCode < Q931E_NO_ERROR) {
			/* Initialize response message */

			/* TODO: Add Error codes for Mandatory IE missing and error */
			switch (RetCode) {
			case Q931E_MANDATORY_IE_MISSING:
				/* Send STATUS with cause 96 */
				RetCode = Q931StatusEnquiryResponse(trunk, call, Q850_CAUSE_MANDATORY_IE_MISSING);
				break;

			case Q931E_ILLEGAL_IE:			/* ?? */
				/* Send STATUS with cause 96 */
				RetCode = Q931StatusEnquiryResponse(trunk, call, Q850_CAUSE_IE_NONEXIST);
				break;

			case Q931E_UNKNOWN_IE:
				/* Send STATUS with cause 100 */
				RetCode = Q931StatusEnquiryResponse(trunk, call, Q850_CAUSE_INVALID_IE_CONTENTS);
				break;

			case Q931E_UNEXPECTED_MESSAGE:
			case Q931E_UNRECOGNIZED_MESSAGE:
				/* Ignore in State 0 */
				if ((Q931CallGetState(call) & 0xff) == Q931_U0)
					goto out;

				/* Ignore for RELEASE and RELEASE COMPLETE */
				if (m->MesType == Q931mes_RELEASE || m->MesType == Q931mes_RELEASE_COMPLETE)
					goto out;

				if (Q931TrunkGetStatusEnquiry(trunk)) {
					/* Send STATUS ENQUIRY */
					Q931InitMesGeneric(m);
					m->CRV     = Q931CallGetCRV(call);
					m->CRVFlag = Q931CallIsOutgoing(call) ? 0 : 1;
					m->MesType = Q931mes_STATUS_ENQUIRY;

					/* TODO: check and / or fire up T322 */
					RetCode = Q931Tx32(trunk, 0, m, m->Size);
				} else {
					/* Send STATUS with cause X */
					RetCode = Q931StatusEnquiryResponse(trunk, call,
								(RetCode == Q931E_UNEXPECTED_MESSAGE) ? Q850_CAUSE_WRONG_MESSAGE : Q850_CAUSE_MESSAGE_TYPE_NONEXIST);
				}
				break;

			default:	/* unhandled / unknown error (ignore) */
				break;
			}
		}
		break;

	default:
		break;
	}

out:
	if (call) {
		/* Send queued call events to layer 4 (if any) */
		Q931CallSendQueuedEvents(call);

		/*
		 * Release CRV if call is still (or again) in Null state,
		 * and send an event to L4 before doing so
		 *
		 * TODO: Remove Q931CallRelease() calls from Q931Proc_*
		 *       functions
		 */
		if (!Q931CallIsGlobal(call)) {
			if ((Q931CallGetState(call) & 0xff) == Q931_U0) {
				/* Send event, so L4 can clean up */
				struct Q931_CallEvent event;

				event.id    = Q931_EVENT_RELEASE_CRV;
				event.type  = Q931_EVENT_TYPE_CRV;
				event.error = 0;

				Q931CallSendEvent(call, &event);

				Q931Log(trunk, Q931_LOG_DEBUG, "Garbage-collecting call %d\n", Q931CallGetCRV(call));

				/* Release call */
				Q931CallRelease(call);
			}
		}
	}
	return RetCode;
}


/**
 * Send message to Layer 4
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call
 * @param[in]	msg	Message to send (Q931mes_Generic *)
 * @param[in]	size	Size of message
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 * @note	most likely going to deprecate this function and use events
 */
L3INT Q931Tx34(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, int size)
{
	Q931Log(trunk, Q931_LOG_DEBUG, "Sending message to Layer4 (size: %d)\n", size);

	if (trunk->Q931Tx34CBProc) {
		return trunk->Q931Tx34CBProc(trunk->PrivateData34, call, msg, size);
	}
	return Q931E_MISSING_CB;
}


/**
 * Receive and process message from Layer 4 (application)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	msg	Message buffer (Q931mes_Generic *)
 * @param[in]	size	Size of message
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 */
Q931_API L3INT Q931Rx43(Q931_TrunkInfo_t *trunk, Q931mes_Generic *msg, int size)
{
	L3INT RetCode = Q931E_NO_ERROR;
	struct Q931_Call *call = NULL;

	Q931Log(trunk, Q931_LOG_DEBUG, "Receiving message from Layer4 (size: %d, type: %d [0x%x], CRV: %d, CRVFlag: %d)\n", size, msg->MesType, msg->MesType, msg->CRV, msg->CRVFlag);

	/*
	 * Check if message is valid in current state & dialect
	 */

	/* Get call handle */
	if (msg->CRV) {
		/* Find the call using CRV */
		call = Q931GetCallByCRV(trunk, msg->CRV);
		if(!call)
			return Q931E_INVALID_CRV;

		/* SAFEGUARD: make sure the CRVFlag is set correctly */
		msg->CRVFlag = Q931CallIsOutgoing(call) ? 0 : 1;
	} else {
		/* global call handle */
		call = Q931GetGlobalCall(trunk);
	}

	/* Check if message can be handled */
	if (!Q931DialectGetMesProc(trunk->Dialect, msg->MesType)) {
		RetCode = Q931E_UNRECOGNIZED_MESSAGE;

		Q931Log(trunk, Q931_LOG_WARNING, "Unrecognized message from Layer 4 (size: %d, type: %d)\n", size, msg->MesType);

	        /* Ignore unknown message? */
	        if (Q931TrunkIsSetFlag(trunk, Q931_TFLAG_IGNORE_UNKNOWN_MSG)) {
			RetCode = Q931E_NO_ERROR;
		}
		goto out;
	}

	/* Check if message is valid in current call state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, Q931_MSG_FROM_L4)) {
		RetCode = Q931E_UNEXPECTED_MESSAGE;

		Q931Log(trunk, Q931_LOG_ERROR, "Unexpected message (invalid in current state) from Layer 4 (size: %d, type: %d [0x%x])\n", size, msg->MesType, msg->MesType);
		goto out;
	}

	Q931Log(trunk, Q931_LOG_NOTICE, "Handling message from Layer4 in %s call state: %d\n", Q931CallIsGlobal(call) ? "global" : "", Q931CallGetState(call));

	RetCode = Q931Proc(trunk, call, msg, Q931_MSG_FROM_L4);
out:
	if (call) {
		/*
		 * Release CRV if call is still (or again) in Null state,
		 * and send an event to L4 before doing so
		 *
		 * TODO: Remove Q931CallRelease() calls from Q931Proc_*
		 *       functions
		 */
		if (!Q931CallIsGlobal(call)) {
			if ((Q931CallGetState(call) & 0xff) == Q931_U0) {
				/* Send event, so L4 can clean up */
				struct Q931_CallEvent event;

				event.id    = Q931_EVENT_RELEASE_CRV;
				event.type  = Q931_EVENT_TYPE_CRV;
				event.error = 0;

				Q931CallSendEvent(call, &event);

				Q931Log(trunk, Q931_LOG_DEBUG, "Garbage-collecting call %d\n", Q931CallGetCRV(call));

				/* Release call */
				Q931CallRelease(call);
			}
		}
	}

	Q931Log(trunk, Q931_LOG_DEBUG, "Q931Rx43 return code: %d\n", RetCode);
	return RetCode;
}


/**
 * Encode and send message to Layer 2
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	bcast	Broadcast message to all TEIs
 * @param[in]	msg	Message buffer
 * @param[in]	size	Size of message
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 * @note	Only for use by *Proc_*() functions
 */
L3INT Q931Tx32(Q931_TrunkInfo_t *trunk, L3UCHAR bcast, Q931mes_Generic *msg, int size)
{
	L3INT RetCode = Q931E_NO_ERROR;
	L3INT Offset = bcast ? (Q931L2HeaderSpace - 1) : Q931L2HeaderSpace;
	L3INT OSize;
	L3UCHAR tei = 0;

	Q931Log(trunk, Q931_LOG_DEBUG, "Sending message to Q.921 (size: %d)\n", size);

	memset(trunk->L2Buf, 0, sizeof(trunk->L2Buf));

	if (msg->CRV) {
		struct Q931_Call *call;

		/* Find the call using CRV */
		call = Q931GetCallByCRV(trunk, msg->CRV);
		if(!call)
			return Q931E_INVALID_CRV;

		tei = call->Tei;

		/* SAFEGUARD: make sure the CRVFlag is set correctly */
		msg->CRVFlag = Q931CallIsOutgoing(call) ? 0 : 1;
	}

	/* Call pack function through table. */
	RetCode = Q931Pmes(trunk, msg->MesType, msg, size, &trunk->L2Buf[Offset], &OSize);
	if (RetCode >= Q931E_NO_ERROR) {
		if (trunk->Q931Tx32CBProc) {
			RetCode = trunk->Q931Tx32CBProc(trunk->PrivateData32, bcast ? Q921_DL_UNIT_DATA : Q921_DL_DATA, tei, trunk->L2Buf, OSize + Offset);
			if (RetCode > 0) {
				/* L2 returns number of packets sent */
				RetCode = Q931E_NO_ERROR;
			}
		} else {
			RetCode = Q931E_MISSING_CB;
		}
	}

	return RetCode;
}


/**
 * Used by the stack to indicate an error
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	ErrID	ID of error (Q931E_*)
 * @param[in]	ErrPar1	Parameter1
 * @param[in]	ErrPar2	Parameter2
 * @internal
 */
void Q931SetError(Q931_TrunkInfo_t *trunk,L3INT ErrID, L3INT ErrPar1, L3INT ErrPar2)
{
	if (trunk->Q931ErrorCBProc) {
		trunk->Q931ErrorCBProc(trunk->PrivateData34, ErrID, ErrPar1, ErrPar2);
	}
}


/**
 * Release a CRV (= call)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	crv	CRV to release
 * @return	Q931E_NO_ERROR on success, Q931E_INVALID_CRV if call does not exist
 */
Q931_API L3INT Q931ReleaseCRV(Q931_TrunkInfo_t *trunk, L3INT crv)
{
	struct Q931_Call *call;

	call = Q931GetCallByCRV(trunk, crv);
	if (!call)
		return Q931E_INVALID_CRV;

	call->InUse = 0;
	return Q931E_NO_ERROR;
}

/**
 * Get current time in milliseconds
 * @return	time in milliseconds
 * @internal
 * @todo	handle wrapping of timestamp
 * 		move (q931_timer.c)
 */
L3ULONG Q931GetTime()
{
	L3ULONG tNow = 0;
	static L3ULONG tLast = 0;

	if (Q931GetTimeProc != NULL) {
		tNow = Q931GetTimeProc();
		if (tNow < tLast) {	/* wrapped */
			/* TODO */
		}
		tLast = tNow;
	}
	return tNow;
}


/**
 * Set timer callback
 * @param[in]	callback	callback function used to get the current time in milliseconds
 * @note	Required, timers won't work without a valid callback
 */
Q931_API void Q931SetGetTimeCB(L3ULONG (*callback)(void))
{
	Q931GetTimeProc = callback;
}


/**
 * Textual description of error codes
 * @internal
 * @todo	Improve these
 */
static const char *q931_error_names[] = {
	"Q931E_NO_ERROR",			/* 0 */

	"Q931E_UNKNOWN_MESSAGE",		/* -3001 */
	"Q931E_ILLEGAL_IE",			/* -3002 */
	"Q931E_UNKNOWN_IE",			/* -3003 */
	"Q931E_BEARERCAP",			/* -3004 */
	"Q931E_HLCOMP",				/* -3005 */
	"Q931E_LLCOMP",				/* -3006 */
	"Q931E_INTERNAL",			/* -3007 */
	"Q931E_MISSING_CB",			/* -3008 */
	"Q931E_UNEXPECTED_MESSAGE",		/* -3009 */
	"Q931E_ILLEGAL_MESSAGE",		/* -3010 */
	"Q931E_TOMANYCALLS",			/* -3011 */
	"Q931E_INVALID_CRV",			/* -3012 */
	"Q931E_CALLID",				/* -3013 */
	"Q931E_CALLSTATE",			/* -3014 */
	"Q931E_CALLEDSUB",			/* -3015 */
	"Q931E_CALLEDNUM",			/* -3016 */
	"Q931E_CALLINGNUM",			/* -3017 */
	"Q931E_CALLINGSUB",			/* -3018 */
	"Q931E_CAUSE",				/* -3019 */
	"Q931E_CHANID",				/* -3020 */
	"Q931E_DATETIME",			/* -3021 */
	"Q931E_DISPLAY",			/* -3022 */
	"Q931E_KEYPADFAC",			/* -3023 */
	"Q931E_NETFAC",				/* -3024 */
	"Q931E_NOTIFIND",			/* -3025 */
	"Q931E_PROGIND",			/* -3026 */
	"Q931E_RESTARTIND",			/* -3027 */
	"Q931E_SEGMENT",			/* -3028 */
	"Q931E_SIGNAL",				/* -3029 */
	"Q931E_GENERIC_DIGITS",			/* -3030 */

	"Q931E_MANDATORY_IE_MISSING",		/* -3031 */
	"Q931E_UNRECOGNIZED_MESSAGE",		/* -3032 */
};

#define Q931_MAX_ERROR 32

/**
 * Return description of error code
 * @param[in]	error	Error code
 * @return	Description of error code or empty string on error
 */
Q931_API const char *q931_error_to_name(q931_error_t error)
{
	int index = 0;
	if ((int)error < 0) {
		index = (((int)error * -1) -3000);
	}
	if (index < 0 || index > Q931_MAX_ERROR) {
		return "";
	}
	return q931_error_names[index];
}

/*
 * Logging
 */
#include <stdarg.h>

/**
 * Log formatted string
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	level	Log level of message
 * @param[in]	fmt	Format string
 * @param[in]	...	Optional arguments
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 * @note	Needs a working log callback set
 */
L3INT Q931Log(const Q931_TrunkInfo_t *trunk, Q931LogLevel_t level, const char *fmt, ...)
{
	char buf[Q931_LOGBUFSIZE];
	L3INT len;
	va_list ap;

	if (!trunk->Q931LogCBProc)
		return 0;

	if (trunk->loglevel < level)
		return 0;

	va_start(ap, fmt);

	len = vsnprintf(buf, sizeof(buf)-1, fmt, ap);
	if (len <= 0) {
		/* TODO: error handling */
		return -1;
	}
	if (len >= sizeof(buf))
		len = sizeof(buf) - 1;

	buf[len] = '\0';

	va_end(ap);

	return trunk->Q931LogCBProc(trunk->PrivateDataLog, level, buf, len);
}

L3INT Q931LogRaw(const Q931_TrunkInfo_t *trunk, Q931LogLevel_t level, const char *str)
{
	if (!trunk->Q931LogCBProc)
		return 0;

	if (trunk->loglevel < level)
		return 0;

	return trunk->Q931LogCBProc(trunk->PrivateDataLog, level, str, strlen(str));
}

/**
 * Set Logging callback function and private data
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	func	Log callback function
 * @param[in]	priv	Optional private data passed to logging function
 */
Q931_API void Q931SetLogCB(Q931_TrunkInfo_t *trunk, Q931LogCB_t func, void *priv)
{
	trunk->Q931LogCBProc  = func;
	trunk->PrivateDataLog = priv;
}


/**
 * Set log level
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	level	Log level
 */
Q931_API void Q931SetLogLevel(Q931_TrunkInfo_t *trunk, Q931LogLevel_t level)
{
	if(!trunk) {
		return;
	}

	if (level < Q931_LOG_NONE) {
		level = Q931_LOG_NONE;
	} else if (level > Q931_LOG_DEBUG) {
		level = Q931_LOG_DEBUG;
	}
	trunk->loglevel = level;
}


/**
 * Get log level
 * @param[in]	trunk	Q.931 trunk
 * @return	current log level set on trunk
 */
Q931_API Q931LogLevel_t Q931GetLogLevel(Q931_TrunkInfo_t *trunk)
{
	if(!trunk) {
		return Q931_LOG_NONE;
	}
	return trunk->loglevel;
}

/**
 * Textual names of log level
 * @internal
 */
static const char *loglevel_names[] = {
	[Q931_LOG_DEBUG]   = "debug",
	[Q931_LOG_INFO]    = "info",
	[Q931_LOG_NOTICE]  = "notice",
	[Q931_LOG_WARNING] = "warning",
	[Q931_LOG_ERROR]   = "error",
	[Q931_LOG_ALERT]   = "alert",
	[Q931_LOG_CRIT]    = "critical",
	[Q931_LOG_EMERG]   = "emergency"
};

/**
 * Get name of log level set on trunk
 * @param[in]	trunk	Q.931 trunk
 * @todo	Make this a generic function e.g.:
 *		const char *Q931GetLogLevelName(q931_loglevel_t level)
 */
Q931_API const char *Q931GetLogLevelName(Q931_TrunkInfo_t *trunk)
{
	if(!trunk) {
		return "unknown";
	}
	if (trunk->loglevel == Q931_LOG_NONE) {
		return "none";
	}
	return loglevel_names[trunk->loglevel];
}

/**
 * Dummy handler for timeouts
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call
 * @return	Always returns Q931E_NO_ERROR
 * @internal
 * @todo	move?
 */
L3INT Q931TimeoutDummy(Q931_TrunkInfo_t *trunk, struct Q931_Call *call)
{
	Q931Log(trunk, Q931_LOG_DEBUG, "Timer %d of call %d timed out\n", call->TimerID, call->CRV);
	return Q931E_NO_ERROR;
}


/**************************************************************************************
 *  Trunk helper functions
 **************************************************************************************/
#include <assert.h>

/**
 * Helper function to set a trunk flag
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	flag	Flag (mask) to set
 * @internal
 */
void Q931TrunkSetFlag(Q931_TrunkInfo_t *trunk, L3INT flag)
{
	trunk->flags |= flag;
}

/**
 * Helper function to unset a trunk flag
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	flag	Flag (mask) to clear
 * @internal
 */
void Q931TrunkClearFlag(Q931_TrunkInfo_t *trunk, L3INT flag)
{
	trunk->flags &= ~flag;
}

/**
 * Helper function to clear all flags of a trunk
 * @param[in]	trunk	Q.931 trunk
 * @internal
 */
void Q931TrunkClearAllFlags(Q931_TrunkInfo_t *trunk)
{
	trunk->flags = 0;
}

/**
 * Helper function to check if a trunk flag is set
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	flag	Flag (mask) to check
 * @return	0 = flag is not set, 1 = it is
 * @internal
 */
L3BOOL Q931TrunkIsSetFlag(const Q931_TrunkInfo_t *trunk, const L3INT flag)
{
	return ((trunk->flags & flag) != 0);
}

/**
 * Macro used for mass-defining flag set functions
 * @param[in]	_name	Name of flag
 * @param[in]	_flag	Value of flag
 * @internal
 */
#define Q931_TRUNK_FLAG_SET_FUNC(_name, _flag) 			\
	Q931_API void _name(Q931_TrunkInfo_t *trunk, L3BOOL enable) {	\
		assert(trunk);					\
		if (enable) {					\
			Q931TrunkSetFlag(trunk, _flag);		\
		} else {					\
			Q931TrunkClearFlag(trunk, _flag);	\
		}						\
	}

/**
 * Macro used for mass-defining flag get functions
 * @param[in]	_name	Name of flag
 * @param[in]	_flag	Value of flag
 * @internal
 */
#define Q931_TRUNK_FLAG_GET_FUNC(_name, _flag) 			\
	Q931_API L3BOOL _name(Q931_TrunkInfo_t *trunk) { 	\
		assert(trunk);					\
		return Q931TrunkIsSetFlag(trunk, _flag);	\
	}


/*
 * definition of actual accessor functions
 */
Q931_TRUNK_FLAG_SET_FUNC(Q931TrunkSetAutoSetupAck,     Q931_TFLAG_AUTO_SETUP_ACK)
Q931_TRUNK_FLAG_SET_FUNC(Q931TrunkSetAutoServiceAck,   Q931_TFLAG_AUTO_SERVICE_ACK)
Q931_TRUNK_FLAG_SET_FUNC(Q931TrunkSetAutoConnectAck,   Q931_TFLAG_AUTO_CONNECT_ACK)
Q931_TRUNK_FLAG_SET_FUNC(Q931TrunkSetAutoRestartAck,   Q931_TFLAG_AUTO_RESTART_ACK)
Q931_TRUNK_FLAG_SET_FUNC(Q931TrunkSetStatusEnquiry,    Q931_TFLAG_STATUS_ENQUIRY)
Q931_TRUNK_FLAG_SET_FUNC(Q931TrunkSetIgnoreUnknownMsg, Q931_TFLAG_IGNORE_UNKNOWN_MSG)
Q931_TRUNK_FLAG_SET_FUNC(Q931TrunkSetIgnoreUnknownIEs, Q931_TFLAG_IGNORE_UNKNOWN_IE)
Q931_TRUNK_FLAG_SET_FUNC(Q931TrunkSetIgnoreIllegalIEs, Q931_TFLAG_IGNORE_ILLEGAL_IE)

Q931_TRUNK_FLAG_GET_FUNC(Q931TrunkGetAutoSetupAck,     Q931_TFLAG_AUTO_SETUP_ACK)
Q931_TRUNK_FLAG_GET_FUNC(Q931TrunkGetAutoServiceAck,   Q931_TFLAG_AUTO_SERVICE_ACK)
Q931_TRUNK_FLAG_GET_FUNC(Q931TrunkGetAutoConnectAck,   Q931_TFLAG_AUTO_CONNECT_ACK)
Q931_TRUNK_FLAG_GET_FUNC(Q931TrunkGetAutoRestartAck,   Q931_TFLAG_AUTO_RESTART_ACK)
Q931_TRUNK_FLAG_GET_FUNC(Q931TrunkGetStatusEnquiry,    Q931_TFLAG_STATUS_ENQUIRY)
Q931_TRUNK_FLAG_GET_FUNC(Q931TrunkGetIgnoreUnknownMsg, Q931_TFLAG_IGNORE_UNKNOWN_MSG)
Q931_TRUNK_FLAG_GET_FUNC(Q931TrunkGetIgnoreUnknownIEs, Q931_TFLAG_IGNORE_UNKNOWN_IE)
Q931_TRUNK_FLAG_GET_FUNC(Q931TrunkGetIgnoreIllegalIEs, Q931_TFLAG_IGNORE_ILLEGAL_IE)


/**
 * Relax IE and Message checks: incoming unknown messages and unknown/invalid IEs will be ignored
 * @param[in]	trunk	Q.931 Trunk handle
 * @param[in]	enable	Enable or disable Relaxed mode (Q931_TRUE / Q931_FALSE)
 * @note	Use with care!
 */
Q931_API void Q931TrunkSetRelaxedMode(Q931_TrunkInfo_t *trunk, L3BOOL enable)
{
	int flags = (Q931_TFLAG_IGNORE_UNKNOWN_MSG | Q931_TFLAG_IGNORE_UNKNOWN_IE | Q931_TFLAG_IGNORE_ILLEGAL_IE);
	assert(trunk);

	if (enable) {
		Q931TrunkSetFlag(trunk, flags);
	} else {
		Q931TrunkClearFlag(trunk, flags);
	}
}


/**
 * Invoke message specific handler function
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call handle
 * @param[in]	msg	Decoded message to handle (Q931mes_Generic *)
 * @param[in]	from	Source of message (Q931_MSG_FROM_L2 or Q931_MSG_FROM_L4)
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 * @todo	move?
 */
L3INT Q931Proc(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	q931proc_func_t *func = Q931DialectGetMesProc(trunk->Dialect, msg->MesType);

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;

	if (func) {
		L3INT retval = func(trunk, call, msg, from);

		switch (retval) {
		case Q931E_UNEXPECTED_MESSAGE:
			{
				Q931Log(trunk, Q931_LOG_ERROR, "Unexpected message 0x%02x for call 0x%04x in state '%s'\n",
						msg->MesType,
						Q931CallGetCRV(call),
						Q931CallGetStateName(call));
			}
			break;
		default:
			break;
		}
		return retval;
	}

        /* Ignore unknown message? */
        if (Q931TrunkIsSetFlag(trunk, Q931_TFLAG_IGNORE_UNKNOWN_MSG)) {
		return Q931E_NO_ERROR;
        }

	func = Q931DialectGetUnknownMesProc(trunk->Dialect);
	if (func) {
		return func(trunk, call, msg, from);
	}
	return Q931E_INTERNAL;
}


/**
 * Invoke invalid message handling function
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call handle
 * @param[in]	msg	Decoded message to handle (Q931mes_Generic *)
 * @param[in]	from	Source of message (Q931_MSG_FROM_L2 or Q931_MSG_FROM_L4)
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 * @todo	move?
 */
L3INT Q931ProcInvalid(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	q931proc_func_t *func = Q931DialectGetInvalidMesProc(trunk->Dialect);

	if (func) {
		return func(trunk, call, msg, from);
	}
	return Q931E_INTERNAL;
}


/**
 * Invoke message specific decoder function
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	id	ID of message (type)
 * @param[in]	IBuf	Input buffer (raw encoded message)
 * @param[out]	OBuf	Output buffer
 * @param[in]	IOff	Input buffer offset
 * @param[in]	Size	Size of input message
 * @param[out]	errs	Message errors reported by parser
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 * @todo	move?
 */
L3INT Q931Umes(Q931_TrunkInfo_t *trunk, L3UCHAR id, L3UCHAR *IBuf, Q931mes_Generic *OBuf, L3INT IOff, L3INT Size, struct Q931MesgErrors *errs)
{
	q931umes_func_t *func = Q931DialectGetUmesProc(trunk->Dialect, id);

	if (func) {
		return func(trunk, IBuf, OBuf, IOff, Size, errs);
	}

	/* unknown message */
/*	Q931MesgErrorsAdd(errs, Q931_MSGE_UNKNOWN); */

        /* Ignore unknown message? */
        if (Q931TrunkIsSetFlag(trunk, Q931_TFLAG_IGNORE_UNKNOWN_MSG)) {
		return Q931E_NO_ERROR;
        }

	return Q931E_INTERNAL;
}


/**
 * Invoke message specific encoder function
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	id	ID of message (type)
 * @param[in]	IBuf	Input buffer
 * @param[in]	ISize	Size of input message
 * @param[out]	OBuf	Output buffer
 * @param[out]	OSize	Output message size
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 * @todo	move?
 */
L3INT Q931Pmes(Q931_TrunkInfo_t *trunk, L3UCHAR id, Q931mes_Generic *IBuf, L3INT ISize, L3UCHAR *OBuf, L3INT *OSize)
{
	q931pmes_func_t *func = Q931DialectGetPmesProc(trunk->Dialect, id);

	if (func) {
		return func(trunk, IBuf, ISize, OBuf, OSize);
	}
	return Q931E_INTERNAL;
}


/**
 * Invoke IE specific decoder function
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	id	ID of information element (type)
 * @param[in]	pMes	Message
 * @param[in]	IBuf	Input buffer
 * @param[out]	OBuf	Output buffer
 * @param[in,out]	IOff	Input buffer offset
 * @param[in,out]	OOff	Output buffer offset
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 * @todo	move?
 */
L3INT Q931Uie(Q931_TrunkInfo_t *trunk, L3UCHAR id, Q931mes_Generic *pMes, L3UCHAR *IBuf, L3UCHAR *OBuf, L3INT *IOff, L3INT *OOff)
{
	q931uie_func_t *func = Q931DialectGetUieProc(trunk->Dialect, id);

	if (func) {
		return func(trunk, pMes, IBuf, OBuf, IOff, OOff);
	}
	return Q931E_INTERNAL;
}


/**
 * Invoke IE specific encoder function
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	id	ID of information element (type)
 * @param[in]	IBuf	Input buffer
 * @param[out]	OBuf	Output buffer
 * @param[in,out]	Octet	Input buffer offset
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 * @todo	move?
 */
L3INT Q931Pie(Q931_TrunkInfo_t *trunk, L3UCHAR id, L3UCHAR *IBuf, L3UCHAR *OBuf, L3INT *Octet)
{
	q931pie_func_t *func = Q931DialectGetPieProc(trunk->Dialect, id);

	if (func) {
		return func(trunk, IBuf, OBuf, Octet);
	}
	return Q931E_INTERNAL;
}

/**
 * Invoke IE specific debug function
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	ie	Pointer to Information Element
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 * @todo	move?
 */
L3INT Q931Die(Q931_TrunkInfo_t *trunk, Q931ie_Generic *ie, struct strstream *ostream)
{
	q931die_func_t *func = Q931DialectGetDieProc(trunk->Dialect, ie->IEId);

	if (func) {
		return func(trunk, ie, ostream);
	}
	return Q931E_INTERNAL;
}

/**
 * Check whether information element has a decoder function available
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	id	Information element id to check
 * @return	Q931_FALSE if this IE can be handled or Q931_TRUE if it can not
 * @internal
 * @todo	move
 */
L3BOOL Q931UieIsNull(Q931_TrunkInfo_t *trunk, L3UCHAR id)
{
	return (!Q931DialectGetUieProc(trunk->Dialect, id));
}

/**
 * Check whether information element has a encoder function available
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	id	Information element id to check
 * @return	Q931_FALSE if this IE can be handled or Q931_TRUE if it can not
 * @internal
 * @todo	move
 */
L3BOOL Q931PieIsNull(Q931_TrunkInfo_t *trunk, L3UCHAR id)
{
	return (!Q931DialectGetPieProc(trunk->Dialect, id));
}

/**
 * Check whether information element has a debug function available
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	id	Information element id to check
 * @return	Q931_FALSE if this IE can be handled or Q931_TRUE if it can not
 * @internal
 * @todo	move
 */
L3BOOL Q931DieIsNull(Q931_TrunkInfo_t *trunk, L3UCHAR id)
{
	return (!Q931DialectGetDieProc(trunk->Dialect, id));
}

/**
 * Invoke timeout handler function
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call
 * @param[in]	timer	Timer to handle
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 * @todo	move?
 */
L3INT Q931Timeout(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, L3UCHAR timer)
{
	q931timeout_func_t *func = Q931DialectGetTimerProc(trunk->Dialect, timer);

	if (func) {
		return func(trunk, call);
	}
	return Q931E_INTERNAL;
}


/**
 * Handle unknown message
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call handle
 * @param[in]	msg	Input buffer (Q931mes_Generic *)
 * @param[in]	from	Source of message (Q931_MSG_FROM_L2 or Q931_MSG_FROM_L4)
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 * @todo	move?
 */
L3INT Q931ProcUnknownMessage(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	Q931Log(trunk, Q931_LOG_NOTICE, "Unknown message %#02hhx received\n", msg->MesType);
	return Q931E_NO_ERROR;
}

/**
 * Handle unexpected message
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call handle
 * @param[in]	msg	Input buffer (Q931mes_Generic *)
 * @param[in]	from	Source of message (Q931_MSG_FROM_L2 or Q931_MSG_FROM_L4)
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 * @todo	move?
 */
L3INT Q931ProcUnexpectedMessage(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	Q931Log(trunk, Q931_LOG_NOTICE, "Unexpected/Invalid message %#02hhx received\n", msg->MesType);
	return Q931E_NO_ERROR;
}

