/*****************************************************************************

  FileName:		q931StateNT.c

  Contents:		Q.931 State Engine for NT (Network Mode).

			The controlling state engine for Q.931 is the state engine
			on the NT side. The state engine on the TE side is a slave
			of this. The TE side maintain it's own states as described in
			ITU-T Q931, but will in	raise conditions be overridden by
			the NT side.

  License/Copyright:

  Copyright (c) 2007, Jan Vidar Berger, Case Labs, Ltd. All rights reserved.
  email:janvb@caselaboratories.com

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are
  met:

	* Redistributions of source code must retain the above copyright notice,
	  this list of conditions and the following disclaimer.
	* Redistributions in binary form must reproduce the above copyright notice,
	  this list of conditions and the following disclaimer in the documentation
	  and/or other materials provided with the distribution.
	* Neither the name of the Case Labs, Ltd nor the names of its contributors
	  may be used to endorse or promote products derived from this software
	  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.
*****************************************************************************/
#include <stdlib.h>

#include "Q931.h"
#include "Q931priv.h"
#include "Q932.h"

extern L3INT Q931L4HeaderSpace;
extern struct Q931MessageIE Q931MessageIEs[];

/*****************************************************************************
  Function:		Q931CreateNT

  Description:	Will create the Q931 NT as a Dialect in the stack. The first
				bulk set up the message handlers, the second bulk the IE
				encoders/coders, and the last bulk set up the state table.

  Parameters:	i		Dialect index
*****************************************************************************/
void Q931CreateNT(struct Q931Dialect *d)
{
	Q931DialectSetName(d, "Q.931 NT");

	/*
	 * Flags
	 */
	Q931DialectSetFlag(d, 0);

	/*
	 * Codesets
	 */
	Q931DialectAddCodeset(d, Q931_CODESET_0);

	/*
	 * Q.931 Message encoder/decoder/handler table
	 */
	Q931DialectSetMesProc(d, Q931mes_ALERTING,             Q931ProcAlertingNT,          Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_CALL_PROCEEDING,      Q931ProcCallProceedingNT,    Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_CONNECT,              Q931ProcConnectNT,           Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_CONNECT_ACKNOWLEDGE,  Q931ProcConnectAckNT,        Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_PROGRESS,             Q931ProcProgressNT,          Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_SETUP,                Q931ProcSetupNT,             Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_SETUP_ACKNOWLEDGE,    Q931ProcSetupAckNT,          Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_RESUME,               Q931ProcResumeNT,            Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_RESUME_ACKNOWLEDGE,   Q931ProcResumeAckNT,         Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_RESUME_REJECT,        Q931ProcResumeRejectNT,      Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_SUSPEND,              Q931ProcSuspendNT,           Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_SUSPEND_ACKNOWLEDGE,  Q931ProcSuspendAckNT,        Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_SUSPEND_REJECT,       Q931ProcSuspendRejectNT,     Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_USER_INFORMATION,     Q931ProcUserInformationNT,   Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_DISCONNECT,           Q931ProcDisconnectNT,        Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_RELEASE,              Q931ProcReleaseNT,           Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_RELEASE_COMPLETE,     Q931ProcReleaseCompleteNT,   Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_RESTART,              Q931ProcRestartNT,           Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_RESTART_ACKNOWLEDGE,  Q931ProcRestartAckNT,        Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_CONGESTION_CONTROL,   Q931ProcCongestionControlNT, Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_INFORMATION,          Q931ProcInformationNT,       Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_NOTIFY,               Q931ProcNotifyNT,            Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_STATUS,               Q931ProcStatusNT,            Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_STATUS_ENQUIRY,       Q931ProcStatusEnquiryNT,     Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_SEGMENT,              Q931ProcSegmentNT,           Q931Umes_Generic, Q931Pmes_Generic);

	/*
	 * Q.932
	 */
	Q931DialectSetMesProc(d, Q932mes_FACILITY,             Q932ProcFacilityNT,          Q932Umes_Facility,          Q932Pmes_Facility);
	Q931DialectSetMesProc(d, Q932mes_HOLD,                 Q932ProcHoldNT,              Q932Umes_Hold,              Q932Pmes_Hold);
	Q931DialectSetMesProc(d, Q932mes_HOLD_ACKNOWLEDGE,     Q932ProcHoldAckNT,           Q932Umes_HoldAck,           Q932Pmes_HoldAck);
	Q931DialectSetMesProc(d, Q932mes_HOLD_REJECT,          Q932ProcHoldRejectNT,        Q932Umes_HoldReject,        Q932Pmes_HoldReject);
	Q931DialectSetMesProc(d, Q932mes_REGISTER,             Q932ProcRegisterNT,          Q932Umes_Register,          Q932Pmes_Register);
	Q931DialectSetMesProc(d, Q932mes_RETRIEVE,             Q932ProcRetrieveNT,          Q932Umes_Retrieve,          Q932Pmes_Retrieve);
	Q931DialectSetMesProc(d, Q932mes_RETRIEVE_ACKNOWLEDGE, Q932ProcRetrieveAckNT,       Q932Umes_RetrieveAck,       Q932Pmes_RetrieveAck);
	Q931DialectSetMesProc(d, Q932mes_RETRIEVE_REJECT,      Q932ProcRetrieveRejectNT,    Q932Umes_RetrieveReject,    Q932Pmes_RetrieveReject);


	/*
	 * Unknown / Invalid (Unexpected) Message handler
	 */
	Q931DialectSetUnknownMesProc(d, Q931ProcUnknownMessage);
	Q931DialectSetInvalidMesProc(d, Q931ProcUnexpectedMessage);


	/*
	 * Q.931 IE encoder/decoder table
	 */
	Q931DialectSetIEProc(d, Q931ie_SEGMENTED_MESSAGE,                Q931Pie_Segment,     Q931Uie_Segment,      Q931Die_Segment);
	Q931DialectSetIEProc(d, Q931ie_BEARER_CAPABILITY,                Q931Pie_BearerCap,   Q931Uie_BearerCap,    Q931Die_BearerCap);
	Q931DialectSetIEProc(d, Q931ie_CAUSE,                            Q931Pie_Cause,       Q931Uie_Cause,        Q931Die_Cause);
	Q931DialectSetIEProc(d, Q931ie_CALL_IDENTITY,                    Q931Pie_CallID,      Q931Uie_CallID,       Q931Die_CallID);
	Q931DialectSetIEProc(d, Q931ie_CALL_STATE,                       Q931Pie_CallState,   Q931Uie_CallState,    Q931Die_CallState);
	Q931DialectSetIEProc(d, Q931ie_CHANNEL_IDENTIFICATION,           Q931Pie_ChanID,      Q931Uie_ChanID,       Q931Die_ChanID);
	Q931DialectSetIEProc(d, Q931ie_PROGRESS_INDICATOR,               Q931Pie_ProgInd,     Q931Uie_ProgInd,      Q931Die_ProgInd);
	Q931DialectSetIEProc(d, Q931ie_NETWORK_SPECIFIC_FACILITIES,      Q931Pie_NetFac,      Q931Uie_NetFac,       Q931Die_Generic);
	Q931DialectSetIEProc(d, Q931ie_NOTIFICATION_INDICATOR,           Q931Pie_NotifInd,    Q931Uie_NotifInd,     Q931Die_Generic);
	Q931DialectSetIEProc(d, Q931ie_DISPLAY,                          Q931Pie_Display,     Q931Uie_Display,      Q931Die_Display);
	Q931DialectSetIEProc(d, Q931ie_DATETIME,                         Q931Pie_DateTime,    Q931Uie_DateTime,     Q931Die_DateTime);
	Q931DialectSetIEProc(d, Q931ie_KEYPAD_FACILITY,                  Q931Pie_KeypadFac,   Q931Uie_KeypadFac,    Q931Die_Generic);
	Q931DialectSetIEProc(d, Q931ie_SIGNAL,                           Q931Pie_Signal,      Q931Uie_Signal,       Q931Die_Signal);
	Q931DialectSetIEProc(d, Q931ie_TRANSIT_DELAY_SELECTION_AND_IND,  Q931Pie_TransNetSel, Q931Uie_TransNetSel,  Q931Die_Generic);
	Q931DialectSetIEProc(d, Q931ie_CALLING_PARTY_NUMBER,             Q931Pie_CallingNum,  Q931Uie_CallingNum,   Q931Die_CallingNum);
	Q931DialectSetIEProc(d, Q931ie_CALLING_PARTY_SUBADDRESS,         Q931Pie_CallingSub,  Q931Uie_CallingSub,   Q931Die_CallingSub);
	Q931DialectSetIEProc(d, Q931ie_CALLED_PARTY_NUMBER,              Q931Pie_CalledNum,   Q931Uie_CalledNum,    Q931Die_CalledNum);
	Q931DialectSetIEProc(d, Q931ie_CALLED_PARTY_SUBADDRESS,          Q931Pie_CalledSub,   Q931Uie_CalledSub,    Q931Die_CalledSub);
	Q931DialectSetIEProc(d, Q931ie_TRANSIT_NETWORK_SELECTION,        Q931Pie_TransNetSel, Q931Uie_TransNetSel,  Q931Die_Generic);
	Q931DialectSetIEProc(d, Q931ie_RESTART_INDICATOR,                Q931Pie_RestartInd,  Q931Uie_RestartInd,   Q931Die_RestartInd);
	Q931DialectSetIEProc(d, Q931ie_LOW_LAYER_COMPATIBILITY,          Q931Pie_LLComp,      Q931Uie_LLComp,       Q931Die_LLComp);
	Q931DialectSetIEProc(d, Q931ie_HIGH_LAYER_COMPATIBILITY,         Q931Pie_HLComp,      Q931Uie_HLComp,       Q931Die_HLComp);
	Q931DialectSetIEProc(d, Q931ie_USER_USER,                        Q931Pie_UserUser,    Q931Uie_UserUser,     Q931Die_UserUser);
	Q931DialectSetIEProc(d, Q931ie_SENDING_COMPLETE,                 Q931Pie_SendComplete,Q931Uie_SendComplete, Q931Die_SendComplete);
	Q931DialectSetIEProc(d, Q931ie_CONNECTED_NUMBER,                 Q931Pie_Generic,     Q931Uie_Generic,      Q931Die_Generic);
	Q931DialectSetIEProc(d, Q931ie_CONNECTED_SUBADDRESS,             Q931Pie_Generic,     Q931Uie_Generic,      Q931Die_Generic);

	/*
	 * Q.932
	 */
	Q931DialectSetIEProc(d, Q932ie_FACILITY,                       Q931Pie_Generic,     Q931Uie_Generic,    Q931Die_Generic);
	Q931DialectSetIEProc(d, Q932ie_EXTENDED_FACILITY,              Q931Pie_Generic,     Q931Uie_Generic,    Q931Die_Generic);
	Q931DialectSetIEProc(d, Q932ie_FEATURE_ACTIVATION,             Q931Pie_Generic,     Q931Uie_Generic,    Q931Die_Generic);
	Q931DialectSetIEProc(d, Q932ie_FEATURE_INDICATION,             Q931Pie_Generic,     Q931Uie_Generic,    Q931Die_Generic);
	Q931DialectSetIEProc(d, Q932ie_SERVICE_PROFILE_IDENTIFICATION, Q931Pie_Generic,     Q931Uie_Generic,    Q931Die_Generic);
	Q931DialectSetIEProc(d, Q932ie_ENDPOINT_IDENTIFIER,            Q931Pie_Generic,     Q931Uie_Generic,    Q931Die_Generic);

	/*
	 * Q.952
	 */
	Q931DialectSetIEProc(d, Q931ie_REDIRECTING_NUMBER,             Q931Pie_Generic,     Q931Uie_Generic,    Q931Die_Generic);
	Q931DialectSetIEProc(d, Q931ie_REDIRECTION_NUMBER,             Q931Pie_Generic,     Q931Uie_Generic,    Q931Die_Generic);

	/*
	 * State -> Allowed message table
	 */
	/* Any State */
	Q931DialectAddStateEntry(d, Q931_NANY, Q931mes_STATUS,           MSGF_FROM_BOTH);
	Q931DialectAddStateEntry(d, Q931_NANY, Q931mes_STATUS_ENQUIRY,   MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_UANY, Q931mes_RESTART,             MSGF_FROM_BOTH);
	Q931DialectAddStateEntry(d, Q931_UANY, Q931mes_RESTART_ACKNOWLEDGE, MSGF_FROM_BOTH);

	/* State 0 Idle */
	Q931DialectAddStateEntry(d, Q931_N0, Q931mes_SETUP,            MSGF_FROM_BOTH);
	Q931DialectAddStateEntry(d, Q931_N0, Q931mes_RELEASE,          MSGF_FROM_BOTH);
	Q931DialectAddStateEntry(d, Q931_N0, Q931mes_RELEASE_COMPLETE, MSGF_FROM_BOTH);
	Q931DialectAddStateEntry(d, Q931_N0, Q931mes_RESUME,           MSGF_FROM_L2);

	/* State 1 Call Initiating */
	Q931DialectAddStateEntry(d, Q931_N1, Q931mes_DISCONNECT,        MSGF_FROM_BOTH);
	Q931DialectAddStateEntry(d, Q931_N1, Q931mes_SETUP_ACKNOWLEDGE, MSGF_FROM_L4);
	Q931DialectAddStateEntry(d, Q931_N1, Q931mes_RELEASE_COMPLETE,  MSGF_FROM_L4);
	Q931DialectAddStateEntry(d, Q931_N1, Q931mes_CALL_PROCEEDING,   MSGF_FROM_L4);
	Q931DialectAddStateEntry(d, Q931_N1, Q931mes_ALERTING,          MSGF_FROM_L4); /* ITU-T Q.931 Annex D */
	Q931DialectAddStateEntry(d, Q931_N1, Q931mes_CONNECT,           MSGF_FROM_L4); /* ITU-T Q.931 Annex D */

	/* State 2 Overlap Sending */
	Q931DialectAddStateEntry(d, Q931_N2, Q931mes_DISCONNECT,      MSGF_FROM_BOTH);
	Q931DialectAddStateEntry(d, Q931_N2, Q931mes_INFORMATION,     MSGF_FROM_BOTH);
	Q931DialectAddStateEntry(d, Q931_N2, Q931mes_CALL_PROCEEDING, MSGF_FROM_L4);
	Q931DialectAddStateEntry(d, Q931_N2, Q931mes_ALERTING,        MSGF_FROM_L4);
	Q931DialectAddStateEntry(d, Q931_N2, Q931mes_PROGRESS,        MSGF_FROM_L4);
	Q931DialectAddStateEntry(d, Q931_N2, Q931mes_CONNECT,         MSGF_FROM_L4);
/*	Q931DialectAddStateEntry(d, Q931_N2, Q931mes_RELEASE,         MSGF_FROM_L4); ??? */

	/* State 3 Outgoing Call Proceeding */
	Q931DialectAddStateEntry(d, Q931_N3, Q931mes_DISCONNECT,     MSGF_FROM_BOTH);
	Q931DialectAddStateEntry(d, Q931_N3, Q931mes_PROGRESS,       MSGF_FROM_L4);
	Q931DialectAddStateEntry(d, Q931_N3, Q931mes_ALERTING,       MSGF_FROM_L4);
	Q931DialectAddStateEntry(d, Q931_N3, Q931mes_CONNECT,        MSGF_FROM_L4);
/*	Q931DialectAddStateEntry(d, Q931_N3, Q931mes_RELEASE,        MSGF_FROM_L4); ??? */
	Q931DialectAddStateEntry(d, Q931_N3, Q931mes_INFORMATION,    MSGF_FROM_BOTH);

	/* State 4 Call Delivered */
	Q931DialectAddStateEntry(d, Q931_N4, Q931mes_DISCONNECT,     MSGF_FROM_BOTH);
	Q931DialectAddStateEntry(d, Q931_N4, Q931mes_CONNECT,        MSGF_FROM_L4);
	Q931DialectAddStateEntry(d, Q931_N4, Q931mes_PROGRESS,       MSGF_FROM_L4);
	Q931DialectAddStateEntry(d, Q931_N4, Q931mes_INFORMATION,    MSGF_FROM_BOTH);

	/* State 6 Call Present */
	Q931DialectAddStateEntry(d, Q931_N6, Q931mes_CONNECT,           MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_N6, Q931mes_RELEASE_COMPLETE,  MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_N6, Q931mes_SETUP_ACKNOWLEDGE, MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_N6, Q931mes_INFORMATION,       MSGF_FROM_L4);
	Q931DialectAddStateEntry(d, Q931_N6, Q931mes_ALERTING,          MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_N6, Q931mes_CALL_PROCEEDING,   MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_N6, Q931mes_RELEASE,           MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_N6, Q931mes_DISCONNECT,        MSGF_FROM_L2);

	/* State 7 Call Received */
	Q931DialectAddStateEntry(d, Q931_N7, Q931mes_CONNECT,        MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_N7, Q931mes_DISCONNECT,     MSGF_FROM_BOTH);
	Q931DialectAddStateEntry(d, Q931_N7, Q931mes_INFORMATION,    MSGF_FROM_BOTH);

	/* State 8 Connect request */
	Q931DialectAddStateEntry(d, Q931_N8, Q931mes_CONNECT_ACKNOWLEDGE, MSGF_FROM_L4);
	Q931DialectAddStateEntry(d, Q931_N8, Q931mes_DISCONNECT,          MSGF_FROM_BOTH);
	Q931DialectAddStateEntry(d, Q931_N8, Q931mes_INFORMATION,         MSGF_FROM_BOTH);

	/* State 9 Incoming Call Proceeding */
	Q931DialectAddStateEntry(d, Q931_N9, Q931mes_CONNECT,        MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_N9, Q931mes_ALERTING,       MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_N9, Q931mes_PROGRESS,       MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_N9, Q931mes_DISCONNECT,     MSGF_FROM_BOTH);
	Q931DialectAddStateEntry(d, Q931_N9, Q931mes_INFORMATION,    MSGF_FROM_BOTH);

	/* State 10 Active */
	Q931DialectAddStateEntry(d, Q931_N10, Q931mes_DISCONNECT,          MSGF_FROM_BOTH);
	Q931DialectAddStateEntry(d, Q931_N10, Q931mes_CONNECT_ACKNOWLEDGE, MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_N10, Q931mes_SUSPEND,             MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_N10, Q931mes_NOTIFY,              MSGF_FROM_BOTH);
	Q931DialectAddStateEntry(d, Q931_N10, Q931mes_INFORMATION,         MSGF_FROM_BOTH);
	Q931DialectAddStateEntry(d, Q931_N10, Q932mes_FACILITY,            MSGF_FROM_BOTH);

	/* State 11 Disconnect Request */
	Q931DialectAddStateEntry(d, Q931_N11, Q931mes_RELEASE,        MSGF_FROM_L4);
	Q931DialectAddStateEntry(d, Q931_N11, Q931mes_DISCONNECT,     MSGF_FROM_L2);
/*	Q931DialectAddStateEntry(d, Q931_N11, Q931mes_NOTIFY,         MSGF_FROM_L2); ??? */
	Q931DialectAddStateEntry(d, Q931_N11, Q931mes_INFORMATION,    MSGF_FROM_BOTH);

	/* State 12 Disconnect Ind */
	Q931DialectAddStateEntry(d, Q931_N12, Q931mes_RELEASE,        MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_N12, Q931mes_DISCONNECT,     MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_N12, Q931mes_INFORMATION,    MSGF_FROM_BOTH);

	/* State 15 Suspend Request */
	Q931DialectAddStateEntry(d, Q931_N15, Q931mes_SUSPEND_ACKNOWLEDGE, MSGF_FROM_L4);
	Q931DialectAddStateEntry(d, Q931_N15, Q931mes_SUSPEND_REJECT,      MSGF_FROM_L4);
	Q931DialectAddStateEntry(d, Q931_N15, Q931mes_DISCONNECT,          MSGF_FROM_BOTH);
/*	Q931DialectAddStateEntry(d, Q931_N15, Q931mes_RELEASE,             MSGF_FROM_L2); ??? */
	Q931DialectAddStateEntry(d, Q931_N15, Q931mes_INFORMATION,         MSGF_FROM_BOTH);

	/* State 17 Resume Request */
	Q931DialectAddStateEntry(d, Q931_N17, Q931mes_RESUME_ACKNOWLEDGE,  MSGF_FROM_L4);
	Q931DialectAddStateEntry(d, Q931_N17, Q931mes_RESUME_REJECT,       MSGF_FROM_L4);
	Q931DialectAddStateEntry(d, Q931_N17, Q931mes_DISCONNECT,          MSGF_FROM_BOTH);

	/* State 19 Release Request */
	Q931DialectAddStateEntry(d, Q931_N19, Q931mes_RELEASE_COMPLETE,    MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_N19, Q931mes_RELEASE,             MSGF_FROM_L2);

	/* State 25 Overlap Receiving */
/*	Q931DialectAddStateEntry(d, Q931_N25, Q931mes_SETUP_ACKNOWLEDGE,   MSGF_FROM_L4); */
/*	Q931DialectAddStateEntry(d, Q931_N25, Q931mes_RELEASE_COMPLETE,    MSGF_FROM_L4); */
	Q931DialectAddStateEntry(d, Q931_N25, Q931mes_CALL_PROCEEDING,     MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_N25, Q931mes_ALERTING,            MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_N25, Q931mes_PROGRESS,            MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_N25, Q931mes_CONNECT,             MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_N25, Q931mes_INFORMATION,         MSGF_FROM_L4);
	Q931DialectAddStateEntry(d, Q931_N25, Q931mes_DISCONNECT,          MSGF_FROM_BOTH);


	/*
	 * Timer default values
	 */
	Q931DialectSetTimerProcAll(d, Q931TimeoutDummy);

	/* Per-timer timeout callbacks, per-mode for now, until i know if they can be shared (and which) */
#if __WIP__
	Q931DialectSetTimerProc(d, Q931_TIMER_T301, Q931ProcTimeoutT301NT);
	Q931DialectSetTimerProc(d, Q931_TIMER_T302, Q931ProcTimeoutT302NT);
#endif
	Q931DialectSetTimerProc(d, Q931_TIMER_T303, Q931ProcTimeoutT303NT);
#if __WIP__
	Q931DialectSetTimerProc(d, Q931_TIMER_T304, Q931ProcTimeoutT304NT);
	Q931DialectSetTimerProc(d, Q931_TIMER_T305, Q931ProcTimeoutT305NT);
	Q931DialectSetTimerProc(d, Q931_TIMER_T306, Q931ProcTimeoutT305NT);
	Q931DialectSetTimerProc(d, Q931_TIMER_T307, Q931ProcTimeoutT305NT);
	Q931DialectSetTimerProc(d, Q931_TIMER_T308, Q931ProcTimeoutT308NT);
	Q931DialectSetTimerProc(d, Q931_TIMER_T309, Q931ProcTimeoutT309NT);
	Q931DialectSetTimerProc(d, Q931_TIMER_T310, Q931ProcTimeoutT310NT);
	Q931DialectSetTimerProc(d, Q931_TIMER_T312, Q931ProcTimeoutT311NT);
	Q931DialectSetTimerProc(d, Q931_TIMER_T314, Q931ProcTimeoutT314NT);
	Q931DialectSetTimerProc(d, Q931_TIMER_T316, Q931ProcTimeoutT316NT);
	Q931DialectSetTimerProc(d, Q931_TIMER_T317, Q931ProcTimeoutT317NT);
	Q931DialectSetTimerProc(d, Q931_TIMER_T320, Q931ProcTimeoutT319NT);
	Q931DialectSetTimerProc(d, Q931_TIMER_T321, Q931ProcTimeoutT321NT);
	Q931DialectSetTimerProc(d, Q931_TIMER_T322, Q931ProcTimeoutT322NT);
#endif

	Q931DialectSetTimeout(d, Q931_TIMER_T301, 180000);	/* T301: 180s */
	Q931DialectSetTimeout(d, Q931_TIMER_T302,  15000);	/* T302:  15s */
	Q931DialectSetTimeout(d, Q931_TIMER_T303,   4000);	/* T303:   4s */
	Q931DialectSetTimeout(d, Q931_TIMER_T304,  20000);	/* T304:  20s */
	Q931DialectSetTimeout(d, Q931_TIMER_T305,  30000);	/* T305:  30s */
	Q931DialectSetTimeout(d, Q931_TIMER_T306,  30000);	/* T306:  30s */
	Q931DialectSetTimeout(d, Q931_TIMER_T307, 180000);	/* T307: 180s */
	Q931DialectSetTimeout(d, Q931_TIMER_T308,   4000);	/* T308:   4s */
	Q931DialectSetTimeout(d, Q931_TIMER_T309,  60000);	/* T309:  60s */
	Q931DialectSetTimeout(d, Q931_TIMER_T310,  10000);	/* T310:  10s */
	Q931DialectSetTimeout(d, Q931_TIMER_T312,  12000);	/* T312:  12s */
	Q931DialectSetTimeout(d, Q931_TIMER_T314,   4000);	/* T314:   4s */
	Q931DialectSetTimeout(d, Q931_TIMER_T316, 120000);	/* T316: 120s */
	Q931DialectSetTimeout(d, Q931_TIMER_T317,  90000);	/* T317:  90s */
	Q931DialectSetTimeout(d, Q931_TIMER_T320,  30000);	/* T320:  30s */
	Q931DialectSetTimeout(d, Q931_TIMER_T321,  30000);	/* T321:  30s */
	Q931DialectSetTimeout(d, Q931_TIMER_T322,   4000);	/* T322:   4s */

	/*
	 * Q.931 message <-> ie table
	 */
	Q931DialectSetMesIEMap(d, Q931MessageIEs);
}

/*****************************************************************************

  Function:		Q931ProcAlertingNT

*****************************************************************************/
L3INT Q931ProcAlertingNT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV (TODO: move one level up,
	 * change function signature(s) to use Q931_Call * instead of
	 * Q931_TrunkInfo_t *)
	 */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state (TODO: move one level up) */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from)) {
		Q931Log(trunk, Q931_LOG_DEBUG, "Message %d from %d not allowed in this state %s\n", msg->MesType, from, Q931CallGetStateName(call));
		return Q931E_UNEXPECTED_MESSAGE;
	}
#endif
	switch (from) {
	case Q931_MSG_FROM_L4:
		switch (Q931CallGetState(call)) {
		case Q931_N2:  /* Alerting request */
			/* Stop T302 */
			Q931CallStopTimer(call, Q931_TIMER_T302);
		case Q931_N3:
			/* Send ALERTING */
			ret = Q931Tx32(trunk, 0, msg, msg->Size);

			/* => N4 Call delivered */
			Q931CallSetState(call, Q931_N4);
			break;
		default:
			break;
		}
		break;
	case Q931_MSG_FROM_L2:
		switch (Q931CallGetState(call)) {
		case Q931_N6:
			/* Stop T303 */
			Q931CallStopTimer(call, Q931_TIMER_T303);

			/* => N7 Call received */
			Q931CallSetState(call, Q931_N7);

			/* TODO: Invoke new event CB: Alerting indication */
			ret = Q931Tx34(trunk, call, msg, msg->Size);

			/* Start T301 */
			Q931CallStartTimer(call, Q931_TIMER_T301);
			break;

		case Q931_N9:
			/* Stop T310 */
			Q931CallStopTimer(call, Q931_TIMER_T310);

			/* => N7 Call received */
			Q931CallSetState(call, Q931_N7);

			/* TODO: Invoke new event CB: Alerting indication */
			ret = Q931Tx34(trunk, call, msg, msg->Size);

			/* Start T301 */
			Q931CallStartTimer(call, Q931_TIMER_T301);
			break;

		default:
			break;
		}
		break;
	default:
		ret = Q931E_INTERNAL;
	}
	return ret;
}

/*****************************************************************************

  Function:		Q931ProcCallProceedingNT

*****************************************************************************/
L3INT Q931ProcCallProceedingNT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV (TODO: move one level up,
	 * change function signature(s) to use Q931_Call * instead of
	 * Q931_TrunkInfo_t *)
	 */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state (TODO: move one level up) */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from)) {
		Q931Log(trunk, Q931_LOG_DEBUG, "Message %d from %d not allowed in this state %s\n", msg->MesType, from, Q931CallGetStateName(call));
		return Q931E_UNEXPECTED_MESSAGE;
	}
#endif
	switch (from) {
	case Q931_MSG_FROM_L4:
		switch (Q931CallGetState(call)) {
		case Q931_N1:	/* Call proceeding request */
			/* Send CALL PROCEEDING */
			ret = Q931Tx32(trunk, 0, msg, msg->Size);

			/* => N3 Outgoing call proceeding */
			Q931CallSetState(call, Q931_N3);
			break;

		case Q931_N2:
			/* Stop T302 */
			Q931CallStopTimer(call, Q931_TIMER_T302);

			/* Send CALL PROCEEDING */
			ret = Q931Tx32(trunk, 0, msg, msg->Size);

			/* => N3 Outgoing call proceeding */
			Q931CallSetState(call, Q931_N3);
			break;
		default:
			break;
		}
		break;
	case Q931_MSG_FROM_L2:
		switch (Q931CallGetState(call)) {
		case Q931_N6:
			/* Stop T303 */
			Q931CallStopTimer(call, Q931_TIMER_T303);

			/* Start T310 */
			Q931CallStartTimer(call, Q931_TIMER_T310);

			/* => N9 Incoming call proceeding */
			Q931CallSetState(call, Q931_N9);

			/* TODO: Invoke new event CB: Proceeding indication */
			ret = Q931Tx34(trunk, call, msg, msg->Size);
			break;
		default:
			break;
		}
		break;
	default:
		ret = Q931E_INTERNAL;
	}
	return ret;
}

/*****************************************************************************

  Function:		Q931ProcConnectNT

*****************************************************************************/
L3INT Q931ProcConnectNT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from)) {
		Q931Log(trunk, Q931_LOG_DEBUG, "Message %d from %d not allowed in this state %s\n", msg->MesType, from, Q931CallGetStateName(call));
		return Q931E_UNEXPECTED_MESSAGE;
	}
#endif
	switch (from) {
	case Q931_MSG_FROM_L4:
		switch (Q931CallGetState(call)) {
		case Q931_N2:	/* Setup response */
			/* Stop T302 */
			Q931CallStopTimer(call, Q931_TIMER_T302);

		case Q931_N3:
		case Q931_N4:
			/* Send CONNECT */
			ret = Q931Tx32(trunk, 0, msg, msg->Size);

			/* => N10 Active */
			Q931CallSetState(call, Q931_N10);
			break;
		default:
			break;
		}
		break;
	case Q931_MSG_FROM_L2:
		switch (Q931CallGetState(call)) {
		case Q931_N6:
			/* Stop T303 */
			Q931CallStopTimer(call, Q931_TIMER_T303);

		case Q931_N7:
			/* Stop T301 */
			Q931CallStopTimer(call, Q931_TIMER_T301);

		case Q931_N9:
			/* Stop T310 */
			Q931CallStopTimer(call, Q931_TIMER_T310);

			/* => N8 Connect request */
			Q931CallSetState(call, Q931_N8);

			/* TODO: Send Setup confirm */
			ret = Q931Tx34(trunk, call, msg, msg->Size);
			break;

		default:
			break;
		}
		break;
	default:
		ret = Q931E_INTERNAL;
	}
	return ret;
}

/*****************************************************************************

  Function:		Q931ProcConnectAckNT

*****************************************************************************/
L3INT Q931ProcConnectAckNT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from)) {
		Q931Log(trunk, Q931_LOG_DEBUG, "Message %d from %d not allowed in this state %s\n", msg->MesType, from, Q931CallGetStateName(call));
		return Q931E_UNEXPECTED_MESSAGE;
	}
#endif
	switch (from) {
	case Q931_MSG_FROM_L4:
		switch (Q931CallGetState(call)) {
		case Q931_N8:	/* Setup comp request */
			/* Send CONNECT ACKNOWLEDGE */
			ret = Q931Tx32(trunk, 0, msg, msg->Size);

			/* => N10 Active */
			Q931CallSetState(call, Q931_N10);
			break;
		default:
			break;
		}
		break;
	case Q931_MSG_FROM_L2:
		switch (Q931CallGetState(call)) {
		case Q931_N10:
			/* == N10 Active */
			break;

		default:
			break;
		}
		break;
	default:
		ret = Q931E_INTERNAL;
	}
	return ret;
}

/*****************************************************************************

  Function:		Q931ProcProgressNT

*****************************************************************************/
L3INT Q931ProcProgressNT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from)) {
		Q931Log(trunk, Q931_LOG_DEBUG, "Message %d from %d not allowed in this state %s\n", msg->MesType, from, Q931CallGetStateName(call));
		return Q931E_UNEXPECTED_MESSAGE;
	}
#endif
	switch (from) {
	case Q931_MSG_FROM_L4:
		switch (Q931CallGetState(call)) {
		case Q931_N2:	/* Progress request */
			/* Stop T302 */
			Q931CallStopTimer(call, Q931_TIMER_T302);

		case Q931_N3:
		case Q931_N4:
			/* Send PROGRESS */
			ret = Q931Tx34(trunk, call, msg, msg->Size);

			/* == N2 Overlap sending
			 * == N3 Outgoing call proceeding
			 * == N4 Call delivered
			 */
			break;
		default:
			break;
		}
		break;
	case Q931_MSG_FROM_L2:
		switch (Q931CallGetState(call)) {
		case Q931_N9:
			/* Send proceeding indication */
			ret = Q931Tx34(trunk, call, msg, msg->Size);

			/* == N9 Incoming call proceeding */
			break;
		default:
			break;
		}
		break;
	default:
		ret = Q931E_INTERNAL;
	}
	return ret;
}

/*****************************************************************************

  Function:		Q931ProcSetupNT

  Description:	Process a SETUP message.

 *****************************************************************************/
L3INT Q931ProcSetupNT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* NOTE: handled by upper layers */

	/* Find the call using CRV */
	if (msg->CRV) {
		call = Q931GetCallByCRV(trunk, msg->CRV);
		if (call && Q931CallGetState(call) != Q931_N0) {
			Q931Log(trunk, Q931_LOG_DEBUG, "SETUP on existing call\n");

			/* Reject SETUP on existing calls */
			Q931Disconnect(trunk, from, msg->CRV, 81);
			return Q931E_UNEXPECTED_MESSAGE;
		}
	}

	if (call) {
		/* check if message is valid in this state */
		if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from)) {
			Q931Log(trunk, Q931_LOG_DEBUG, "Message %d from %d not allowed in this state %s\n", msg->MesType, from, Q931CallGetStateName(call));
			return Q931E_UNEXPECTED_MESSAGE;
		}
	}
#endif
	switch (from) {
	case Q931_MSG_FROM_L4:		/* outgoing call (Setup request) */
		if (!call) {
			/* CR selection */
			call = Q931CallNew(trunk);
			if (!call) {
				return Q931E_INTERNAL;
			}
			msg->CRV = Q931CallGetCRV(call);
		}

		/* Broadcast SETUP message if in PTMP mode */
		ret = Q931Tx32(trunk, Q931_IS_PTP(trunk) ? 0 : 1, msg, msg->Size);
		if (ret)
			return ret;

		/* Start T303 */
		Q931CallStartTimer(call, Q931_TIMER_T303);

		/* => N6 Call present */
		Q931CallSetState(call, Q931_N6);
		break;

	case Q931_MSG_FROM_L2:		/* incoming call */
		if (!call) {
			call = Q931CallNewIncoming(trunk, msg->CRV);
			if (!call) {
				/* Not possible to allocate CRV entry, so must reject call */
				Q931Disconnect(trunk, from, msg->CRV, 42);
				return Q931E_INTERNAL;
			}
		}

		/* Store TEI */
		call->Tei = msg->Tei;

		/* => N1 Call initiated */
		Q931CallSetState(call, Q931_N1);

		/* Send setup indication to user */
		ret = Q931Tx34(trunk, call, msg, msg->Size);
		if (ret != Q931E_NO_ERROR) {
			if (Q931TrunkIsSetFlag(trunk, Q931_TFLAG_AUTO_SETUP_ACK)) {
				Q931AckSetup(trunk, msg);
			}
			return ret;
		} else {
			/* Must be full queue, meaning we can't process the call */
			/* so we must disconnect */
			Q931Disconnect(trunk, from, msg->CRV, 81);
			return ret;
		}
		break;

	default:
		ret = Q931E_INTERNAL;
	}
	return ret;
}

/*****************************************************************************

  Function:		Q931ProcSetupAckNT

  Description:	Used to acknowedge a SETUP. Usually the first initial
				response recevide back used to buy some time.

				Note that ChanID (B Channel Assignment) might come here from
				NT side.

*****************************************************************************/
L3INT Q931ProcSetupAckNT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from)) {
		Q931Log(trunk, Q931_LOG_DEBUG, "Message %d from %d not allowed in this state %s\n", msg->MesType, from, Q931CallGetStateName(call));
		return Q931E_UNEXPECTED_MESSAGE;
	}
#endif
	switch (from) {
	case Q931_MSG_FROM_L4:
		switch (Q931CallGetState(call)) {
		case Q931_N1:	/* More info request */
			/* Send SETUP ACK */
			ret = Q931Tx32(trunk, 0, msg, msg->Size);

			/* Start T302 */
			Q931CallStartTimer(call, Q931_TIMER_T302);

			/* => N2 Overlap sending */
			Q931CallSetState(call, Q931_N2);
			break;
		default:
			break;
		}
		break;
	case Q931_MSG_FROM_L2:
		switch (Q931CallGetState(call)) {
		case Q931_N6:
			/* Stop T303 */
			Q931CallStopTimer(call, Q931_TIMER_T303);

			/* TODO: More information indication */

			/* Start T304 */
			Q931CallStartTimer(call, Q931_TIMER_T304);

			/* => N25 Overlap receiving */
			Q931CallSetState(call, Q931_N25);
			break;
		default:
			break;
		}
		break;
	default:
		ret = Q931E_INTERNAL;
	}
	return ret;
}

/*****************************************************************************

  Function:		Q931ProcResumeNT

*****************************************************************************/
L3INT Q931ProcResumeNT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;

	switch (from) {
	case Q931_MSG_FROM_L2:
		/* Find the call using CRV */
		call = Q931GetCallByCRV(trunk, msg->CRV);
		if (!call) {
			call = Q931CallNew(trunk);
			if (!call) {
				return Q931E_INTERNAL;
			}

			/* Send resume indication to user */
			ret = Q931Tx34(trunk, call, msg, msg->Size);
			if (ret != Q931E_NO_ERROR)
				return ret;

			/* => N17 Resume request */
			Q931CallSetState(call, Q931_N17);
		} else {
			return Q931E_ILLEGAL_MESSAGE;
		}
		break;

	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}

/*****************************************************************************

  Function:		Q931ProcResumeAckNT

*****************************************************************************/
L3INT Q931ProcResumeAckNT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from)) {
		Q931Log(trunk, Q931_LOG_DEBUG, "Message %d from %d not allowed in this state %s\n", msg->MesType, from, Q931CallGetStateName(call));
		return Q931E_UNEXPECTED_MESSAGE;
	}
#endif
	switch (from) {
	case Q931_MSG_FROM_L4: /* Resume response */
		/* Send RESUME ACKNOWLEDGE */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);

		/* Stop T307 */
		Q931CallStopTimer(call, Q931_TIMER_T307);

		/* Release CallID??? */

		/* => N10 Active */
		Q931CallSetState(call, Q931_N10);
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}

/*****************************************************************************

  Function:		Q931ProcResumeRejectNT

*****************************************************************************/
L3INT Q931ProcResumeRejectNT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from)) {
		Q931Log(trunk, Q931_LOG_DEBUG, "Message %d from %d not allowed in this state %s\n", msg->MesType, from, Q931CallGetStateName(call));
		return Q931E_UNEXPECTED_MESSAGE;
	}
#endif
	switch (from) {
	case Q931_MSG_FROM_L4: /* Resume response */
		/* Send RESUME REJECT */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);

		/* Release CRV */
		Q931CallSetState(call, Q931_N0);
		Q931CallRelease(call);
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}

/*****************************************************************************

  Function:		Q931ProcSuspendNT

*****************************************************************************/
L3INT Q931ProcSuspendNT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from)) {
		Q931Log(trunk, Q931_LOG_DEBUG, "Message %d from %d not allowed in this state %s\n", msg->MesType, from, Q931CallGetStateName(call));
		return Q931E_UNEXPECTED_MESSAGE;
	}
#endif
	switch (from) {
	case Q931_MSG_FROM_L2:
		/* Send Suspend indication */
		ret = Q931Tx34(trunk, call, msg, msg->Size);

		/* => N15 Suspend request */
		Q931CallSetState(call, Q931_N15);
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}

/*****************************************************************************

  Function:		Q931ProcSuspendAckNT

*****************************************************************************/
L3INT Q931ProcSuspendAckNT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from)) {
		Q931Log(trunk, Q931_LOG_DEBUG, "Message %d from %d not allowed in this state %s\n", msg->MesType, from, Q931CallGetStateName(call));
		return Q931E_UNEXPECTED_MESSAGE;
	}
#endif
	switch (from) {
	case Q931_MSG_FROM_L4:	/* Suspend response */
		/* Send SUSPEND ACKNOWLEDGE */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);

		/* Start T307 */
		Q931CallStartTimer(call, Q931_TIMER_T307);

		/* TODO: Spec says to release the CRV, but that'd mean we loose the call handle
		 * ==> Add per-call flag Q931_CALL_SUSPENDED
		 */
		/* Q931CallSetFlag(call, Q931_CALL_SUSPENDED); */

		/* => N0 Null */
		Q931CallSetState(call, Q931_N0);
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}

/*****************************************************************************

  Function:		Q931ProcSuspendRejectNT

*****************************************************************************/
L3INT Q931ProcSuspendRejectNT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from)) {
		Q931Log(trunk, Q931_LOG_DEBUG, "Message %d from %d not allowed in this state %s\n", msg->MesType, from, Q931CallGetStateName(call));
		return Q931E_UNEXPECTED_MESSAGE;
	}
#endif
	switch (from) {
	case Q931_MSG_FROM_L4:	/* Suspend reject request */
		/* Send SUSPEND REJECT */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);

		/* == N10 Active */
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}

/*****************************************************************************

  Function:		Q931ProcInformationNT

*****************************************************************************/
L3INT Q931ProcUserInformationNT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;

	/* TODO: ???? */
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch (from) {
	case Q931_MSG_FROM_L4:
		/* Send INFORMATION */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);

		/* Start / Restart T304 */
		Q931CallStartTimer(call, Q931_TIMER_T304);

		/* == U2 (Overlap sending) */
		break;
	case Q931_MSG_FROM_L2:
		/* TODO: send Info indication */
		ret = Q931Tx34(trunk, call, msg, msg->Size);

		/* Start T302 */
		Q931CallStartTimer(call, Q931_TIMER_T302);

		/* == U25 (Overlap receiving) */
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}

/*****************************************************************************

  Function:		Q931ProcDisconnectNT

*****************************************************************************/
L3INT Q931ProcDisconnectNT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;

	Q931Log(trunk, Q931_LOG_DEBUG, "Processing DISCONNECT message from %s for CRV: %d (%#hx)\n",
						 from == 4 ? "Local" : "Remote", msg->CRV, msg->CRV);

	/* TODO: not complete... */
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch (from) {
	case Q931_MSG_FROM_L4:
		switch (Q931CallGetState(call)) {
		case Q931_N0:	/* Disconnect request */
		case Q931_N6:
		case Q931_N11:
		case Q931_N12:
		case Q931_N19:
		case Q931_N22:
			/* TODO: should never occur? */
			break;

		case Q931_N7:
		case Q931_N8:
		case Q931_N9:
		default:
			/* Stop all timers */
			Q931CallStopAllTimers(call);

			/* Tones are handled by L4!! */

			/* Disconnect */
			ret = Q931Tx32(trunk, 0, msg, msg->Size);

			/* Start T305 */
			Q931CallStartTimer(call, Q931_TIMER_T305);

			/* => N12 Disconnect indication */
			Q931CallSetState(call, Q931_N12);
		}
		break;
	case Q931_MSG_FROM_L2:
		switch (Q931CallGetState(call)) {
		case Q931_N0:
		case Q931_N11:
		case Q931_N12:
		case Q931_N19:
		case Q931_N22:
			/* TODO: should never occur? */
			break;

#if 0
		/* TODO: Move to release request */
		case Q931_N12:
			/* Stop T305 or T306 */
			Q931CallStopTimer(call, Q931_TIMER_T305);
			Q931CallStopTimer(call, Q931_TIMER_T306);

			/* TODO: Send RELEASE */

			/* Start T308 */
			Q931CallStartTimer(call, Q931_TIMER_T308);

			/* => N19 Release request */
			Q931CallSetState(call, Q931_N19);
			break;
#endif

		case Q931_N1:
			/* B-Channel disconnect for states > N1 handled by L4!! */
		case Q931_N7:
		case Q931_N8:
		case Q931_N9:
		default:
			/* Stop all timers */
			Q931CallStopAllTimers(call);

			/* TODO: Send Disconnect indication */
			ret = Q931Tx34(trunk, call, msg, msg->Size);

			/* => N11 Disconnect request */
			Q931CallSetState(call, Q931_N11);
		}
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}

/*****************************************************************************

  Function:		Q931ProcReleaseNT

*****************************************************************************/
L3INT Q931ProcReleaseNT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;

	/* TODO: not complete... */
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch (from) {
	case Q931_MSG_FROM_L4:
		switch (Q931CallGetState(call)) {
		case Q931_N11:	/* Release request */
			/* Send RELEASE */
			ret = Q931Tx32(trunk, 0, msg, msg->Size);

			/* Start T308 */
			Q931CallStartTimer(call, Q931_TIMER_T308);

			/* => N19 Release request */
			Q931CallSetState(call, Q931_N19);
			break;

		default:
			break;
		}
		break;
	case Q931_MSG_FROM_L2:
		switch (Q931CallGetState(call)) {
		case Q931_N0:
			/* Send RELEASE COMPLETE reply */
			ret = Q931ReleaseComplete(trunk, call, 0);

			/* Release Call */
			Q931CallRelease(call);
			break;

		case Q931_N6:
			/* Page 22 NC */
			/* Stop all timers */
			Q931CallStopAllTimers(call);

			/* TODO: Send DISCONNECT */

			/* Start T305 or T306 */
			Q931CallStartTimer(call, Q931_TIMER_T305);

			/* => N12 Disconnect indication */
			Q931CallSetState(call, Q931_N12);
			break;

		case Q931_N12:
			/* Stop T305 or T306 */
			Q931CallStopTimer(call, Q931_TIMER_T305);
			Q931CallStopTimer(call, Q931_TIMER_T306);

			/* Send Relase indication */
			ret = Q931Tx34(trunk, call, msg, msg->Size);

			/* B-Channel release handled by L4!! */

			/* Send RELEASE COMPLETE reply */
			ret = Q931ReleaseComplete(trunk, call, 0);

			/* Release CRV */
			Q931CallSetState(call, Q931_N0);
			Q931CallRelease(call);
			break;

		case Q931_N19:
			/* Stop T308 */
			Q931CallStopTimer(call, Q931_TIMER_T308);

			/* TODO: Release confirm */
			ret = Q931Tx34(trunk, call, msg, msg->Size);

			/* B-channel release handled by L4!! */

			/* Release CRV  */
			Q931CallSetState(call, Q931_N0);
			Q931CallRelease(call);
			break;

		default:
			/* Stop all timers */
			Q931CallStopAllTimers(call);

			/* TODO: Release indication */
			ret = Q931Tx34(trunk, call, msg, msg->Size);

			/* B-channel release handled by L4!! */

			/* Send RELEASE COMPLETE reply */
			ret = Q931ReleaseComplete(trunk, call, 0);

			/* Release CRV  */
			Q931CallSetState(call, Q931_N0);
			Q931CallRelease(call);
			break;
		}
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}

/*****************************************************************************

  Function:		Q931ProcReleaseCompleteNT

*****************************************************************************/
L3INT Q931ProcReleaseCompleteNT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch (from) {
	case Q931_MSG_FROM_L4:
		switch (Q931CallGetState(call)) {
		case Q931_N1:	/* Reject request */
			/* Send RELEASE COMPLETE */
			ret = Q931Tx32(trunk, 0, msg, msg->Size);

			/* Release CRV (?) */
			Q931CallSetState(call, Q931_N0);
			Q931CallRelease(call);
			break;
		default:
			break;
		}
		break;
	case Q931_MSG_FROM_L2:
		switch (Q931CallGetState(call)) {
		case Q931_N0:
			/* Release call */
			Q931CallRelease(call);
			break;

		case Q931_N6:
			/* Stop T303 */
			Q931CallStopTimer(call, Q931_TIMER_T303);

			/* TODO: Send Reject indication */
			ret = Q931Tx34(trunk, call, msg, msg->Size);

			/* Release CRV */
			Q931CallSetState(call, Q931_N0);
			Q931CallRelease(call);
			break;

		case Q931_N19:
			/* Stop T308 */
			Q931CallStopTimer(call, Q931_TIMER_T308);

			/* TODO: Send release confirm */
			ret = Q931Tx34(trunk, call, msg, msg->Size);

			/* Release CRV */
			Q931CallSetState(call, Q931_N0);
			Q931CallRelease(call);
			break;
		default:
			/* TODO: Send release indication */
			ret = Q931Tx34(trunk, call, msg, msg->Size);

			/* B-channel release handled by L4!! */

			/* Release CRV */
			Q931CallSetState(call, Q931_N0);
			Q931CallRelease(call);
			break;
		}

		/* Release CRV */
		Q931CallSetState(call, Q931_N0);
		Q931CallRelease(call);
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}

/*****************************************************************************

  Function:		Q931ProcRestartNT

*****************************************************************************/
L3INT Q931ProcRestartNT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	if (msg->CRV)
		return Q931E_INVALID_CRV;

	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	/* TODO: T317, proper handling etc. */

	switch (from) {
	case Q931_MSG_FROM_L4:
		/* TODO Add proc here */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);
		break;
	case Q931_MSG_FROM_L2:
		/* TODO Add proc here */
		ret = Q931Tx34(trunk, call, msg, msg->Size);

		if (Q931TrunkIsSetFlag(trunk, Q931_TFLAG_AUTO_RESTART_ACK)) {
			Q931AckRestart(trunk, msg);
		}
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}

/*****************************************************************************

  Function:		Q931ProcRestartAckNT

*****************************************************************************/
L3INT Q931ProcRestartAckNT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
#if 0
	if (msg->CRV) {
		/* Find the call using CRV */
		ret = Q931FindCRV(trunk, msg->CRV, &callIndex);
		if (ret != Q931E_NO_ERROR)
			return ret;
		/* TODO - Set correct timer here */
		Q931StartTimer(trunk, callIndex, 303);
	}
#endif

	switch (from) {
	case Q931_MSG_FROM_L4:
		/* TODO Add proc here */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);
		break;
	case Q931_MSG_FROM_L2:
		/* TODO Add proc here */
		ret = Q931Tx34(trunk, call, msg, msg->Size);
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}

/*****************************************************************************

  Function:		Q931ProcCongestionControlNT

*****************************************************************************/
L3INT Q931ProcCongestionControlNT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch (from) {
	case Q931_MSG_FROM_L4:
		/* TODO Add proc here */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);
		break;
	case Q931_MSG_FROM_L2:
		/* TODO Add proc here */
		ret = Q931Tx34(trunk, call, msg, msg->Size);
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}

/*****************************************************************************

  Function:		Q931ProcInformationNT

*****************************************************************************/
L3INT Q931ProcInformationNT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch (from) {
	case Q931_MSG_FROM_L4:
		/* TODO Add proc here */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);
		break;
	case Q931_MSG_FROM_L2:
		/* TODO Add proc here */
		ret = Q931Tx34(trunk, call, msg, msg->Size);
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}

/*****************************************************************************

  Function:		Q931ProcNotifyNT

*****************************************************************************/
L3INT Q931ProcNotifyNT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch (from) {
	case Q931_MSG_FROM_L4:
		/* TODO Add proc here */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);
		break;
	case Q931_MSG_FROM_L2:
		/* TODO Add proc here */
		ret = Q931Tx34(trunk, call, msg, msg->Size);
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}

/*****************************************************************************

  Function:		Q931ProcStatusNT

*****************************************************************************/
L3INT Q931ProcStatusNT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch (from) {
	case Q931_MSG_FROM_L4:
		/* TODO Add proc here */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);
		break;
	case Q931_MSG_FROM_L2:
		/* TODO Add proc here */
		ret = Q931Tx34(trunk, call, msg, msg->Size);
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}

/*****************************************************************************

  Function:		Q931ProcStatusEnquiryNT

*****************************************************************************/
L3INT Q931ProcStatusEnquiryNT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch (from) {
#if 0
	case Q931_MSG_FROM_L4:
		/* TODO Add proc here */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);
		break;
#endif
	case Q931_MSG_FROM_L2:
		ret = Q931StatusEnquiryResponse(trunk, call, Q850_CAUSE_RESPONSE_TO_STATUS_ENQUIRY);
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}

/*****************************************************************************

  Function:		Q931ProcSegmentNT

*****************************************************************************/
L3INT Q931ProcSegmentNT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch (from) {
	case Q931_MSG_FROM_L4:
		/* TODO Add proc here */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);
		break;
	case Q931_MSG_FROM_L2:
		/* TODO Add proc here */
		ret = Q931Tx34(trunk, call, msg, msg->Size);
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}

/****************************************************************************/
/******************* Q.932 - Supplementary Services *************************/
/****************************************************************************/

/*****************************************************************************

  Function:		Q932ProcFacilityNT

*****************************************************************************/
L3INT Q932ProcFacilityNT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch (from) {
	case Q931_MSG_FROM_L4:
		/* TODO Add proc here */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);
		break;
	case Q931_MSG_FROM_L2:
		/* TODO Add proc here */
		ret = Q931Tx34(trunk, call, msg, msg->Size);
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}

/*****************************************************************************

  Function:		Q932ProcHoldNT

*****************************************************************************/
L3INT Q932ProcHoldNT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch (from) {
	case Q931_MSG_FROM_L4:
		/* TODO Add proc here */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);
		break;
	case Q931_MSG_FROM_L2:
		/* TODO Add proc here */
		ret = Q931Tx34(trunk, call, msg, msg->Size);
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}

/*****************************************************************************

  Function:		Q932ProcHoldAckNT

*****************************************************************************/
L3INT Q932ProcHoldAckNT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch (from) {
	case Q931_MSG_FROM_L4:
		/* TODO Add proc here */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);
		break;
	case Q931_MSG_FROM_L2:
		/* TODO Add proc here */
		ret = Q931Tx34(trunk, call, msg, msg->Size);
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}

/*****************************************************************************

  Function:		Q932ProcHoldRejectNT

*****************************************************************************/
L3INT Q932ProcHoldRejectNT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch (from) {
	case Q931_MSG_FROM_L4:
		/* TODO Add proc here */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);
		break;
	case Q931_MSG_FROM_L2:
		/* TODO Add proc here */
		ret = Q931Tx34(trunk, call, msg, msg->Size);
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}

/*****************************************************************************

  Function:		Q932ProcRegisterTE

*****************************************************************************/
L3INT Q932ProcRegisterNT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch (from) {
	case Q931_MSG_FROM_L4:
		/* TODO Add proc here */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);
		break;
	case Q931_MSG_FROM_L2:
		/* TODO Add proc here */
		ret = Q931Tx34(trunk, call, msg, msg->Size);
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}

/*****************************************************************************

  Function:		Q932ProcRetrieveNT

*****************************************************************************/
L3INT Q932ProcRetrieveNT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch (from) {
	case Q931_MSG_FROM_L4:
		/* TODO Add proc here */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);
		break;
	case Q931_MSG_FROM_L2:
		/* TODO Add proc here */
		ret = Q931Tx34(trunk, call, msg, msg->Size);
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}

/*****************************************************************************

  Function:		Q931ProcRetrieveAckNT

*****************************************************************************/
L3INT Q932ProcRetrieveAckNT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch (from) {
	case Q931_MSG_FROM_L4:
		/* TODO Add proc here */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);
		break;
	case Q931_MSG_FROM_L2:
		/* TODO Add proc here */
		ret = Q931Tx34(trunk, call, msg, msg->Size);
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}

/*****************************************************************************

  Function:		Q931ProcRetrieveRejectNT

*****************************************************************************/
L3INT Q932ProcRetrieveRejectNT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch (from) {
	case Q931_MSG_FROM_L4:
		/* TODO Add proc here */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);
		break;
	case Q931_MSG_FROM_L2:
		/* TODO Add proc here */
		ret = Q931Tx34(trunk, call, msg, msg->Size);
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}


/************************************************************************************
 * Timer callbacks (NT side)
 ************************************************************************************/

L3INT Q931ProcTimeoutT301NT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call)
{
	Q931Log(trunk, Q931_LOG_NOTICE, "Timer T301 timed out for call %d\n", call->CRV);

	if (Q931CallGetState(call) != Q931_N7) {
		Q931Log(trunk, Q931_LOG_WARNING, "Ignoring timeout of T301 in state %s (wrong state)\n", Q931CallGetStateName(call));
		return Q931E_NO_ERROR;
	}

	/* TODO: Send TIMEOUT indication */

	return Q931E_NO_ERROR;
}

L3INT Q931ProcTimeoutT302NT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call)
{
	Q931Log(trunk, Q931_LOG_NOTICE, "Timer T302 timed out for call %d\n", call->CRV);

	if (Q931CallGetState(call) != Q931_N2) {
		Q931Log(trunk, Q931_LOG_WARNING, "Ignoring timeout of T302 in state %s (wrong state)\n", Q931CallGetStateName(call));
		return Q931E_NO_ERROR;
	}

	/* TODO: Send TIMEOUT indication */

	return Q931E_NO_ERROR;
}

L3INT Q931ProcTimeoutT303NT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call)
{
	L3UCHAR cnt = Q931CallGetTimerExpireCount(call);	/* T303 uses the counter */

	Q931Log(trunk, Q931_LOG_NOTICE, "Timer T303 timed out for call %d (number of times: %hhu)\n", call->CRV, cnt);

	if (Q931CallGetState(call) != Q931_N6) {
		Q931Log(trunk, Q931_LOG_WARNING, "Ignoring timeout of T303 in state %s (wrong state)\n", Q931CallGetStateName(call));
		return Q931E_NO_ERROR;
	}

	if (cnt == 1) {
		/* TODO: Resend SETUP */

		/* Restart T303 */
		Q931CallRestartTimer(call, Q931_TIMER_T303);

		/* No state change */
	} else {
		struct Q931_CallEvent event;

		/* (TODO:) Send RELEASE INDICATION */
		Q931CallInitEvent(&event);
		event.id    = Q931_EVENT_RELEASE_INDICATION;
		event.type  = Q931_EVENT_TYPE_TIMER;
		event.data.timer.id = Q931_TIMER_T303;

		/* Send event immediately to layer 4 */
		Q931CallSendEvent(call, &event);

		/* Release CRV */
		Q931CallSetState(call, Q931_N0);
		Q931CallRelease(call);
	}
	return Q931E_NO_ERROR;
}

L3INT Q931ProcTimeoutT304NT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call)
{
	Q931Log(trunk, Q931_LOG_NOTICE, "Timer T304 timed out for call %d\n", call->CRV);

	if (Q931CallGetState(call) != Q931_N25) {
		Q931Log(trunk, Q931_LOG_WARNING, "Ignoring timeout of T304 in state %s (wrong state)\n", Q931CallGetStateName(call));
		return Q931E_NO_ERROR;
	}

	/* TODO: Send TIMEOUT indication */

	return Q931E_NO_ERROR;
}

L3INT Q931ProcTimeoutT305NT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call)
{
	Q931Log(trunk, Q931_LOG_NOTICE, "Timer T305 timed out for call %d\n", call->CRV);

	if (Q931CallGetState(call) != Q931_N12) {
		Q931Log(trunk, Q931_LOG_WARNING, "Ignoring timeout of T305 in state %s (wrong state)\n", Q931CallGetStateName(call));
		return Q931E_NO_ERROR;
	}

	/* TODO: Send RELEASE */

	/* Start T308 */
	Q931CallStartTimer(call, Q931_TIMER_T308);

	/* => 19: Release request */
	Q931CallSetState(call, Q931_N19);

	return Q931E_NO_ERROR;
}

L3INT Q931ProcTimeoutT306NT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call)
{
	Q931Log(trunk, Q931_LOG_NOTICE, "Timer T306 timed out for call %d\n", call->CRV);

	if (Q931CallGetState(call) != Q931_N12) {
		Q931Log(trunk, Q931_LOG_WARNING, "Ignoring timeout of T306 in state %s (wrong state)\n", Q931CallGetStateName(call));
		return Q931E_NO_ERROR;
	}

	/* TODO: Send RELEASE */

	/* Start T308 */
	Q931CallStartTimer(call, Q931_TIMER_T308);

	/* => 19: Release request */
	Q931CallSetState(call, Q931_N19);

	return Q931E_NO_ERROR;
}

L3INT Q931ProcTimeoutT307NT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call)
{
	Q931Log(trunk, Q931_LOG_NOTICE, "Timer T307 timed out for call %d\n", call->CRV);

	if (Q931CallGetState(call) != Q931_N0) {
		Q931Log(trunk, Q931_LOG_WARNING, "Ignoring timeout of T307 in state %s (wrong state)\n", Q931CallGetStateName(call));
		return Q931E_NO_ERROR;
	}

	/* Release CRV */
	Q931CallSetState(call, Q931_N0);
	Q931CallRelease(call);

	/* TODO: Release B-Channel */

	return Q931E_NO_ERROR;
}

L3INT Q931ProcTimeoutT308NT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call)
{
	L3UCHAR cnt = Q931CallGetTimerExpireCount(call);	/* T308 uses the counter */

	Q931Log(trunk, Q931_LOG_NOTICE, "Timer T308 timed out for call %d (number of times %hhu)\n", call->CRV, cnt);

	if (Q931CallGetState(call) != Q931_N19) {
		Q931Log(trunk, Q931_LOG_WARNING, "Ignoring timeout of T308 in state %s (wrong state)\n", Q931CallGetStateName(call));
		return Q931E_NO_ERROR;
	}

	if (cnt == 1) {
		/* TODO: Send RELEASE */

		/* Restart T308 */
		Q931CallRestartTimer(call, Q931_TIMER_T308);

		/* No state change */
	} else {
		/* TODO: Place B-Channel in maintenance */

		/* TODO: Send RELEASE CONFIRM (error) indication */

		/* Release CRV */
		Q931CallSetState(call, Q931_N0);
		Q931CallRelease(call);
	}
	return Q931E_NO_ERROR;
}

L3INT Q931ProcTimeoutT309NT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call)
{
	Q931Log(trunk, Q931_LOG_NOTICE, "Timer T309 timed out for call %d\n", call->CRV);
	return Q931E_NO_ERROR;
}

L3INT Q931ProcTimeoutT310NT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call)
{
	Q931Log(trunk, Q931_LOG_NOTICE, "Timer T310 timed out for call %d\n", call->CRV);

	if (Q931CallGetState(call) != Q931_N9) {
		Q931Log(trunk, Q931_LOG_WARNING, "Ignoring timeout of T310 in state %s (wrong state)\n", Q931CallGetStateName(call));
		return Q931E_NO_ERROR;
	}

	/* Stop all timers */
	Q931CallStopAllTimers(call);

	/* TODO: Tones option? */
	if (1) {
		/* TODO: Start tone */

		/* TODO: Send DISCONNECT */

		/* Start T306 */
		Q931CallStartTimer(call, Q931_TIMER_T306);
	} else {
		/* TODO: Disconnect B-Channel */

		/* TOOD: Send DISCONNECT */

		/* Start T305 */
		Q931CallStartTimer(call, Q931_TIMER_T305);
	}

	/* => 12: Disconnect indication */
	Q931CallSetState(call, Q931_N12);

	return Q931E_NO_ERROR;
}

L3INT Q931ProcTimeoutT312NT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call)
{
	Q931Log(trunk, Q931_LOG_NOTICE, "Timer T312 timed out for call %d\n", call->CRV);
	return Q931E_NO_ERROR;
}

L3INT Q931ProcTimeoutT313NT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call)
{
	Q931Log(trunk, Q931_LOG_NOTICE, "Timer T313 timed out for call %d\n", call->CRV);
	return Q931E_NO_ERROR;
}

L3INT Q931ProcTimeoutT314NT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call)
{
	Q931Log(trunk, Q931_LOG_NOTICE, "Timer T314 timed out for call %d\n", call->CRV);
	return Q931E_NO_ERROR;
}

L3INT Q931ProcTimeoutT316NT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call)
{
	Q931Log(trunk, Q931_LOG_NOTICE, "Timer T316 timed out for call %d\n", call->CRV);
	return Q931E_NO_ERROR;
}

L3INT Q931ProcTimeoutT317NT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call)
{
	Q931Log(trunk, Q931_LOG_NOTICE, "Timer T317 timed out for call %d\n", call->CRV);
	return Q931E_NO_ERROR;
}

L3INT Q931ProcTimeoutT320NT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call)
{
	Q931Log(trunk, Q931_LOG_NOTICE, "Timer T320 timed out for call %d\n", call->CRV);
	return Q931E_NO_ERROR;
}

L3INT Q931ProcTimeoutT321NT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call)
{
	Q931Log(trunk, Q931_LOG_NOTICE, "Timer T321 timed out for call %d\n", call->CRV);
	return Q931E_NO_ERROR;
}

L3INT Q931ProcTimeoutT322NT(Q931_TrunkInfo_t *trunk, struct Q931_Call *call)
{
	Q931Log(trunk, Q931_LOG_NOTICE, "Timer T322 timed out for call %d\n", call->CRV);
	return Q931E_NO_ERROR;
}

