/*****************************************************************************

  FileName:		q931StateTE.c

  Contents:		Q.931 State Engine for TE (User Mode).

			The controlling state engine for Q.931 is the state engine
			on the NT side. The state engine on the TE side is a slave
			of this. The TE side maintain it's own states as described in
			ITU-T Q931, but will in	raise conditions be overridden by
			the NT side.

			This reference implementation uses a process per message,
			meaning that each message must check call states. This
			is easier for dialect maintenance as each message proc
			can be replaced individually. A new TE variant only
			need to copy the Q931CreateTE and replace those procs or
			need to override.

  License/Copyright:

  Copyright (c) 2007, Jan Vidar Berger, Case Labs, Ltd. All rights reserved.
  email:janvb@caselaboratories.com

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are
  met:

	* Redistributions of source code must retain the above copyright notice,
	  this list of conditions and the following disclaimer.
	* Redistributions in binary form must reproduce the above copyright notice,
	  this list of conditions and the following disclaimer in the documentation
	  and/or other materials provided with the distribution.
	* Neither the name of the Case Labs, Ltd nor the names of its contributors
	  may be used to endorse or promote products derived from this software
	  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.
*****************************************************************************/
#include <stdlib.h>

#include "Q931.h"
#include "Q931priv.h"
#include "Q932.h"

extern L3INT Q931L4HeaderSpace;
extern struct Q931MessageIE Q931MessageIEs[];

/**
 * @defgroup	q931_te		Q.931 TE State machine
 * @{
 */

/**
 * Initialize Q.931 TE dialect
 * @param[in]	d	Q.931 Dialect handle
 */
void Q931CreateTE(struct Q931Dialect *d)
{
	Q931DialectSetName(d, "Q.931 TE");

	/*
	 * Flags
	 */
	Q931DialectSetFlag(d, 0);

	/*
	 * Codesets
	 */
	Q931DialectAddCodeset(d, Q931_CODESET_0);

	/*
	 * Q.931 Message encoder/decoder/handler table
	 */
	Q931DialectSetMesProc(d, Q931mes_ALERTING,             Q931ProcAlertingTE,          Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_CALL_PROCEEDING,      Q931ProcCallProceedingTE,    Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_CONNECT,              Q931ProcConnectTE,           Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_CONNECT_ACKNOWLEDGE,  Q931ProcConnectAckTE,        Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_PROGRESS,             Q931ProcProgressTE,          Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_SETUP,                Q931ProcSetupTE,             Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_SETUP_ACKNOWLEDGE,    Q931ProcSetupAckTE,          Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_RESUME,               Q931ProcResumeTE,            Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_RESUME_ACKNOWLEDGE,   Q931ProcResumeAckTE,         Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_RESUME_REJECT,        Q931ProcResumeRejectTE,      Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_SUSPEND,              Q931ProcSuspendTE,           Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_SUSPEND_ACKNOWLEDGE,  Q931ProcSuspendAckTE,        Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_SUSPEND_REJECT,       Q931ProcSuspendRejectTE,     Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_USER_INFORMATION,     Q931ProcUserInformationTE,   Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_DISCONNECT,           Q931ProcDisconnectTE,        Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_RELEASE,              Q931ProcReleaseTE,           Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_RELEASE_COMPLETE,     Q931ProcReleaseCompleteTE,   Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_RESTART,              Q931ProcRestartTE,           Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_RESTART_ACKNOWLEDGE,  Q931ProcRestartAckTE,        Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_CONGESTION_CONTROL,   Q931ProcCongestionControlTE, Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_INFORMATION,          Q931ProcInformationTE,       Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_NOTIFY,               Q931ProcNotifyTE,            Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_STATUS,               Q931ProcStatusTE,            Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_STATUS_ENQUIRY,       Q931ProcStatusEnquiryTE,     Q931Umes_Generic, Q931Pmes_Generic);
	Q931DialectSetMesProc(d, Q931mes_SEGMENT,              Q931ProcSegmentTE,           Q931Umes_Generic, Q931Pmes_Generic);

	/*
	 * Q.932
	 */
	Q931DialectSetMesProc(d, Q932mes_FACILITY,             Q932ProcFacilityTE,       Q932Umes_Facility,       Q932Pmes_Facility);
	Q931DialectSetMesProc(d, Q932mes_HOLD,                 Q932ProcHoldTE,           Q932Umes_Hold,           Q932Pmes_Hold);
	Q931DialectSetMesProc(d, Q932mes_HOLD_ACKNOWLEDGE,     Q932ProcHoldAckTE,        Q932Umes_HoldAck,        Q932Pmes_HoldAck);
	Q931DialectSetMesProc(d, Q932mes_HOLD_REJECT,          Q932ProcHoldRejectTE,     Q932Umes_HoldReject,     Q932Pmes_HoldReject);
	Q931DialectSetMesProc(d, Q932mes_REGISTER,             Q932ProcRegisterTE,       Q932Umes_Register,       Q932Pmes_Register);
	Q931DialectSetMesProc(d, Q932mes_RETRIEVE,             Q932ProcRetrieveTE,       Q932Umes_Retrieve,       Q932Pmes_Retrieve);
	Q931DialectSetMesProc(d, Q932mes_RETRIEVE_ACKNOWLEDGE, Q932ProcRetrieveAckTE,    Q932Umes_RetrieveAck,    Q932Pmes_RetrieveAck);
	Q931DialectSetMesProc(d, Q932mes_RETRIEVE_REJECT,      Q932ProcRetrieveRejectTE, Q932Umes_RetrieveReject, Q932Pmes_RetrieveReject);

	/*
	 * Unknown / Invalid (Unexpected) Message handler
	 */
	Q931DialectSetUnknownMesProc(d, Q931ProcUnknownMessage);
	Q931DialectSetInvalidMesProc(d, Q931ProcUnexpectedMessage);

	/*
	 * Q.931 IE encoder/decoder table
	 */
	Q931DialectSetIEProc(d, Q931ie_SEGMENTED_MESSAGE,                Q931Pie_Segment,     Q931Uie_Segment,      Q931Die_Segment);
	Q931DialectSetIEProc(d, Q931ie_BEARER_CAPABILITY,                Q931Pie_BearerCap,   Q931Uie_BearerCap,    Q931Die_BearerCap);
	Q931DialectSetIEProc(d, Q931ie_CAUSE,                            Q931Pie_Cause,       Q931Uie_Cause,        Q931Die_Cause);
	Q931DialectSetIEProc(d, Q931ie_CALL_IDENTITY,                    Q931Pie_CallID,      Q931Uie_CallID,       Q931Die_CallID);
	Q931DialectSetIEProc(d, Q931ie_CALL_STATE,                       Q931Pie_CallState,   Q931Uie_CallState,    Q931Die_CallState);
	Q931DialectSetIEProc(d, Q931ie_CHANNEL_IDENTIFICATION,           Q931Pie_ChanID,      Q931Uie_ChanID,       Q931Die_ChanID);
	Q931DialectSetIEProc(d, Q931ie_PROGRESS_INDICATOR,               Q931Pie_ProgInd,     Q931Uie_ProgInd,      Q931Die_ProgInd);
	Q931DialectSetIEProc(d, Q931ie_NETWORK_SPECIFIC_FACILITIES,      Q931Pie_NetFac,      Q931Uie_NetFac,       Q931Die_Generic);
	Q931DialectSetIEProc(d, Q931ie_NOTIFICATION_INDICATOR,           Q931Pie_NotifInd,    Q931Uie_NotifInd,     Q931Die_Generic);
	Q931DialectSetIEProc(d, Q931ie_DISPLAY,                          Q931Pie_Display,     Q931Uie_Display,      Q931Die_Display);
	Q931DialectSetIEProc(d, Q931ie_DATETIME,                         Q931Pie_DateTime,    Q931Uie_DateTime,     Q931Die_DateTime);
	Q931DialectSetIEProc(d, Q931ie_KEYPAD_FACILITY,                  Q931Pie_KeypadFac,   Q931Uie_KeypadFac,    Q931Die_Generic);
	Q931DialectSetIEProc(d, Q931ie_SIGNAL,                           Q931Pie_Signal,      Q931Uie_Signal,       Q931Die_Signal);
	Q931DialectSetIEProc(d, Q931ie_TRANSIT_DELAY_SELECTION_AND_IND,  Q931Pie_TransNetSel, Q931Uie_TransNetSel,  Q931Die_Generic);
	Q931DialectSetIEProc(d, Q931ie_CALLING_PARTY_NUMBER,             Q931Pie_CallingNum,  Q931Uie_CallingNum,   Q931Die_CallingNum);
	Q931DialectSetIEProc(d, Q931ie_CALLING_PARTY_SUBADDRESS,         Q931Pie_CallingSub,  Q931Uie_CallingSub,   Q931Die_CallingSub);
	Q931DialectSetIEProc(d, Q931ie_CALLED_PARTY_NUMBER,              Q931Pie_CalledNum,   Q931Uie_CalledNum,    Q931Die_CalledNum);
	Q931DialectSetIEProc(d, Q931ie_CALLED_PARTY_SUBADDRESS,          Q931Pie_CalledSub,   Q931Uie_CalledSub,    Q931Die_CalledSub);
	Q931DialectSetIEProc(d, Q931ie_TRANSIT_NETWORK_SELECTION,        Q931Pie_TransNetSel, Q931Uie_TransNetSel,  Q931Die_Generic);
	Q931DialectSetIEProc(d, Q931ie_RESTART_INDICATOR,                Q931Pie_RestartInd,  Q931Uie_RestartInd,   Q931Die_RestartInd);
	Q931DialectSetIEProc(d, Q931ie_LOW_LAYER_COMPATIBILITY,          Q931Pie_LLComp,      Q931Uie_LLComp,       Q931Die_LLComp);
	Q931DialectSetIEProc(d, Q931ie_HIGH_LAYER_COMPATIBILITY,         Q931Pie_HLComp,      Q931Uie_HLComp,       Q931Die_HLComp);
	Q931DialectSetIEProc(d, Q931ie_USER_USER,                        Q931Pie_UserUser,    Q931Uie_UserUser,     Q931Die_UserUser);
	Q931DialectSetIEProc(d, Q931ie_SENDING_COMPLETE,                 Q931Pie_SendComplete,Q931Uie_SendComplete, Q931Die_SendComplete);
	Q931DialectSetIEProc(d, Q931ie_CONNECTED_NUMBER,                 Q931Pie_Generic,     Q931Uie_Generic,      Q931Die_Generic);
	Q931DialectSetIEProc(d, Q931ie_CONNECTED_SUBADDRESS,             Q931Pie_Generic,     Q931Uie_Generic,      Q931Die_Generic);

	/*
	 * Q.932
	 */
	Q931DialectSetIEProc(d, Q932ie_FACILITY,                       Q931Pie_Generic,     Q931Uie_Generic,    Q931Die_Generic);
	Q931DialectSetIEProc(d, Q932ie_EXTENDED_FACILITY,              Q931Pie_Generic,     Q931Uie_Generic,    Q931Die_Generic);
	Q931DialectSetIEProc(d, Q932ie_FEATURE_ACTIVATION,             Q931Pie_Generic,     Q931Uie_Generic,    Q931Die_Generic);
	Q931DialectSetIEProc(d, Q932ie_FEATURE_INDICATION,             Q931Pie_Generic,     Q931Uie_Generic,    Q931Die_Generic);
	Q931DialectSetIEProc(d, Q932ie_SERVICE_PROFILE_IDENTIFICATION, Q931Pie_Generic,     Q931Uie_Generic,    Q931Die_Generic);
	Q931DialectSetIEProc(d, Q932ie_ENDPOINT_IDENTIFIER,            Q931Pie_Generic,     Q931Uie_Generic,    Q931Die_Generic);

	/*
	 * Q.952
	 */
	Q931DialectSetIEProc(d, Q931ie_REDIRECTING_NUMBER,             Q931Pie_Generic,     Q931Uie_Generic,    Q931Die_Generic);
	Q931DialectSetIEProc(d, Q931ie_REDIRECTION_NUMBER,             Q931Pie_Generic,     Q931Uie_Generic,    Q931Die_Generic);

	/*
	 * State -> Allowed message table
	 */
	/* Any State */
	Q931DialectAddStateEntry(d, Q931_UANY, Q931mes_STATUS,         MSGF_FROM_BOTH);
	Q931DialectAddStateEntry(d, Q931_UANY, Q931mes_STATUS_ENQUIRY, MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_UANY, Q931mes_RESTART,             MSGF_FROM_BOTH);
	Q931DialectAddStateEntry(d, Q931_UANY, Q931mes_RESTART_ACKNOWLEDGE, MSGF_FROM_BOTH);

	/* State 0 Idle */
	Q931DialectAddStateEntry(d, Q931_U0, Q931mes_SETUP,            MSGF_FROM_BOTH);
	Q931DialectAddStateEntry(d, Q931_U0, Q931mes_RELEASE,          MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U0, Q931mes_RELEASE_COMPLETE, MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U0, Q931mes_RESUME,           MSGF_FROM_L4);

	/* State 1 Call Initiating */
	Q931DialectAddStateEntry(d, Q931_U1, Q931mes_DISCONNECT,        MSGF_FROM_L4);
	Q931DialectAddStateEntry(d, Q931_U1, Q931mes_SETUP_ACKNOWLEDGE, MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U1, Q931mes_RELEASE_COMPLETE,  MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U1, Q931mes_CALL_PROCEEDING,   MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U1, Q931mes_ALERTING,          MSGF_FROM_L2);	/* ITU-T Q.931 Annex D */
	Q931DialectAddStateEntry(d, Q931_U1, Q931mes_CONNECT,           MSGF_FROM_L2);	/* ITU-T Q.931 Annex D */

	/* State 2 Overlap Sending */
	Q931DialectAddStateEntry(d, Q931_U2, Q931mes_DISCONNECT,       MSGF_FROM_BOTH);
	Q931DialectAddStateEntry(d, Q931_U2, Q931mes_INFORMATION,      MSGF_FROM_BOTH);
	Q931DialectAddStateEntry(d, Q931_U2, Q931mes_CALL_PROCEEDING,  MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U2, Q931mes_ALERTING,         MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U2, Q931mes_PROGRESS,         MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U2, Q931mes_CONNECT,          MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U2, Q931mes_RELEASE_COMPLETE, MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U2, Q931mes_RELEASE,          MSGF_FROM_BOTH);

	/* State 3 Outgoing Call Proceeding */
	Q931DialectAddStateEntry(d, Q931_U3, Q931mes_DISCONNECT,       MSGF_FROM_BOTH);
	Q931DialectAddStateEntry(d, Q931_U3, Q931mes_PROGRESS,         MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U3, Q931mes_ALERTING,         MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U3, Q931mes_CONNECT,          MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U3, Q931mes_RELEASE_COMPLETE, MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U3, Q931mes_RELEASE,          MSGF_FROM_BOTH);
	Q931DialectAddStateEntry(d, Q931_U3, Q931mes_INFORMATION,      MSGF_FROM_BOTH);

	/* State 4 Call Delivered */
	Q931DialectAddStateEntry(d, Q931_U4, Q931mes_RELEASE_COMPLETE, MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U4, Q931mes_RELEASE,          MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U4, Q931mes_DISCONNECT,       MSGF_FROM_BOTH);
	Q931DialectAddStateEntry(d, Q931_U4, Q931mes_CONNECT,          MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U4, Q931mes_INFORMATION,      MSGF_FROM_BOTH);

	/* State 6 Call Precent */
	Q931DialectAddStateEntry(d, Q931_U6, Q931mes_INFORMATION,      MSGF_FROM_L4);
	Q931DialectAddStateEntry(d, Q931_U6, Q931mes_ALERTING,         MSGF_FROM_L4);
	Q931DialectAddStateEntry(d, Q931_U6, Q931mes_CALL_PROCEEDING,  MSGF_FROM_L4);
	Q931DialectAddStateEntry(d, Q931_U6, Q931mes_CONNECT,          MSGF_FROM_L4);
	Q931DialectAddStateEntry(d, Q931_U6, Q931mes_RELEASE_COMPLETE, MSGF_FROM_BOTH);
	Q931DialectAddStateEntry(d, Q931_U6, Q931mes_RELEASE,          MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U6, Q931mes_DISCONNECT,       MSGF_FROM_L2);

	/* State 7 Call Received */
	Q931DialectAddStateEntry(d, Q931_U7, Q931mes_RELEASE_COMPLETE, MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U7, Q931mes_RELEASE,          MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U7, Q931mes_DISCONNECT,       MSGF_FROM_BOTH);
	Q931DialectAddStateEntry(d, Q931_U7, Q931mes_CONNECT,          MSGF_FROM_L4);
	Q931DialectAddStateEntry(d, Q931_U7, Q931mes_INFORMATION,      MSGF_FROM_BOTH);

	/* State 8 Connect request */
	Q931DialectAddStateEntry(d, Q931_U8, Q931mes_RELEASE_COMPLETE,    MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U8, Q931mes_RELEASE,             MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U8, Q931mes_DISCONNECT,          MSGF_FROM_BOTH);
	Q931DialectAddStateEntry(d, Q931_U8, Q931mes_CONNECT_ACKNOWLEDGE, MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U8, Q931mes_INFORMATION,         MSGF_FROM_BOTH);

	/* State 9 Incoming Call Proceeding */
	Q931DialectAddStateEntry(d, Q931_U9, Q931mes_RELEASE_COMPLETE,  MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U9, Q931mes_RELEASE,           MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U9, Q931mes_DISCONNECT,        MSGF_FROM_BOTH);
	Q931DialectAddStateEntry(d, Q931_U9, Q931mes_CONNECT,           MSGF_FROM_L4);
	Q931DialectAddStateEntry(d, Q931_U9, Q931mes_ALERTING,          MSGF_FROM_L4);
	Q931DialectAddStateEntry(d, Q931_U9, Q931mes_PROGRESS,          MSGF_FROM_L4);
	Q931DialectAddStateEntry(d, Q931_U9, Q931mes_INFORMATION,       MSGF_FROM_BOTH);

	/* State 10 Active */
	Q931DialectAddStateEntry(d, Q931_U10, Q931mes_RELEASE_COMPLETE, MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U10, Q931mes_RELEASE,          MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U10, Q931mes_DISCONNECT,       MSGF_FROM_BOTH);
	Q931DialectAddStateEntry(d, Q931_U10, Q931mes_SUSPEND,          MSGF_FROM_L4);
	Q931DialectAddStateEntry(d, Q931_U10, Q931mes_NOTIFY,           MSGF_FROM_BOTH);
	Q931DialectAddStateEntry(d, Q931_U10, Q931mes_INFORMATION,      MSGF_FROM_BOTH);
	Q931DialectAddStateEntry(d, Q931_U10, Q932mes_FACILITY,         MSGF_FROM_BOTH);

	/* State 11 Disconnect Request */
	Q931DialectAddStateEntry(d, Q931_U11, Q931mes_RELEASE_COMPLETE, MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U11, Q931mes_RELEASE,          MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U11, Q931mes_DISCONNECT,       MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U11, Q931mes_NOTIFY,           MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U11, Q931mes_INFORMATION,      MSGF_FROM_BOTH);

	/* State 12 Disconnect Ind */
	Q931DialectAddStateEntry(d, Q931_U12, Q931mes_RELEASE_COMPLETE, MSGF_FROM_BOTH);
	Q931DialectAddStateEntry(d, Q931_U12, Q931mes_RELEASE,          MSGF_FROM_BOTH);
	Q931DialectAddStateEntry(d, Q931_U12, Q931mes_INFORMATION,      MSGF_FROM_BOTH);

	/* State 15 Suspend Request */
	Q931DialectAddStateEntry(d, Q931_U15, Q931mes_RELEASE_COMPLETE,    MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U15, Q931mes_SUSPEND_ACKNOWLEDGE, MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U15, Q931mes_SUSPEND_REJECT,      MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U15, Q931mes_DISCONNECT,          MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U15, Q931mes_RELEASE,             MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U15, Q931mes_INFORMATION,         MSGF_FROM_BOTH);

	/* State 17 Resume Request */
	Q931DialectAddStateEntry(d, Q931_U17, Q931mes_RELEASE_COMPLETE,   MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U17, Q931mes_RESUME_ACKNOWLEDGE, MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U17, Q931mes_RESUME_REJECT,      MSGF_FROM_L2);

	/* State 19 Release Request */
	Q931DialectAddStateEntry(d, Q931_U19, Q931mes_RELEASE_COMPLETE, MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U19, Q931mes_RELEASE,          MSGF_FROM_L2);

	/* State 25 Overlap Receiving */
	Q931DialectAddStateEntry(d, Q931_U25, Q931mes_DISCONNECT,        MSGF_FROM_BOTH);
	Q931DialectAddStateEntry(d, Q931_U25, Q931mes_SETUP_ACKNOWLEDGE, MSGF_FROM_L4);
	Q931DialectAddStateEntry(d, Q931_U25, Q931mes_RELEASE_COMPLETE,  MSGF_FROM_BOTH);
	Q931DialectAddStateEntry(d, Q931_U25, Q931mes_RELEASE,           MSGF_FROM_L2);
	Q931DialectAddStateEntry(d, Q931_U25, Q931mes_CALL_PROCEEDING,   MSGF_FROM_L4);
	Q931DialectAddStateEntry(d, Q931_U25, Q931mes_ALERTING,          MSGF_FROM_L4);
	Q931DialectAddStateEntry(d, Q931_U25, Q931mes_CONNECT,           MSGF_FROM_L4);
	Q931DialectAddStateEntry(d, Q931_U25, Q931mes_INFORMATION,       MSGF_FROM_L2);

	/*
	 * Timer default values & callbacks
	 */
	Q931DialectSetTimerProcAll(d, Q931TimeoutDummy);

	/* Per-timer timeout callbacks, per-mode for now, until i know if they can be shared (and which) */
/*
	Q931DialectSetTimerProc(d, Q931_TIMER_T301, Q931ProcTimeoutT301TE);
	Q931DialectSetTimerProc(d, Q931_TIMER_T302, Q931ProcTimeoutT302TE);
*/
	Q931DialectSetTimerProc(d, Q931_TIMER_T303, Q931ProcTimeoutT303TE);
	Q931DialectSetTimerProc(d, Q931_TIMER_T304, Q931ProcTimeoutT304TE);
	Q931DialectSetTimerProc(d, Q931_TIMER_T305, Q931ProcTimeoutT305TE);
	Q931DialectSetTimerProc(d, Q931_TIMER_T308, Q931ProcTimeoutT308TE);
/*	Q931DialectSetTimerProc(d, Q931_TIMER_T309, Q931ProcTimeoutT309TE); */
	Q931DialectSetTimerProc(d, Q931_TIMER_T310, Q931ProcTimeoutT310TE);
/*	Q931DialectSetTimerProc(d, Q931_TIMER_T311, Q931ProcTimeoutT311TE); */
	Q931DialectSetTimerProc(d, Q931_TIMER_T313, Q931ProcTimeoutT313TE);
/*
	Q931DialectSetTimerProc(d, Q931_TIMER_T314, Q931ProcTimeoutT314TE);
	Q931DialectSetTimerProc(d, Q931_TIMER_T316, Q931ProcTimeoutT316TE);
	Q931DialectSetTimerProc(d, Q931_TIMER_T317, Q931ProcTimeoutT317TE);
*/
	Q931DialectSetTimerProc(d, Q931_TIMER_T318, Q931ProcTimeoutT318TE);
/*
	Q931DialectSetTimerProc(d, Q931_TIMER_T319, Q931ProcTimeoutT319TE);
	Q931DialectSetTimerProc(d, Q931_TIMER_T321, Q931ProcTimeoutT321TE);
	Q931DialectSetTimerProc(d, Q931_TIMER_T322, Q931ProcTimeoutT322TE);
*/

	Q931DialectSetTimeout(d, Q931_TIMER_T301, 180000);	/* T301: 180s */
	Q931DialectSetTimeout(d, Q931_TIMER_T302,  15000);	/* T302:  15s */
	Q931DialectSetTimeout(d, Q931_TIMER_T303,   4000);	/* T303:   4s */
	Q931DialectSetTimeout(d, Q931_TIMER_T304,  30000);	/* T304:  30s */
	Q931DialectSetTimeout(d, Q931_TIMER_T305,  30000);	/* T305:  30s */
	Q931DialectSetTimeout(d, Q931_TIMER_T308,   4000);	/* T308:   4s */
	Q931DialectSetTimeout(d, Q931_TIMER_T309,  60000);	/* T309:  60s */
	Q931DialectSetTimeout(d, Q931_TIMER_T310,  60000);	/* T310:  60s */
	Q931DialectSetTimeout(d, Q931_TIMER_T313,   4000);	/* T313:   4s */
	Q931DialectSetTimeout(d, Q931_TIMER_T314,   4000);	/* T314:   4s */
	Q931DialectSetTimeout(d, Q931_TIMER_T316, 120000);	/* T316: 120s */
	Q931DialectSetTimeout(d, Q931_TIMER_T317,  90000);	/* T317:  90s */
	Q931DialectSetTimeout(d, Q931_TIMER_T318,   4000);	/* T318:   4s */
	Q931DialectSetTimeout(d, Q931_TIMER_T319,   4000);	/* T319:   4s */
	Q931DialectSetTimeout(d, Q931_TIMER_T321,  30000);	/* T321:  30s */
	Q931DialectSetTimeout(d, Q931_TIMER_T322,   4000);	/* T322:   4s */

	/*
	 * Q.931 message <-> ie table
	 */
	Q931DialectSetMesIEMap(d, Q931MessageIEs);
}


/**
 * @defgroup	q931_te_msg_proc	Q.931 TE Message handling
 * @{
 */

/**
 * Handle ALERTING message (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call handle
 * @param[in]	msg	Incoming message
 * @param[in]	from	Message coming from Layer 2 or 4?
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931ProcAlertingTE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV (TODO: move one level up,
	 * change function signature(s) to use Q931_Call * instead of
	 * Q931_TrunkInfo_t *)
	 */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state (TODO: move one level up) */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch(from) {
	case Q931_MSG_FROM_L4:
		switch(Q931CallGetState(call)) {
		case Q931_U6:
		case Q931_U9:
			/* Send ALERTING */
			ret = Q931Tx32(trunk, 0, msg, msg->Size);

			/* => U7 Call received */
			Q931CallSetState(call, Q931_U7);
			break;
		default:
			break;
		}
		break;
	case Q931_MSG_FROM_L2:
		switch(Q931CallGetState(call)) {
		case Q931_U1:
			/* Stop T303 */
			Q931CallStopTimer(call, Q931_TIMER_T303);

			/* TODO: Invoke new event CB: Alerting indication */
			ret = Q931Tx34(trunk, call, msg, msg->Size);

			{
				/* Enqueue event */
				struct Q931_CallEvent *event = Q931CallNewEvent(call);

				event->id   = Q931_EVENT_ALERTING_INDICATION;
				event->type = Q931_EVENT_TYPE_MESSAGE;
				event->data.message.type = msg->MesType;
				event->data.message.data = msg;

				Q931CallQueueEvent(call, event);
			}

			/* => U4: Call delivered */
			Q931CallSetState(call, Q931_U4);
			break;

		case Q931_U2:
			/* Stop T304 */
			Q931CallStopTimer(call, Q931_TIMER_T304);

			/* TODO: Invoke new event CB: Alerting indication */
			ret = Q931Tx34(trunk, call, msg, msg->Size);

			{
				/* Enqueue event */
				struct Q931_CallEvent *event = Q931CallNewEvent(call);

				event->id   = Q931_EVENT_ALERTING_INDICATION;
				event->type = Q931_EVENT_TYPE_MESSAGE;
				event->data.message.type = msg->MesType;
				event->data.message.data = msg;

				Q931CallQueueEvent(call, event);
			}

			/* => U4: Call delivered */
			Q931CallSetState(call, Q931_U4);
			break;

		case Q931_U3:
			/* Stop T310, Start T301 */
			Q931CallStartTimer(call, Q931_TIMER_T301);

			/* TODO: Invoke new event CB: Alerting indication */
			ret = Q931Tx34(trunk, call, msg, msg->Size);

			{
				/* Enqueue event */
				struct Q931_CallEvent *event = Q931CallNewEvent(call);

				event->id   = Q931_EVENT_ALERTING_INDICATION;
				event->type = Q931_EVENT_TYPE_MESSAGE;
				event->data.message.type = msg->MesType;
				event->data.message.data = msg;

				Q931CallQueueEvent(call, event);
			}

			/* => U4: Call delivered */
			Q931CallSetState(call, Q931_U4);
			break;

		default:
			break;
		}
		break;
	default:
		ret = Q931E_INTERNAL;
	}
	return ret;
}


/**
 * Handle PROCEEDING message (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call handle
 * @param[in]	msg	Incoming message
 * @param[in]	from	Message coming from Layer 2 or 4?
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931ProcCallProceedingTE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV (TODO: move one level up,
	 * change function signature(s) to use Q931_Call * instead of
	 * Q931_TrunkInfo_t *)
	 */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state (TODO: move one level up) */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch(from) {
	case Q931_MSG_FROM_L4:
		switch(Q931CallGetState(call)) {
		case Q931_U6:
			/* Send CALL PROCEEDING */
			ret = Q931Tx32(trunk, 0, msg, msg->Size);

			/* => U9 Incoming call proceeding */
			Q931CallSetState(call, Q931_U9);
			break;
		default:
			break;
		}
		break;
	case Q931_MSG_FROM_L2:
		switch(Q931CallGetState(call)) {
		case Q931_U1:
		case Q931_U2:
			/* Stop T303/T304, Start T310 */
			Q931CallStartTimer(call, Q931_TIMER_T310);

			/* TODO: Invoke new event CB: Proceeding indication */
			ret = Q931Tx34(trunk, call, msg, msg->Size);

			{
				/* Enqueue event */
				struct Q931_CallEvent *event = Q931CallNewEvent(call);

				event->id   = Q931_EVENT_PROCEEDING_INDICATION;
				event->type = Q931_EVENT_TYPE_MESSAGE;
				event->data.message.type = msg->MesType;
				event->data.message.data = msg;

				Q931CallQueueEvent(call, event);
			}

			/* => U3 Outgoing call proceeding */
			Q931CallSetState(call, Q931_U3);
			break;
		default:
			break;
		}
		break;
	default:
		ret = Q931E_INTERNAL;
	}
	return ret;
}


/**
 * Handle CONNECT message (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call handle
 * @param[in]	msg	Incoming message
 * @param[in]	from	Message coming from Layer 2 or 4?
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931ProcConnectTE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch(from) {
	case Q931_MSG_FROM_L4:
		switch(Q931CallGetState(call)) {
		case Q931_U6:	/* Setup response */
		case Q931_U7:
		case Q931_U9:
			/* Send CONNECT */
			ret = Q931Tx32(trunk, 0, msg, msg->Size);

			/* Start T313 */
			Q931CallStartTimer(call, Q931_TIMER_T313);

			/* => U8 Connect request */
			Q931CallSetState(call, Q931_U8);
			break;
		default:
			break;
		}
		break;
	case Q931_MSG_FROM_L2:
		switch(Q931CallGetState(call)) {
		case Q931_U1:
			/* Stop T303 */
			Q931CallStopTimer(call, Q931_TIMER_T303);

		case Q931_U2:
			/* Stop T304 */
			Q931CallStopTimer(call, Q931_TIMER_T304);

		case Q931_U3:
			/* Stop T310 */
			Q931CallStopTimer(call, Q931_TIMER_T310);

		case Q931_U4:
			/* Stop T301 (something missing in the spec huh?) */
			Q931CallStopTimer(call, Q931_TIMER_T301);

			/* TODO: Send Setup confirm */
			ret = Q931Tx34(trunk, call, msg, msg->Size);

			{
				/* Enqueue event */
				struct Q931_CallEvent *event = Q931CallNewEvent(call);

				event->id   = Q931_EVENT_SETUP_CONFIRM;
				event->type = Q931_EVENT_TYPE_MESSAGE;
				event->data.message.type = msg->MesType;
				event->data.message.data = msg;

				Q931CallQueueEvent(call, event);
			}

			if (Q931TrunkIsSetFlag(trunk, Q931_TFLAG_AUTO_CONNECT_ACK)) {
				Q931AckConnect(call);
			}

			/* => U10 Active */
			Q931CallSetState(call, Q931_U10);
			break;

		default:
			break;
		}
		break;
	default:
		ret = Q931E_INTERNAL;
	}
	return ret;
}


/**
 * Handle CONNECT ACKNOWLEDGE message (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call handle
 * @param[in]	msg	Incoming message
 * @param[in]	from	Message coming from Layer 2 or 4?
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931ProcConnectAckTE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch(from) {
#if 0
	case Q931_MSG_FROM_L4:
		/* TODO Add proc here */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);
		break;
#endif
	case Q931_MSG_FROM_L2:
		switch(Q931CallGetState(call)) {
		case Q931_U8:
			/* Stop T313 */
			Q931CallStopTimer(call, Q931_TIMER_T313);

			/* TODO: Setup complete indication */
			ret = Q931Tx34(trunk, call, msg, msg->Size);

			{
				/* Enqueue event */
				struct Q931_CallEvent *event = Q931CallNewEvent(call);

				event->id   = Q931_EVENT_SETUP_COMPLETE_INDICATION;
				event->type = Q931_EVENT_TYPE_MESSAGE;
				event->data.message.type = msg->MesType;
				event->data.message.data = msg;

				Q931CallQueueEvent(call, event);
			}

			/* => U10 Active */
			Q931CallSetState(call, Q931_U10);
			break;
		default:
			break;
		}
		break;
	default:
		ret = Q931E_INTERNAL;
	}
	return ret;
}


/**
 * Handle PROGRESS message (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call handle
 * @param[in]	msg	Incoming message
 * @param[in]	from	Message coming from Layer 2 or 4?
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931ProcProgressTE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch(from) {
	case Q931_MSG_FROM_L4:
		switch(Q931CallGetState(call)) {
		case Q931_U9:
			/* Send PROGRESS */
			ret = Q931Tx32(trunk, 0, msg, msg->Size);

			/* == U9 Incoming call proceeding */
			break;
		default:
			break;
		}
		break;
	case Q931_MSG_FROM_L2:
		switch(Q931CallGetState(call)) {
		case Q931_U2:
			/* Stop T304 (?) */
			Q931CallStopTimer(call, Q931_TIMER_T304);

			/* Send proceeding indication */
			ret = Q931Tx34(trunk, call, msg, msg->Size);

			{
				/* Enqueue event */
				struct Q931_CallEvent *event = Q931CallNewEvent(call);

				event->id   = Q931_EVENT_PROCEEDING_INDICATION;
				event->type = Q931_EVENT_TYPE_MESSAGE;
				event->data.message.type = msg->MesType;
				event->data.message.data = msg;

				Q931CallQueueEvent(call, event);
			}

			/* == U2 Overlap sending */
			break;

		case Q931_U3:
			/* Stop T310 */
			Q931CallStopTimer(call, Q931_TIMER_T310);

			/* Send proceeding indication */
			ret = Q931Tx34(trunk, call, msg, msg->Size);

			{
				/* Enqueue event */
				struct Q931_CallEvent *event = Q931CallNewEvent(call);

				event->id   = Q931_EVENT_PROCEEDING_INDICATION;
				event->type = Q931_EVENT_TYPE_MESSAGE;
				event->data.message.type = msg->MesType;
				event->data.message.data = msg;

				Q931CallQueueEvent(call, event);
			}

			/* == U3 Outgoing call proceeding */
			break;
		default:
			break;
		}
		break;
	default:
		ret = Q931E_INTERNAL;
	}
	return ret;
}


/**
 * Handle SETUP message (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call handle
 * @param[in]	msg	Incoming message
 * @param[in]	from	Message coming from Layer 2 or 4?
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931ProcSetupTE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#ifdef __UNUSED__
	/* NOTE: this is handled by upper layers */

	/* Find the call using CRV */
	if (msg->CRV) {
		call = Q931GetCallByCRV(trunk, msg->CRV);
		if (call && Q931CallGetState(call) != Q931_U0) {
			/* Reject SETUP on existing calls */
			Q931Disconnect(trunk, from, msg->CRV, 81);
			return Q931E_UNEXPECTED_MESSAGE;
		}
	}

	if (call) {
		/* check if message is valid in this state */
		if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
			return Q931E_UNEXPECTED_MESSAGE;
	}
#endif
	switch(from) {
	case Q931_MSG_FROM_L4:		/* outgoing call */
		if (!call) {
			call = Q931CallNew(trunk);
			if (!call) {
				return Q931E_INTERNAL;
			}
			msg->CRV = Q931CallGetCRV(call);
		}

		Q931Log(trunk, Q931_LOG_INFO, "Creating new outbound call with CRV: %d\n", msg->CRV);

		ret = Q931Tx32(trunk, 0, msg, msg->Size);
		if (ret < Q931E_NO_ERROR) {
			Q931Log(trunk, Q931_LOG_INFO, "ProcSETUP Q931Tx32 retval: %d\n", ret);
			return ret;
		}

		Q931CallStartTimer(call, Q931_TIMER_T303);

		Q931CallSetState(call, Q931_U1);
		break;

	case Q931_MSG_FROM_L2:		/* incoming call */
		if (!call) {
			call = Q931CallNewIncoming(trunk, msg->CRV);
			if (!call) {
				/* Not possible to allocate CRV entry, so must reject call */
				Q931Disconnect(trunk, from, msg->CRV, 42);
				return Q931E_INTERNAL;
			}
		}

		/* Send setup indication to user */
		{
			/* Enqueue event */
			struct Q931_CallEvent *event = Q931CallNewEvent(call);

			event->id   = Q931_EVENT_SETUP_INDICATION;
			event->type = Q931_EVENT_TYPE_MESSAGE;
			event->data.message.type = msg->MesType;
			event->data.message.data = msg;

			Q931CallQueueEvent(call, event);
		}

		/* Set state U6 */
		Q931CallSetState(call, Q931_U6);

		/* old way */
		ret = Q931Tx34(trunk, call, msg, msg->Size);
		if (ret >= Q931E_NO_ERROR) {
			if (Q931TrunkIsSetFlag(trunk, Q931_TFLAG_AUTO_SETUP_ACK)) {
				Q931AckSetup(trunk, msg);
			}
			return ret;
		} else {
			/* Must be full queue, meaning we can't process the call */
			/* so we must disconnect */
			Q931Disconnect(trunk, from, msg->CRV, 81);
			return ret;
		}
		break;

	default:
		ret = Q931E_INTERNAL;
	}
	return ret;
}


/**
 * Handle SETUP ACKNOWLEDGE message (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call handle
 * @param[in]	msg	Incoming message
 * @param[in]	from	Message coming from Layer 2 or 4?
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931ProcSetupAckTE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch(from) {
	case Q931_MSG_FROM_L4:
		switch(Q931CallGetState(call)) {
		case Q931_U6:	/* More info request */
			/* Send SETUP ACK */
			ret = Q931Tx32(trunk, 0, msg, msg->Size);

			/* Start T302 */
			Q931CallStartTimer(call, Q931_TIMER_T302);

			/* => U25 Overlap receiving */
			Q931CallSetState(call, Q931_U25);
			break;
		default:
			break;
		}
		break;
	case Q931_MSG_FROM_L2:
		switch(Q931CallGetState(call)) {
		case Q931_U1:
			/* Stop T303 */
			Q931CallStopTimer(call, Q931_TIMER_T303);

			/* TODO: More information indication */
			{
				/* Enqueue event */
				struct Q931_CallEvent *event = Q931CallNewEvent(call);

				event->id   = Q931_EVENT_MORE_INFO_INDICATION;
				event->type = Q931_EVENT_TYPE_MESSAGE;
				event->data.message.type = msg->MesType;
				event->data.message.data = msg;

				Q931CallQueueEvent(call, event);
			}

			/* Stop T303, Start T304 */
			Q931CallStartTimer(call, Q931_TIMER_T304);

			/* => U2 Overlap sending */
			Q931CallSetState(call, Q931_U2);
			break;
		default:
			break;
		}
		break;
	default:
		ret = Q931E_INTERNAL;
	}
	return ret;
}


/**
 * Handle RESUME message (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call handle
 * @param[in]	msg	Incoming message
 * @param[in]	from	Message coming from Layer 2 or 4?
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931ProcResumeTE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;

	switch(from) {
	case Q931_MSG_FROM_L4:
		/* Find the call using CRV */
		call = Q931GetCallByCRV(trunk, msg->CRV);
		if (!call) {
			call = Q931CallNew(trunk);
			if (!call) {
				return Q931E_INTERNAL;
			}

			/* Send RESUME to network */
			ret = Q931Tx32(trunk, 0, msg, msg->Size);
			if (ret != Q931E_NO_ERROR)
				return ret;

			/* Start timer T318 */
			Q931CallStartTimer(call, Q931_TIMER_T318);

			/* set state U17 */
			Q931CallSetState(call, Q931_U17);
		} else {
			return Q931E_ILLEGAL_MESSAGE;
		}
		break;

	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}


/**
 * Handle RESUME ACKNOWLEDGE message (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call handle
 * @param[in]	msg	Incoming message
 * @param[in]	from	Message coming from Layer 2 or 4?
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931ProcResumeAckTE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch(from) {
	case Q931_MSG_FROM_L4:
		/* TODO Add proc here */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);
		break;
	case Q931_MSG_FROM_L2:
		/* Stop T318 */
		Q931CallStopTimer(call, Q931_TIMER_T318);

		/* => State U10 (Active) */
		Q931CallSetState(call, Q931_U10);

		/* TODO: Send Resume confirm */
		{
			/* Enqueue event */
			struct Q931_CallEvent *event = Q931CallNewEvent(call);

			event->id   = Q931_EVENT_RESUME_CONFIRM;
			event->type = Q931_EVENT_TYPE_MESSAGE;
			event->data.message.type = msg->MesType;
			event->data.message.data = msg;

			Q931CallQueueEvent(call, event);
		}

		ret = Q931Tx34(trunk, call, msg, msg->Size);
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}


/**
 * Handle RESUME REJECT message (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call handle
 * @param[in]	msg	Incoming message
 * @param[in]	from	Message coming from Layer 2 or 4?
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931ProcResumeRejectTE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch(from) {
	case Q931_MSG_FROM_L4:
		/* TODO Add proc here */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);
		break;
	case Q931_MSG_FROM_L2:
		/* Stop T318 */
		Q931CallStopTimer(call, Q931_TIMER_T318);

		/* TODO: send resume confirm error */
		ret = Q931Tx34(trunk, call, msg, msg->Size);

		{
			/* Enqueue event */
			struct Q931_CallEvent *event = Q931CallNewEvent(call);

			event->id    = Q931_EVENT_RESUME_CONFIRM;
			event->type  = Q931_EVENT_TYPE_MESSAGE;
			event->error = 1;
			event->data.message.type = msg->MesType;
			event->data.message.data = msg;

			Q931CallQueueEvent(call, event);
		}

		/* TODO: release CRV */
		Q931CallSetState(call, Q931_U0);
		Q931CallRelease(call);
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}


/**
 * Handle SUSPEND message (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call handle
 * @param[in]	msg	Incoming message
 * @param[in]	from	Message coming from Layer 2 or 4?
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931ProcSuspendTE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch(from) {
	case Q931_MSG_FROM_L4:
		/* Start T319 */
		Q931CallStartTimer(call, Q931_TIMER_T319);

		/* Send message */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);

		/* => State U15 (Suspend Request) */
		Q931CallSetState(call, Q931_U15);
		break;
#if 0
	case Q931_MSG_FROM_L2:
		/* TODO Add proc here */
		ret = Q931Tx34(trunk, call, msg, msg->Size);
		break;
#endif
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}


/**
 * Handle SUSPEND ACKNOWLEDGE message (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call handle
 * @param[in]	msg	Incoming message
 * @param[in]	from	Message coming from Layer 2 or 4?
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931ProcSuspendAckTE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch(from) {
#if 0
	case Q931_MSG_FROM_L4:
		/* TODO Add proc here */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);
		break;
#endif
	case Q931_MSG_FROM_L2:
		/* Stop T319 */
		Q931CallStopTimer(call, Q931_TIMER_T319);

		/* TODO: send Suspend confirm */
		ret = Q931Tx34(trunk, call, msg, msg->Size);

		{
			/* Enqueue event */
			struct Q931_CallEvent *event = Q931CallNewEvent(call);

			event->id   = Q931_EVENT_SUSPEND_CONFIRM;
			event->type = Q931_EVENT_TYPE_MESSAGE;
			event->data.message.type = msg->MesType;
			event->data.message.data = msg;

			Q931CallQueueEvent(call, event);
		}

		/* Release CRV */
		Q931CallSetState(call, Q931_U0);
		Q931CallRelease(call);
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}


/**
 * Handle SUSPEND REJECT message (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call handle
 * @param[in]	msg	Incoming message
 * @param[in]	from	Message coming from Layer 2 or 4?
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931ProcSuspendRejectTE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch(from) {
#if 0
	case Q931_MSG_FROM_L4:
		/* TODO Add proc here */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);
		break;
#endif
	case Q931_MSG_FROM_L2:
		/* Stop T319 */
		Q931CallStopTimer(call, Q931_TIMER_T319);

		/* TODO: Send Suspend confirm error */
		ret = Q931Tx34(trunk, call, msg, msg->Size);

		{
			/* Enqueue event */
			struct Q931_CallEvent *event = Q931CallNewEvent(call);

			event->id    = Q931_EVENT_SUSPEND_CONFIRM;
			event->type  = Q931_EVENT_TYPE_MESSAGE;
			event->error = 1;
			event->data.message.type = msg->MesType;
			event->data.message.data = msg;

			Q931CallQueueEvent(call, event);
		}

		/* => State U10 (Active) */
		Q931CallSetState(call, Q931_U10);
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}


/**
 * Handle USER INFORMATION message (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call handle
 * @param[in]	msg	Incoming message
 * @param[in]	from	Message coming from Layer 2 or 4?
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931ProcUserInformationTE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;

	/* TODO: ????? */
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch(from) {
	case Q931_MSG_FROM_L4:
		switch (Q931CallGetState(call)) {
		case Q931_U0:
		case Q931_U1:
		case Q931_U6:
		case Q931_U17:
		case Q931_U19:
			ret = Q931E_ILLEGAL_MESSAGE;
			break;

		case Q931_U2:
			/* Send INFORMATION */
			ret = Q931Tx32(trunk, 0, msg, msg->Size);

			/* Start / Restart T304 */
			Q931CallStartTimer(call, Q931_TIMER_T304);

			/* == U2 (Overlap sending) */
			break;

		default:
			/* Send INFORMATION */
			ret = Q931Tx32(trunk, 0, msg, msg->Size);
			break;
		}
		break;
	case Q931_MSG_FROM_L2:
		switch (Q931CallGetState(call)) {
		case Q931_U0:
		case Q931_U1:
		case Q931_U6:
		case Q931_U17:
		case Q931_U19:
			ret = Q931E_ILLEGAL_MESSAGE;
			break;

		case Q931_U25:
			/* TODO: send Info indication */
			ret = Q931Tx34(trunk, call, msg, msg->Size);

			{
				/* Enqueue event */
				struct Q931_CallEvent *event = Q931CallNewEvent(call);

				event->id   = Q931_EVENT_INFORMATION_INDICATION;
				event->type = Q931_EVENT_TYPE_MESSAGE;
				event->data.message.type = msg->MesType;
				event->data.message.data = msg;

				Q931CallQueueEvent(call, event);
			}

			/* Start T302 */
			Q931CallStartTimer(call, Q931_TIMER_T302);

			/* == U25 (Overlap receiving) */
			break;

		default:
			/* TODO: send Info indication */
			ret = Q931Tx34(trunk, call, msg, msg->Size);

			{
				/* Enqueue event */
				struct Q931_CallEvent *event = Q931CallNewEvent(call);

				event->id   = Q931_EVENT_INFORMATION_INDICATION;
				event->type = Q931_EVENT_TYPE_MESSAGE;
				event->data.message.type = msg->MesType;
				event->data.message.data = msg;

				Q931CallQueueEvent(call, event);
			}
			break;
		}
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}


/**
 * Handle DISCONNECT message (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call handle
 * @param[in]	msg	Incoming message
 * @param[in]	from	Message coming from Layer 2 or 4?
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931ProcDisconnectTE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;

	Q931Log(trunk, Q931_LOG_DEBUG, "Processing DISCONNECT message from %s for CRV: %d (%#hx)\n",
						 from == Q931_MSG_FROM_L4 ? "Local" : "Remote", msg->CRV, msg->CRV);
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch(from) {
	case Q931_MSG_FROM_L4:
		switch(Q931CallGetState(call)) {
		case Q931_U0:
		case Q931_U6:
		case Q931_U11:
		case Q931_U12:
		case Q931_U15:
		case Q931_U17:
		case Q931_U19:
			/* TODO: should never occur? */
			break;

		case Q931_U1:
			/* Disconnect */
			ret = Q931Tx32(trunk, 0, msg, msg->Size);

			/* Stop T303, Start T305 */
			Q931CallStartTimer(call, Q931_TIMER_T305);

			/* => U11 Disconnect request */
			Q931CallSetState(call, Q931_U11);
			break;

		default:
			/* Stop all timers */
			Q931CallStopAllTimers(call);

			/* Disconnect */
			ret = Q931Tx32(trunk, 0, msg, msg->Size);

			/* Start T305 */
			Q931CallStartTimer(call, Q931_TIMER_T305);

			/* => U11 Disconnect request */
			Q931CallSetState(call, Q931_U11);
		}
		break;
	case Q931_MSG_FROM_L2:
		switch(Q931CallGetState(call)) {
		case Q931_U0:
		case Q931_U1:
		case Q931_U12:
		case Q931_U17:
		case Q931_U19:
			/* TODO: should never occur? */
			break;

		case Q931_U6:
			/* TODO: Send Disconnect indication */
			ret = Q931Tx34(trunk, call, msg, msg->Size);

			{
				/* Enqueue event */
				struct Q931_CallEvent *event = Q931CallNewEvent(call);

				event->id   = Q931_EVENT_DISCONNECT_INDICATION;
				event->type = Q931_EVENT_TYPE_MESSAGE;
				event->data.message.type = msg->MesType;
				event->data.message.data = msg;

				Q931CallQueueEvent(call, event);
			}

			/* => U12 Disconnect indication */
			Q931CallSetState(call, Q931_U12);
			break;

		case Q931_U11:
			/* Stop T305 */
			Q931CallStopTimer(call, Q931_TIMER_T305);

			/* TODO: Send Release reply */
			Q931Log(trunk, Q931_LOG_ERROR, "OOOPS!! We should really send a RELEASE request here!\n");

			/* Start T308 */
			Q931CallStartTimer(call, Q931_TIMER_T308);

			/* => U19 Release request */
			Q931CallSetState(call, Q931_U19);
			break;

		case Q931_U15:
			/* Stop T319... */
			Q931CallStopAllTimers(call);

			/* TODO: Send Disconnect indication */
			ret = Q931Tx34(trunk, call, msg, msg->Size);

			{
				/* Enqueue event */
				struct Q931_CallEvent *event = Q931CallNewEvent(call);

				event->id   = Q931_EVENT_DISCONNECT_INDICATION;
				event->type = Q931_EVENT_TYPE_MESSAGE;
				event->data.message.type = msg->MesType;
				event->data.message.data = msg;

				Q931CallQueueEvent(call, event);
			}

			/* => U12 Disconnect indication */
			Q931CallSetState(call, Q931_U12);
			break;

		default:
			/* TODO: Send Disconnect indication */
			ret = Q931Tx34(trunk, call, msg, msg->Size);

			{
				/* Enqueue event */
				struct Q931_CallEvent *event = Q931CallNewEvent(call);

				event->id   = Q931_EVENT_DISCONNECT_INDICATION;
				event->type = Q931_EVENT_TYPE_MESSAGE;
				event->data.message.type = msg->MesType;
				event->data.message.data = msg;

				Q931CallQueueEvent(call, event);
			}

			/* Stop all timers */
			Q931CallStopAllTimers(call);

			/* => U12 Disconnect indication */
			Q931CallSetState(call, Q931_U12);
		}
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}


/**
 * Handle RELEASE message (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call handle
 * @param[in]	msg	Incoming message
 * @param[in]	from	Message coming from Layer 2 or 4?
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931ProcReleaseTE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch(from) {
	case Q931_MSG_FROM_L4:
		switch(Q931CallGetState(call)) {
		case Q931_U2:
			/* Send RELEASE (cause 6) */
			ret = Q931Tx32(trunk, 0, msg, msg->Size);

			/* Stop T304, Start T308 */
			Q931CallStartTimer(call, Q931_TIMER_T308);

			/* => U19 Release Request */
			Q931CallSetState(call, Q931_U19);
			break;

		case Q931_U3:
			/* Stop T310 */
			Q931CallStopTimer(call, Q931_TIMER_T310);

			/* Send RELEASE (cause 6) */
			ret = Q931Tx32(trunk, 0, msg, msg->Size);

			/* Stop Start T308 */
			Q931CallStartTimer(call, Q931_TIMER_T308);

			/* => U19 Release Request */
			Q931CallSetState(call, Q931_U19);
			break;

		case Q931_U12:
			/* Send RELEASE */
			ret = Q931Tx32(trunk, 0, msg, msg->Size);

			/* Stop T304, Start T308 */
			Q931CallStartTimer(call, Q931_TIMER_T308);

			/* => U19 Release Request */
			Q931CallSetState(call, Q931_U19);
			break;

		default:
			break;
		}
		break;
	case Q931_MSG_FROM_L2:
		switch(Q931CallGetState(call)) {
		case Q931_U1:
		case Q931_U17:
			/* Should never happen? */
			break;

		case Q931_U0:
			/* Release indication */
			Q931Tx34(trunk, call, msg, msg->Size);

			{
				/* Enqueue event */
				struct Q931_CallEvent *event = Q931CallNewEvent(call);

				event->id   = Q931_EVENT_RELEASE_INDICATION;
				event->type = Q931_EVENT_TYPE_MESSAGE;
				event->data.message.type = msg->MesType;
				event->data.message.data = msg;

				Q931CallQueueEvent(call, event);
			}

			/* Send RELEASE COMPLETE reply */
			ret = Q931ReleaseComplete(trunk, call, 0);
			break;

		case Q931_U6:
		case Q931_U12:
			/* Release indication */
			Q931Tx34(trunk, call, msg, msg->Size);

			{
				/* Enqueue event */
				struct Q931_CallEvent *event = Q931CallNewEvent(call);

				event->id   = Q931_EVENT_RELEASE_INDICATION;
				event->type = Q931_EVENT_TYPE_MESSAGE;
				event->data.message.type = msg->MesType;
				event->data.message.data = msg;

				Q931CallQueueEvent(call, event);
			}

			/* Send RELEASE COMPLETE reply */
			ret = Q931ReleaseComplete(trunk, call, 0);
			break;

		case Q931_U11:
			/* Stop T305 */
			Q931CallStopTimer(call, Q931_TIMER_T305);

			/* Release indication */
			Q931Tx34(trunk, call, msg, msg->Size);

			{
				/* Enqueue event */
				struct Q931_CallEvent *event = Q931CallNewEvent(call);

				event->id   = Q931_EVENT_RELEASE_INDICATION;
				event->type = Q931_EVENT_TYPE_MESSAGE;
				event->data.message.type = msg->MesType;
				event->data.message.data = msg;

				Q931CallQueueEvent(call, event);
			}

			/* Send RELEASE COMPLETE reply */
			ret = Q931ReleaseComplete(trunk, call, 0);
			break;

		case Q931_U15:
			/* Stop T319 */
			Q931CallStopTimer(call, Q931_TIMER_T319);

			/* Release indication */
			Q931Tx34(trunk, call, msg, msg->Size);

			{
				/* Enqueue event */
				struct Q931_CallEvent *event = Q931CallNewEvent(call);

				event->id   = Q931_EVENT_RELEASE_INDICATION;
				event->type = Q931_EVENT_TYPE_MESSAGE;
				event->data.message.type = msg->MesType;
				event->data.message.data = msg;

				Q931CallQueueEvent(call, event);
			}

			/* Send RELEASE COMPLETE reply */
			ret = Q931ReleaseComplete(trunk, call, 0);
			break;

		case Q931_U19:
			/* Stop T308 */
			Q931CallStopTimer(call, Q931_TIMER_T308);

			/* TODO: Send release confirm */
			Q931Tx34(trunk, call, msg, msg->Size);

			{
				/* Enqueue event */
				struct Q931_CallEvent *event = Q931CallNewEvent(call);

				event->id   = Q931_EVENT_RELEASE_CONFIRM;
				event->type = Q931_EVENT_TYPE_MESSAGE;
				event->data.message.type = msg->MesType;
				event->data.message.data = msg;

				Q931CallQueueEvent(call, event);
			}
			break;

		default:
			/* Stop all timers */
			Q931CallStopAllTimers(call);

			/* TODO: Send release indication */
			Q931Tx34(trunk, call, msg, msg->Size);

			{
				/* Enqueue event */
				struct Q931_CallEvent *event = Q931CallNewEvent(call);

				event->id   = Q931_EVENT_RELEASE_INDICATION;
				event->type = Q931_EVENT_TYPE_MESSAGE;
				event->data.message.type = msg->MesType;
				event->data.message.data = msg;

				Q931CallQueueEvent(call, event);
			}

			/* Send RELEASE COMPLETE reply */
			ret = Q931ReleaseComplete(trunk, call, 0);
		}

		/* Release CRV */
		Q931CallSetState(call, Q931_U0);
		Q931CallRelease(call);
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}


/**
 * Handle RELEASE COMPLETE message (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call handle
 * @param[in]	msg	Incoming message
 * @param[in]	from	Message coming from Layer 2 or 4?
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931ProcReleaseCompleteTE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch(from) {
	case Q931_MSG_FROM_L4:
		switch(Q931CallGetState(call)) {
		case Q931_U6:	/* Reject request */
			/* Send RELEASE COMPLETE */
			ret = Q931ReleaseComplete(trunk, call, 0);

			/* Release CRV (?) */
			Q931CallSetState(call, Q931_U0);
			Q931CallRelease(call);
			break;

		default:
			break;
		}
		break;
	case Q931_MSG_FROM_L2:
		switch(Q931CallGetState(call)) {
		case Q931_U0:
			/* Release call */
			Q931CallRelease(call);
			break;

		case Q931_U1:
			/* Stop T303 */
			Q931CallStopTimer(call, Q931_TIMER_T303);

			/* TODO: Send reject indication */
			Q931Tx34(trunk, call, msg, msg->Size);

			{
				/* Enqueue event */
				struct Q931_CallEvent *event = Q931CallNewEvent(call);

				event->id   = Q931_EVENT_REJECT_INDICATION;
				event->type = Q931_EVENT_TYPE_MESSAGE;
				event->data.message.type = msg->MesType;
				event->data.message.data = msg;

				Q931CallQueueEvent(call, event);
			}
			break;

		case Q931_U19:
			/* Stop T308 */
			Q931CallStopTimer(call, Q931_TIMER_T308);

			/* TODO: Send release confirm */
			Q931Tx34(trunk, call, msg, msg->Size);

			{
				/* Enqueue event */
				struct Q931_CallEvent *event = Q931CallNewEvent(call);

				event->id   = Q931_EVENT_RELEASE_CONFIRM;
				event->type = Q931_EVENT_TYPE_MESSAGE;
				event->data.message.type = msg->MesType;
				event->data.message.data = msg;

				Q931CallQueueEvent(call, event);
			}
			break;

		default:
			/* Stop all timers */
			Q931CallStopAllTimers(call);

			/* TODO: Release indication */
			ret = Q931Tx34(trunk, call, msg, msg->Size);

			{
				/* Enqueue event */
				struct Q931_CallEvent *event = Q931CallNewEvent(call);

				event->id   = Q931_EVENT_RELEASE_INDICATION;
				event->type = Q931_EVENT_TYPE_MESSAGE;
				event->data.message.type = msg->MesType;
				event->data.message.data = msg;

				Q931CallQueueEvent(call, event);
			}
		}

		/* Release CRV */
		Q931CallSetState(call, Q931_U0);
		Q931CallRelease(call);
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}


/**
 * Handle RESTART message (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call handle
 * @param[in]	msg	Incoming message
 * @param[in]	from	Message coming from Layer 2 or 4?
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931ProcRestartTE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;

	if (msg->CRV)
		return Q931E_INVALID_CRV;

#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	/* TODO: T317, proper handling etc. */

	switch(from) {
	case Q931_MSG_FROM_L4:
		/* TODO Add proc here */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);
		break;
	case Q931_MSG_FROM_L2:
		/* TODO Add proc here */
		ret = Q931Tx34(trunk, call, msg, msg->Size);

		if (Q931TrunkIsSetFlag(trunk, Q931_TFLAG_AUTO_RESTART_ACK)) {
			Q931AckRestart(trunk, msg);
		}

		/* TODO: Restart?? */
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}


/**
 * Handle RESTART ACKNOWLEDGE message (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call handle
 * @param[in]	msg	Incoming message
 * @param[in]	from	Message coming from Layer 2 or 4?
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931ProcRestartAckTE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;

#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
#if 0
	if (msg->CRV) {
		/* Find the call using CRV */
		ret = Q931FindCRV(trunk, msg->CRV, &callIndex);
		if (ret != Q931E_NO_ERROR)
			return ret;
		/* TODO - Set correct timer here */
		Q931StartTimer(trunk, callIndex, Q931_TIMER_T303);
	}
#endif

	switch(from) {
	case Q931_MSG_FROM_L4:
		/* TODO Add proc here */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);
		break;
	case Q931_MSG_FROM_L2:
		/* TODO Add proc here */
		ret = Q931Tx34(trunk, call, msg, msg->Size);

		/* TODO: RestartACK?? */
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}


/**
 * Handle CONGESTION CONTROL message (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call handle
 * @param[in]	msg	Incoming message
 * @param[in]	from	Message coming from Layer 2 or 4?
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931ProcCongestionControlTE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch(from) {
	case Q931_MSG_FROM_L4:
		/* TODO Add proc here */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);
		break;
	case Q931_MSG_FROM_L2:
		/* TODO Add proc here */
		ret = Q931Tx34(trunk, call, msg, msg->Size);
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}


/**
 * Handle INFORMATION message (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call handle
 * @param[in]	msg	Incoming message
 * @param[in]	from	Message coming from Layer 2 or 4?
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931ProcInformationTE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch(from) {
	case Q931_MSG_FROM_L4:
		/* TODO Add proc here */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);
		break;
	case Q931_MSG_FROM_L2:
		/* TODO Add proc here */
		ret = Q931Tx34(trunk, call, msg, msg->Size);
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}


/**
 * Handle NOTIFY message (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call handle
 * @param[in]	msg	Incoming message
 * @param[in]	from	Message coming from Layer 2 or 4?
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931ProcNotifyTE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch(from) {
	case Q931_MSG_FROM_L4:
		/* TODO Add proc here */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);
		break;
	case Q931_MSG_FROM_L2:
		/* TODO Add proc here */
		ret = Q931Tx34(trunk, call, msg, msg->Size);
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}


/**
 * Handle STATUS message (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call handle
 * @param[in]	msg	Incoming message
 * @param[in]	from	Message coming from Layer 2 or 4?
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931ProcStatusTE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch(from) {
	case Q931_MSG_FROM_L4:
		/* TODO Add proc here */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);
		break;
	case Q931_MSG_FROM_L2:
		{
			Q931ie_Cause *cause = Q931GetIEPtr(msg->Cause, msg->buf);
			int state = Q931_U0;

			if (Q931IsIEPresent(msg->CallState)) {
				Q931ie_CallState *cs = Q931GetIEPtr(msg->CallState, msg->buf);
				state = cs->CallState;
			}

			switch (Q931CallGetState(call)) {
			case Q931_U0:
				if (Q931CallIsGlobal(call)) {
					return Q931E_NO_ERROR;
				} else {
					/* Reply with RELEASE COMPLETE */
					L3UCHAR tmp[Q931L4BUF];
					Q931mes_Generic *gen = (Q931mes_Generic *)tmp;

					Q931InitMesGeneric(gen);
					gen->MesType = Q931mes_DISCONNECT;
					gen->CRV     = Q931CallGetCRV(call);

					cause->Value = Q850_CAUSE_WRONG_CALL_STATE;
					gen->Cause = Q931AppendIE(gen, (L3UCHAR *)cause);

					Q931Rx43(trunk, gen, gen->Size);
				}
				break;

			case Q931_U19:
				if (Q931CallIsGlobal(call)) {
					return Q931E_NO_ERROR;
				}

				if (state == Q931_U0) {
					/* Terminate call */

					/* STATUS INDICATION */
					ret = Q931Tx34(trunk, call, msg, msg->Size);

					{
						/* Enqueue event */
						struct Q931_CallEvent *event = Q931CallNewEvent(call);

						event->id   = Q931_EVENT_RELEASE_INDICATION;
						event->type = Q931_EVENT_TYPE_MESSAGE;
						event->data.message.type = msg->MesType;
						event->data.message.data = msg;

						Q931CallQueueEvent(call, event);
					}

					/* Release call ref */
					Q931CallSetState(call, Q931_U0);
					Q931CallRelease(call);
				}
				break;

			/* TODO: RESTART / RESTART_REQ state */

			default:
				if (Q931CallIsGlobal(call)) {
					return Q931E_NO_ERROR;
				}

				if (state == Q931_U0) {
					/* Terminate call */

					/* STATUS INDICATION */
					ret = Q931Tx34(trunk, call, msg, msg->Size);

					{
						/* Enqueue event */
						struct Q931_CallEvent *event = Q931CallNewEvent(call);

						event->id   = Q931_EVENT_RELEASE_INDICATION;
						event->type = Q931_EVENT_TYPE_MESSAGE;
						event->data.message.type = msg->MesType;
						event->data.message.data = msg;

						Q931CallQueueEvent(call, event);
					}

					/* Release call ref */
					Q931CallSetState(call, Q931_U0);
					Q931CallRelease(call);
				}
				else if (state == Q931CallGetState(call)) {
					/* "Compatible" call state */

					switch (cause->Value) {
					case Q850_CAUSE_MANDATORY_IE_MISSING:
					case Q850_CAUSE_MESSAGE_TYPE_NONEXIST:
					case Q850_CAUSE_WRONG_MESSAGE:
					case Q850_CAUSE_IE_NONEXIST:
					case Q850_CAUSE_INVALID_IE_CONTENTS:
						/* fatal */
						break;

					default:
						return Q931E_NO_ERROR;
					}

					/* STATUS indication (error) */
					ret = Q931Tx34(trunk, call, msg, msg->Size);

					{
						/* Enqueue event */
						struct Q931_CallEvent *event = Q931CallNewEvent(call);

						event->id   = Q931_EVENT_STATUS_INDICATION;
						event->type = Q931_EVENT_TYPE_MESSAGE;
						event->error= 1;
						event->data.message.type = msg->MesType;
						event->data.message.data = msg;

						Q931CallQueueEvent(call, event);
					}

					/* Send DISCONNECT */
					{
						L3UCHAR tmp[Q931L4BUF];
						Q931mes_Generic *gen = (Q931mes_Generic *)tmp;

						Q931InitMesGeneric(gen);
						gen->MesType = Q931mes_DISCONNECT;
						gen->CRV     = Q931CallGetCRV(call);

						gen->Cause = Q931AppendIE(gen, (L3UCHAR *)cause);

						Q931Rx43(trunk, gen, gen->Size);
					}

					/* Start T305 */
					Q931CallStartTimer(call, Q931_TIMER_T305);

					/* => Disconnect request */
					Q931CallSetState(call, Q931_U11);
				}
				else {
					/* Call state not compatible */

					/* STATUS indication (error) */
					ret = Q931Tx34(trunk, call, msg, msg->Size);

					{
						/* Enqueue event */
						struct Q931_CallEvent *event = Q931CallNewEvent(call);

						event->id   = Q931_EVENT_STATUS_INDICATION;
						event->type = Q931_EVENT_TYPE_MESSAGE;
						event->error= 1;
						event->data.message.type = msg->MesType;
						event->data.message.data = msg;

						Q931CallQueueEvent(call, event);
					}

					/* Send DISCONNECT */
					{
						L3UCHAR tmp[Q931L4BUF];
						Q931mes_Generic *gen = (Q931mes_Generic *)tmp;

						Q931InitMesGeneric(gen);
						gen->MesType = Q931mes_DISCONNECT;
						gen->CRV     = Q931CallGetCRV(call);

						cause->Value = Q850_CAUSE_WRONG_CALL_STATE;
						gen->Cause = Q931AppendIE(gen, (L3UCHAR *)cause);

						Q931Rx43(trunk, gen, gen->Size);
					}

					/* Start T305 */
					Q931CallStartTimer(call, Q931_TIMER_T305);

					/* => Disconnect request */
					Q931CallSetState(call, Q931_U11);
				}
			}
		}
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}


/**
 * Handle STATUS ENQUIRY message (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call handle
 * @param[in]	msg	Incoming message
 * @param[in]	from	Message coming from Layer 2 or 4?
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931ProcStatusEnquiryTE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch(from) {
#if 0
	case Q931_MSG_FROM_L4:
		/* TODO Add proc here */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);
		break;
#endif
	case Q931_MSG_FROM_L2:
		ret = Q931StatusEnquiryResponse(trunk, call, Q850_CAUSE_RESPONSE_TO_STATUS_ENQUIRY);
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}


/**
 * Handle SEGMENT message (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call handle
 * @param[in]	msg	Incoming message
 * @param[in]	from	Message coming from Layer 2 or 4?
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931ProcSegmentTE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch(from) {
	case Q931_MSG_FROM_L4:
		/* TODO Add proc here */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);
		break;
	case Q931_MSG_FROM_L2:
		/* TODO Add proc here */
		ret = Q931Tx34(trunk, call, msg, msg->Size);
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}

/** @} */

/****************************************************************************/
/******************* Q.932 - Supplementary Services *************************/
/****************************************************************************/

/**
 * @defgroup	q932_te_msg_proc	Q.932 TE Message handling
 * @{
 */

L3INT Q932ProcFacilityTE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch(from) {
	case Q931_MSG_FROM_L4:
		/* TODO Add proc here */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);
		break;
	case Q931_MSG_FROM_L2:
		/* TODO Add proc here */
		ret = Q931Tx34(trunk, call, msg, msg->Size);
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}

L3INT Q932ProcHoldTE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch(from) {
	case Q931_MSG_FROM_L4:
		/* TODO Add proc here */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);
		break;
	case Q931_MSG_FROM_L2:
		/* TODO Add proc here */
		ret = Q931Tx34(trunk, call, msg, msg->Size);
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}

L3INT Q932ProcHoldAckTE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch(from) {
	case Q931_MSG_FROM_L4:
		/* TODO Add proc here */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);
		break;
	case Q931_MSG_FROM_L2:
		/* TODO Add proc here */
		ret = Q931Tx34(trunk, call, msg, msg->Size);
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}

L3INT Q932ProcHoldRejectTE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch(from) {
	case Q931_MSG_FROM_L4:
		/* TODO Add proc here */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);
		break;
	case Q931_MSG_FROM_L2:
		/* TODO Add proc here */
		ret = Q931Tx34(trunk, call, msg, msg->Size);
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}

L3INT Q932ProcRegisterTE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch(from) {
	case Q931_MSG_FROM_L4:
		/* TODO Add proc here */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);
		break;
	case Q931_MSG_FROM_L2:
		/* TODO Add proc here */
		ret = Q931Tx34(trunk, call, msg, msg->Size);
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}

L3INT Q932ProcRetrieveTE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch(from) {
	case Q931_MSG_FROM_L4:
		/* TODO Add proc here */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);
		break;
	case Q931_MSG_FROM_L2:
		/* TODO Add proc here */
		ret = Q931Tx34(trunk, call, msg, msg->Size);
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}

L3INT Q932ProcRetrieveAckTE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch(from) {
	case Q931_MSG_FROM_L4:
		/* TODO Add proc here */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);
		break;
	case Q931_MSG_FROM_L2:
		/* TODO Add proc here */
		ret = Q931Tx34(trunk, call, msg, msg->Size);
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}

L3INT Q932ProcRetrieveRejectTE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, Q931mes_Generic *msg, q931_msg_from_t from)
{
	L3INT ret = Q931E_NO_ERROR;
#if 0
	/* Find the call using CRV */
	call = Q931GetCallByCRV(trunk, msg->CRV);
	if (!call)
		return Q931E_INVALID_CRV;

	/* check if message is valid in this state */
	if (!Q931DialectIsEventLegal(trunk->Dialect, Q931CallGetState(call), msg->MesType, from))
		return Q931E_UNEXPECTED_MESSAGE;
#endif
	switch(from) {
	case Q931_MSG_FROM_L4:
		/* TODO Add proc here */
		ret = Q931Tx32(trunk, 0, msg, msg->Size);
		break;
	case Q931_MSG_FROM_L2:
		/* TODO Add proc here */
		ret = Q931Tx34(trunk, call, msg, msg->Size);
		break;
	default:
		ret = Q931E_ILLEGAL_MESSAGE;
	}
	return ret;
}

/** @} */

/************************************************************************************
 * Timer callbacks (TE side)
 ************************************************************************************/

/**
 * @defgroup	q931_te_timer	Q.931 TE Timeout handler
 * @{
 */

/**
 * Handle T301 timer expiration (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931ProcTimeoutT301TE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call)
{
	Q931Log(trunk, Q931_LOG_NOTICE, "Timer T301 timed out for call %d\n", call->CRV);

	if (Q931CallGetState(call) != Q931_U4) {
		Q931Log(trunk, Q931_LOG_WARNING, "Ignoring timeout of T301 in state %s (wrong state)\n", Q931CallGetStateName(call));
		return Q931E_NO_ERROR;
	}

	return Q931E_NO_ERROR;
}

/**
 * Handle T302 timer expiration (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931ProcTimeoutT302TE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call)
{
	Q931Log(trunk, Q931_LOG_NOTICE, "Timer T302 timed out for call %d\n", call->CRV);

	if (Q931CallGetState(call) != Q931_U25) {
		Q931Log(trunk, Q931_LOG_WARNING, "Ignoring timeout of T302 in state %s (wrong state)\n", Q931CallGetStateName(call));
		return Q931E_NO_ERROR;
	}

	/* TODO: Send Timeout indication */
	{
		struct Q931_CallEvent event;

		Q931CallInitEvent(&event);
		event.id   = Q931_EVENT_TIMEOUT_INDICATION;
		event.type = Q931_EVENT_TYPE_TIMER;
		event.data.timer.id = Q931_TIMER_T302;

		Q931CallSendEvent(call, &event);
	}

	return Q931E_NO_ERROR;
}


/**
 * Handle T303 timer expiration (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931ProcTimeoutT303TE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call)
{
	L3UCHAR cnt = Q931CallGetTimerExpireCount(call);	/* T303 uses the counter */

	Q931Log(trunk, Q931_LOG_NOTICE, "Timer T303 timed out for call %d (number of times: %hhu)\n", call->CRV, cnt);

	if (Q931CallGetState(call) != Q931_U1) {
		Q931Log(trunk, Q931_LOG_WARNING, "Ignoring timeout of T303 in state %s (wrong state)\n", Q931CallGetStateName(call));
		return Q931E_NO_ERROR;
	}

	if (cnt == 1) {
		/* TODO: Resend SETUP */

		/* Restart T303 */
		Q931CallRestartTimer(call, Q931_TIMER_T303);

		/* No state change */
	} else {
		struct Q931_CallEvent event;

		/* (TODO:) Send SETUP CONFIRM Error */
		Q931CallInitEvent(&event);
		event.id    = Q931_EVENT_SETUP_CONFIRM;
		event.type  = Q931_EVENT_TYPE_TIMER;
		event.error = 1;
		event.data.timer.id = Q931_TIMER_T303;

		/* Send event immediately to layer 4 */
		Q931CallSendEvent(call, &event);

		/* Release CRV */
		Q931CallSetState(call, Q931_U0);
		Q931CallRelease(call);
	}
	return Q931E_NO_ERROR;
}


/**
 * Handle T304 timer expiration (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931ProcTimeoutT304TE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call)
{
	Q931Log(trunk, Q931_LOG_NOTICE, "Timer T304 timed out for call %d\n", call->CRV);

	if (Q931CallGetState(call) != Q931_U2) {
		Q931Log(trunk, Q931_LOG_WARNING, "Ignoring timeout of T304 in state %s (wrong state)\n", Q931CallGetStateName(call));
		return Q931E_NO_ERROR;
	}

	/* TODO: Send DISCONNECT */
	{
		L3UCHAR tmp[Q931L4BUF];
		Q931mes_Generic *gen = (Q931mes_Generic *)tmp;
		Q931ie_Cause cause;

		Q931InitMesGeneric(gen);
		gen->MesType = Q931mes_DISCONNECT;
		gen->CRV     = Q931CallGetCRV(call);

		/* TODO: use the correct cause code */
		cause.IEId = Q931ie_CAUSE;
		cause.Size = sizeof(Q931ie_Cause);
		cause.CodStand = Q931_CODING_ITU;	/* ITU */
		cause.Location = 1;	/* private network */
		cause.Recom    = 1;	/* */
		cause.Value = Q850_CAUSE_RECOVERY_ON_TIMER_EXPIRE;

		gen->Cause = Q931AppendIE(gen, (L3UCHAR *)&cause);

		Q931Tx32(trunk, 0, gen, gen->Size);
	}

	/* (TODO:) Send SETUP CONFIRM error */
	{
		struct Q931_CallEvent event;

		Q931CallInitEvent(&event);
		event.id    = Q931_EVENT_SETUP_CONFIRM;
		event.type  = Q931_EVENT_TYPE_TIMER;
		event.error = 1;
		event.data.timer.id = Q931_TIMER_T304;

		/* Send event immediately to layer 4 */
		Q931CallSendEvent(call, &event);
	}

	/* Start T305 */
	Q931CallStartTimer(call, Q931_TIMER_T305);

	/* => 11: Disconnect request */
	Q931CallSetState(call, Q931_U11);

	return Q931E_NO_ERROR;
}


/**
 * Handle T305 timer expiration (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931ProcTimeoutT305TE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call)
{
	Q931Log(trunk, Q931_LOG_NOTICE, "Timer T305 timed out for call %d\n", call->CRV);

	if (Q931CallGetState(call) != Q931_U11) {
		Q931Log(trunk, Q931_LOG_WARNING, "Ignoring timeout of T305 in state %s (wrong state)\n", Q931CallGetStateName(call));
		return Q931E_NO_ERROR;
	}

	/* Send RELEASE (cause) */
	{
		L3UCHAR tmp[Q931L4BUF];
		Q931mes_Generic *gen = (Q931mes_Generic *)tmp;
		Q931ie_Cause cause;

		Q931InitMesGeneric(gen);
		gen->MesType = Q931mes_RELEASE;
		gen->CRV     = Q931CallGetCRV(call);

		/* TODO: use the original cause */
		cause.IEId = Q931ie_CAUSE;
		cause.Size = sizeof(Q931ie_Cause);
		cause.CodStand = Q931_CODING_ITU;	/* ITU */
		cause.Location = 1;	/* private network */
		cause.Recom    = 1;	/* */
		cause.Value = Q850_CAUSE_RECOVERY_ON_TIMER_EXPIRE;

		gen->Cause = Q931AppendIE(gen, (L3UCHAR *)&cause);

		Q931Tx32(trunk, 0, gen, gen->Size);
	}

	/* Start T308 */
	Q931CallStartTimer(call, Q931_TIMER_T308);

	/* => 19: Release request */
	Q931CallSetState(call, Q931_U19);

	return Q931E_NO_ERROR;
}


/**
 * Handle T308 timer expiration (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931ProcTimeoutT308TE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call)
{
	L3UCHAR cnt = Q931CallGetTimerExpireCount(call);	/* T308 uses the counter */

	Q931Log(trunk, Q931_LOG_NOTICE, "Timer T308 timed out for call %d (number of times %hhu)\n", call->CRV, cnt);

	if (Q931CallGetState(call) != Q931_U19) {
		Q931Log(trunk, Q931_LOG_WARNING, "Ignoring timeout of T308 in state %s (wrong state)\n", Q931CallGetStateName(call));
		return Q931E_NO_ERROR;
	}

	if (cnt == 1) {
		/* TODO: Send RELEASE */
		{
			L3UCHAR tmp[Q931L4BUF];
			Q931mes_Generic *gen = (Q931mes_Generic *)tmp;
			Q931ie_Cause cause;

			Q931InitMesGeneric(gen);
			gen->MesType = Q931mes_RELEASE;
			gen->CRV     = Q931CallGetCRV(call);

			/* TODO: use the original cause */
			cause.IEId = Q931ie_CAUSE;
			cause.Size = sizeof(Q931ie_Cause);
			cause.CodStand = Q931_CODING_ITU;	/* ITU */
			cause.Location = 1;	/* private network */
			cause.Recom    = 1;	/* */
			cause.Value = Q850_CAUSE_RECOVERY_ON_TIMER_EXPIRE;

			gen->Cause = Q931AppendIE(gen, (L3UCHAR *)&cause);

			Q931Tx32(trunk, 0, gen, gen->Size);
		}

		/* Restart T308 */
		Q931CallRestartTimer(call, Q931_TIMER_T308);

		/* No state change */
	} else {
		/* TODO: Place B-Channel in maintenance (option, ?) */

		/* (TODO:) Send RELEASE CONFIRM (error) indication */
		{
			struct Q931_CallEvent event;

			Q931CallInitEvent(&event);
			event.id    = Q931_EVENT_RELEASE_CONFIRM;
			event.type  = Q931_EVENT_TYPE_TIMER;
			event.error = 1;
			event.data.timer.id = Q931_TIMER_T308;

			/* Send event immediately to layer 4 */
			Q931CallSendEvent(call, &event);
		}

		/* Release CRV */
		Q931CallSetState(call, Q931_U0);
		Q931CallRelease(call);
	}
	return Q931E_NO_ERROR;
}


/**
 * Handle T309 timer expiration (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931ProcTimeoutT309TE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call)
{
	Q931Log(trunk, Q931_LOG_NOTICE, "Timer T309 timed out for call %d\n", call->CRV);

	/* TODO: Send DATA LINK FAILURE indication */

	/* Release CRV */
	Q931CallSetState(call, Q931_U0);
	Q931CallRelease(call);

	return Q931E_NO_ERROR;
}


/**
 * Handle T310 timer expiration (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931ProcTimeoutT310TE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call)
{
	Q931Log(trunk, Q931_LOG_NOTICE, "Timer T310 timed out for call %d\n", call->CRV);

	if (Q931CallGetState(call) != Q931_U3) {
		Q931Log(trunk, Q931_LOG_WARNING, "Ignoring timeout of T310 in state %s (wrong state)\n", Q931CallGetStateName(call));
		return Q931E_NO_ERROR;
	}

	/* TODO: Send DISCONNECT */
	{
		L3UCHAR tmp[Q931L4BUF];
		Q931mes_Generic *gen = (Q931mes_Generic *)tmp;
		Q931ie_Cause cause;

		Q931InitMesGeneric(gen);
		gen->MesType = Q931mes_DISCONNECT;
		gen->CRV     = Q931CallGetCRV(call);

		/* TODO: use the correct cause code */
		cause.IEId = Q931ie_CAUSE;
		cause.Size = sizeof(Q931ie_Cause);
		cause.CodStand = Q931_CODING_ITU;	/* ITU */
		cause.Location = 1;	/* private network */
		cause.Recom    = 1;	/* */
		cause.Value = Q850_CAUSE_RECOVERY_ON_TIMER_EXPIRE;

		gen->Cause = Q931AppendIE(gen, (L3UCHAR *)&cause);

		Q931Tx32(trunk, 0, gen, gen->Size);
	}

	/* (TODO:) Send SETUP CONFIRM error */
	{
		struct Q931_CallEvent event;

		Q931CallInitEvent(&event);
		event.id    = Q931_EVENT_SETUP_CONFIRM;
		event.type  = Q931_EVENT_TYPE_TIMER;
		event.error = 1;
		event.data.timer.id = Q931_TIMER_T310;

		/* Send event immediately to layer 4 */
		Q931CallSendEvent(call, &event);
	}

	/* Start T305 */
	Q931CallStartTimer(call, Q931_TIMER_T305);

	/* => 11: Disconnect request */
	Q931CallSetState(call, Q931_U11);

	return Q931E_NO_ERROR;
}


/**
 * Handle T311 timer expiration (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931ProcTimeoutT311TE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call)
{
	Q931Log(trunk, Q931_LOG_NOTICE, "Timer T311 timed out for call %d\n", call->CRV);
	return Q931E_NO_ERROR;
}


/**
 * Handle T313 timer expiration (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931ProcTimeoutT313TE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call)
{
	Q931Log(trunk, Q931_LOG_NOTICE, "Timer T313 timed out for call %d\n", call->CRV);

	if (Q931CallGetState(call) != Q931_U8) {
		Q931Log(trunk, Q931_LOG_WARNING, "Ignoring timeout of T313 in state %s (wrong state)\n", Q931CallGetStateName(call));
		return Q931E_NO_ERROR;
	}

	/* TODO: Send DISCONNECT */
	{
		L3UCHAR tmp[Q931L4BUF];
		Q931mes_Generic *gen = (Q931mes_Generic *)tmp;
		Q931ie_Cause cause;

		Q931InitMesGeneric(gen);
		gen->MesType = Q931mes_DISCONNECT;
		gen->CRV     = Q931CallGetCRV(call);

		/* TODO: use the correct cause code */
		cause.IEId = Q931ie_CAUSE;
		cause.Size = sizeof(Q931ie_Cause);
		cause.CodStand = Q931_CODING_ITU;	/* ITU */
		cause.Location = 1;	/* private network */
		cause.Recom    = 1;	/* */
		cause.Value = Q850_CAUSE_RECOVERY_ON_TIMER_EXPIRE;

		gen->Cause = Q931AppendIE(gen, (L3UCHAR *)&cause);

		Q931Tx32(trunk, 0, gen, gen->Size);
	}

	/* (TODO:) Send SETUP COMPLETE error */
	{
		struct Q931_CallEvent event;

		Q931CallInitEvent(&event);
		event.id    = Q931_EVENT_SETUP_COMPLETE_INDICATION;
		event.type  = Q931_EVENT_TYPE_TIMER;
		event.error = 1;
		event.data.timer.id = Q931_TIMER_T313;

		/* Send event immediately to layer 4 */
		Q931CallSendEvent(call, &event);
	}

	/* Start T305 */
	Q931CallStartTimer(call, Q931_TIMER_T305);

	/* => 11: Disconnect request */
	Q931CallSetState(call, Q931_U11);

	return Q931E_NO_ERROR;
}


/**
 * Handle T314 timer expiration (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 * @todo	unhandled
 */
L3INT Q931ProcTimeoutT314TE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call)
{
	Q931Log(trunk, Q931_LOG_NOTICE, "Timer T314 timed out for call %d\n", call->CRV);
	return Q931E_NO_ERROR;
}


/**
 * Handle T316 timer expiration (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931ProcTimeoutT316TE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call)
{
	L3UCHAR cnt = Q931CallGetTimerExpireCount(call);

	Q931Log(trunk, Q931_LOG_NOTICE, "Timer T316 timed out for call %d (number of times: %hhu)\n", call->CRV, cnt);

	/* TODO: Only valid in state REST1 */

	if (cnt < 2) {
		/* TODO: Send RESTART */

		/* Restart T316 */
		Q931CallRestartTimer(call, Q931_TIMER_T316);
	} else {
		/* TODO: Maintenance action (?) */
	}
	return Q931E_NO_ERROR;
}


/**
 * Handle T317 timer expiration (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931ProcTimeoutT317TE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call)
{
	Q931Log(trunk, Q931_LOG_NOTICE, "Timer T317 timed out for call %d\n", call->CRV);

	/* TODO: Only valid in states REST1 + REST2 */

	/* TODO: Maintenance action (?) */

	return Q931E_NO_ERROR;
}


/**
 * Handle T318 timer expiration (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931ProcTimeoutT318TE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call)
{
	Q931Log(trunk, Q931_LOG_NOTICE, "Timer T318 timed out for call %d\n", call->CRV);

	if (Q931CallGetState(call) != Q931_U17) {
		Q931Log(trunk, Q931_LOG_WARNING, "Ignoring timeout of T318 in state %s (wrong state)\n", Q931CallGetStateName(call));
		return Q931E_NO_ERROR;
	}

	/* Send RELEASE (cause) */
	{
		L3UCHAR tmp[Q931L4BUF];
		Q931mes_Generic *gen = (Q931mes_Generic *)tmp;
		Q931ie_Cause cause;

		Q931InitMesGeneric(gen);
		gen->MesType = Q931mes_RELEASE;
		gen->CRV     = Q931CallGetCRV(call);

		/* TODO: use the original cause */
		cause.IEId = Q931ie_CAUSE;
		cause.Size = sizeof(Q931ie_Cause);
		cause.CodStand = Q931_CODING_ITU;	/* ITU */
		cause.Location = 1;	/* private network */
		cause.Recom    = 1;	/* */
		cause.Value = Q850_CAUSE_RECOVERY_ON_TIMER_EXPIRE;

		gen->Cause = Q931AppendIE(gen, (L3UCHAR *)&cause);

		Q931Tx32(trunk, 0, gen, gen->Size);
	}

	/* Send RESUME CONFIRM (error) indication */
	{
		struct Q931_CallEvent event;

		Q931CallInitEvent(&event);
		event.id    = Q931_EVENT_RESUME_CONFIRM;
		event.type  = Q931_EVENT_TYPE_TIMER;
		event.error = 1;
		event.data.timer.id = Q931_TIMER_T318;

		/* Send event immediately to layer 4 */
		Q931CallSendEvent(call, &event);
	}

	/* Start T308 */
	Q931CallStartTimer(call, Q931_TIMER_T308);

	/* => 19: Release request */
	Q931CallSetState(call, Q931_U19);

	return Q931E_NO_ERROR;
}


/**
 * Handle T319 timer expiration (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931ProcTimeoutT319TE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call)
{
	Q931Log(trunk, Q931_LOG_NOTICE, "Timer T319 timed out for call %d\n", call->CRV);

	if (Q931CallGetState(call) != Q931_U15) {
		Q931Log(trunk, Q931_LOG_WARNING, "Ignoring timeout of T319 in state %s (wrong state)\n", Q931CallGetStateName(call));
		return Q931E_NO_ERROR;
	}

	/* (TODO:) Send SUSPEND CONFIRM (error) indication */
	{
		struct Q931_CallEvent event;

		Q931CallInitEvent(&event);
		event.id    = Q931_EVENT_SUSPEND_CONFIRM;
		event.type  = Q931_EVENT_TYPE_TIMER;
		event.error = 1;
		event.data.timer.id = Q931_TIMER_T319;

		/* Send event immediately to layer 4 */
		Q931CallSendEvent(call, &event);
	}

	/* => 10: Active */
	Q931CallSetState(call, Q931_U10);

	return Q931E_NO_ERROR;
}


/**
 * Handle T321 timer expiration (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 * @todo	unhandled
 */
L3INT Q931ProcTimeoutT321TE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call)
{

	Q931Log(trunk, Q931_LOG_NOTICE, "Timer T321 timed out for call %d\n", call->CRV);

	/* TODO: ??? */
	{
		struct Q931_CallEvent event;

		Q931CallInitEvent(&event);
		event.id   = Q931_EVENT_TIMEOUT_INDICATION;
		event.type = Q931_EVENT_TYPE_TIMER;
		event.data.timer.id = Q931_TIMER_T321;

		Q931CallSendEvent(call, &event);
	}

	return Q931E_NO_ERROR;
}


/**
 * Handle T322 timer expiration (TE)
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	call	Q.931 call
 * @return	Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 * @todo	unhandled
 */
L3INT Q931ProcTimeoutT322TE(Q931_TrunkInfo_t *trunk, struct Q931_Call *call)
{
	Q931Log(trunk, Q931_LOG_NOTICE, "Timer T322 timed out for call %d\n", call->CRV);

	/* TODO: ??? */
	{
		struct Q931_CallEvent event;

		Q931CallInitEvent(&event);
		event.id   = Q931_EVENT_TIMEOUT_INDICATION;
		event.type = Q931_EVENT_TYPE_TIMER;
		event.data.timer.id = Q931_TIMER_T322;

		/* Send event immediately to layer 4 */
		Q931CallSendEvent(call, &event);
	}

	return Q931E_NO_ERROR;
}

/** @} */

/** @} */
