/*****************************************************************************

  FileName:		Q931api.c

  Contents:		api (Application Programming Interface) functions.
				See	q931.h for description.

  License/Copyright:

  Copyright (c) 2007, Jan Vidar Berger, Case Labs, Ltd. All rights reserved.
  email:janvb@caselaboratories.com

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are
  met:

    * Redistributions of source code must retain the above copyright notice,
	  this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
	  this list of conditions and the following disclaimer in the documentation
	  and/or other materials provided with the distribution.
    * Neither the name of the Case Labs, Ltd nor the names of its contributors
	  may be used to endorse or promote products derived from this software
	  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.
*****************************************************************************/
#include <stdlib.h>
#include <string.h>

#include "Q931.h"
#include "Q931priv.h"

extern L3INT Q931L4HeaderSpace;

Q931_API L3INT Q931InitTrunk(Q931_TrunkInfo_t *trunk,
						Q931Dialect_t Dialect,
						Q931NetUser_t NetUser,
						Q931_TrunkType_t TrunkType,
						Q931Tx34CB_t Q931Tx34CBProc,
						Q931Tx32CB_t Q931Tx32CBProc,
						Q931ErrorCB_t Q931ErrorCBProc,
						void *PrivateData32,
						void *PrivateData34)
{
	int y, dchannel, maxchans, has_sync = 0;

	memset(trunk, 0, sizeof(Q931_TrunkInfo_t));

	switch(TrunkType) {
	case Q931_TrType_E1:
		dchannel = 16;
		maxchans = 31;
		has_sync = 1;
		break;

	case Q931_TrType_T1:
	case Q931_TrType_J1:
		dchannel = 24;
		maxchans = 24;
		break;

	case Q931_TrType_BRI:
	case Q931_TrType_BRI_PTMP:
		dchannel = 3;
		maxchans = 3;
		break;

	default:
		return -1;
	}

	trunk->Q931Tx34CBProc  = Q931Tx34CBProc;
	trunk->Q931Tx32CBProc  = Q931Tx32CBProc;
	trunk->Q931ErrorCBProc = Q931ErrorCBProc;
	trunk->PrivateData32   = PrivateData32;
	trunk->PrivateData34   = PrivateData34;

	trunk->LastCRV        = 0;
	trunk->Enabled        = 0;
	trunk->TrunkState     = 0;
#if 0
	trunk->autoRestartAck = 0;
#endif

	/* Set dialect */
	trunk->Dialect = Q931DialectGet(Dialect, NetUser);
	if (!trunk->Dialect) {
		return -1;
	}
	trunk->TrunkType = TrunkType;
	trunk->NetUser   = NetUser;

	/* Init channel list */
	for (y = 0; y < Q931MAXCHPERTRUNK; y++) {
		trunk->ch[y].Available = 1;

		if (has_sync && y == 0) {
			trunk->ch[y].ChanType = Q931_ChType_Sync;
		}
		else if (y == dchannel) {
			trunk->ch[y].ChanType = Q931_ChType_D;
		}
		else if(y > maxchans) {
			trunk->ch[y].ChanType = Q931_ChType_NotUsed;
		}
		else {
			trunk->ch[y].ChanType = Q931_ChType_B;
		}
	}

	/* Init global call */
	trunk->call[0].Trunk = trunk;
	trunk->call[0].InUse = 1;

	return 0;
}

/**
 * @brief	Start Q.931 Trunk (Send RESTART)
 */
Q931_API L3INT Q931Start(Q931_TrunkInfo_t *trunk)
{
	L3UCHAR buf[Q931L4BUF];
	Q931mes_Generic *msg = (Q931mes_Generic *)buf;
	Q931ie_RestartInd resind;
	Q931ie_ChanID chanID;

	/*
	 * TODO: Only restart Interface on
	 * Init -> Up trunk state change (after receiving a L2_ESTABLISH indication)
	 */
	Q931InitMesGeneric(msg);
	msg->MesType = Q931mes_RESTART;
	msg->CRV     = 0;

	/* Restart Indicator IE */
	resind.IEId  = Q931ie_RESTART_INDICATOR;
	resind.Class = 0x06;	/* Restart this interfaces (?) */
	resind.Size  = sizeof(Q931ie_RestartInd);

	msg->RestartInd = Q931AppendIE(msg, (L3UCHAR *)&resind);

	/* Channel ID IE */
	Q931InitIEChanID(&chanID);
	chanID.CodStand = 0;
	chanID.IntType = Q931_IS_PRI(trunk) ? 1 : 0;

	msg->ChanID = Q931AppendIE(msg, (L3UCHAR *)&chanID);

	return Q931Proc(trunk, Q931GetGlobalCall(trunk), msg, Q931_MSG_FROM_L4);
}

L3INT Q931GetMesSize(Q931mes_Generic *msg)
{
	return ((L3INT)(&msg->buf[0] - (L3UCHAR *)msg));
}

/*****************************************************************************

  Function:     q931AppendIE

  Description:  Append IE to the message.

  Parameters:   pm      Ptr to message.
                pi      Ptr to information element

  Return Value  ie setting

*****************************************************************************/
Q931_API ie Q931AppendIE(Q931mes_Generic *msg, L3UCHAR *pi)
{
	L3UCHAR *ptr = (L3UCHAR *)msg;
	L3INT iISize = ((Q931ie_BearerCap *)pi)->Size;
	L3INT Off    = msg->Size - (&msg->buf[0] - ptr);

	memcpy(ptr + msg->Size, pi, iISize);
	msg->Size += iISize;

	return (ie)(Off | 0x8000);
}

/*****************************************************************************
*****************************************************************************/

L3INT Q931GetUniqueCRV(Q931_TrunkInfo_t *trunk)
{
	L3INT max = (Q931_IS_BRI(trunk)) ? Q931_BRI_MAX_CRV : Q931_PRI_MAX_CRV;
	L3INT attempts = 50;

	do {
		trunk->LastCRV++;
		trunk->LastCRV = (trunk->LastCRV <= max) ? trunk->LastCRV : 1;

	} while(Q931GetCallByCRV(trunk, trunk->LastCRV) && --attempts > 0);

	return trunk->LastCRV;
}

Q931_API L3INT Q931InitMesGeneric(Q931mes_Generic *msg)
{
	memset(msg, 0, sizeof(*msg));
	msg->ProtDisc = 0x08;
	msg->Size     = Q931GetMesSize(msg);
	return 0;
}

Q931_API L3INT Q931InitMesResume(Q931mes_Generic *msg)
{
	memset(msg, 0, sizeof(*msg));
	msg->ProtDisc = 0x08;
	msg->MesType  = Q931mes_RESUME;
	msg->Size     = Q931GetMesSize(msg);
	return 0;
}

Q931_API L3INT Q931InitMesRestartAck(Q931mes_Generic *msg)
{
	memset(msg, 0, sizeof(*msg));
	msg->ProtDisc = 0x08;
	msg->MesType  = Q931mes_RESTART_ACKNOWLEDGE;
	msg->Size     = Q931GetMesSize(msg);
	return 0;
}

Q931_API L3INT Q931InitIEBearerCap(Q931ie_BearerCap *ie)
{
	memset(ie, 0, sizeof(*ie));
	ie->IEId = Q931ie_BEARER_CAPABILITY;
	ie->Size = sizeof(Q931ie_BearerCap);
	ie->ITR  = 0x10;
	return 0;
}

Q931_API L3INT Q931InitIEChanID(Q931ie_ChanID *ie)
{
	memset(ie, 0, sizeof(*ie));
	ie->IEId = Q931ie_CHANNEL_IDENTIFICATION;
	ie->Size = sizeof(Q931ie_ChanID);
	return 0;
}

Q931_API L3INT Q931InitIEProgInd(Q931ie_ProgInd *ie)
{
	memset(ie, 0, sizeof(*ie));
	ie->IEId = Q931ie_PROGRESS_INDICATOR;
	ie->Size = sizeof(Q931ie_ProgInd);
	return 0;
}

Q931_API L3INT Q931InitIENetFac(Q931ie_NetFac *ie)
{
	memset(ie, 0, sizeof(*ie));
	ie->IEId = Q931ie_NETWORK_SPECIFIC_FACILITIES;
	ie->Size = sizeof(Q931ie_NetFac);
	return 0;
}

Q931_API L3INT Q931InitIEDisplay(Q931ie_Display *ie)
{
	memset(ie, 0, sizeof(*ie));
	ie->IEId = Q931ie_DISPLAY;
	ie->Size = sizeof(Q931ie_Display);
	return 0;
}

Q931_API L3INT Q931InitIEDateTime(Q931ie_DateTime *ie)
{
	memset(ie, 0, sizeof(*ie));
	ie->IEId = Q931ie_DATETIME;
	ie->Size = sizeof(Q931ie_DateTime);
	return 0;
}

Q931_API L3INT Q931InitIEKeypadFac(Q931ie_KeypadFac *ie)
{
	memset(ie, 0, sizeof(*ie));
	ie->IEId = Q931ie_KEYPAD_FACILITY;
	ie->Size = sizeof(Q931ie_KeypadFac);
	return 0;
}

Q931_API L3INT Q931InitIESignal(Q931ie_Signal *ie)
{
	memset(ie, 0, sizeof(*ie));
	ie->IEId = Q931ie_SIGNAL;
	ie->Size = sizeof(Q931ie_Signal);
	return 0;
}

Q931_API L3INT Q931InitIECallingNum(Q931ie_CallingNum * ie)
{
	memset(ie, 0, sizeof(*ie));
	ie->IEId = Q931ie_CALLING_PARTY_NUMBER;
	ie->Size = sizeof(Q931ie_CallingNum);
	return 0;
}

Q931_API L3INT Q931InitIECallingSub(Q931ie_CallingSub *ie)
{
	memset(ie, 0, sizeof(*ie));
	ie->IEId = Q931ie_CALLING_PARTY_SUBADDRESS;
	ie->Size = sizeof(Q931ie_CallingSub);
	return 0;
}

Q931_API L3INT Q931InitIECalledNum(Q931ie_CalledNum *ie)
{
	memset(ie, 0, sizeof(*ie));
	ie->IEId = Q931ie_CALLED_PARTY_NUMBER;
	ie->Size = sizeof(Q931ie_CalledNum);
	return 0;
}

Q931_API L3INT Q931InitIECalledSub(Q931ie_CalledSub *ie)
{
	memset(ie, 0, sizeof(*ie));
	ie->IEId = Q931ie_CALLED_PARTY_SUBADDRESS;
	ie->Size = sizeof(Q931ie_CalledSub);
	return 0;
}

Q931_API L3INT Q931InitIETransNetSel(Q931ie_TransNetSel *ie)
{
	memset(ie, 0, sizeof(*ie));
	ie->IEId = Q931ie_TRANSIT_NETWORK_SELECTION;
	ie->Size = sizeof(Q931ie_TransNetSel);
	return 0;
}

Q931_API L3INT Q931InitIELLComp(Q931ie_LLComp *ie)
{
	memset(ie, 0, sizeof(*ie));
	ie->IEId = Q931ie_LOW_LAYER_COMPATIBILITY;
	ie->Size = sizeof(Q931ie_LLComp);
	return 0;
}

Q931_API L3INT Q931InitIEHLComp(Q931ie_HLComp *ie)
{
	memset(ie, 0, sizeof(*ie));
	ie->IEId = Q931ie_HIGH_LAYER_COMPATIBILITY;
	ie->Size = sizeof(Q931ie_HLComp);
	return 0;
}

L3INT Q931Disconnect(Q931_TrunkInfo_t *trunk, L3INT iTo, L3INT iCRV, L3INT iCause)
{
	/* TODO:  Unhandled paramaters */
	(void)trunk;
	(void)iTo;
	(void)iCRV;
	(void)iCause;
	return 0;
}

L3INT Q931Release(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, L3UCHAR causeval)
{
	L3UCHAR buf[Q931L4BUF];
	Q931mes_Generic *msg = (Q931mes_Generic *)buf;
	Q931ie_Cause cause;

	Q931InitMesGeneric(msg);
	msg->MesType = Q931mes_RELEASE;
	msg->CRV     = Q931CallGetCRV(call);
	msg->CRVFlag = (Q931CallGetDirection(call) == Q931_DIRECTION_INBOUND) ? 1 : 0;

	if (causeval) {
		/* Inititalize cause IE */
		cause.IEId = Q931ie_CAUSE;
		cause.Size = sizeof(Q931ie_Cause);
		cause.CodStand = Q931_CODING_ITU;
		cause.Location = 1;
		cause.Recom    = 1;
		cause.Value    = causeval;
		*cause.Diag    = '\0';

		msg->Cause = Q931AppendIE(msg, (L3UCHAR *)&cause);
	}

	return Q931Tx32(trunk, 0, msg, msg->Size);
}

L3INT Q931ReleaseComplete(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, L3UCHAR causeval)
{
	L3UCHAR buf[Q931L4BUF];
	Q931mes_Generic *msg = (Q931mes_Generic *)buf;
	Q931ie_Cause cause;

	Q931InitMesGeneric(msg);
	msg->MesType = Q931mes_RELEASE_COMPLETE;
	msg->CRV     = Q931CallGetCRV(call);
	msg->CRVFlag = (Q931CallGetDirection(call) == Q931_DIRECTION_INBOUND) ? 1 : 0;

	if (causeval) {
		/* Inititalize cause IE */
		cause.IEId = Q931ie_CAUSE;
		cause.Size = sizeof(Q931ie_Cause);
		cause.CodStand = Q931_CODING_ITU;
		cause.Location = 1;
		cause.Recom    = 1;
		cause.Value    = causeval;
		*cause.Diag    = '\0';

		msg->Cause = Q931AppendIE(msg, (L3UCHAR *)&cause);
	}

	return Q931Tx32(trunk, 0, msg, msg->Size);
}

L3INT Q931StatusEnquiryResponse(Q931_TrunkInfo_t *trunk, struct Q931_Call *call, L3UCHAR causeval)
{
	L3UCHAR buf[Q931L4BUF];
	Q931mes_Generic *msg = (Q931mes_Generic *)buf;
	Q931ie_CallState state;
	Q931ie_Cause cause;

	Q931InitMesGeneric(msg);
	msg->MesType = Q931mes_STATUS;
	msg->CRV     = Q931CallGetCRV(call);
	msg->CRVFlag = (Q931CallGetDirection(call) == Q931_DIRECTION_INBOUND) ? 1 : 0;

	/* Inititalize cause IE */
	cause.IEId = Q931ie_CAUSE;
	cause.Size = sizeof(Q931ie_Cause);
	cause.CodStand = Q931_CODING_ITU;
	cause.Location = 1;
	cause.Recom    = 1;
	cause.Value    = causeval;
	*cause.Diag    = '\0';

	msg->Cause = Q931AppendIE(msg, (L3UCHAR *)&cause);

	/* Initialize callstate IE */
	state.IEId = Q931ie_CALL_STATE;
	state.Size = sizeof(Q931ie_CallState);
	state.CodStand  = Q931_CODING_ITU;
	state.CallState = (L3UCHAR)(Q931CallGetState(call) & 0xff);

	msg->CallState = Q931AppendIE(msg, (L3UCHAR *)&state);

	return Q931Tx32(trunk, 0, msg, msg->Size);
}

/**
 * Q931CallSendStatus
 * @brief	Send STATUS message with cause code
 * @param[in]	call		Q.931 call handle
 * @param[in]	causeval	Q.850 cause code
 * @return	>= 0 on success, < 0 on error
 * @note	Let's see... this is how the L4 api stuff is supposed to look (custom IEs missing here)
 * @todo	Naming s*cks a bit... (maybe use a new prefix (instead of "Q931Call") for layer4 call control related stuff)
 */
L3INT Q931CallSendStatus(const struct Q931_Call *call, const L3UCHAR causeval)
{
	L3UCHAR buf[Q931L4BUF];
	Q931mes_Generic *msg = (Q931mes_Generic *)buf;
	Q931_TrunkInfo_t *trunk;
	Q931ie_CallState state;
	Q931ie_Cause cause;

	trunk = Q931CallGetTrunk(call);

	Q931InitMesGeneric(msg);
	msg->MesType = Q931mes_STATUS;
	msg->CRVFlag = Q931CallIsOutgoing(call) ? 0 : 1;

	/* Inititalize cause IE */
	cause.IEId = Q931ie_CAUSE;
	cause.Size = sizeof(Q931ie_Cause);
	cause.CodStand = Q931_CODING_ITU;
	cause.Location = 1;
	cause.Recom    = 1;
	cause.Value    = causeval;
	*cause.Diag    = '\0';

	msg->Cause = Q931AppendIE(msg, (L3UCHAR *)&cause);

	/* Initialize callstate IE */
	state.IEId = Q931ie_CALL_STATE;
	state.Size = sizeof(Q931ie_CallState);
	state.CodStand  = Q931_CODING_ITU;
	state.CallState = (L3UCHAR)(Q931CallGetState(call) & 0xff);

	msg->CallState = Q931AppendIE(msg, (L3UCHAR *) &state);

	return Q931Tx32(trunk, 0, msg, msg->Size);
}

L3INT Q931AckRestart(Q931_TrunkInfo_t *trunk, Q931mes_Generic *msg)
{
	L3INT RetCode;

	msg->MesType = Q931mes_RESTART_ACKNOWLEDGE;
	msg->CRVFlag = !(msg->CRVFlag);

	/* Restart Indicator */
	if (!msg->RestartInd) {
		Q931ie_RestartInd resind;

		resind.IEId  = Q931ie_RESTART_INDICATOR;
		resind.Class = 0x06;	/* Restart this interface (?) */

		msg->RestartInd = Q931AppendIE(msg, (L3UCHAR *)&resind);
	}

	RetCode = Q931Proc(trunk, Q931GetGlobalCall(trunk), msg, Q931_MSG_FROM_L4);
	return RetCode;
}

L3INT Q931AckSetup(Q931_TrunkInfo_t *trunk, Q931mes_Generic *msg)
{
	msg->MesType = Q931mes_SETUP_ACKNOWLEDGE;

	return Q931Proc(trunk, Q931GetGlobalCall(trunk), msg, Q931_MSG_FROM_L4);
}

L3INT Q931Send(struct Q931_Call *call, Q931mes_Generic *msg, L3INT bcast)
{
	Q931_TrunkInfo_t *trunk = Q931CallGetTrunk(call);

	return Q931Tx32(trunk, bcast, msg, msg->Size);
}

L3INT Q931AckConnect(struct Q931_Call *call)
{
	L3UCHAR buf[Q931L4BUF];
	Q931mes_Generic *msg = (Q931mes_Generic *)buf;
	Q931_TrunkInfo_t *trunk = Q931CallGetTrunk(call);

	Q931InitMesGeneric(msg);
	msg->MesType = Q931mes_CONNECT_ACKNOWLEDGE;
	msg->CRVFlag = Q931CallIsOutgoing(call) ? 0 : 1;
	msg->CRV     = Q931CallGetCRV(call);

	return Q931Tx32(trunk, 0, msg, msg->Size);
}

L3INT Q931AckService(Q931_TrunkInfo_t *trunk, Q931mes_Generic *msg)
{
	msg->MesType = Q931mes_SERVICE_ACKNOWLEDGE;
	if (msg->CRV)
		msg->CRVFlag = !(msg->CRVFlag);

	return Q931Proc(trunk, Q931GetGlobalCall(trunk), msg, Q931_MSG_FROM_L4);
}


/***************************************************************************************************************************
 *
 * Q.931 Simple API
 *
 ***************************************************************************************************************************/
#if 0
#include <stdarg.h>

/*
 * Tag definition
 */
typedef enum {
	Q931_TAG_TYPE_END = 0,			/*!< End of list */
	Q931_TAG_TYPE_SKIP,			/*!< Skip over entry */

	/* CID */
	Q931_TAG_TYPE_CALLED_NUMBER,
	Q931_TAG_TYPE_CALLING_NUMBER,

	/* Cause */
	Q931_TAG_TYPE_CAUSE_CODE,

	/* Always last */
	Q931_TAG_TYPE_MAX
} q931_tag_type_t;

#define Q931_TAG_IF(cond, x)		!(cond) ? Q931_TAG_TYPE_SKIP : (x)

#define Q931_TAG_END			Q931_TAG_TYPE_END

#define Q931_TAG_CALLED_NUMBER(x)	Q931_TAG_TYPE_CALLED_NUMBER, (x)
#define Q931_TAG_CALLING_NUMBER(x)	Q931_TAG_TYPE_CALLING_NUMBER, (x)

#define Q931_TAG_CAUSE(x)		Q931_TAG_TYPE_CAUSE_CODE, (x)


#define Q931_NUMBER_LEN		32
#define Q931_NAME_LEN		32

//struct Q931_CallData {
//};

struct Q931_CallProfile {
	L3CHAR	ton;
	L3CHAR	screen;
	L3CHAR	pres;
	L3CHAR	number_plan;
	L3CHAR	number[Q931_NUMBER_LEN];
	L3CHAR	  name[Q931_NAME_LEN];
};

#define Q931_STRLEN_ZERO(x)	(!x || !x[0])

enum {
	NUMBER_TAG_NUMBER = 0,	/*!< Number */
	NUMBER_TAG_NAME,	/*!< Name */
	NUMBER_TAG_TON,		/*!< Type of number */
	NUMBER_TAG_NUMPLAN,	/*!< Numbering plan */
	NUMBER_TAG_SCREEN,	/*!< Screening */
	NUMBER_TAG_PRES		/*!< Presentation */
};

static L3INT Q931Tag_setup_number_params(struct Q931_CallProfile *profile, int opt, ap_list *ap)
{
	const char *strval;
	const int   intval;

	if (!profile || !ap)
		return -1;

	switch (opt) {
	case NUMBER_TAG_NUMBER:
		strval = va_arg(*ap, char *);
		if (!Q931_STRLEN_ZERO(strval))
			strncpy(profile->number, strval, Q931_NUMBER_LEN);
		break;

	case NUMBER_TAG_NAME:
		strval = va_arg(*ap, char *);
		if (!Q931_STRLEN_ZERO(strval))
			strncpy(profile->name, strval, Q931_NAME_LEN);
		break;

	case NUMBER_TAG_TON:
		intval = va_arg(*ap, int);
		if (intval < 0 || intval >= Q931_TON_RESERVED)
			return -1;

		profile->ton = intval;
		break;

	case NUMBER_TAG_NUMPLAN:
		intval = va_arg(*ap, int);
		if (intval < 0 || intval >= Q931_NUMPLAN_RESERVED)
			return -1;

		profile->number_plan = intval;
		break;

	case NUMBER_TAG_SCREEN:
		intval = va_arg(*ap, int);
		if (intval < 0 || intval > Q931_SCREEN_NETWORK)
			return -1;

		profile->screen = intval;
		break;

	case NUMBER_TAG_PRES:
		intval = va_arg(*ap, int);
		if (intval < 0 || intval >= Q931_PRES_RESERVED)
			return -1;

		profile->pres = intval;
		break;

	default:
		return -1;
	}

	return 0;
}

L3INT Q931Simple_Setup(Q931_TrunkInfo_t *trunk, struct Q931_Call **call, ...)
{
	struct Q931_Call *new_call = NULL;
	va_list ap;
	q931_tag_type_t tag;
	int ret = 0;

	if (!trunk || !call)
		return -1;

	/*
	 * Allocate + init new outgoing call
	 */
	new_call = Q931CallNew(trunk);
	if (!new_call)
		return -1;

#if __TODO__
	/* Probably going to put all call parameters into a sub structure */
	memset(&call->called_number,  0, sizeof(struct Q931_CallProfile));
	memset(&call->calling_number, 0, sizeof(struct Q931_CallProfile));
#endif

	/*
	 * Parse tags
	 */
	va_start(ap, call);
	while (!ret) {
		tag = va_arg(ap, q931_tag_type_t);
		if (tag == Q931_TAG_TYPE_END)
			break;

		switch (tag) {
		case Q931_TAG_TYPE_SKIP:
			continue;

		case Q931_TAG_TYPE_CALLED_NUMBER:
//			ret = Q931Tag_setup_number_params(&new_call->called_number,  NUMBER_TAG_NUMBER, &ap);
			break;

		case Q931_TAG_TYPE_CALLING_NUMBER:
//			ret = Q931Tag_setup_number_params(&new_call->calling_number, NUMBER_TAG_NUMBER, &ap);
			break;

		default:
			Q931Log(trunk, Q931_LOG_ERROR, "Unknown/invalid tag %d\n", tag);
			ret = -1;
		}
	}
	va_end(ap);

	if (ret < 0) {
		Q931CallRelease(new_call);
		return ret;
	}

	/*
	 * Minimum requirements: Destination number
	 * (unless we're going to support overlap dialing)
	 */
#if __TODO__
	if (Q931_STRLEN_ZERO(call->called_number.number)) {
		Q931Log(trunk, Q931_LOG_ERROR, "Need a destination number\n");
		goto error;
	}
#endif

	*call = new_call;
	return ret;

error:
	Q931Release(new_call);
	return -1;
}

L3INT Q931Simple_Hangup(struct Q931_Call *call, ...)
{
	Q931_TrunkInfo_t *trunk = NULL;
	va_list ap;
	q931_tag_type_t tag;
	int ret = 0;
	int intval;
	int cause = Q850_CAUSE_NORMAL_CLEARING;

	if (!call)
		return -1;

	if (!(trunk = Q931CallGetTrunk(call)))
		return -1;

	/*
	 * Parse tags
	 */
	va_start(ap, call);

	while (!ret) {
		tag = va_arg(ap, q931_tag_type_t);
		if (tag == Q931_TAG_TYPE_END)
			break;

		switch (tag) {
		case Q931_TAG_TYPE_SKIP:
			continue;

		case Q931_TAG_TYPE_CAUSE_CODE:
			intval = va_arg(ap, int);
			if (intval <= Q850_CAUSE_NONE || intval > Q850_CAUSE_INTERWORKING) {
				ret = -1;
				break;
			}

			cause = intval;
			break;

		default:
			Q931Log(trunk, Q931_LOG_ERROR, "Unknown/invalid tag %d\n", tag);
			ret = -1;
		}
	}
	va_end(ap);

	if (ret < 0)
		return ret;

	/*
	 * Hangup
	 * --- TODO ---
	 */

	return ret;
}
#endif /* 0 */
