/*****************************************************************************

  FileName:		Q931call.c

  Contents:		call related api functions.
			See q931.h for description.

  License/Copyright:

  Copyright (c) 2008, Stefan Knoblich, axsentis GmbH. All rights reserved.
  email: s.knoblich@axsentis.de

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are
  met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.
    * Neither the name of the axsentis GmbH nor the names of its contributors
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.
*****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "Q931.h"
#include "Q931priv.h"

/*
 * State handling
 */
void Q931CallSetState(struct Q931_Call *call, const L3INT state)
{
	assert(call);

	if (Q931CallIsGlobal(call))
		return;

	Q931Log(Q931CallGetTrunk(call), Q931_LOG_DEBUG, "Call going from state %d -> %d\n",
					 (call->State & 0xff), (state & 0xff));

	call->State = state;
}

Q931_API L3INT Q931CallGetState(const struct Q931_Call *call)
{
	assert(call);

	return call->State;
}


/*
 * Allocate / Destroy call handle
 */
static struct Q931_Call *Q931FindFreeCall(const Q931_TrunkInfo_t *trunk)
{
	struct Q931_Call *call = NULL;
	L3INT i;

	/* call[0] reserved for global CRV */
	for (i = 1; i < Q931MAXCALLPERTRUNK; i++) {
		if (!trunk->call[i].InUse) {
			call = (struct Q931_Call *)&trunk->call[i];
			break;
		}
	}
	return call;
}

/**
 * Q931CallNew
 * @brief	Request new outgoing call handle & crv
 * @param[in]	trunk		Q931 trunk
 * @return	NULL on error, pointer to call handle otherwise
 *
 * @todo	Does direction really make sense here? Primary use is to
 *		request a new outgoing call...
 */
struct Q931_Call *Q931CallNew(Q931_TrunkInfo_t *trunk)
{
	struct Q931_Call *call = NULL;
	L3INT crv = 0;

	/* find free call handle */
	call = Q931FindFreeCall(trunk);
	if (!call) {
		/* No free call handle found */
		return NULL;
	}
	call->InUse = 1;

	Q931Log(trunk, Q931_LOG_INFO, "Found a free call, trying to find a new CRV\n");

	/* get new call reference */
	crv = Q931GetUniqueCRV(trunk);
	if (!crv) {
		goto crverror;
	}

	Q931Log(trunk, Q931_LOG_INFO, "Using crv %d for new call\n", crv);

	/* set parameters & state */
	call->Direction = Q931_DIRECTION_OUTBOUND;
	call->CRV   = crv;
	call->Trunk = trunk;		/* yah, sucks, but makes the api easier */

	Q931CallSetState(call, (trunk->NetUser == Q931_NT) ? Q931_N0 : Q931_U0);

	return call;

crverror:
	call->InUse = 0;
	return NULL;
}

struct Q931_Call *Q931CallNewIncoming(Q931_TrunkInfo_t *trunk, const L3INT crv)
{
	struct Q931_Call *call = NULL;

	/* find free call handle */
	call = Q931FindFreeCall(trunk);
	if (!call) {
		/* No free call handle found */
		return NULL;
	}
	call->InUse = 1;

	/* set parameters & state */
	call->Direction = Q931_DIRECTION_INBOUND;
	call->CRV   = crv;
	call->Trunk = trunk;		/* yah, sucks, but makes the api easier */

	Q931CallSetState(call, (trunk->NetUser == Q931_NT) ? Q931_N0 : Q931_U0);

	return call;
}


/**
 * Q931CallRelease
 * @brief	Release call handle and CRV
 * @param[in]	call	The call to release
 * @return	0 on success, < 0 otherwise
 *
 * @note	Use only after the call has ended
 */
L3INT Q931CallRelease(struct Q931_Call *call)
{
	L3INT state;

	assert(call);

	if (Q931CallIsGlobal(call))
		return 0;

	/* add some sanity checks here */
	state = Q931CallGetState(call);

	if (state != Q931_U0 || state != Q931_N0) {
		/* error */
		return -1;
	}

	/* reset completely */
	memset(call, 0, sizeof(struct Q931_Call));

	return 0;
}


/*
 * Accessor functions for private data
 */
Q931_API void Q931CallSetPrivate(struct Q931_Call *call, const void *priv)
{
	assert(call);

	call->pvt = (void *)priv;
}

Q931_API void * Q931CallGetPrivate(const struct Q931_Call *call)
{
	assert(call);

	return call->pvt;
}

Q931_API Q931_TrunkInfo_t * Q931CallGetTrunk(const struct Q931_Call *call)
{
	assert(call);
	assert(call->Trunk);

	return call->Trunk;
}

Q931_API L3INT Q931CallGetCRV(const struct Q931_Call *call)
{
	assert(call);

	return call->CRV;
}

Q931_API L3INT Q931CallGetDirection(const struct Q931_Call *call)
{
	assert(call);

	return call->Direction;
}

Q931_API L3BOOL Q931CallIsOutgoing(const struct Q931_Call *call)
{
	assert(call);

	return call->Direction == Q931_DIRECTION_OUTBOUND;
}


/*
 *
 */
struct Q931CallStateName
{
	L3INT id;
	const L3CHAR *name;

} Q931CallStateNames[] = {
	{ Q931_U0, "Null" },
	{ Q931_U1, "Call initiated" },
	{ Q931_U2, "Overlap sending" },
	{ Q931_U3, "Outgoing call proceeding" },
	{ Q931_U4, "Call delivered" },
	{ Q931_U6, "Call present" },
	{ Q931_U7, "Call received" },
	{ Q931_U8, "Connect request" },
	{ Q931_U9, "Incoming call proceeding" },
	{ Q931_U10, "Active" },
	{ Q931_U11, "Disconnect request" },
	{ Q931_U12, "Disconnect indication" },
	{ Q931_U15, "Suspend request" },
	{ Q931_U17, "Resume request" },
	{ Q931_U19, "Release request" },
	{ Q931_U25, "Overlap receiving" },

	{ Q931_N0, "Null" },
	{ Q931_N1, "Call initiated" },
	{ Q931_N2, "Overlap sending" },
	{ Q931_N3, "Outgoing call proceeding" },
	{ Q931_N4, "Call delivered" },
	{ Q931_N6, "Call present" },
	{ Q931_N7, "Call received" },
	{ Q931_N8, "Connect request" },
	{ Q931_N9, "Incoming call proceeding" },
	{ Q931_N10, "Active" },
	{ Q931_N11, "Disconnect request" },
	{ Q931_N12, "Disconnect indication" },
	{ Q931_N15, "Suspend request" },
	{ Q931_N17, "Resume request" },
	{ Q931_N19, "Release request" },
	{ Q931_N22, "N22" },			/* whatever it's name is... */
	{ Q931_N25, "Overlap receiving" },

	{ 0, NULL }
};

Q931_API const char * Q931CallGetStateName(const struct Q931_Call *call)
{
	struct Q931CallStateName *ptr = Q931CallStateNames;
	L3INT state;

	assert(call);

	state = Q931CallGetState(call);

	while (ptr->name) {
		if (ptr->id == state) {
			return ptr->name;
		}
		ptr++;
	}
	return "Unknown/Invalid";
}

struct Q931CallEventName
{
	L3INT id;
	const L3CHAR *name;

} Q931CallEventNames[] = {
	{ Q931_EVENT_SETUP_INDICATION,          "Setup indication" },
	{ Q931_EVENT_SETUP_CONFIRM,             "Setup confirm" },
	{ Q931_EVENT_SETUP_COMPLETE_INDICATION, "Setup complete indication" },
	{ Q931_EVENT_MORE_INFO_INDICATION,      "More information indication" },
	{ Q931_EVENT_REJECT_INDICATION,         "Reject indication" },
	{ Q931_EVENT_PROCEEDING_INDICATION,     "Proceeding indication" },
	{ Q931_EVENT_ALERTING_INDICATION,       "Alerting indication" },
	{ Q931_EVENT_PROGRESS_INDICATION,       "Progress indication" },
	{ Q931_EVENT_NOTIFY_INDICATION,         "Notify indication" },
	{ Q931_EVENT_RELEASE_INDICATION,        "Release indication" },
	{ Q931_EVENT_SUSPEND_CONFIRM,           "Suspend confirm" },
	{ Q931_EVENT_RESUME_CONFIRM,            "Resume confirm" },
	{ Q931_EVENT_DISCONNECT_INDICATION,     "Disconnect indication" },
	{ Q931_EVENT_INFORMATION_INDICATION,    "Information indication" },
	{ Q931_EVENT_STATUS_INDICATION,         "Status indication" },
	{ Q931_EVENT_RESTART_CONFIRM,           "Restart confirm" },

	/* Requests (L4 -> L3) (?) */
	{ Q931_EVENT_SETUP_RESPONSE,      "Setup response" },
	{ Q931_EVENT_REJECT_REQUEST,      "Reject request" },
	{ Q931_EVENT_INFORMATION_REQUEST, "Information request" },
	{ Q931_EVENT_ALERTING_REQUEST,    "Alerting request" },
	{ Q931_EVENT_PROGRESS_REQUEST,    "Progress request" },
	{ Q931_EVENT_PROCEEDING_REQUEST,  "Proceeding request" },
	{ Q931_EVENT_NOTIFY_REQUEST,      "Notify request" },
	{ Q931_EVENT_SUSPEND_REQUEST,     "Suspend request" },
	{ Q931_EVENT_DISCONNECT_REQUEST,  "Disconnect request" },
	{ Q931_EVENT_RESTART_REQUEST,     "Restart request" },

	/* misc events (L2/L3 -> L4) */
	{ Q931_EVENT_TIMEOUT_INDICATION,    "Timeout indication" },
	{ Q931_EVENT_DL_FAILURE_INDICATION, "Datalink Layer failure indication" },

	{ 0, NULL },
};

Q931_API const char * Q931CallGetEventName(const L3INT event)
{
	struct Q931CallEventName *ptr = Q931CallEventNames;

	while (ptr->name) {
		if (ptr->id == event) {
			return ptr->name;
		}
		ptr++;
	}
	return "Unknown";
}

struct Q931CallTimerName
{
	L3INT id;
	const L3CHAR *name;

} Q931CallTimerNames[] = {
	{ Q931_TIMER_T300, "T300" },
	{ Q931_TIMER_T301, "T301" },
	{ Q931_TIMER_T302, "T302" },
	{ Q931_TIMER_T303, "T303" },
	{ Q931_TIMER_T304, "T304" },
	{ Q931_TIMER_T305, "T305" },
	{ Q931_TIMER_T306, "T306" },
	{ Q931_TIMER_T307, "T307" },
	{ Q931_TIMER_T308, "T308" },
	{ Q931_TIMER_T309, "T309" },
	{ Q931_TIMER_T310, "T310" },
	{ Q931_TIMER_T311, "T311" },
	{ Q931_TIMER_T312, "T312" },
	{ Q931_TIMER_T313, "T313" },
	{ Q931_TIMER_T314, "T314" },
	{ Q931_TIMER_T315, "T315" },
	{ Q931_TIMER_T316, "T316" },
	{ Q931_TIMER_T317, "T317" },
	{ Q931_TIMER_T318, "T318" },
	{ Q931_TIMER_T319, "T319" },
	{ Q931_TIMER_T320, "T320" },
	{ Q931_TIMER_T321, "T321" },
	{ Q931_TIMER_T322, "T322" },

	{ 0, NULL }
};

Q931_API const char * Q931CallGetTimerName(const L3USHORT timer)
{
	struct Q931CallTimerName *ptr = Q931CallTimerNames;

	while (ptr->name) {
		if (ptr->id == timer) {
			return ptr->name;
		}
		ptr++;
	}
	return "Unknown";
}



Q931_API struct Q931_Call * Q931GetCallByCRV(const Q931_TrunkInfo_t *trunk, const L3INT crv)
{
	struct Q931_Call *call = NULL;
	int i = 0;

	assert(trunk);

	if (!crv) {
		/* global call context */
		return (struct Q931_Call *)&trunk->call[0];
	}

	/* call[0] reserved for global CRV */
	for (i = 1; i < Q931MAXCALLPERTRUNK; i++) {
		call = (struct Q931_Call *)&trunk->call[i];

		if (call->InUse && call->CRV == crv) {
			return call;
		}
	}
	return NULL;
}

struct Q931_Call *Q931GetCallByIndex(const Q931_TrunkInfo_t *trunk, const L3INT idx)
{
	struct Q931_Call *call = NULL;

	assert(trunk);

	call = (struct Q931_Call *)&trunk->call[idx];
	if (call->InUse) {
		return call;
	}
	return NULL;
}

Q931_API struct Q931_Call * Q931GetGlobalCall(const Q931_TrunkInfo_t *trunk)
{
	return (struct Q931_Call *)&trunk->call[0];
}

Q931_API L3BOOL Q931CallIsGlobal(const struct Q931_Call *call)
{
	return (call->CRV == 0);
}

/**
 * Q931CallStartTimer
 * @brief	Start a call timer
 * @param	call	Q.931 call
 * @param	timer	ID of timer
 */
void Q931CallStartTimer(struct Q931_Call *call, L3USHORT timer)
{
	Q931_TrunkInfo_t *trunk = NULL;
	L3ULONG duration;

	assert(call);

	trunk = Q931CallGetTrunk(call);
	assert(trunk);

	duration = Q931DialectGetTimeout(trunk->Dialect, timer);

	if (duration) {
		call->Timer    = Q931GetTime() + duration;
		call->TimerID  = timer;
		call->TimerCnt = 0;

		Q931Log(Q931CallGetTrunk(call), Q931_LOG_NOTICE, "Starting timer %d (timeout: %d ms) for call %d [0x%x]\n", timer, duration, Q931CallGetCRV(call), Q931CallGetCRV(call));
	}
}

/**
 * Q931CallRestartTimer
 * @brief	Restart a call timer
 * @param	call	Q.931 call
 * @param	timer	ID of timer
 * @note	Does not reset the expire counter
 */
void Q931CallRestartTimer(struct Q931_Call *call, L3USHORT timer)
{
	Q931_TrunkInfo_t *trunk = NULL;
	L3ULONG duration;

	assert(call);

	trunk = Q931CallGetTrunk(call);
	assert(trunk);

	duration = Q931DialectGetTimeout(trunk->Dialect, timer);

	if (duration) {
		call->Timer    = Q931GetTime() + duration;
		call->TimerID  = timer;

		Q931Log(Q931CallGetTrunk(call), Q931_LOG_NOTICE, "Restarting timer %d (timeout: %d ms) for call %d [0x%x]\n", timer, duration, Q931CallGetCRV(call), Q931CallGetCRV(call));
	}
}

/**
 * Q931CallStopTimer
 * @brief	Stop a call timer
 * @param	call	Q.931 call
 * @param	timer	ID of timer
 */
void Q931CallStopTimer(struct Q931_Call *call, L3USHORT timer)
{
	assert(call);

        if (call->TimerID == timer) {
                call->TimerID = Q931_TIMER_NONE;

		Q931Log(Q931CallGetTrunk(call), Q931_LOG_NOTICE, "Stopping timer %d for call %d [0x%x]\n", timer, Q931CallGetCRV(call), Q931CallGetCRV(call));
	}
}

/**
 * Q931CallStopAllTimers
 * @brief	Stop all call timers
 * @param	call	Q.931 call
 */
void Q931CallStopAllTimers(struct Q931_Call *call)
{
	assert(call);

	call->TimerID = Q931_TIMER_NONE;

	Q931Log(Q931CallGetTrunk(call), Q931_LOG_NOTICE, "Stopped all timers for call %d [0x%x]\n", Q931CallGetCRV(call), Q931CallGetCRV(call));
}

/**
 * Q931CallGetTimerID
 * @brief	Return ID of the currently active timer
 * @param	call	Q.931 call
 * @return	ID of the active timer or Q931_TIMER_NONE if no timer is active
 */
Q931_API L3USHORT Q931CallGetTimerID(const struct Q931_Call *call)
{
	return call->TimerID;
}

/**
 * Q931CallGetTimerExpireCount
 * @brief	Return number of times the current timer has expired
 * @param	call	Q.931 call
 * @return	counter value
 * @note	Only used by the timer callbacks to fetch the counter value
 *		(Q931CallStartTimer resets the counter)
 */
Q931_API L3UCHAR Q931CallGetTimerExpireCount(const struct Q931_Call *call)
{
	return call->TimerCnt;
}


/**
 * Q931CallIncrementTimerExpireCount
 * @brief	Increment expire counter for current timer
 * @param	call	Q.931 call
 * @note	Only used by Q931TimerTick
 */
void Q931CallIncrementTimerExpireCount(struct Q931_Call *call)
{
	call->TimerCnt++;
}

/*
 * Event
 */

/**
 * @brief	Set call event callback and (optional) private data
 * @param	trunk	Q.931 Trunk handle
 * @param	cb	Call event callback function
 * @param	priv	Optional private data for callback function
 */
Q931_API void Q931SetCallEventCB(Q931_TrunkInfo_t *trunk, const Q931CallEventCB_t cb, const void *priv)
{
	assert(trunk);
	assert(cb);

	trunk->CallEventCBProc = (Q931CallEventCB_t)cb;
	trunk->PrivateDataCallEventCB = (void *)priv;
}

/**
 * @brief	Send call event to Layer 4
 * @param	call	Q.931 Call handle
 * @param	event	Call event to send
 */
void Q931CallSendEvent(const struct Q931_Call *call, const struct Q931_CallEvent *event)
{
	Q931_TrunkInfo_t *trunk = Q931CallGetTrunk(call);

	assert(trunk);
	assert(event);

	if (!event || event->id == Q931_EVENT_NONE || event->type == Q931_EVENT_TYPE_NONE)
		return;

	if (trunk->CallEventCBProc) {
		Q931Log(trunk, Q931_LOG_DEBUG, "Sending call event to Layer4 [call: %d, id: %d, type: %d]\n", Q931CallGetCRV(call), event->id, event->type);

		trunk->CallEventCBProc((struct Q931_Call *)call, (struct Q931_CallEvent *)event, trunk->PrivateDataCallEventCB);
	}
}

/**
 * @brief	Initialize call event
 */
void Q931CallInitEvent(struct Q931_CallEvent *event)
{
	memset(event, 0, sizeof(struct Q931_CallEvent));
}

/************************************************************************************************
 *  V2 - Enqueue events in the call handle and send them to layer 4, after processing everything
 ************************************************************************************************/

/**
 * @brief	Get new free event slot from call handle
 */
struct Q931_CallEvent *Q931CallNewEvent(struct Q931_Call *call)
{
	assert(call);

	if (call->numevents == Q931_CALL_MAX_EVENTS)
		return NULL;

	return &call->events[call->numevents];
}

/**
 * @brief	"Enqueue" call event
 * @note	Dummy function, actually just increases the call's event counter
 */
void Q931CallQueueEvent(struct Q931_Call *call, struct Q931_CallEvent *event)
{
	assert(call);

	call->numevents++;
}

/**
 * @brief	Clear list of queued events
 */
void Q931CallFlushEvents(struct Q931_Call *call)
{
	assert(call);

	memset(call->events, 0, Q931_CALL_MAX_EVENTS * sizeof(struct Q931_CallEvent));
	call->numevents = 0;
}

/**
 * @brief	Send queued call events to layer 4
 */
void Q931CallSendQueuedEvents(struct Q931_Call *call)
{
	int i;

	assert(call);

	for (i = 0; i < call->numevents; i++) {
		Q931CallSendEvent(call, &call->events[i]);
	}

	Q931CallFlushEvents(call);
}

/**
 * @brief	Get event type
 */
Q931_API L3INT Q931CallEventGetType(const struct Q931_CallEvent *event)
{
	return event->type;
}

/**
 * @brief	Get event id
 */
Q931_API L3INT Q931CallEventGetId(const struct Q931_CallEvent *event)
{
	return event->id;
}

/**
 * @brief	Get pointer to event data
 * @note	You'll have to do the right casting yourself
 */
Q931_API void * Q931CallEventGetData(const struct Q931_CallEvent *event)
{
	switch (event->type) {
	case Q931_EVENT_TYPE_TIMER:
		return (void *)&event->data.timer;

	case Q931_EVENT_TYPE_MESSAGE:
		return (void *)&event->data.message;

	default:
		return NULL;
	}
}


/************************************************************************************************
 * Debug information
 ************************************************************************************************/

#include <stdio.h>

#define TMP_BUFSIZE	4096
#define TMP_PUTS(fmt) \
	strncat(tmp + offset, fmt, size - offset - 1);	\
	offset += strlen(tmp + offset);
#define TMP_APPEND(fmt, args...) \
	offset += snprintf(tmp + offset, size - offset, fmt, ##args);

/**
 * Dump call information
 */
static L3INT __Q931CallDumpInformation(const struct Q931_Call *call, char *tmp, const int size)
{
	int offset = 0;
	int x;

	TMP_APPEND("\tGlobal call:   %s\n",      Q931CallIsGlobal(call) ? "Yes (CRV: 0)" : "No");
	TMP_APPEND("\tDirection:     %s\n",      Q931CallIsOutgoing(call) ? "Outbound" : "Inbound");
	TMP_APPEND("\tState:         %d '%s'\n", Q931CallGetState(call), Q931CallGetStateName(call));
	TMP_APPEND("\tL4 User data:  %p\n",      Q931CallGetPrivate(call));

	/* Timer information */
	if (call->TimerID != Q931_TIMER_NONE) {
		TMP_APPEND("\tTimer:         '%s', time left: %5lu ms, expired: %d time(s)\n",
				Q931CallGetTimerName(call->TimerID), (call->Timer - Q931GetTime()),
				Q931CallGetTimerExpireCount(call));
	}

	/* Dump events queued */
	TMP_APPEND("\tEvents queued: %d\n", call->numevents);

	for (x = 0; x < call->numevents; x++) {
		const struct Q931_CallEvent *event = &call->events[x];

		TMP_APPEND("\t\t<%2d> %s\n", x, Q931CallGetEventName(event->id));
	}

	return offset;
}

/**
 * Dump information about call
 * @param	call	Call to dump
 * @return	0 on success
 */
Q931_API L3INT Q931CallDump(const struct Q931_Call *call)
{
	Q931_TrunkInfo_t *trunk = Q931CallGetTrunk(call);
	char tmp[TMP_BUFSIZE >> 2] = { 0 };
	int offset = 0;
	int size   = sizeof(tmp);

	/* header */
	TMP_PUTS("Call Information:\n");
	TMP_APPEND("============================== Call: %3d (%#02x) =============================\n", Q931CallGetCRV(call), Q931CallGetCRV(call));

	offset += __Q931CallDumpInformation(call, tmp + offset, size - offset);

	/* footer */
	TMP_PUTS("---------------------------------------------------------------------------------\n");

	Q931Log(trunk, Q931_LOG_DEBUG, "%s\n", tmp);
	return Q931E_NO_ERROR;
}

/**
 * Dump information about all active calls
 * @param	trunk	Q.931 Trunk handle
 * @return	0 on success
 */
Q931_API L3INT Q931DumpAllCalls(const Q931_TrunkInfo_t *trunk)
{
	struct Q931_Call *call;
	char tmp[TMP_BUFSIZE] = { 0 };
	int offset = 0;
	int count  = 0;
	int size   = sizeof(tmp);
	int x;

	/* header */
	TMP_PUTS("Information of all active calls:\n");
	TMP_PUTS("==================================== Calldump ===================================\n");

	for (x = 0; x < Q931MAXCALLPERTRUNK; x++) {
		if (!(call = Q931GetCallByIndex(trunk, x))) {
			continue;
		}

		/* skip inactive calls */
		if ((Q931CallGetState(call) & 0xff) == Q931_U0) {
			continue;
		}

		/* per-call header */
		TMP_APPEND("* CRV: %3d (%#02x)\n", Q931CallGetCRV(call), Q931CallGetCRV(call));

		/* Add call's information */
		offset += __Q931CallDumpInformation(call, tmp + offset, size - offset);
		count++;
	}

	/* footer */
	TMP_APPEND("-----------------------------------------------------------------[ %2d call(s) ]--\n", count);

	Q931Log(trunk, Q931_LOG_DEBUG, "%s\n", tmp);
	return Q931E_NO_ERROR;
}

