#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include "Q931.h"
#include "Q931priv.h"

/*
 * Internal data structures
 */
struct Q931Dialect {
	L3INT id;
	L3INT flags;				/*!< flags */
	L3UCHAR mode;				/*!< TE or NT mode */
	L3UCHAR	codesets;			/*!< bitmap of allowed codesets */

	const char *name;			/*!< name */

	/* Tables */
	struct Q931MessageIE     *mesIEMap;		/*!< Message -> IE map */
	Q931State		 states[Q931MAXSTATE];	/*!< Allowed states */

	/* Callbacks */
	q931proc_func_t		*proc_unknown_cb;
	q931proc_func_t		*proc_invalid_cb;

	q931proc_func_t		*proc_cb[Q931MAXMES];
	q931umes_func_t		*umes_cb[Q931MAXMES];
	q931pmes_func_t		*pmes_cb[Q931MAXMES];

	q931uie_func_t		*uie_cb[Q931MAXIE];
	q931pie_func_t		*pie_cb[Q931MAXIE];
	q931die_func_t		*die_cb[Q931MAXIE];

	q931timeout_func_t	*timeout_cb[Q931MAXTIMER];
	q931timer_t		 timeout[Q931MAXTIMER];
};

/*
 * Private global
 */
static struct Q931Dialect dialects[Q931MAXDLCT];
static L3BOOL initialized = L3FALSE;

/*
 *
 */
void Q931DialectRegistryInit(void)
{
	int x;

	if (initialized) {
		return;
	}

	memset(dialects, 0, sizeof(dialects));

	for (x = 0; x < Q931MAXDLCT; x++) {
		dialects[x].id = -1;
	}

	initialized = L3TRUE;
}

void Q931DialectRegister(L3UCHAR id, L3INT mode, void (* create_func)(struct Q931Dialect *dialect))
{
	int x;

	for (x = 0; x < Q931MAXDLCT; x++) {
		if (dialects[x].id == -1) {
			continue;
		}
		if (dialects[x].id == id && dialects[x].mode == mode) {
			return;
		}
	}

	for (x = 0; x < Q931MAXDLCT; x++) {
		if (dialects[x].id == -1) {
			dialects[x].id   = id;
			dialects[x].mode = mode;
			create_func(&dialects[x]);
			break;
		}
	}
}

struct Q931Dialect *Q931DialectGet(L3UCHAR id, L3INT mode)
{
	int x;

	for (x = 0; x < Q931MAXDLCT; x++) {
		if (dialects[x].id == -1) {
			continue;
		}
		if (dialects[x].id == id && dialects[x].mode == mode) {
			return &dialects[x];
		}
	}
	return NULL;
}

void Q931DialectSetMesIEMap(struct Q931Dialect *dialect, struct Q931MessageIE *map)
{
	assert(dialect);
	assert(map);

	dialect->mesIEMap = map;
}

struct Q931MessageIE* Q931DialectGetMesIEMap(struct Q931Dialect *dialect)
{
	assert(dialect);
	assert(dialect->mesIEMap);

	return dialect->mesIEMap;
}

/*
 * Dialect codesets
 */
void Q931DialectAddCodeset(struct Q931Dialect *dialect, L3UCHAR codeset)
{
	assert(dialect);

	if (codeset == Q931_CODESET_ALL) {
		dialect->codesets = Q931_CODESET_ALL;
	} else {
		dialect->codesets |= (1 << codeset);
	}
}

L3BOOL Q931DialectIsValidCodeset(struct Q931Dialect *dialect, L3UCHAR codeset)
{
	assert(dialect);

	return ((dialect->codesets & codeset) != 0);
}

/*
 * Dialect flags
 */
void Q931DialectSetFlag(struct Q931Dialect *dialect, L3INT flag)
{
	assert(dialect);

	dialect->flags |= flag;
}

L3BOOL Q931DialectIsSetFlag(struct Q931Dialect *dialect, L3INT flag)
{
	assert(dialect);

	return ((dialect->flags & flag) != 0);
}

void Q931DialectSetMesProc(struct Q931Dialect *dialect, L3UCHAR mes, q931proc_func_t *Q931ProcFunc, q931umes_func_t *Q931UmesFunc, q931pmes_func_t *Q931PmesFunc)
{
	assert(dialect);

	dialect->proc_cb[mes] = Q931ProcFunc;
	dialect->umes_cb[mes] = Q931UmesFunc;
	dialect->pmes_cb[mes] = Q931PmesFunc;
}

/*
 * IE cb table
 */
void Q931DialectSetIEProc(struct Q931Dialect *dialect, L3UCHAR iec, q931pie_func_t *PieProc, q931uie_func_t *UieProc, q931die_func_t *DieProc)
{
	assert(dialect);

	dialect->uie_cb[iec] = UieProc;
	dialect->pie_cb[iec] = PieProc;
	dialect->die_cb[iec] = DieProc;
}

/*
 * Timer
 */
void Q931DialectSetTimerProc(struct Q931Dialect *dialect, L3UCHAR timer, q931timeout_func_t *Q931TimeoutProc)
{
	assert(dialect);

	dialect->timeout_cb[timer] = Q931TimeoutProc;
}

void Q931DialectSetTimerProcAll(struct Q931Dialect *dialect, q931timeout_func_t *Q931TimeoutProc)
{
	int i;

	assert(dialect);

	for (i = 0; i < Q931MAXTIMER; i++) {
		dialect->timeout_cb[i] = Q931TimeoutProc;
	}
}

void Q931DialectSetTimeout(struct Q931Dialect *dialect, L3UCHAR timer, q931timer_t timeout)
{
	assert(dialect);

	dialect->timeout[timer] = timeout;
}

/*
 * State table
 */
void Q931DialectAddStateEntry(struct Q931Dialect *dialect, L3INT state, L3INT mes, L3UCHAR flags)
{
	int x;

	assert(dialect);
	assert(dialect->states);

	for (x = 0; x < Q931MAXSTATE; x++) {
		if (dialect->states[x].Message == 0) {
			dialect->states[x].State   = state;
			dialect->states[x].Message = mes;
			dialect->states[x].Flags   = flags;
			break;
		}
	}
}

L3BOOL Q931DialectIsEventLegal(struct Q931Dialect *dialect, L3INT state, L3INT mes, L3UCHAR dirflags)
{
	int x;

	assert(dialect);
	assert(dialect->states);

	/* strip other flags */
	dirflags &= MSGF_FROM_BOTH;

	for (x = 0; x < Q931MAXSTATE; x++) {
		if(!dialect->states[x].Message) {
			break;
		}
		if ((dialect->states[x].State & state) == state &&
		     dialect->states[x].Message == mes &&
		     dialect->states[x].Flags & dirflags) {
			return L3TRUE;
		}
	}
	return L3FALSE;
}


/*
 * Callback helpers
 */
void Q931DialectSetUnknownMesProc(struct Q931Dialect *dialect, q931proc_func_t proc)
{
	dialect->proc_unknown_cb = proc;
}

q931proc_func_t *Q931DialectGetUnknownMesProc(const struct Q931Dialect *dialect)
{
	return dialect->proc_unknown_cb;
}

void Q931DialectSetInvalidMesProc(struct Q931Dialect *dialect, q931proc_func_t proc)
{
	dialect->proc_invalid_cb = proc;
}

q931proc_func_t *Q931DialectGetInvalidMesProc(const struct Q931Dialect *dialect)
{
	return dialect->proc_invalid_cb;
}


q931proc_func_t *Q931DialectGetMesProc(const struct Q931Dialect *dialect, L3UCHAR type)
{
	return dialect->proc_cb[type];
}

q931umes_func_t *Q931DialectGetUmesProc(const struct Q931Dialect *dialect, L3UCHAR type)
{
	return dialect->umes_cb[type];
}

q931pmes_func_t *Q931DialectGetPmesProc(const struct Q931Dialect *dialect, L3UCHAR type)
{
	return dialect->pmes_cb[type];
}


q931uie_func_t *Q931DialectGetUieProc(const struct Q931Dialect *dialect, L3UCHAR type)
{
	return dialect->uie_cb[type];
}

q931pie_func_t *Q931DialectGetPieProc(const struct Q931Dialect *dialect, L3UCHAR type)
{
	return dialect->pie_cb[type];
}

q931die_func_t *Q931DialectGetDieProc(const struct Q931Dialect *dialect, L3UCHAR type)
{
	return dialect->die_cb[type];
}


q931timeout_func_t *Q931DialectGetTimerProc(const struct Q931Dialect *dialect, L3UCHAR timer)
{
	return dialect->timeout_cb[timer];
}

q931timer_t Q931DialectGetTimeout(const struct Q931Dialect *dialect, L3UCHAR timer)
{
	return dialect->timeout[timer];
}

/*
 * Toolfunctions
 */
L3INT Q931DialectGetID(const struct Q931Dialect *dialect)
{
	return dialect->id;
}

Q931Dialect_t Q931DialectGetType(const struct Q931Dialect *dialect)
{
	return (dialect->id & (~Q931_NT));
}

Q931NetUser_t Q931DialectGetMode(const struct Q931Dialect *dialect)
{
	return (dialect->id & Q931_NT);
}

const char * Q931DialectGetName(const struct Q931Dialect *dialect)
{
	return dialect->name;
}

void Q931DialectSetName(struct Q931Dialect *dialect, const char *name)
{
	dialect->name = name;
}
