/*****************************************************************************

  FileName:	 Q931ie.c

  Contents:	 Information Element Pack/Unpack functions.

		These functions will pack out a Q931 message from the bit
		packed original format into structs that are easier to process
		and pack the same structs back into bit fields when sending
		messages out.

		The messages contains a short for each possible IE. The MSB
		bit flags the precense of an IE, while the remaining bits
		are the offset into a buffer to find the actual IE.

		Each IE are supported by 3 functions:

		Q931Pie_XXX	 Pack struct into Q.931 IE
		Q931Uie_XXX	 Unpack Q.931 IE into struct
		Q931InitIEXXX   Initialize IE (see Q931api.c).

  Dialect Note: This file will only contain standard DSS1 IE. Other IE as
		used in QSIG, NI2, Q.932 etc are located in separate files.

		See	q931.h for description.

  License/Copyright:

  Copyright (c) 2007, Jan Vidar Berger, Case Labs, Ltd. All rights reserved.
  email:janvb@caselaboratories.com

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are
  met:

	* Redistributions of source code must retain the above copyright notice,
	  this list of conditions and the following disclaimer.
	* Redistributions in binary form must reproduce the above copyright notice,
	  this list of conditions and the following disclaimer in the documentation
	  and/or other materials provided with the distribution.
	* Neither the name of the Case Labs, Ltd nor the names of its contributors
	  may be used to endorse or promote products derived from this software
	  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.
*****************************************************************************/
#include <stdio.h>
#include <string.h>

#if defined(__WIN32__) || defined(__WIN64__)
#else
#include <netinet/in.h>	/* X.213 NPAS, IANA IPC - RFC4548 */
#include <arpa/inet.h>	/* X.213 NPAS, IANA IPC - RFC4548 */
#endif

#include "Q931.h"
#include "Q931priv.h"

#include "Q850.h"
#include "X213.h"

#include "utils/common.h"
#include "utils/strstream.h"

/****************************************************************************
 * Helper functions for tracing output
 ****************************************************************************/

static const char *q931_network_id_type(const unsigned char type)
{
	switch (type) {
	case 0x00:
		return "User specified";
	case 0x02:
		return "National network identification";
	case 0x03:
		return "International network identification";
	default:
		return "reserved";
	}
}

static const char *q931_network_id_plan(const unsigned char plan)
{
	switch (plan) {
	case 0x00:
		return "Unknown";
	case 0x01:
		return "Carrier Identification Code (CIC)";
	case 0x03:
		return "Data network identification code (X.121)";
	default:
		return "reserved";
	}
}

static const char *q931_signal_desc(unsigned char value)
{
	switch (value) {
	case 0x00:
		return "Dial tone on";
	case 0x01:
		return "Ring back tone on";
	case 0x02:
		return "Intercept tone on";
	case 0x03:
		return "Network congestion tone on";
	case 0x04:
		return "Busy tone on";
	case 0x05:
		return "Confirm tone on";
	case 0x06:
		return "Answer tone on";
	case 0x07:
		return "Call waiting tone on";
	case 0x08:
		return "Off-hook warning tone on";
	case 0x09:
		return "Pre-emption tone on";
	case 0x3f:
		return "Tones off";

	case 0x40:
		return "Alerting on, pattern 0";
	case 0x41:
		return "Alerting on, pattern 1";
	case 0x42:
		return "Alerting on, pattern 2";
	case 0x43:
		return "Alerting on, pattern 3";
	case 0x44:
		return "Alerting on, pattern 4";
	case 0x45:
		return "Alerting on, pattern 5";
	case 0x46:
		return "Alerting on, pattern 6";
	case 0x47:
		return "Alerting on, pattern 7";
	case 0x4f:
		return "Alerting off";

	default:
		return "reserved";
	}
}

static const char *q931_progdesc_type(unsigned char type)
{
	switch (type) {
	case 0x01:
		return "Call is not end-to-end ISDN, progress in-band";
	case 0x02:
		return "Destination address is non-ISDN";
	case 0x03:
		return "Origination address is non-ISDN";
	case 0x04:
		return "Call has returned to the ISDN";
	case 0x05:
		return "Interworking has occured";
	case 0x08:
		return "In-band information or an appropriate pattern is now available";
	default:
		return "reserved";
	}
}

static const char *q931_subaddr_type(unsigned char type)
{
	switch (type) {
	case 0x00:
		return "NSAP";
	case 0x02:
		return "user specified";
	default:
		return "reserved";
	}
}

static const char *q931_coding_type(unsigned char type)
{
	switch (type) {
	case 0x00:
		return "ITU-T";
	case 0x01:
		return "ISO/IEC";
	case 0x02:
		return "national standard";
	case 0x03:
		return "specific to location";
	default:
		return "unknown/invalid";
	}
}

static const char *q931_location_type(unsigned char type)
{
	switch (type) {
	case 0x00:
		return "user";
	case 0x01:
		return "private network serving the local user";
	case 0x02:
		return "public network serving the local user";
	case 0x03:
		return "transit network";
	case 0x04:
		return "public network serving the remote user";
	case 0x05:
		return "private network serving the remote user";
	case 0x07:
		return "international network";
	case 0x0a:
		return "network beyond interworking point";
	case 0x0c:
	case 0x0d:
	case 0x0e:
	case 0x0f:
		return "reserved";
	default:
		return "invalid/unknown";
	}
}

static const char *q931_ton_type(const unsigned char ton)
{
	switch (ton) {
	case 0x00:
		return "Unknown";
	case 0x01:
		return "International number";
	case 0x02:
		return "National number";
	case 0x03:
		return "Network specific number";
	case 0x04:
		return "Subscriber number";
	case 0x06:
		return "Abbreviated number";
	case 0x07:
		return "Reserved for extension";
	default:
		return "reserved";
	}
}

static const char *q931_numbering_plan_type(const unsigned char plan)
{
	switch (plan) {
	case 0x00:
		return "Unknown";
	case 0x01:
		return "ISDN/telephony numbering plan (E.164)";
	case 0x03:
		return "Data numbering plan (X.121)";
	case 0x04:
		return "Telext numbering plan (F.69)";
	case 0x08:
		return "National standard numbering plan";
	case 0x09:
		return "Private numbering plan";
	case 0x0f:
		return "Reserved for extension";
	default:
		return "reserved";
	}
}

static const char *q931_presentation_type(const unsigned char type)
{
	switch (type) {
	case 0x00:
		return "Presentation allowed";
	case 0x01:
		return "Presentation restricted";
	case 0x02:
		return "Number not available due to interworking";
	case 0x03:
		return "Reserved";
	default:
		return "unknown/invalid";
	}
}

static const char *q931_screening_type(const unsigned char type)
{
	switch (type) {
	case 0x00:
		return "User-provided, not screened";
	case 0x01:
		return "User-provided, verified and passed";
	case 0x02:
		return "User-provided, verified and failed";
	case 0x03:
		return "Network provided";
	default:
		return "unknown/invalid";
	}
}

/********************************************************************************************************************************
 * CRV Decoding
 ********************************************************************************************************************************/

L3USHORT Q931Uie_CRV(Q931_TrunkInfo_t *trunk, L3UCHAR *IBuf, L3UCHAR *OBuf, L3INT *IOff, L3INT *OOff)
{
	L3USHORT CRV = 0;
	L3INT Octet = *IOff;
	L3INT l = IBuf[Octet++];

	if (l == 1) {	/* One octet CRV */
		CRV = IBuf[Octet++] & 0x7F;
	}
	else if (l == 2) {	/* two octet CRV */
		CRV  = (IBuf[Octet++] & 0x7f) << 8;
		CRV |=  IBuf[Octet++];
	}
	else {
		/* Long CRV is not used, so we skip this */
		/* TODO: is it right to set to 0 here? */
		CRV = 0;
		Octet += l;
	}

	*IOff = Octet;
	return CRV;
}


/*****************************************************************************

  Macro:		Q931MoreIE

  Description:  Local helper macro detecting if there is more IE space left
		based on the 3 standard parameters Octet, Off and IESpace.
		This can be used to test if the IE is completed to avoid
		that the header of the next IE is interpreted as a part of
		the current IE.

*****************************************************************************/
#define Q931MoreIE() (Octet + Off - 2 < IESize)

#define Q931IESizeTest(x) {\
	if (Octet + Off - 2 != IESize) {\
		Q931SetError(trunk, x, Octet, Off);\
		return x;\
	}\
}

/*****************************************************************************

  Function:	 Q931ReadExt

  Description:  Many of the octets in the standard have an MSB 'ext.1'. This
				means that the octet usually is the latest octet, but that a
				futhure standard may extend the octet. A stack must be able
				to handle such extensions by skipping the extension octets.

				This function will increase the offset counter with 1 for
				each octet with an MSB of zero. This will allow the stack to
				skip extensions wihout knowing anything about them.

  Parameters:   IBuf	ptr to octet array.
				Off	 Starting offset counter

  Return Value: New offset value.

*****************************************************************************/

L3INT Q931ReadExt(L3UCHAR * IBuf, L3INT Off)
{
	L3INT c = 0;
	while ((IBuf[c] & 0x80) == 0) {
		c++;
	}
	return Off + c;
}


/********************************************************************************************************************************
 * Bearer Capability IE
 ********************************************************************************************************************************/

L3INT Q931Uie_BearerCap(Q931_TrunkInfo_t *trunk, Q931mes_Generic *pMsg, L3UCHAR * IBuf, L3UCHAR * OBuf, L3INT *IOff, L3INT *OOff)
{
	Q931ie_BearerCap *pie = (Q931ie_BearerCap*)OBuf;
	ie *pIE = &pMsg->BearerCap;
	L3INT Off = 0;
	L3INT Octet = 0;
	L3INT IESize;
	*pIE = 0;

	/* Octet 1 */
	pie->IEId = IBuf[Octet++];

	/* Octet 2 */
	IESize = IBuf[Octet++];

	/* Octet 3 */
	pie->CodStand = ieGetOctet((IBuf[Octet] & 0x60) >> 5);
	pie->ITC      = ieGetOctet(IBuf[Octet] & 0x1f);
	Off           = Q931ReadExt(&IBuf[Octet], Off);
	Octet++;

	/* Octet 4 */
	pie->TransMode = ieGetOctet((IBuf[Octet + Off] & 0x60) >> 5);
	pie->ITR       = ieGetOctet(IBuf[Octet + Off] & 0x1f);
	Off            = Q931ReadExt(&IBuf[Octet + Off], Off);
	Octet++;

	/* Octet 4.1. Rate multiplier is only present if ITR = Multirate		*/
	if (pie->ITR == 0x18) {
		pie->RateMul = ieGetOctet(IBuf[Octet + Off] & 0x7f);
		Off = Q931ReadExt(&IBuf[Octet + Off], Off);
		Off ++;
	}

	/* Octet 5 */
	if ((IBuf[Octet + Off] & 0x60) == 0x20 && Q931MoreIE()) {
		pie->Layer1Ident = ieGetOctet((IBuf[Octet + Off] & 0x60) >> 5);
		pie->UIL1Prot    = ieGetOctet(IBuf[Octet + Off] & 0x1f);
		Octet++;

		/* Octet 5a. The octet may be present if ITC is unrestrictd digital info
		 * and UIL1Prot is either V.110, I.460 and X.30 or V.120. It may also
		 * be present if ITC = 3.1 kHz audio and UIL1Prot is G.711.
		 * Bit 8 of Octet 5 = 0 indicates that 5a is present.
		 */

		if (IsQ931Ext(IBuf[Octet + Off - 1])) {
			if (((pie->ITC == 0x08) && (pie->UIL1Prot == 0x01 || pie->UIL1Prot == 0x08))
			|| ((pie->ITC == 0x10) && (pie->UIL1Prot == 0x02 || pie->UIL1Prot == 0x03))) {
				pie->SyncAsync = ieGetOctet((IBuf[Octet + Off] & 0x40) >> 6);
				pie->Negot     = ieGetOctet((IBuf[Octet + Off] & 0x20) >> 5);
				pie->UserRate  = ieGetOctet(IBuf[Octet + Off] & 0x1f);
				Off ++;
			}
			else {
				/* We have detected bit 8 = 0, but no setting that require the  */
				/* additional octets ??? */
				Q931SetError(trunk, Q931E_BEARERCAP, 5,Off);
				return Q931E_BEARERCAP;
			}

			/* Octet 5b. Two different structures used. */
			if (IsQ931Ext(IBuf[Octet + Off - 1])) {
				if (pie->UIL1Prot == 0x01) { /* ITU V.110, I.460 and X.30 */
					pie->InterRate = ieGetOctet((IBuf[Octet + Off] & 0x60) >> 5);
					pie->NIConTx   = ieGetOctet((IBuf[Octet + Off] & 0x10) >> 4);
					pie->NIConRx   = ieGetOctet((IBuf[Octet + Off] & 0x08) >> 3);
					pie->FlowCtlTx = ieGetOctet((IBuf[Octet + Off] & 0x04) >> 2);
					pie->FlowCtlRx = ieGetOctet((IBuf[Octet + Off] & 0x20) >> 1);
					Off++;
				}
				else if (pie->UIL1Prot == 0x08) { /* ITU V.120 */
					pie->HDR        = ieGetOctet((IBuf[Octet + Off] & 0x40) >> 6);
					pie->MultiFrame = ieGetOctet((IBuf[Octet + Off] & 0x20) >> 5);
					pie->Mode       = ieGetOctet((IBuf[Octet + Off] & 0x10) >> 4);
					pie->LLInegot   = ieGetOctet((IBuf[Octet + Off] & 0x08) >> 3);
					pie->Assignor   = ieGetOctet((IBuf[Octet + Off] & 0x04) >> 2);
					pie->InBandNeg  = ieGetOctet((IBuf[Octet + Off] & 0x02) >> 1);
					Off++;
				}
				else {
					Q931SetError(trunk,Q931E_BEARERCAP, 5,Off);
					return Q931E_BEARERCAP;
				}

				/* Octet 5c */
				if (IsQ931Ext(IBuf[Octet + Off - 1])) {
					pie->NumStopBits = ieGetOctet((IBuf[Octet + Off] & 0x60) >> 5);
					pie->NumDataBits = ieGetOctet((IBuf[Octet + Off] & 0x18) >> 3);
					pie->Parity      = ieGetOctet(IBuf[Octet + Off] & 0x07);
					Off++;

					/* Octet 5d */
					if (IsQ931Ext(IBuf[Octet + Off - 1])) {
						pie->DuplexMode = ieGetOctet((IBuf[Octet + Off] & 0x40) >> 6);
						pie->ModemType  = ieGetOctet(IBuf[Octet + Off] & 0x3f);
						Off ++;
					}
				}
			}
		}
	}

	/* Octet 6 */
	if ((IBuf[Octet + Off] & 0x60) == 0x40 && Q931MoreIE()) {
		pie->Layer2Ident = ieGetOctet((IBuf[Octet + Off] & 0x60) >> 5);
		pie->UIL2Prot    = ieGetOctet(IBuf[Octet + Off] & 0x1f);

		Off = Q931ReadExt(&IBuf[Octet + Off], Off);
		Octet ++;
	}

	/* Octet 7 */
	if ((IBuf[Octet + Off] & 0x60) == 0x60 && Q931MoreIE()) {
		pie->Layer3Ident = ieGetOctet((IBuf[Octet + Off] & 0x60) >> 5);
		pie->UIL3Prot    = ieGetOctet(IBuf[Octet + Off] & 0x1f);
		Octet++;

		/* Octet 7a */
		if (IsQ931Ext(IBuf[Octet + Off - 1])) {
			if (pie->UIL3Prot == 0x0c) {
				pie->AL3Info1 = ieGetOctet(IBuf[Octet + Off] & 0x0f);
				Off++;

				/* Octet 7b */
				if (IsQ931Ext(IBuf[Octet + Off])) {
					pie->AL3Info2 = ieGetOctet(IBuf[Octet + Off] & 0x0f);
					Off++;
				}
			}
			else {
				Q931SetError(trunk,Q931E_BEARERCAP, 7, Off);
				return Q931E_BEARERCAP;

			}
		}
	}

	Q931IESizeTest(Q931E_BEARERCAP);
	Q931SetIE(*pIE, *OOff);

	*IOff = (*IOff) + Octet + Off;
	*OOff = (*OOff) + sizeof(Q931ie_BearerCap);
	pie->Size = sizeof(Q931ie_BearerCap);

	return Q931E_NO_ERROR;
}


L3INT Q931Pie_BearerCap(Q931_TrunkInfo_t *trunk, L3UCHAR *IBuf, L3UCHAR *OBuf, L3INT *Octet)
{
	Q931ie_BearerCap *pIE = (Q931ie_BearerCap*)IBuf;
	L3INT rc = 0;
	L3INT Beg = *Octet; /* remember current offset */
	L3INT li;

	Q931Log(trunk, Q931_LOG_DEBUG, "Encoding Bearer Capability IE\n");

	OBuf[(*Octet)++] = Q931ie_BEARER_CAPABILITY ;
	li = (*Octet)++; /* remember length position */

	/* Octet 3 - Coding standard / Information transfer capability */
	OBuf[(*Octet)++] = 0x80 | ((pIE->CodStand << 5) & 0x60) | (pIE->ITC & 0x1f);

	/* Octet 4 - Transfer mode / Information transfer rate */
	OBuf[(*Octet)++] = 0x80 | ((pIE->TransMode << 5) & 0x60) | (pIE->ITR & 0x1f);

	if (pIE->ITR == 0x18) {
		/* Octet 4.1 - Rate Multiplier */
		OBuf[(*Octet)++] = 0x80 | (pIE->RateMul & 0x7f);
	}

	/* Octet 5 - Layer 1 Ident / User information layer 1 protocol */
	if (pIE->Layer1Ident == 0x01) {
		if (((pIE->ITC == 0x08) && (pIE->UIL1Prot == 0x01 || pIE->UIL1Prot == 0x08)) ||
		   ((pIE->ITC == 0x10) && (pIE->UIL1Prot == 0x02 || pIE->UIL1Prot == 0x03))) {
			OBuf[(*Octet)++] = 0x00 | ((pIE->Layer1Ident << 5) & 0x60) | (pIE->UIL1Prot & 0x15);

			/* Octet 5a - SyncAsync/Negot/UserRate */
			OBuf[(*Octet)++] = 0x00 | ((pIE->SyncAsync << 6) & 0x40) | ((pIE->Negot << 5) & 0x20) | (pIE->UserRate & 0x1f);

			/* Octet 5b - one of two types */
			if (pIE->UIL1Prot == 0x01) { /* ITU V.110, I.460 and X.30	*/
				/* Octet 5b - Intermed rate/ Nic on Tx/Nix on Rx/FlowCtlTx/FlowCtlRx */
				OBuf[(*Octet)++] = 0x00
						| ((pIE->InterRate << 6) & 0x60)
						| ((pIE->NIConTx   << 4) & 0x10)
						| ((pIE->NIConRx   << 3) & 0x08)
						| ((pIE->FlowCtlTx << 2) & 0x04)
						| ((pIE->FlowCtlRx << 1) & 0x02);
			}
			else if (pIE->UIL1Prot == 0x08) { /* ITU V.120 */
				/* Octet 5b - HDR/Multiframe/Mode/LLINegot/Assignor/Inbandneg*/
				OBuf[(*Octet)++] = 0x00
						| ((pIE->InterRate  << 6) & 0x60)
						| ((pIE->MultiFrame << 5) & 0x20)
						| ((pIE->Mode       << 4) & 0x10)
						| ((pIE->LLInegot   << 3) & 0x08)
						| ((pIE->Assignor   << 2) & 0x04)
						| ((pIE->InBandNeg  << 1) & 0x02);
			}

			/* Octet 5c - NumStopBits/NumStartBits/Parity					*/
			OBuf[(*Octet)++] = 0x00
					| ((pIE->NumStopBits << 5) & 0x60)
					| ((pIE->NumDataBits << 3) & 0x18)
					| (pIE->Parity & 0x07);

			/* Octet 5d - Duplex Mode/Modem Type */
			OBuf[(*Octet)++] = 0x80 | ((pIE->DuplexMode << 6) & 0x40) | (pIE->ModemType & 0x3f);
		}
		else {
			OBuf[(*Octet)++] = 0x80 | ((pIE->Layer1Ident << 5) & 0x60) | (pIE->UIL1Prot & 0x1f);
		}
	}

	/* Octet 6 - Layer2Ident/User information layer 2 prtocol */
	if (pIE->Layer2Ident == 0x02) {
		OBuf[(*Octet)++] = 0x80 | ((pIE->Layer2Ident << 5) & 0x60) | (pIE->UIL2Prot & 0x1f);
	}

	/* Octet 7 - Layer 3 Ident/ User information layer 3 protocol */
	if (pIE->Layer3Ident == 0x03) {
		if (pIE->UIL3Prot == 0x0c) {
			OBuf[(*Octet)++] = 0x00 | ((pIE->Layer3Ident << 5) & 0x60) | (pIE->UIL3Prot & 0x1f);

			/* Octet 7a - Additional information layer 3 msb */
			OBuf[(*Octet)++] = 0x00 | (pIE->AL3Info1 & 0x0f);

			/* Octet 7b - Additional information layer 3 lsb */
			OBuf[(*Octet)++] = 0x80 | (pIE->AL3Info2 & 0x0f);
		}
		else {
			OBuf[(*Octet)++] = 0x80 | ((pIE->Layer3Ident << 5) & 0x60) | (pIE->UIL3Prot & 0x1f);
		}
	}

	OBuf[li] = (L3UCHAR)((*Octet) - Beg) - 2;
	return rc;
}


static const char *q931_information_transfer_capability_type(const unsigned char type)
{
	switch (type) {
	case 0x00:
		return "Speech";
	case 0x08:
		return "Unrestricted digital information";
	case 0x09:
		return "Restricted digital information";
	case 0x10:
		return "3.1KHz Audio";
	case 0x11:
		return "Unrestricted digital information with tones/announcements";
	case 0x18:
		return "Video";
	default:
		return "reserved";
	}
}

static const char *q931_transfer_mode_type(const unsigned char type)
{
	switch (type) {
	case 0x00:
		return "Circuit mode";
	case 0x02:
		return "Packet mode";
	default:
		return "reserved";
	}
}

static const char *q931_information_transfer_rate_type(const unsigned char type)
{
	switch (type) {
	case 0x00:
		return "- Packet mode -";
	case 0x10:
		return "64 kbit/s";
	case 0x11:
		return "2 x 64 kbit/s";
	case 0x13:
		return "384 kbit/s";
	case 0x15:
		return "1536 kbit/s";
	case 0x17:
		return "1920 kbit/s";
	case 0x18:
		return "Multirate (64 kbit/s base rate)";
	default:
		return "reserved";
	}
}

static const char *q931_user_information_layer1_type(const unsigned char type)
{
	switch (type) {
	case 0x01:
		return "ITU-T V.110 / I.460 / X.30";
	case 0x02:
		return "G.711 u-law";
	case 0x03:
		return "G.711 a-law";
	case 0x04:
		return "G.721 32 kbit/s ADPCM (I.460)";
	case 0x05:
		return "H.221 and H.242";
	case 0x06:
		return "H.223 and H.245";
	case 0x07:
		return "Non-ITU-T";
	case 0x08:
		return "ITU-T V.120";
	case 0x09:
		return "ITU-T X.31 HDLC flag stuffing";
	default:
		return "reserved";
	}
}

static const char *q931_user_information_layer2_type(const unsigned char type)
{
	switch (type) {
	case 0x02:
		return "Q.921 / I.441";
	case 0x06:
		return "X.25 link-layer";
	case 0x0c:
		return "LAN logical link control (ISO/IEC 8802-2)";
	default:
		return "reserved";
	}
}

static const char *q931_user_information_layer3_type(const unsigned char type)
{
	switch (type) {
	case 0x02:
		return "Q.931";
	case 0x06:
		return "X.25 packet layer";
	case 0x0b:
		return "ISO/IEC TR 9577";
	default:
		return "reserved";
	}
}

static const char *q931_tr9577_layer3_type(const unsigned char n1, const unsigned char n2)
{
	unsigned char value = (n1 << 4) | n2;

	switch (value) {
	case 0xcc:
		return "Internet Protocol (RFC 791, ISO/IEC TR 9577)";
	case 0xcf:
		return "Point-to-point Protocol (RFC 1548)";
	default:
		return "unknown/invalid";
	}
}

L3INT Q931Die_BearerCap(Q931_TrunkInfo_t *trunk, Q931ie_Generic *ie, struct strstream *ostream)
{
	Q931ie_BearerCap *pie = (Q931ie_BearerCap *)ie;
	strstream_printf(ostream,
			"\tBearer Capability IE [%#02x]:\n"
			"\t\tCoding:                              %s\n"
			"\t\tInformation transfer cap.:           %s\n"
			"\t\tTransfer mode:                       %s\n"
			"\t\tInformation transfer rate:           %s\n",
			pie->IEId,
			q931_coding_type(pie->CodStand),
			q931_information_transfer_capability_type(pie->ITC),
			q931_transfer_mode_type(pie->TransMode),
			q931_information_transfer_rate_type(pie->ITR));

	switch (pie->ITR) {
	case 0x18:
		strstream_printf(ostream,
			 "\t\tRate-multiplier:                    %d (%d kbit/s)\n", pie->RateMul, pie->RateMul << 7);
	default:
		break;
	}

	strstream_printf(ostream,
			"\t\tUser information layer 1 protocol:   %s\n"
			"\t\tUser information layer 2 protocol:   %s\n"
			"\t\tUser information layer 3 protocol:   %s\n",
			q931_user_information_layer1_type(pie->UIL1Prot),
			q931_user_information_layer2_type(pie->UIL2Prot),
			q931_user_information_layer3_type(pie->UIL3Prot));

	if (pie->UIL3Prot == 0x0b) {
		strstream_printf(ostream,
			"\t\tLayer 3 protocol:                    %s\n",
			q931_tr9577_layer3_type(pie->AL3Info1, pie->AL3Info2));
	}
	return Q931E_NO_ERROR;
}


/********************************************************************************************************************************
 * Call ID IE
 ********************************************************************************************************************************/

L3INT Q931Uie_CallID(Q931_TrunkInfo_t *trunk, Q931mes_Generic *pMsg, L3UCHAR * IBuf, L3UCHAR * OBuf, L3INT *IOff, L3INT *OOff)
{
	Q931ie_CallID *pie = (Q931ie_CallID*)OBuf;
	ie *pIE = &pMsg->CallID;
	L3INT Off = 0;
	L3INT Octet = 0;
	L3INT x = 0;
	L3INT IESize;

	*pIE = 0;

	/* Octet 1 */
	pie->IEId = IBuf[Octet++];

	/* Octet 2 */
	IESize = IBuf[Octet++];

	/* Octet 3 */
	do {
		pie->CallId[x] = IBuf[Octet + Off] & 0x7f;
		Off++;
		x++;
	} while (Q931MoreIE());

	Q931IESizeTest(Q931E_CALLID);
	Q931SetIE(*pIE, *OOff);

	*IOff = (*IOff) + Octet + Off;
	*OOff = (*OOff) + sizeof(Q931ie_CallID) + x - 1;
	pie->Size = (L3UCHAR)(sizeof(Q931ie_CallID) + x - 1);

	return Q931E_NO_ERROR;
}

L3INT Q931Pie_CallID(Q931_TrunkInfo_t *trunk, L3UCHAR *IBuf, L3UCHAR *OBuf, L3INT *Octet)
{
	Q931ie_CallID *pIE = (Q931ie_CallID*)IBuf;
	L3INT rc = 0;
	L3INT Beg = *Octet;/* remember current offset */
	L3INT li;
	L3INT sCI = pIE->Size - sizeof(Q931ie_CallID) + 1;
	L3INT x;

	OBuf[(*Octet)++] = Q931ie_CALL_IDENTITY ;
	li = (*Octet)++; /* remember length position */

	for (x = 0; x < sCI; x++) {
		OBuf[(*Octet)++] = pIE->CallId[x];
	}

	OBuf[(*Octet) - 1] |= 0x80; /* set complete flag at last octet*/

	OBuf[li] = (L3UCHAR)((*Octet) - Beg) - 2;
	return rc;
}

L3INT Q931Die_CallID(Q931_TrunkInfo_t *trunk, Q931ie_Generic *ie, struct strstream *ostream)
{
	Q931ie_CallID *pie = (Q931ie_CallID *)ie;
	strstream_printf(ostream,
			"\tCall ID IE [%#02x]:\n"
			"\t\tID:    [ ",
			pie->IEId);
	strstream_printhex(ostream, (char *)pie->CallId, pie->Size - sizeof(*pie));
	strstream_puts(ostream, " ]\n");
	return Q931E_NO_ERROR;
}


/********************************************************************************************************************************
 * Call State IE
 ********************************************************************************************************************************/

L3INT Q931Uie_CallState(Q931_TrunkInfo_t *trunk, Q931mes_Generic *pMsg, L3UCHAR * IBuf, L3UCHAR * OBuf, L3INT *IOff, L3INT *OOff)
{
	Q931ie_CallState *pie = (Q931ie_CallState*)OBuf;
	ie *pIE = &pMsg->CallState;
	L3INT Off = 0;
	L3INT Octet = 0;
	L3INT IESize;

	*pIE = 0;

	/* Octet 1 */
	pie->IEId = IBuf[Octet++];

	/* Octet 2 */
	IESize = IBuf[Octet++];

	/* Octet 3 */
	pie->CodStand  = (IBuf[Octet + Off] >> 6) & 0x03;
	pie->CallState =  IBuf[Octet + Off] & 0x3f;
	Octet++;

	Q931IESizeTest(Q931E_CALLSTATE);
	Q931SetIE(*pIE, *OOff);

	*IOff = (*IOff) + Octet + Off;
	*OOff = (*OOff) + sizeof(Q931ie_CallState);
	pie->Size = sizeof(Q931ie_CallState);

	return Q931E_NO_ERROR;
}

L3INT Q931Pie_CallState(Q931_TrunkInfo_t *trunk, L3UCHAR *IBuf, L3UCHAR *OBuf, L3INT *Octet)
{
	Q931ie_CallState *pIE = (Q931ie_CallState*)IBuf;
	L3INT rc = 0;
	L3INT Beg = *Octet; /* remember current offset */
	L3INT li;

	OBuf[(*Octet)++] = Q931ie_CALL_STATE;
	li = (*Octet)++; /* remember length position */

	OBuf[(*Octet)++] = ((pIE->CodStand & 0x03) << 6) | (pIE->CallState & 0x3f);

	OBuf[li] = (L3UCHAR)((*Octet) - Beg) - 2;
	return rc;
}

static const struct {
	const unsigned char id;
	const char *name;
} q931_callstate_names[] = {
	/* common */
	{ 0x00, "Idle"			},

	/* local */
	{ 0x01, "Call initiated"	},
	{ 0x02, "Overlap sending"	},
	{ 0x03, "Outgoing call proceeding" },
	{ 0x04, "Call delivered"	},
	{ 0x06, "Call present"		},
	{ 0x07, "Call received"		},
	{ 0x08, "Connect request"	},
	{ 0x09, "Incoming call proceeding" },
	{ 0x0a, "Active"		},
	{ 0x0b, "Disconnect request"	},
	{ 0x0c, "Disconnect indication"	},
	{ 0x0f, "Suspend request"	},
	{ 0x11, "Resume request"	},
	{ 0x13, "Release request"	},
	{ 0x16, "Call abort"		},
	{ 0x19, "Overlap receiving"	},

	/* global */
	{ 0x3d, "Restart request"	},
	{ 0x3e, "Restart"		},
};

static const char *q931_callstate_name(const unsigned char state)
{
	for (int i = 0; i < ARRAY_SIZE(q931_callstate_names); i++) {
		if (q931_callstate_names[i].id == state)
			return q931_callstate_names[i].name;
	}
	return "invalid";
}

L3INT Q931Die_CallState(Q931_TrunkInfo_t *trunk, Q931ie_Generic *ie, struct strstream *ostream)
{
	Q931ie_CallState *pie = (Q931ie_CallState *)ie;
	strstream_printf(ostream,
			"\tCall State IE [%#02x]:\n"
			"\t\tCoding:     %s\n"
			"\t\tState:      %s\n",
			pie->IEId,
			q931_coding_type(pie->CodStand),
			q931_callstate_name(pie->CallState));
	return Q931E_NO_ERROR;
}


/********************************************************************************************************************************
 * Called SubAddress IE
 ********************************************************************************************************************************/

L3INT Q931Uie_CalledSub(Q931_TrunkInfo_t *trunk, Q931mes_Generic *pMsg, L3UCHAR * IBuf, L3UCHAR * OBuf, L3INT *IOff, L3INT *OOff)
{
	Q931ie_CalledSub *pie = (Q931ie_CalledSub*)OBuf;
	ie *pIE = &pMsg->CalledSub;
	L3INT Off = 0;
	L3INT Octet = 0;
	L3INT x;
	L3INT IESize;

	*pIE = 0;

	/* Octet 1 */
	pie->IEId = IBuf[Octet++];

	/* Octet 2 */
	IESize = IBuf[Octet++];

	/* Octet 3 */
	pie->TypNum     = (IBuf[Octet + Off] >> 4) & 0x07;
	pie->OddEvenInd = (IBuf[Octet + Off] >> 3) & 0x01;
	Octet++;

	/* Octet 4 */
	x = 0;
	do {
		pie->Digit[x] = IBuf[Octet + Off] & 0x7f;
		Off++;
		x++;
	} while (Q931MoreIE() && x < 20);

	Q931IESizeTest(Q931E_CALLEDSUB);
	Q931SetIE(*pIE, *OOff);

	*IOff = (*IOff) + Octet + Off;
	*OOff = (*OOff) + sizeof(Q931ie_CalledSub) + x - 1;
	pie->Size = (L3UCHAR)(sizeof(Q931ie_CalledSub) + x - 1);

	return Q931E_NO_ERROR;
}

L3INT Q931Pie_CalledSub(Q931_TrunkInfo_t *trunk, L3UCHAR *IBuf, L3UCHAR *OBuf, L3INT *Octet)
{
	Q931ie_CalledSub *pIE = (Q931ie_CalledSub*)IBuf;
	L3INT rc = 0;
	L3INT Beg = *Octet;
	L3INT li;
	L3INT sN = pIE->Size - sizeof(Q931ie_CalledSub) + 1;
	L3INT x;

	/* Octet 1 */
	OBuf[(*Octet)++] = Q931ie_CALLED_PARTY_SUBADDRESS;
	li = (*Octet)++;

	/* Octet 3 */
	OBuf[(*Octet)++] = 0x80 | (pIE->TypNum << 4) | (pIE->OddEvenInd << 3);

	/* Octet 4 */
	for (x = 0; x<sN; x++) {
		OBuf[(*Octet)++] = pIE->Digit[x];
	}

	OBuf[(*Octet) - 1] |= 0x80; /* Terminate bit */

	OBuf[li] = (L3UCHAR)((*Octet) - Beg) - 2;
	return rc;
}

L3INT Q931Die_CalledSub(Q931_TrunkInfo_t *trunk, Q931ie_Generic *ie, struct strstream *ostream)
{
	Q931ie_CalledSub *pie = (Q931ie_CalledSub *)ie;
	char tmp[50] = { 0 };
	int ret = 0;

	if (nsap_afi_valid(pie->Digit[0])) {
		struct nsap_addr nsap = { 0 };
		int offset = 0;

		/* Decode NSAP Address */
		ret = nsap_decode(&nsap, (char *)&pie->Digit[0], pie->Size - sizeof(*pie));
		if (ret != NSAPE_NO_ERROR) {
			Q931Log(trunk, Q931_LOG_ERROR, "Failed to decode NSAP address\n");
			return Q931E_INTERNAL;
		}
		nsap_print(&nsap, &tmp[offset], sizeof(tmp) - offset);
	}
	strstream_printf(ostream,
			"\tCalled Party Subaddress IE [%#02x]:\n"
			"\t\tType:        %s\n"
			"\t\tOdd/Even:    %s\n"
			"\t\tAddress:     %s\n",
			pie->IEId, q931_subaddr_type(pie->TypNum), pie->OddEvenInd ? "Odd" : "Even", tmp);
	return Q931E_NO_ERROR;
}


/********************************************************************************************************************************
 * Called Number IE
 ********************************************************************************************************************************/

L3INT Q931Uie_CalledNum(Q931_TrunkInfo_t *trunk, Q931mes_Generic *pMsg, L3UCHAR * IBuf, L3UCHAR * OBuf, L3INT *IOff, L3INT *OOff)
{
	Q931ie_CalledNum *pie = (Q931ie_CalledNum*)OBuf;
	ie *pIE = &pMsg->CalledNum;
	L3INT Off = 0;
	L3INT Octet = 0;
	L3INT x;
	L3INT IESize; /* # digits in this case */

	*pIE = 0;

	/* Octet 1 */
	pie->IEId = IBuf[Octet++];

	/* Octet 2 */
	IESize = IBuf[Octet++];

	/* Octet 3 */
	pie->TypNum    = (IBuf[Octet + Off] >> 4) & 0x07;
	pie->NumPlanID =  IBuf[Octet + Off] & 0x0f;
	Octet++;

	/* Octet 4*/
	x = 0;
	do {
		pie->Digit[x] = IBuf[Octet + Off] & 0x7f;
		Off++;
		x++;
	} while ((IBuf[Octet + Off]&0x80) == 0 && Q931MoreIE());

	pie->Digit[x] = '\0';

	Q931SetIE(*pIE, *OOff);

	*IOff = (*IOff) + Octet + Off;
	*OOff = (*OOff) + sizeof(Q931ie_CalledNum) + x;
	pie->Size = (L3UCHAR)(sizeof(Q931ie_CalledNum) + x);

	return Q931E_NO_ERROR;
}

L3INT Q931Pie_CalledNum(Q931_TrunkInfo_t *trunk, L3UCHAR *IBuf, L3UCHAR *OBuf, L3INT *Octet)
{
	Q931ie_CalledNum *pIE = (Q931ie_CalledNum*)IBuf;
	L3INT rc = 0;
	L3INT Beg = *Octet;
	L3INT li;
	L3INT sN = pIE->Size - sizeof(Q931ie_CalledNum);
	L3INT x;

	/* Octet 1 */
	OBuf[(*Octet)++] = Q931ie_CALLED_PARTY_NUMBER;

	/* Octet 2 */
	li = (*Octet)++;

	/* Octet 3 */
	OBuf[(*Octet)++] = 0x80 | (pIE->TypNum << 4) | (pIE->NumPlanID);

	/* Octet 4 */
	for (x = 0; x<sN; x++) {
		OBuf[(*Octet)++] = pIE->Digit[x];
	}

	OBuf[li] = (L3UCHAR)((*Octet) - Beg) - 2;
	return rc;
}

L3INT Q931Die_CalledNum(Q931_TrunkInfo_t *trunk, Q931ie_Generic *ie, struct strstream *ostream)
{
	Q931ie_CalledNum *pie = (Q931ie_CalledNum *)ie;
	strstream_printf(ostream,
			"\tCalled Party Number IE [%#02x]:\n"
			"\t\tType of number:    %s\n"
			"\t\tNumbering plan:    %s\n"
			"\t\tNumber:            %.*s\n",
			pie->IEId, q931_ton_type(pie->TypNum),
			q931_numbering_plan_type(pie->NumPlanID),
			pie->Size - sizeof(*pie), pie->Digit);
	return Q931E_NO_ERROR;
}


/********************************************************************************************************************************
 * Calling Number IE
 ********************************************************************************************************************************/

L3INT Q931Uie_CallingNum(Q931_TrunkInfo_t *trunk, Q931mes_Generic *pMsg, L3UCHAR * IBuf, L3UCHAR * OBuf, L3INT *IOff, L3INT *OOff)
{
	Q931ie_CallingNum *pie = (Q931ie_CallingNum*)OBuf;
	ie *pIE = &pMsg->CallingNum;
	L3INT Off = 0;
	L3INT Octet = 0;
	L3INT x;
	L3INT IESize;

	*pIE = 0;

	/* Octet 1 */
	pie->IEId = IBuf[Octet++];

	/* Octet 2 */
	IESize = IBuf[Octet++];

	/* Octet 3 */
	pie->TypNum    = (IBuf[Octet + Off] >> 4) & 0x07;
	pie->NumPlanID =  IBuf[Octet + Off] & 0x0f;

	/* Octet 3a */
	if ((IBuf[Octet + Off] & 0x80) == 0) {
		Off++;
		pie->PresInd   = (IBuf[Octet + Off] >> 5) & 0x03;
		pie->ScreenInd =  IBuf[Octet + Off] & 0x03;
	}
	Octet++;

	/* Octet 4 */
	x = 0;
	while (Q931MoreIE()) {
		pie->Digit[x++] = IBuf[Octet + Off] & 0x7f;

		if ((IBuf[Octet + Off] & 0x80) != 0) {
			break;
		}
		Off++;
	}
	pie->Digit[x] = '\0';

	Q931IESizeTest(Q931E_CALLINGNUM);
	Q931SetIE(*pIE, *OOff);

	*IOff = (*IOff) + Octet + Off;
	*OOff = (*OOff) + sizeof(Q931ie_CallingNum) + x;
	pie->Size = (L3UCHAR)(sizeof(Q931ie_CallingNum) + x);

	return Q931E_NO_ERROR;
}

L3INT Q931Pie_CallingNum(Q931_TrunkInfo_t *trunk, L3UCHAR *IBuf, L3UCHAR *OBuf, L3INT *Octet)
{
	Q931ie_CallingNum *pIE = (Q931ie_CallingNum*)IBuf;
	L3INT rc = 0;
	L3INT Beg = *Octet;
	L3INT li;
	L3INT sN = pIE->Size - sizeof(Q931ie_CallingNum);
	L3INT x;

	/* Octet 1 */
	OBuf[(*Octet)++] = Q931ie_CALLING_PARTY_NUMBER;

	/* Octet 2 */
	li = (*Octet)++;

	/* Octet 3 */
	OBuf[(*Octet)++] = 0x00 | (pIE->TypNum << 4) | (pIE->NumPlanID);

	/* Octet 4 */
	OBuf[(*Octet)++] = 0x80;

	/* Octet 5 */
	for (x = 0; x<sN; x++) {
		OBuf[(*Octet)++] = pIE->Digit[x];
	}

	OBuf[li] = (L3UCHAR)((*Octet) - Beg) - 2;
	return rc;
}

L3INT Q931Die_CallingNum(Q931_TrunkInfo_t *trunk, Q931ie_Generic *ie, struct strstream *ostream)
{
	Q931ie_CallingNum *pie = (Q931ie_CallingNum *)ie;
	strstream_printf(ostream,
			"\tCalling Party Number IE [%#02x]:\n"
			"\t\tType of number:    %s\n"
			"\t\tNumbering plan:    %s\n"
			"\t\tPresentation:      %s\n"
			"\t\tScreening:         %s\n"
			"\t\tNumber:            %.*s\n",
			pie->IEId, q931_ton_type(pie->TypNum),
			q931_numbering_plan_type(pie->NumPlanID),
			q931_presentation_type(pie->PresInd),
			q931_screening_type(pie->ScreenInd),
			pie->Size - sizeof(*pie), pie->Digit);
	return Q931E_NO_ERROR;
}


/********************************************************************************************************************************
 * Calling SubAddress IE
 ********************************************************************************************************************************/

L3INT Q931Uie_CallingSub(Q931_TrunkInfo_t *trunk, Q931mes_Generic *pMsg, L3UCHAR * IBuf, L3UCHAR * OBuf, L3INT *IOff, L3INT *OOff)
{
	Q931ie_CallingSub *pie = (Q931ie_CallingSub*)OBuf;
	ie *pIE = &pMsg->CallingSub;
	L3INT Off = 0;
	L3INT Octet = 0;
	L3INT x;
	L3INT IESize;

	*pIE = 0;

	/* Octet 1 */
	pie->IEId = IBuf[Octet++];

	/* Octet 2 */
	IESize = IBuf[Octet++];

	/* Octet 3 */
	pie->TypNum     = (IBuf[Octet + Off] >> 4) & 0x07;
	pie->OddEvenInd = (IBuf[Octet + Off] >> 3) & 0x01;
	Octet++;

	/* Octet 4*/
	x = 0;
	do {
		pie->Digit[x] = IBuf[Octet + Off] & 0x7f;
		Off++;
		x++;
	} while (Q931MoreIE() && x < 20);

	Q931IESizeTest(Q931E_CALLINGSUB);
	Q931SetIE(*pIE, *OOff);

	*IOff = (*IOff) + Octet + Off;
	*OOff = (*OOff) + sizeof(Q931ie_CallingSub) + x -1;
	pie->Size = (L3UCHAR)(sizeof(Q931ie_CallingSub) + x -1);

	return Q931E_NO_ERROR;
}

L3INT Q931Pie_CallingSub(Q931_TrunkInfo_t *trunk, L3UCHAR *IBuf, L3UCHAR *OBuf, L3INT *Octet)
{
	Q931ie_CallingSub *pIE = (Q931ie_CallingSub*)IBuf;
	L3INT rc = 0;
	L3INT Beg = *Octet;
	L3INT li;
	L3INT sN = pIE->Size - sizeof(Q931ie_CallingSub) + 1;
	L3INT x;

	/* Octet 1 */
	OBuf[(*Octet)++] = Q931ie_CALLING_PARTY_SUBADDRESS;
	li = (*Octet)++;

	/* Octet 3 */
	OBuf[(*Octet)++] = 0x80 | (pIE->TypNum << 4) | (pIE->OddEvenInd << 3);

	/* Octet 4 */
	for (x = 0; x<sN; x++) {
		OBuf[(*Octet)++] = pIE->Digit[x];
	}

	OBuf[(*Octet) - 1] |= 0x80; /* Terminate bit */

	OBuf[li] = (L3UCHAR)((*Octet) - Beg) - 2;
	return rc;
}

L3INT Q931Die_CallingSub(Q931_TrunkInfo_t *trunk, Q931ie_Generic *ie, struct strstream *ostream)
{
	Q931ie_CallingSub *pie = (Q931ie_CallingSub *)ie;
	char tmp[50] = { 0 };

	strstream_printf(ostream,
			"\tCalling Party Subaddress IE [%#02x]:\n",
			"\t\tType:      %s\n"
			"\t\tOdd/Even:  %s\n"
			"\t\tAddress:   %s\n",
			pie->IEId, q931_subaddr_type(pie->TypNum), pie->OddEvenInd ? "Odd" : "Even", tmp);
	return Q931E_NO_ERROR;
}


/********************************************************************************************************************************
 * Cause IE
 ********************************************************************************************************************************/

L3INT Q931Uie_Cause(Q931_TrunkInfo_t *trunk, Q931mes_Generic *pMsg, L3UCHAR * IBuf, L3UCHAR * OBuf, L3INT *IOff, L3INT *OOff)
{
	Q931ie_Cause *pie = (Q931ie_Cause*)OBuf;
	ie *pIE = &pMsg->Cause;
	L3INT Off = 0;
	L3INT Octet = 0;
	L3INT IESize;

	*pIE = 0;

	pie->IEId = IBuf[Octet++];

	/* Octet 2*/
	IESize = IBuf[Octet++];

	/* Octet 3*/
	pie->CodStand = (IBuf[Octet + Off]>>5) & 0x03;
	pie->Location =  IBuf[Octet + Off] & 0x0f;

	/* Octet 3a */
	if ((IBuf[Octet + Off] & 0x80) == 0) {
		Off++;
		pie->Recom = IBuf[Octet + Off] & 0x7f;
	}
	Octet++;

	/* Octet 4 */
	pie->Value = IBuf[Octet + Off] & 0x7f;
	Octet++;

	/* Consume optional Diagnostic bytes */
	while (Q931MoreIE()) {
		Off++;
	};

	Q931IESizeTest(Q931E_CAUSE);
	Q931SetIE(*pIE, *OOff);

	*IOff = (*IOff) + Octet + Off;
	*OOff = (*OOff) + sizeof(Q931ie_Cause);
	pie->Size = sizeof(Q931ie_Cause);

	return Q931E_NO_ERROR;
}

L3INT Q931Pie_Cause(Q931_TrunkInfo_t *trunk, L3UCHAR *IBuf, L3UCHAR *OBuf, L3INT *Octet)
{
	Q931ie_Cause *pIE = (Q931ie_Cause*)IBuf;
	L3INT rc = 0;
	L3INT Beg = *Octet;
	L3INT li;

	OBuf[(*Octet)++] = Q931ie_CAUSE;
	li = (*Octet)++;

	/* Octet 3 */
	OBuf[(*Octet)++] = 0x80 | (pIE->CodStand<<5) | pIE->Location;

	/* Octet 3a - currently not supported in send */

	/* Octet 4 */
	OBuf[(*Octet)++] = 0x80 | pIE->Value;

	/* Octet 5 - diagnostics not supported in send */

	OBuf[li] = (L3UCHAR)((*Octet) - Beg) - 2;
	return rc;
}

L3INT Q931Die_Cause(Q931_TrunkInfo_t *trunk, Q931ie_Generic *ie, struct strstream *ostream)
{
	Q931ie_Cause *pie = (Q931ie_Cause *)ie;
	strstream_printf(ostream,
			"\tCause IE [%#02x]:\n"
			"\t\tCoding:   %s\n"
			"\t\tLocation: %s\n"
			"\t\tCode:     %d (%s)\n",
			pie->IEId, q931_coding_type(pie->CodStand), q931_location_type(pie->Location),
			pie->Value, Q850CauseGetName(pie->Value));
	return Q931E_NO_ERROR;
}


/********************************************************************************************************************************
 * Congestion Level IE
 ********************************************************************************************************************************/

L3INT Q931Uie_CongLevel(Q931_TrunkInfo_t *trunk, Q931mes_Generic *pMsg, L3UCHAR * IBuf, L3UCHAR * OBuf, L3INT *IOff, L3INT *OOff)
{
	Q931ie_CongLevel *pie = (Q931ie_CongLevel*)OBuf;
	ie *pIE = &pMsg->CongestionLevel;
	L3INT Off = 0;
	L3INT Octet = 0;

	*pIE = 0;

	pie->IEId      = IBuf[Octet] & 0xf0;
	pie->CongLevel = IBuf[Octet] & 0x0f;
	Octet ++;

	Q931SetIE(*pIE, *OOff);

	*IOff = (*IOff) + Octet + Off;
	*OOff = (*OOff) + sizeof(Q931ie_CongLevel);
	pie->Size = sizeof(Q931ie_CongLevel);

	return Q931E_NO_ERROR;
}

L3INT Q931Pie_CongLevel(Q931_TrunkInfo_t *trunk, L3UCHAR *IBuf, L3UCHAR *OBuf, L3INT *Octet)
{
	Q931ie_CongLevel *pIE = (Q931ie_CongLevel*)IBuf;
	L3INT rc = 0;
	/* L3INT Beg = *Octet; */

	OBuf[(*Octet)++] = Q931ie_CONGESTION_LEVEL | pIE->CongLevel;

	return rc;
}


static const char *q931_congestion_level_type(const unsigned char type)
{
	switch (type) {
	case 0x00:
		return "Receiver ready";
	case 0x0f:
		return "Receiver not ready";
	default:
		return "reserved";
	}
}

L3INT Q931Die_CongLevel(Q931_TrunkInfo_t *trunk, Q931ie_Generic *ie, struct strstream *ostream)
{
	Q931ie_CongLevel *pie = (Q931ie_CongLevel *)ie;
	strstream_printf(ostream,
			"\tCongestion Level IE [%#02x]:\n"
			"\t\tLevel:   %s\n",
			pie->IEId, q931_congestion_level_type(pie->CongLevel));
	return Q931E_NO_ERROR;
}


/********************************************************************************************************************************
 * Channel ID IE
 ********************************************************************************************************************************/

L3INT Q931Uie_ChanID(Q931_TrunkInfo_t *trunk, Q931mes_Generic *pMsg, L3UCHAR * IBuf, L3UCHAR *OBuf, L3INT *IOff, L3INT *OOff)
{
	Q931ie_ChanID *pie = (Q931ie_ChanID*)OBuf;
	ie *pIE = &pMsg->ChanID;
	L3INT Off = 0;
	L3INT Octet = 0;
	L3INT IESize;
//18 04 e1 80 83 01
	*pIE = 0;

	Q931Log(trunk, Q931_LOG_DEBUG, "Decoding ChanID IE\n");

	/* Octet 1 */
	pie->IEId = IBuf[Octet++];

	/* Octet 2 */
	IESize    = IBuf[Octet++];

	/* Octet 3 */
	pie->IntIDPresent = (IBuf[Octet] >> 6) & 0x01;
	pie->IntType      = (IBuf[Octet] >> 5) & 0x01;
	pie->PrefExcl     = (IBuf[Octet] >> 3) & 0x01;
	pie->DChanInd     = (IBuf[Octet] >> 2) & 0x01;
	pie->InfoChanSel  =  IBuf[Octet] & 0x03;

	Off = Q931ReadExt(&IBuf[Octet++], Off);

	/* Octet 3.1 */
	if (pie->IntIDPresent) {
		pie->InterfaceID = IBuf[Octet + Off] & 0x7f;

		/* Temp fix. Interface id can be extended using the extension bit */
		/* this will read the octets, but do nothing with them. this is done */
		/* because the usage of this field is a little unclear */
		/* 30.jan.2001/JVB */
		Off = Q931ReadExt(&IBuf[Octet + Off], Off);
		Off++;
	}

	if ((Octet + Off - 2) != IESize) {
		/* Octet 3.2 */
		if (pie->IntType == 1) {        /* PRI etc */
			pie->CodStand    = (IBuf[Octet + Off] >> 5) & 0x03;
			pie->NumMap      = (IBuf[Octet + Off] >> 4) & 0x01;
			pie->ChanMapType =  IBuf[Octet + Off] & 0x0f;
			Off++;

			/* Octet 3.3 */
			/* Temp fix. Assume B channel. H channels not supported */
			pie->ChanSlot = IBuf[Octet + Off] & 0x7f;

			/* Some dialects don't follow the extension coding properly for this, but this should be safe for all */
			if ((Octet + Off - 1) != IESize) {
				Off = Q931ReadExt(&IBuf[Octet + Off], Off);
			}
			Off++;
		}
	}

	Q931IESizeTest(Q931E_CHANID);
	Q931SetIE(*pIE, *OOff);

	*IOff = (*IOff) + Octet + Off;
	*OOff = (*OOff) + sizeof(Q931ie_ChanID);
	pie->Size = sizeof(Q931ie_ChanID);

	if (trunk->loglevel == Q931_LOG_DEBUG) {
		const char *iface;
		char tmp[100] = "";

		if (!pie->IntType) {
			switch (pie->InfoChanSel) {
			case 0x0:
				iface = "None";
				break;
			case 0x1:
				iface = "B1";
				break;
			case 0x2:
				iface = "B2";
				break;
			default:
				iface = "Any Channel";
			}

			snprintf(tmp, sizeof(tmp)-1, "InfoChanSel: %d (%s)", pie->InfoChanSel, iface);
		}

		Q931Log(trunk, Q931_LOG_DEBUG,
			"\n-------------------------- Q.931 Channel ID ------------------------\n"
			"    Pref/Excl: %s, Interface Type: %s\n"
			"    %s\n"
			"--------------------------------------------------------------------\n\n",
			((pie->PrefExcl) ? "Preferred" : "Exclusive"),
			((pie->IntType) ? "PRI/Other" : "BRI"),
			tmp);
	}
	return Q931E_NO_ERROR;
}

L3INT Q931Pie_ChanID(Q931_TrunkInfo_t *trunk, L3UCHAR *IBuf, L3UCHAR *OBuf, L3INT *Octet)
{
	Q931ie_ChanID *pIE = (Q931ie_ChanID*)IBuf;
	L3INT rc = Q931E_NO_ERROR;
	L3INT Beg = *Octet;	/* remember current offset */
	L3INT li;

	OBuf[(*Octet)++] = Q931ie_CHANNEL_IDENTIFICATION;
	li = (*Octet)++; /* remember length position */

	/* Octet 3 flags & BRI chan # */
	OBuf[(*Octet)++] = 0x80
			| ((pIE->IntIDPresent << 6) & 0x40)
			| ((pIE->IntType << 5) & 0x20)
			| ((pIE->PrefExcl << 3) & 0x08)
			|  (pIE->InfoChanSel & 0x03);

	/* Octet 3.1 - Interface Identifier */
	if (pIE->IntIDPresent) {
		OBuf[(*Octet)++] = 0x80 | (pIE->InterfaceID & 0x7f);
	}

	/* Octet 3.2 & 3.3 - PRI */
	if (pIE->IntType) {
		OBuf[(*Octet)++]  = 0x80
			| ((pIE->CodStand << 5) & 0x60)
			| ((pIE->NumMap << 4) & 0x10)
			|  (pIE->ChanMapType & 0x0f);		/* TODO: support all possible channel map types */

		/* Octet 3.3 Channel number */
		switch (pIE->ChanMapType) {
		case 0x6:	/* Slot map: H0 Channel Units */	/* unsupported, Octets 3.3.1 - 3.3.3 */
			return Q931E_CHANID;

		case 0x8:	/* Slot map: H11 Channel Units */
		case 0x9:	/* Slot map: H12 Channel Units */
		default:	/* Channel number */
			OBuf[(*Octet)++] = 0x80 | (pIE->ChanSlot & 0x7f);
			break;
		}
	}

	OBuf[li] = (L3UCHAR)((*Octet) - Beg) - 2;
	return rc;
}

L3INT Q931Die_ChanID(Q931_TrunkInfo_t *trunk, Q931ie_Generic *ie, struct strstream *ostream)
{
	Q931ie_ChanID *pie = (Q931ie_ChanID *)ie;
	char tmp[1024] = { 0 };
	int offset = 0;

	if (!pie->IntType) {	/* BRI */
		const char *iface;

		switch (pie->InfoChanSel) {
		case 0x0:
			iface = "None";
			break;
		case 0x1:
			iface = "B1";
			break;
		case 0x2:
			iface = "B2";
			break;
		default:
			iface = "Any Channel";
		}

		offset += snprintf(tmp, sizeof(tmp) - 1, "\t\tInfoChanSel:          %d (%s)", pie->InfoChanSel, iface);
 	} else {	/* PRI */
		const char *iface;

		switch (pie->InfoChanSel) {
		case 0x0:
			iface = "None";
			break;
		case 0x1:
			iface = "As indicated";
			break;
		case 0x2:
			iface = "Reserved";
			break;
		default:
			iface = "Any Channel";
		}

		offset += snprintf(tmp, sizeof(tmp) - 1, "\t\tInfoChanSel:          %d (%s)\n", pie->InfoChanSel, iface);
	}

	if (pie->ChanSlot) {
		offset += snprintf(tmp + offset, sizeof(tmp) - offset - 1, "\t\tChannel:              %d\n", pie->ChanSlot);
	} else if (pie->InterfaceID) {
		offset += snprintf(tmp + offset, sizeof(tmp) - offset - 1, "\t\tInterface:            %d\n", pie->InterfaceID);
	}

	strstream_printf(ostream,
		"\tChannel ID IE [%#02x]:\n"
		"\t\tPreferred/Exclusive:  %s\n"
		"\t\tInterface Type:       %s\n"
		"%s",
		pie->IEId,
		((pie->PrefExcl) ? "Preferred" : "Exclusive"),
		((pie->IntType) ? "PRI/Other" : "BRI"),
		tmp);
	return Q931E_NO_ERROR;
}


/********************************************************************************************************************************
 * DateTime IE
 ********************************************************************************************************************************/

L3INT Q931Uie_DateTime(Q931_TrunkInfo_t *trunk, Q931mes_Generic *pMsg, L3UCHAR *IBuf, L3UCHAR *OBuf, L3INT *IOff, L3INT *OOff)
{
	Q931ie_DateTime * pie = (Q931ie_DateTime*)OBuf;
	ie *pIE = &pMsg->DateTime;
	L3INT Off = 0;
	L3INT Octet = 0;
	L3INT IESize = 0;

	*pIE = 0;

	pie->IEId = IBuf[Octet++];

	/* Octet 2 */
	IESize = IBuf[Octet++];

	/* Octet 3 - Year */
	pie->Year = IBuf[Octet++];

	/* Octet 4 - Month */
	pie->Month = IBuf[Octet++];

	/* Octet 5 - Day */
	pie->Day = IBuf[Octet++];

	/*******************************************************************
		The remaining part of the IE are optioinal, but only the length
		can now tell us wherever these fields are present or not
		(always remember: IESize does not include ID and Size octet)
	********************************************************************/
	pie->Format = 0;

	/* Octet 6 - Hour (optional)*/
	if (IESize >= 4) {
		pie->Format = 1;
		pie->Hour = IBuf[Octet++];

		/* Octet 7 - Minute (optional)*/
		if (IESize >= 5) {
			pie->Format = 2;
			pie->Minute = IBuf[Octet++];

			/* Octet 8 - Second (optional)*/
			if (IESize >= 6) {
				pie->Format = 3;
				pie->Second = IBuf[Octet++];
			}
		}
	}

	Q931IESizeTest(Q931E_DATETIME);
	Q931SetIE(*pIE, *OOff);

	*IOff = (*IOff) + Octet + Off;
	*OOff = (*OOff) + sizeof(Q931ie_DateTime);
	pie->Size = sizeof(Q931ie_DateTime);

	return Q931E_NO_ERROR;
}

L3INT Q931Pie_DateTime(Q931_TrunkInfo_t *trunk, L3UCHAR *IBuf, L3UCHAR *OBuf, L3INT *Octet)
{
	Q931ie_DateTime *pIE = (Q931ie_DateTime*)IBuf;
	L3INT rc = 0;
	L3INT Beg = *Octet;
	L3INT li;

	OBuf[(*Octet)++] = Q931ie_DATETIME;
	li = (*Octet)++;

	OBuf[(*Octet)++] = pIE->Year;
	OBuf[(*Octet)++] = pIE->Month;
	OBuf[(*Octet)++] = pIE->Day;
	if (pIE->Format >= 1) {
		OBuf[(*Octet)++] = pIE->Hour;

		if (pIE->Format >= 2) {
			OBuf[(*Octet)++] = pIE->Minute;

			if (pIE->Format >= 3) {
				OBuf[(*Octet)++] = pIE->Second;
			}
		}
	}

	OBuf[li] = (L3UCHAR)((*Octet)-Beg) - 2;
	return rc;
}

L3INT Q931Die_DateTime(Q931_TrunkInfo_t *trunk, Q931ie_Generic *ie, struct strstream *ostream)
{
	Q931ie_DateTime *pie = (Q931ie_DateTime *)ie;
	char tmp[50] = { 0 };

	switch (pie->Format) {
	case 3:
		sprintf(tmp, "%02u:%02u:%02u [hh:mm:ss]", pie->Hour, pie->Minute, pie->Second);
		break;
	case 2:
		sprintf(tmp, "%02u:%02u [hh:mm]", pie->Hour, pie->Minute);
		break;
	case 1:
		sprintf(tmp, "%02u [hh]", pie->Hour);
		break;
	default:
		strcat(tmp, "N/A");
		break;
	}

	strstream_printf(ostream,
			"\tDateTime IE [%#02x]:\n"
			"\t\tDate:   %02u/%02u/%02u [mm/dd/yy]\n"
			"\t\tTime:   %s\n",
			pie->IEId, pie->Month, pie->Day, pie->Year, tmp);
	return Q931E_NO_ERROR;
}


/********************************************************************************************************************************
 * Display IE
 ********************************************************************************************************************************/

L3INT Q931Uie_Display(Q931_TrunkInfo_t *trunk, Q931mes_Generic *pMsg, L3UCHAR * IBuf, L3UCHAR * OBuf, L3INT *IOff, L3INT *OOff)
{
	Q931ie_Display *pie = (Q931ie_Display*)OBuf;
	ie *pIE = &pMsg->Display;
	L3INT Off = 0;
	L3INT Octet = 0;
	L3INT IESize;
	L3INT x;

	*pIE = 0;

	pie->IEId = IBuf[Octet++];
	IESize    = IBuf[Octet++];

	for (x = 0; x<IESize; x++) {
		pie->Display[x] = IBuf[Octet + Off] & 0x7f;
		Off++;
	}

	Q931IESizeTest(Q931E_DISPLAY);
	Q931SetIE(*pIE, *OOff);

	*IOff = (*IOff) + Octet + Off;
	*OOff = (*OOff) + sizeof(Q931ie_Display) + x - 1;
	pie->Size = (L3UCHAR)(sizeof(Q931ie_Display) + x - 1);

	return Q931E_NO_ERROR;
}

L3INT Q931Pie_Display(Q931_TrunkInfo_t *trunk, L3UCHAR *IBuf, L3UCHAR *OBuf, L3INT *Octet)
{
	Q931ie_Display *pIE = (Q931ie_Display*)IBuf;
	L3INT rc = 0;
	L3INT Beg = *Octet;
	L3INT li;
	L3INT DSize;
	L3INT x;

	OBuf[(*Octet)++] = Q931ie_DISPLAY;
	li = (*Octet)++;

	DSize = pIE->Size - sizeof(Q931ie_Display);

	for (x = 0; x< DSize; x++) {

		OBuf[(*Octet)++] = pIE->Display[x];
	}

	OBuf[li] = (L3UCHAR)((*Octet) - Beg) - 2;
	return rc;
}

L3INT Q931Die_Display(Q931_TrunkInfo_t *trunk, Q931ie_Generic *ie, struct strstream *ostream)
{
	Q931ie_Display *pie = (Q931ie_Display *)ie;
	strstream_printf(ostream,
			"\tDisplay IE [%#02x]:\n"
			"\t\tValue:   %.*s\n",
			pie->IEId, pie->Size - sizeof(*pie), pie->Display);
	return Q931E_NO_ERROR;
}


/********************************************************************************************************************************
 * High-Layer Compat IE
 ********************************************************************************************************************************/

L3INT Q931Uie_HLComp(Q931_TrunkInfo_t *trunk, Q931mes_Generic *pMsg, L3UCHAR * IBuf, L3UCHAR * OBuf, L3INT *IOff, L3INT *OOff)
{
	Q931ie_HLComp * pie = (Q931ie_HLComp*)OBuf;
	ie *pIE = &pMsg->HLComp;
	L3INT Off = 0;
	L3INT Octet = 0;
	L3INT IESize;

	*pIE = 0;

	pie->IEId = IBuf[Octet++];

	/* Octet */
	IESize = IBuf[Octet++];

	/* Octet 3*/
	pie->CodStand  = (IBuf[Octet + Off] >>5) & 0x03;
	pie->Interpret = (IBuf[Octet + Off] >>2) & 0x07;
	pie->PresMeth  =  IBuf[Octet + Off] & 0x03;
	Octet++;

	/* Octet 4 */
	pie->HLCharID = IBuf[Octet + Off] & 0x7f;
	Octet++;

	/* Octet 4a*/
	if ((IBuf[Octet + Off - 1] & 0x80) == 0 && Q931MoreIE()) {
		if (pie->HLCharID == 0x5e || pie->HLCharID == 0x5f) {
			pie->EHLCharID = IBuf[Octet + Off] & 0x7f;
			Off++;
		}
		else if ( pie->HLCharID >= 0xc3 && pie->HLCharID <= 0xcf) {
			pie->EVideoTlfCharID = IBuf[Octet + Off] & 0x7f;
			Off++;
		}
		else {
			/* error Octet 4a indicated, but invalid value in Octet 4. */
			Q931SetError(trunk,Q931E_HLCOMP, 4, Off);
			return Q931E_HLCOMP;
		}
		Off = Q931ReadExt(&IBuf[Octet + Off], Off);
	}

	Q931IESizeTest(Q931E_HLCOMP);
	Q931SetIE(*pIE, *OOff);

	*IOff = (*IOff) + Octet + Off;
	*OOff = (*OOff) + sizeof(Q931ie_HLComp);
	pie->Size = sizeof(Q931ie_HLComp);

	return Q931E_NO_ERROR;
}

L3INT Q931Pie_HLComp(Q931_TrunkInfo_t *trunk, L3UCHAR *IBuf, L3UCHAR *OBuf, L3INT *Octet)
{
	Q931ie_HLComp *pIE = (Q931ie_HLComp*)IBuf;
	L3INT rc = 0;
	L3INT Beg = *Octet;
	L3INT li;

	OBuf[(*Octet)++] = Q931ie_HIGH_LAYER_COMPATIBILITY;
	li = (*Octet)++;

	/* Octet 3 */
	OBuf[(*Octet)++] = 0x80 | ((pIE->CodStand << 5) & 0x60) | ((pIE->Interpret << 2) & 0x1c) | (pIE->PresMeth & 0x03);

	/* Octet 4 */
	OBuf[(*Octet)++] = pIE->HLCharID;

	/* Octet 4a */
	if (pIE->HLCharID == 0x5e || pIE->HLCharID == 0x5f) {
		OBuf[(*Octet)++] = 0x80 | (pIE->EHLCharID & 0x7f);
	}
	else if ( pIE->HLCharID >= 0xc3 && pIE->HLCharID <= 0xcf) {
		OBuf[(*Octet)++] = 0x80 | (pIE->EVideoTlfCharID & 0x7f);
	}
	else {
		OBuf[(*Octet) - 1] |= 0x80;
	}

	OBuf[li] = (L3UCHAR)((*Octet) - Beg) - 2;
	return rc;
}

static const char *q931_hl_characteristics_type(const unsigned char type)
{
	switch (type) {
	case 0x01:
		return "Telephony";
	case 0x04:
		return "Facsimile Group 2/3 (F.182)";
	case 0x21:
		return "Facsimile Group 4 Class I (F.184)";
	case 0x24:
		return "Facsimile Group 4 Classes II and III (F.184)";
	case 0x28:
		return "-";
	case 0x31:
		return "-";
	case 0x32:
		return "Syntax based Videotex (F.300 and T.102)";
	case 0x33:
		return "Internationel Videotex interworking via gateways or interworking units (F.300 and T.101)";
	case 0x35:
		return "Telex service (F.60)";
	case 0x38:
		return "Message Handling System (MHS, X.400-series)";
	case 0x41:
		return "OSI application (X.200-series)";
	case 0x42:
		return "FTAM application (ISO 8571)";
	case 0x5e:
		return "Reserved for maintenance";
	case 0x5f:
		return "Reserved for maintenance";
	case 0x60:
		return "Videotelephony (F.720, F.721 and F.731 profile 1a)";
	case 0x61:
		return "Videoconferenceing (F.702 and F.731 profile 1b)";
	case 0x62:
		return "Audiographic conferencing (F.702 and F.731 profile 2a2 etc.)";
	case 0x63:
	case 0x64:
	case 0x65:
	case 0x66:
	case 0x67:
		return "Reserved for audiovisual service (F.700-series)";
	case 0x68:
		return "Multimedia services (F.700-series)";
	case 0x6b:
	case 0x6c:
	case 0x6d:
	case 0x6e:
	case 0x6f:
		return "Reserved for audiovisual service (F.700-series)";
	default:
		return "reserved";
	}
}

L3INT Q931Die_HLComp(Q931_TrunkInfo_t *trunk, Q931ie_Generic *ie, struct strstream *ostream)
{
	Q931ie_HLComp *pie = (Q931ie_HLComp *)ie;
	strstream_printf(ostream,
			"\tHigh-Layer Compat IE [%#02x]:\n"
			"\t\tCoding:                %s\n"
			"\t\tInterpretation:        %s\n"
			"\t\tPresentation:          %s\n"
			"\t\tHL Characteristics:    %s\n",
			pie->IEId, q931_coding_type(pie->CodStand),
			((pie->Interpret == 0x04) ? "First/only HL characteristic" : "reserved"),
			((pie->PresMeth  == 0x01) ? "HL protocol profile" : "reserved"),
			q931_hl_characteristics_type(pie->HLCharID));
	return Q931E_NO_ERROR;
}


/********************************************************************************************************************************
 * Keypad Facility IE
 ********************************************************************************************************************************/

L3INT Q931Uie_KeypadFac(Q931_TrunkInfo_t *trunk, Q931mes_Generic *pMsg, L3UCHAR * IBuf, L3UCHAR * OBuf, L3INT *IOff, L3INT *OOff)
{
	Q931ie_KeypadFac *pie = (Q931ie_KeypadFac*)OBuf;
	ie *pIE = &pMsg->KeypadFac;
	L3INT Off = 0;
	L3INT Octet = 0;
	L3INT IESize;
	L3INT x;

	*pIE = 0;

	pie->IEId = IBuf[Octet++];
	IESize = IBuf[Octet++];

	for (x = 0; x<IESize; x++) {
		pie->KeypadFac[x] = IBuf[Octet + Off] & 0x7f;
		Off++;
	}

	Q931IESizeTest(Q931E_KEYPADFAC);
	Q931SetIE(*pIE, *OOff);

	*IOff = (*IOff) + Octet + Off;
	*OOff = (*OOff) + sizeof(Q931ie_KeypadFac) + x - 1;
	pie->Size = (L3UCHAR)(sizeof(Q931ie_KeypadFac) + x - 1);

	return Q931E_NO_ERROR;
}

L3INT Q931Pie_KeypadFac(Q931_TrunkInfo_t *trunk, L3UCHAR *IBuf, L3UCHAR *OBuf, L3INT *Octet)
{
	Q931ie_KeypadFac *pIE = (Q931ie_KeypadFac*)IBuf;
	L3INT rc = 0;
	L3INT Beg = *Octet;
	L3INT li;
	L3INT DSize;
	L3INT x;

	OBuf[(*Octet)++] = Q931ie_KEYPAD_FACILITY;
	li = (*Octet)++;

	DSize = pIE->Size - sizeof(Q931ie_KeypadFac) + 1;

	for (x = 0; x< DSize; x++) {
		OBuf[(*Octet)++] = pIE->KeypadFac[x];
	}

	OBuf[li] = (L3UCHAR)((*Octet) - Beg) - 2;
	return rc;
}

L3INT Q931Die_KeypadFac(Q931_TrunkInfo_t *trunk, Q931ie_Generic *ie, struct strstream *ostream)
{
	Q931ie_KeypadFac *pie = (Q931ie_KeypadFac *)ie;
	strstream_printf(ostream,
			"\tKeypad Facility IE [%#02x]:\n"
			"\t\tValue:     %.*s\n",
			pie->IEId, pie->Size - sizeof(*pie), pie->KeypadFac);
	return Q931E_NO_ERROR;
}


/********************************************************************************************************************************
 * Low-Layer Compat IE
 ********************************************************************************************************************************/

L3INT Q931Uie_LLComp(Q931_TrunkInfo_t *trunk, Q931mes_Generic *pMsg, L3UCHAR * IBuf, L3UCHAR * OBuf, L3INT *IOff, L3INT *OOff)
{
	Q931ie_LLComp *pie = (Q931ie_LLComp*)OBuf;
	ie *pIE = &pMsg->LLComp;
	L3INT Off = 0;
	L3INT Octet = 0;
	L3INT IESize;

	*pIE = 0;

	pie->IEId = IBuf[Octet++];

	/* Octet 2 */
	IESize = IBuf[Octet++];

	/* Octet 3 */
	pie->CodStand  = (IBuf[Octet + Off] >> 5) & 0x03;
	pie->ITransCap =  IBuf[Octet + Off] & 0x1f;
	Octet++;

	/* Octet 3a*/
	if (IsQ931Ext(IBuf[Octet + Off - 1])) {
		pie->NegotInd = (IBuf[Octet + Off] >> 6) & 0x01;
		Off++;
	}

	/* Octet 4 */
	pie->TransMode = (IBuf[Octet + Off] >> 5) & 0x03;
	pie->InfoRate  =  IBuf[Octet + Off] & 0x1f;

	Octet++;

	/* Octet 4.1 */
	if (pie->InfoRate == 0x14) { /* Mutirate */
		pie->RateMul = IBuf[Octet + Off] & 0x7f;
		Off++;
	}

	/* Octet 5 - Layer 1 Ident */
	if ((IBuf[Octet + Off] & 0x60) == 0x20) { /* Layer 1 Ident ? */
		pie->Layer1Ident = (IBuf[Octet + Off] >> 5) & 0x03;
		pie->UIL1Prot    =  IBuf[Octet + Off] & 0x1f;
		Octet++;

		/* Octet 5a */
		if (IsQ931Ext(IBuf[Octet + Off - 1])) {
			pie->SyncAsync = (IBuf[Octet + Off] >> 6) & 0x01;
			pie->Negot     = (IBuf[Octet + Off] >> 5) & 0x01;
			pie->UserRate  =  IBuf[Octet + Off] & 0x1f;
			Off++;

			/* Octet 5b - 2 options */
			if (IsQ931Ext(IBuf[Octet + Off - 1])) {
				if (pie->UIL1Prot == 0x01) { /* V.110, I.460 and X.30*/
					pie->InterRate = (IBuf[Octet + Off] >> 5) & 0x03;
					pie->NIConTx   = (IBuf[Octet + Off] >> 4) & 0x01;
					pie->NIConRx   = (IBuf[Octet + Off] >> 3) & 0x01;
					pie->FlowCtlTx = (IBuf[Octet + Off] >> 2) & 0x01;
					pie->FlowCtlRx = (IBuf[Octet + Off] >> 1) & 0x01;
					Off++;
				}
				else if (pie->UIL1Prot == 0x80) { /* V.120 */
					pie->HDR        = (IBuf[Octet + Off] >> 6) & 0x01;
					pie->MultiFrame = (IBuf[Octet + Off] >> 5) & 0x01;
					pie->ModeL1     = (IBuf[Octet + Off] >> 4) & 0x01;
					pie->NegotLLI   = (IBuf[Octet + Off] >> 3) & 0x01;
					pie->Assignor   = (IBuf[Octet + Off] >> 2) & 0x01;
					pie->InBandNeg  = (IBuf[Octet + Off] >> 1) & 0x01;
					Off++;
				}
				else if (pie->UIL1Prot == 0x07) { /* non standard */
					Off = Q931ReadExt(&IBuf[Octet + Off], Off);
					Off++;
				}
				else {
					Q931SetError(trunk,Q931E_LLCOMP, 5,2);
					return Q931E_LLCOMP;
				}

				/* Octet 5c */
				if (IsQ931Ext(IBuf[Octet + Off - 1])) {
					pie->NumStopBits = (IBuf[Octet + Off] >> 5) & 0x03;
					pie->NumDataBits = (IBuf[Octet + Off] >> 3) & 0x03;
					pie->Parity      =  IBuf[Octet + Off] & 0x07;
					Off++;

					/* Octet 5d */
					if (IsQ931Ext(IBuf[Octet + Off - 1])) {
						pie->DuplexMode	= (IBuf[Octet + Off] >> 6) & 0x01;
						pie->ModemType  =  IBuf[Octet + Off] & 0x3f;
						Off = Q931ReadExt(&IBuf[Octet + Off], Off);
						Off++;
					}
				}
			}
		}
	}

	/* Octet 6 - Layer 2 Ident */
	if ((IBuf[Octet + Off] & 0x60) == 0x40) { /* Layer 1 Ident ? */
		pie->Layer2Ident = (IBuf[Octet + Off] >>5) & 0x03;
		pie->UIL2Prot    =  IBuf[Octet + Off] & 0x1f;
		Octet++;

		/* Octet 6a */
		if (IsQ931Ext(IBuf[Octet + Off - 1])) {
			if (pie->UIL2Prot == 0x10) { /* 2nd 6a */
				pie->UsrSpcL2Prot = IBuf[Octet + Off] & 0x7f;
				Off++;
			}
			else { /* assume 1st 6a */
				pie->ModeL2  = (IBuf[Octet + Off] >> 5) & 0x03;
				pie->Q933use =  IBuf[Octet + Off] & 0x03;
				Off++;
			}
			/* Octet 6b */
			if (IsQ931Ext(IBuf[Octet + Off - 1])) {
				pie->WindowSize = IBuf[Octet + Off] & 0x7f;
				Off++;
			}
		}
	}

	/* Octet 7 - layer 3 Ident */
	if ((IBuf[Octet + Off] & 0x60) == 0x60) { /* Layer 3 Ident ? */
		pie->Layer3Ident = (IBuf[Octet + Off] >> 5) & 0x03;
		pie->UIL3Prot    =  IBuf[Octet + Off] & 0x1f;
		Octet++;

		/* Octet 7a */
		if (IsQ931Ext(IBuf[Octet + Off - 1])) {
			if (pie->UIL3Prot == 0x0b) {
				/* Octet 7a + 7b AddL3Info */
				pie->AddL3Info = ((IBuf[Octet + Off] << 4) & 0xf0)
								| (IBuf[Octet + Off + 1] & 0x0f);
				Off += 2;
			}
			else {
				if (pie->UIL3Prot == 0x1f) {
					pie->ModeL3 = (IBuf[Octet + Off] >> 5) & 0x03;
					Off++;
				}
				else {
					pie->OptL3Info = IBuf[Octet + Off] & 0x7f;
					Off++;
				}

				/* Octet 7b*/
				if (IsQ931Ext(IBuf[Octet + Off - 1])) {
					pie->DefPackSize = IBuf[Octet + Off] & 0x0f;
					Off++;

					/* Octet 7c */
					if (IsQ931Ext(IBuf[Octet + Off - 1])) {
						pie->PackWinSize= IBuf[Octet + Off] & 0x7f;
					}
				}
			}
		}
	}

	Q931IESizeTest(Q931E_LLCOMP);
	Q931SetIE(*pIE, *OOff);

	*IOff = (*IOff) + Octet + Off;
	*OOff = (*OOff) + sizeof(Q931ie_LLComp);
	pie->Size = sizeof(Q931ie_LLComp);

	return Q931E_NO_ERROR;
}

L3INT Q931Pie_LLComp(Q931_TrunkInfo_t *trunk, L3UCHAR *IBuf, L3UCHAR *OBuf, L3INT *Octet)
{
	Q931ie_LLComp *pIE = (Q931ie_LLComp*)IBuf;
	L3INT rc = 0;
	L3INT Beg = *Octet;
	L3INT li;

	OBuf[(*Octet)++] = Q931ie_LOW_LAYER_COMPATIBILITY;
	li = (*Octet)++;

	/* Octet 3 */
	OBuf[(*Octet)++] = (pIE->CodStand << 6) | pIE->ITransCap;

	/* Octet 3a */
	OBuf[(*Octet)++] = 0x80 | (pIE->NegotInd << 6);

	/* Octet 4 */
	OBuf[(*Octet)++] = 0x80 | (pIE->TransMode << 5) | pIE->InfoRate;

	/* Octet 4.1 */
	if (pIE->InfoRate == 0x18) {
		OBuf[(*Octet)++] = 0x80 | pIE->RateMul;
	}

	/* Octet 5 */
	if (pIE->Layer1Ident == 0x01) {
		OBuf[(*Octet)++] = (pIE->Layer1Ident << 5) | pIE->UIL1Prot;

		/* Octet 5a */
		if ((pIE->ITransCap == 0x08 && (pIE->UIL1Prot == 0x01 || pIE->UIL1Prot == 0x08))
			|| (pIE->ITransCap == 0x10 && (pIE->UIL1Prot == 0x02 || pIE->UIL1Prot == 0x03))) {
			OBuf[(*Octet)++] = (pIE->SyncAsync<<6) | (pIE->Negot<<5) | pIE->UserRate;

			/* Octet 5b*/
			if (pIE->UIL1Prot == 0x01) {
				OBuf[(*Octet)++] =  (pIE->InterRate << 5)
							| (pIE->NIConTx << 4)
							| (pIE->NIConTx << 3)
							| (pIE->FlowCtlTx << 2)
							| (pIE->FlowCtlRx << 1);
			}
			else if (pIE->UIL1Prot == 0x08) {
				OBuf[(*Octet)++] =  (pIE->HDR << 6)
							| (pIE->MultiFrame << 5)
							| (pIE->ModeL1 << 4)
							| (pIE->NegotLLI << 3)
							| (pIE->Assignor << 2)
							| (pIE->InBandNeg << 1);
			}
			else {
				OBuf[(*Octet) - 1] |= 0x80;
			}

			/* How to detect wherever 5c and 5d is to present is not clear
			 * but they have been inculded as 'standard'
			 * Octet 5c
			 */
			if (pIE->UIL1Prot == 0x01 || pIE->UIL1Prot == 0x08) {
				OBuf[(*Octet)++] = (pIE->NumStopBits << 5) | (pIE->NumDataBits << 3) | pIE->Parity ;

				/* Octet 5d */
				OBuf[(*Octet)++] = 0x80 | (pIE->DuplexMode << 6) | pIE->ModemType;
			}
		}
		else {
			OBuf[(*Octet) - 1] |= 0x80;
		}
	}

	/* Octet 6 */
	if (pIE->Layer2Ident == 0x02) {
		OBuf[(*Octet)++] = (pIE->Layer2Ident << 5) | pIE->UIL2Prot;

		/* Octet 6a*/
		if (pIE->UIL2Prot == 0x02 /* Q.921/I.441 */
		|| pIE->UIL2Prot == 0x06 /* X.25 link layer */
		|| pIE->UIL2Prot == 0x07 /* X.25 multilink */
		|| pIE->UIL2Prot == 0x09 /* HDLC ARM */
		|| pIE->UIL2Prot == 0x0a /* HDLC NRM */
		|| pIE->UIL2Prot == 0x0b /* HDLC ABM */
		|| pIE->UIL2Prot == 0x0d /* X.75 SLP */
		|| pIE->UIL2Prot == 0x0e /* Q.922 */
		|| pIE->UIL2Prot == 0x11) { /* ISO/ECE 7776 DTE-DCE */
			OBuf[(*Octet)++] = (pIE->ModeL2 << 5) | pIE->Q933use;

			/* Octet 6b */
			OBuf[(*Octet)++] = 0x80 | pIE->WindowSize;
		}
		else if (pIE->UIL2Prot == 0x10) { /* User Specific */
			OBuf[(*Octet)++] = 0x80 | pIE->UsrSpcL2Prot;
		}
		else {
			OBuf[(*Octet) - 1] |= 0x80;
		}
	}

	/* Octet 7 */
	if (pIE->Layer3Ident == 0x03) {
		OBuf[(*Octet)++] = (pIE->Layer3Ident << 5) | pIE->UIL3Prot;

		/* Octet 7a - 3 different ones */
		if (pIE->UIL3Prot == 0x10) {
			OBuf[(*Octet++)] = 0x80 | pIE->OptL3Info;
		}
		else if (pIE->UIL3Prot == 0x06
			||  pIE->UIL3Prot == 0x07
			||  pIE->UIL3Prot == 0x08) {
			OBuf[(*Octet)++] = pIE->ModeL3 << 5;

			/* Octet 7b note 7 */
			OBuf[(*Octet)++] = pIE->DefPackSize;

			/* Octet 7c note 7 */
			OBuf[(*Octet)++] = 0x80 | pIE->PackWinSize;
		}
		else if (pIE->UIL3Prot == 0x0b) {
			OBuf[(*Octet)++] = (pIE->AddL3Info >> 4) & 0x0f;
			OBuf[(*Octet)++] = 0x80 | (pIE->AddL3Info & 0x0f);
		}
		else {
			OBuf[(*Octet) - 1] |= 0x80;
		}
	}
	else {
		Q931SetError(trunk,Q931E_LLCOMP, 7,0);
		rc = Q931E_LLCOMP;
	}

	OBuf[li] = (L3UCHAR)((*Octet) - Beg) - 2;
	return rc;
}

/**
 * @todo	Extend
 */
L3INT Q931Die_LLComp(Q931_TrunkInfo_t *trunk, Q931ie_Generic *ie, struct strstream *ostream)
{
	Q931ie_LLComp *pie = (Q931ie_LLComp *)ie;
	strstream_printf(ostream,
			"\tLow-Layer Compat IE [%#02x]:\n"
			"\t\tCoding:        %s\n",
			pie->IEId, q931_coding_type(pie->CodStand));
	return Q931E_NO_ERROR;
}


/********************************************************************************************************************************
 * Network Facility IE
 ********************************************************************************************************************************/

L3INT Q931Uie_NetFac(Q931_TrunkInfo_t *trunk, Q931mes_Generic *pMsg, L3UCHAR * IBuf, L3UCHAR * OBuf, L3INT *IOff, L3INT *OOff)
{
	Q931ie_NetFac *pie = (Q931ie_NetFac*)OBuf;
	ie *pIE = &pMsg->NetFac;
	L3INT Off = 0;
	L3INT Octet = 0;
	L3INT x = 0;
	L3INT IESize;

	*pIE = 0;

	pie->IEId = IBuf[Octet++];

	/* Octet 2 */
	IESize = IBuf[Octet++];

	pie->LenNetID = IBuf[Octet + Off]; /* full octet is used */
	Octet++;

	if (pie->LenNetID > 0) {
		/* Octet 3.1 */
		pie->TypeNetID = (IBuf[Octet + Off] >> 4) & 0x0f;
		pie->NetIDPlan =  IBuf[Octet + Off] & 0x0f;
		Off = Q931ReadExt(&IBuf[Octet], Off);
		Off++;

		/* Octet 3.2*/
		for (x = 0; x < pie->LenNetID; x++) {
			pie->NetID[x] = IBuf[Octet + Off] & 0x7f;
			Off++;
		}
	}

	/* Octet 4*/
	pie->NetFac = IBuf[Octet + Off]; /* Full Octet is used */
	Octet++;

	Q931IESizeTest(Q931E_NETFAC);
	Q931SetIE(*pIE, *OOff);

	*IOff = (*IOff) + Octet + Off;
	*OOff = (*OOff) + sizeof(Q931ie_NetFac) + x - 1;
	pie->Size = (L3UCHAR)(sizeof(Q931ie_NetFac) + x - 1);

	return Q931E_NO_ERROR;
}

L3INT Q931Pie_NetFac(Q931_TrunkInfo_t *trunk, L3UCHAR *IBuf, L3UCHAR *OBuf, L3INT *Octet)
{
	Q931ie_NetFac *pIE = (Q931ie_NetFac*)IBuf;
	L3INT rc = Q931E_NO_ERROR;
	L3INT Beg = *Octet;
	L3INT li;
	L3INT x;

	OBuf[(*Octet)++] = Q931ie_NETWORK_SPECIFIC_FACILITIES;
	li = (*Octet)++;

	/* Octet 3 */
	OBuf[(*Octet)++] = pIE->LenNetID;

	if (pIE->LenNetID > 0) {
		/* Octet 3.1 */
		OBuf[(*Octet)++] = 0x80 | (pIE->TypeNetID << 4) | pIE->NetIDPlan;

		/* Octet 3.2 */
		for (x = 0; x <pIE->LenNetID; x++) {
			OBuf[(*Octet)++] = pIE->NetID[x];
		}
	}

	/* Octet 4 */
	OBuf[(*Octet)++] = pIE->NetFac;

	OBuf[li] = (L3UCHAR)((*Octet) - Beg) - 2;
	return rc;
}

L3INT Q931Die_NetFac(Q931_TrunkInfo_t *trunk, Q931ie_Generic *ie, struct strstream *ostream)
{
	Q931ie_NetFac *pie = (Q931ie_NetFac *)ie;
	strstream_printf(ostream,
			"\tNetwork-specific Facilities IE [%#02x]:\n"
			"\t\tType of Network ID:    %s\n"
			"\t\tNetwork ID plan:       %s\n"
			"\t\tNetwork ID:            %.*s\n"
			"\t\tFacilities:            [ ",
			pie->IEId, q931_network_id_type(pie->TypeNetID), q931_network_id_plan(pie->NetIDPlan),
			pie->LenNetID, pie->NetID);
	strstream_printhex(ostream, (char *)&pie->NetFac, pie->Size - (sizeof(*pie) + pie->LenNetID));
	strstream_puts(ostream, " ]\n");
	return Q931E_NO_ERROR;
}


/********************************************************************************************************************************
 * Notify Indication IE
 ********************************************************************************************************************************/

L3INT Q931Uie_NotifInd(Q931_TrunkInfo_t *trunk, Q931mes_Generic *pMsg, L3UCHAR * IBuf, L3UCHAR * OBuf, L3INT *IOff, L3INT *OOff)
{
	Q931ie_NotifInd *pie = (Q931ie_NotifInd*)OBuf;
	ie *pIE = &pMsg->NotifInd;
	L3INT Off = 0;
	L3INT Octet = 0;
	L3INT IESize;

	*pIE = 0;

	pie->IEId = IBuf[Octet++];

	/* Octet 2*/
	IESize = IBuf[Octet++];

	/* Octet 3 */
	pie->Notification = IBuf[Octet + Off] & 0x7f;

	Off = Q931ReadExt(&IBuf[Octet], Off);
	Octet++;

	Q931IESizeTest(Q931E_NOTIFIND);
	Q931SetIE(*pIE, *OOff);

	*IOff = (*IOff) + Octet + Off;
	*OOff = (*OOff) + sizeof(Q931ie_NotifInd);
	pie->Size = sizeof(Q931ie_NotifInd);

	return Q931E_NO_ERROR;
}

L3INT Q931Pie_NotifInd(Q931_TrunkInfo_t *trunk, L3UCHAR *IBuf, L3UCHAR *OBuf, L3INT *Octet)
{
	Q931ie_NotifInd *pIE = (Q931ie_NotifInd*)IBuf;
	L3INT rc = Q931E_NO_ERROR;
	L3INT Beg = *Octet;
	L3INT li;

	OBuf[(*Octet)++] = Q931ie_NOTIFICATION_INDICATOR;
	li = (*Octet)++;

	/* Octet 3 */
	OBuf[(*Octet)++] = pIE->Notification;

	OBuf[li] = (L3UCHAR)((*Octet) - Beg) - 2;
	return rc;
}


static const char *q931_notification_type(const unsigned char type)
{
	switch (type) {
	case 0x00:
		return "User suspended";
	case 0x01:
		return "User resumed";
	case 0x02:
		return "Bearer service change";
	default:
		return "reserved";
	}
}

L3INT Q931Die_NotifInd(Q931_TrunkInfo_t *trunk, Q931ie_Generic *ie, struct strstream *ostream)
{
	Q931ie_NotifInd *pie = (Q931ie_NotifInd *)ie;
	strstream_printf(ostream,
			"\tNotification Indicator IE [%#02x]:\n"
			"\t\tDescription:  %s\n",
			pie->IEId, q931_notification_type(pie->Notification));
	return Q931E_NO_ERROR;
}


/********************************************************************************************************************************
 * Progress Indication IE
 ********************************************************************************************************************************/

L3INT Q931Uie_ProgInd(Q931_TrunkInfo_t *trunk, Q931mes_Generic *pMsg, L3UCHAR * IBuf, L3UCHAR * OBuf, L3INT *IOff, L3INT *OOff)
{
	Q931ie_ProgInd *pie = (Q931ie_ProgInd*)OBuf;
	ie *pIE = &pMsg->ProgInd;
	L3INT Off = 0;
	L3INT Octet = 0;
	L3INT IESize;

	*pIE = 0;

	pie->IEId = IBuf[Octet++];

	/* Octet 2 */
	IESize = IBuf[Octet++];

	/* Octet 3 */
	pie->CodStand = (IBuf[Octet + Off] >> 5) & 0x03;
	pie->Location =  IBuf[Octet + Off] & 0x0f;

	Off = Q931ReadExt(&IBuf[Octet], Off);
	Octet++;

	/* Octet 4 */
	pie->ProgDesc = IBuf[Octet + Off] & 0x7f;
	Off = Q931ReadExt(&IBuf[Octet], Off);
	Octet++;

	Q931IESizeTest(Q931E_PROGIND);
	Q931SetIE(*pIE, *OOff);

	*IOff = (*IOff) + Octet + Off;
	*OOff = (*OOff) + sizeof(Q931ie_ProgInd);
	pie->Size = sizeof(Q931ie_ProgInd);

	return Q931E_NO_ERROR;
}

L3INT Q931Pie_ProgInd(Q931_TrunkInfo_t *trunk, L3UCHAR *IBuf, L3UCHAR *OBuf, L3INT *Octet)
{
	Q931ie_ProgInd *pIE = (Q931ie_ProgInd*)IBuf;
	L3INT rc = Q931E_NO_ERROR;
	L3INT Beg = *Octet;
	L3INT li;

	OBuf[(*Octet)++] = Q931ie_PROGRESS_INDICATOR;
	li = (*Octet)++;

	/* Octet 3 */
	OBuf[(*Octet)++] = 0x80 | (pIE->CodStand << 5) | pIE->Location;

	/* Octet 4 */
	OBuf[(*Octet)++] = 0x80 | pIE->ProgDesc;

	OBuf[li] = (L3UCHAR)((*Octet) - Beg) - 2;
	return rc;
}

L3INT Q931Die_ProgInd(Q931_TrunkInfo_t *trunk, Q931ie_Generic *ie, struct strstream *ostream)
{
	Q931ie_ProgInd *pie = (Q931ie_ProgInd *)ie;
	strstream_printf(ostream,
			"\tProgress Indicator IE [%#02x]:\n"
			"\t\tCoding:          %s\n"
			"\t\tLocation:        %s\n"
			"\t\tProgress Desc.:  %s\n",
			pie->IEId, q931_coding_type(pie->CodStand), q931_location_type(pie->Location),
			q931_progdesc_type(pie->ProgDesc));
	return Q931E_NO_ERROR;
}


/********************************************************************************************************************************
 * Repeat Indication IE
 ********************************************************************************************************************************/

L3INT Q931Uie_RepeatInd(Q931_TrunkInfo_t *trunk, Q931mes_Generic *pMsg, L3UCHAR * IBuf, L3UCHAR * OBuf, L3INT *IOff, L3INT *OOff)
{
	Q931ie_RepeatInd *pie = (Q931ie_RepeatInd*)OBuf;
	ie *pIE = &pMsg->RepeatInd;
	L3INT Off = 0;
	L3INT Octet = 0;

	*pIE = 0;

	pie->IEId      = IBuf[Octet] & 0xf0;
	pie->RepeatInd = IBuf[Octet] & 0x0f;
	Octet ++;

	Q931SetIE(*pIE, *OOff);

	*IOff = (*IOff) + Octet + Off;
	*OOff = (*OOff) + sizeof(Q931ie_RepeatInd);
	pie->Size = sizeof(Q931ie_RepeatInd);

	return Q931E_NO_ERROR;
}

L3INT Q931Pie_RepeatInd(Q931_TrunkInfo_t *trunk, L3UCHAR *IBuf, L3UCHAR *OBuf, L3INT *Octet)
{
	Q931ie_RepeatInd *pIE = (Q931ie_RepeatInd*)IBuf;
	L3INT rc = 0;
	/* L3INT Beg = *Octet; */

	OBuf[(*Octet)++] = Q931ie_REPEAT_INDICATOR | pIE->RepeatInd;

	return rc;
}

L3INT Q931Die_RepeatInd(Q931_TrunkInfo_t *trunk, Q931ie_Generic *ie, struct strstream *ostream)
{
	Q931ie_RepeatInd *pie = (Q931ie_RepeatInd *)ie;
	strstream_printf(ostream,
			"\tRepeat Indicator IE [%#02x]:\n",
			pie->IEId);
	return Q931E_NO_ERROR;
}


/********************************************************************************************************************************
 * Rev-Charge Indication IE
 ********************************************************************************************************************************/

L3INT Q931Uie_RevChargeInd(Q931_TrunkInfo_t *trunk, Q931mes_Generic *pMsg, L3UCHAR * IBuf, L3UCHAR * OBuf, L3INT *IOff, L3INT *OOff)
{
	ie iE;
	/* ie *pIE = &pMsg->RevChargeInd; */
	Q931SetIE(iE, *OOff);

	return iE;
}

L3INT Q931Pie_RevChargeInd(Q931_TrunkInfo_t *trunk, L3UCHAR *IBuf, L3UCHAR *OBuf, L3INT *Octet)
{
	L3BOOL RetCode = L3FALSE;

	NoWarning(OBuf);
	NoWarning(IBuf);

	return RetCode;
}

L3INT Q931Die_RevChargeInd(Q931_TrunkInfo_t *trunk, Q931ie_Generic *ie, struct strstream *ostream)
{
	strstream_printf(ostream,
			"\tRevCharge Indicator IE [%#02x]:\n",
			ie->IEId);
	return Q931E_NO_ERROR;
}


/********************************************************************************************************************************
 * Restart Indication IE
 ********************************************************************************************************************************/

L3INT Q931Uie_RestartInd(Q931_TrunkInfo_t *trunk, Q931mes_Generic *pMsg, L3UCHAR * IBuf, L3UCHAR * OBuf, L3INT *IOff, L3INT *OOff)
{
	Q931ie_RestartInd *pie = (Q931ie_RestartInd*)OBuf;
	ie *pIE = &pMsg->RestartInd;
	L3INT Off = 0;
	L3INT Octet = 0;
	L3INT IESize;

	*pIE = 0;

	pie->IEId = IBuf[Octet++];

	/* Octet 2 */
	IESize = IBuf[Octet++];

	/* Octet 3 */
	pie->Class = IBuf[Octet + Off] & 0x07;
	pie->Spare = IBuf[Octet + Off] & 0x70;

	Off = Q931ReadExt(&IBuf[Octet], Off);
	Octet++;

	Q931IESizeTest(Q931E_RESTARTIND);
	Q931SetIE(*pIE, *OOff);

	*IOff = (*IOff) + Octet + Off;
	*OOff = (*OOff) + sizeof(Q931ie_RestartInd);
	pie->Size = sizeof(Q931ie_RestartInd);


	return Q931E_NO_ERROR;
}

L3INT Q931Pie_RestartInd(Q931_TrunkInfo_t *trunk, L3UCHAR *IBuf, L3UCHAR *OBuf, L3INT *Octet)
{
	Q931ie_RestartInd *pIE = (Q931ie_RestartInd*)IBuf;
	L3INT rc = Q931E_NO_ERROR;
	L3INT Beg = *Octet;
	L3INT li;

	OBuf[(*Octet)++] = Q931ie_RESTART_INDICATOR;
	li = (*Octet)++;

	/* Octet 3*/
	OBuf[(*Octet)++] = 0x80 | pIE->Class;

	OBuf[li] = (L3UCHAR)((*Octet) - Beg) - 2;
	return rc;
}

L3INT Q931Die_RestartInd(Q931_TrunkInfo_t *trunk, Q931ie_Generic *ie, struct strstream *ostream)
{
	Q931ie_RestartInd *pie = (Q931ie_RestartInd *)ie;
	const char *rclass = NULL;

	switch (pie->Class) {
	case 0x00:
		rclass = "Channel";
		break;
	case 0x06:
		rclass = "Interface";
		break;
	case 0x07:
		rclass = "All interfaces";
		break;
	default:
		rclass = "Invalid/unknown";
	}

	strstream_printf(ostream,
			"\tRestart Indicator IE [%#02x]:\n"
			"\t\tClass: %s\n",
			pie->IEId, rclass);
	return Q931E_NO_ERROR;
}


/********************************************************************************************************************************
 * Segment IE
 ********************************************************************************************************************************/

L3INT Q931Uie_Segment(Q931_TrunkInfo_t *trunk, Q931mes_Generic *pMsg, L3UCHAR * IBuf, L3UCHAR * OBuf, L3INT *IOff, L3INT *OOff)
{
	Q931ie_Segment *pie = (Q931ie_Segment*)OBuf;
	ie *pIE = &pMsg->Segment;
	L3INT Off = 0;
	L3INT Octet = 0;
	L3INT IESize;

	*pIE = 0;

	pie->IEId = IBuf[Octet++];
	Octet++;

	/* Octet 2*/
	IESize = IBuf[Octet++];

	/* Octet 3 */
	pie->FSI       = (IBuf[Octet + Off] & 0x80) >> 7;
	pie->NumSegRem =  IBuf[Octet + Off] & 0x7f;
	Octet++;

	/* Octet 4 */
	pie->SegType = IBuf[Octet + Off] & 0x7f;
	Octet++;

	Q931IESizeTest(Q931E_SEGMENT);
	Q931SetIE(*pIE, *OOff);

	*IOff = (*IOff) + Octet + Off;
	*OOff = (*OOff) + sizeof(Q931ie_Segment);
	pie->Size = sizeof(Q931ie_Segment);

	return Q931E_NO_ERROR;
}

L3INT Q931Pie_Segment(Q931_TrunkInfo_t *trunk, L3UCHAR *IBuf, L3UCHAR *OBuf, L3INT *Octet)
{
	Q931ie_Segment *pIE = (Q931ie_Segment*)IBuf;
	L3INT rc = Q931E_NO_ERROR;
	L3INT Beg = *Octet;
	L3INT li;

	OBuf[(*Octet)++] = Q931ie_SEGMENTED_MESSAGE;
	li = (*Octet)++;

	/* Octet 3 */
	OBuf[(*Octet)++] = (pIE->FSI << 7) | pIE->NumSegRem;

	/* Octet 4 */
	OBuf[(*Octet)++] = pIE->SegType;

	OBuf[li] = (L3UCHAR)((*Octet) - Beg) - 2;
	return rc;
}

L3INT Q931Die_Segment(Q931_TrunkInfo_t *trunk, Q931ie_Generic *ie, struct strstream *ostream)
{
	Q931ie_Segment *pie = (Q931ie_Segment *)ie;
	strstream_printf(ostream,
			"\tSegment IE [%#02x]:\n"
			"\t\tFirst segment:  %s\n"
			"\t\tRemaining:      %d\n"
			"\t\tType:           %#02x (%s)\n",
			pie->IEId, pie->FSI ? "Yes" : "No", pie->NumSegRem,
			pie->SegType, "-N/A-");
	return Q931E_NO_ERROR;
}


/********************************************************************************************************************************
 * Sending Complete IE
 ********************************************************************************************************************************/

/**
 * Decode SENDING COMPLETE information element
 * @param[in]		trunk	Q.931 trunk
 * @param[out]		pMsg	Decoded message
 * @param[in]		IBuf	Input buffer
 * @param[out]		OBuf	Output buffer (pointer to IE struct)
 * @param[in,out]	IOff	Current position in input buffer
 * @param[in,out]	OOff	Current position in output buffer
 * @return		Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931Uie_SendComplete(Q931_TrunkInfo_t *trunk, Q931mes_Generic *pMsg, L3UCHAR * IBuf, L3UCHAR * OBuf, L3INT *IOff, L3INT *OOff)
{
	Q931ie_SendComplete *pie = (Q931ie_SendComplete*)OBuf;
	ie *pIE = &pMsg->SendComplete;
	L3INT Off = 0;
	L3INT Octet = 0;

	*pIE = 0;
	Octet++;

	Q931SetIE(*pIE, *OOff);

	*IOff = (*IOff) + Octet + Off;
	*OOff = (*OOff) + sizeof(Q931ie_SendComplete);
	pie->Size = sizeof(Q931ie_SendComplete);

	return Q931E_NO_ERROR;
}

/**
 * Encode SENDING COMPLETE information element
 * @param[in]		trunk	Q.931 trunk
 * @param[in]		IBuf	Input buffer
 * @param[out]		OBuf	Output buffer
 * @param[in,out]	Octet	Current position in output buffer
 * @return		Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931Pie_SendComplete(Q931_TrunkInfo_t *trunk, L3UCHAR *IBuf, L3UCHAR *OBuf, L3INT *Octet)
{
	/* Q931ie_SendComplete * pIE = (Q931ie_SendComplete*)IBuf; */
	/* L3INT Beg = *Octet; */

	OBuf[(*Octet)++] = 0x80 | (L3UCHAR)Q931ie_SENDING_COMPLETE;

	return Q931E_NO_ERROR;
}

/**
 * Pretty-print SENDING COMPLETE information element
 * @param[in]		trunk	Q.931 trunk
 * @param[in]		ie	Pointer to generic IE struct
 * @return		Q931E_NO_ERROR on success, Q931E_* otherwise
 * @internal
 */
L3INT Q931Die_SendComplete(Q931_TrunkInfo_t *trunk, Q931ie_Generic *ie, struct strstream *ostream)
{
	strstream_printf(ostream,
		"\tSending Complete IE [%#02x]\n",
		ie->IEId);
	return Q931E_NO_ERROR;
}


/********************************************************************************************************************************
 * Signal IE
 ********************************************************************************************************************************/

L3INT Q931Uie_Signal(Q931_TrunkInfo_t *trunk, Q931mes_Generic *pMsg, L3UCHAR * IBuf, L3UCHAR * OBuf, L3INT *IOff, L3INT *OOff)
{
	Q931ie_Signal *pie = (Q931ie_Signal*)OBuf;
	ie *pIE = &pMsg->Signal;
	L3INT Off = 0;
	L3INT Octet = 0;
	L3INT IESize;

	*pIE = 0;

	pie->IEId = IBuf[Octet++];

	/* Octet 2 */
	IESize = IBuf[Octet++];

	/* Octet 3 */
	pie->Signal = IBuf[Octet + Off];
	Octet++;

	Q931IESizeTest(Q931E_SIGNAL);
	Q931SetIE(*pIE, *OOff);

	*IOff = (*IOff) + Octet + Off;
	*OOff = (*OOff) + sizeof(Q931ie_Signal);
	pie->Size = sizeof(Q931ie_Signal);

	return Q931E_NO_ERROR;
}

L3INT Q931Pie_Signal(Q931_TrunkInfo_t *trunk, L3UCHAR *IBuf, L3UCHAR *OBuf, L3INT *Octet)
{
	Q931ie_Signal *pIE = (Q931ie_Signal*)IBuf;
	L3INT rc = Q931E_NO_ERROR;
	L3INT Beg = *Octet;
	L3INT li;

	OBuf[(*Octet)++] = Q931ie_SIGNAL;
	li = (*Octet)++;

	/* Octet 3 */
	OBuf[(*Octet)++] = pIE->Signal;

	OBuf[li] = (L3UCHAR)((*Octet) - Beg) - 2;
	return rc;
}

L3INT Q931Die_Signal(Q931_TrunkInfo_t *trunk, Q931ie_Generic *ie, struct strstream *ostream)
{
	Q931ie_Signal *pie = (Q931ie_Signal *)ie;
	strstream_printf(ostream,
			"\tSignal IE [%#02x]:\n"
			"\t\tSignal:    %#02x (%s)\n",
			pie->IEId, pie->Signal, q931_signal_desc(pie->Signal));
	return Q931E_NO_ERROR;
}


/********************************************************************************************************************************
 * Trans-Net Selector IE
 ********************************************************************************************************************************/

L3INT Q931Uie_TransNetSel(Q931_TrunkInfo_t *trunk, Q931mes_Generic *pMsg, L3UCHAR * IBuf, L3UCHAR * OBuf, L3INT *IOff, L3INT *OOff)
{
	Q931ie_TransNetSel *pie = (Q931ie_TransNetSel*)OBuf;
	ie *pIE = &pMsg->TransNetSel;
	L3INT Off = 0;
	L3INT Octet = 0;
	L3INT x = 0;
	L3INT l;

	*pIE = 0;

	pie->IEId = IBuf[Octet++];

	/* Octet 2 */
	l = IBuf[Octet++] - 3;

	/* Octet 3 */
	pie->Type = (IBuf[Octet + Off] >> 4) & 0x07;

	Off = Q931ReadExt(&IBuf[Octet], Off);
	Octet++;

	for (x = 0; x < l; x++) {
		pie->NetID[x] = IBuf[Octet + Off] & 0x7f;
		Off++;
	}

	Q931SetIE(*pIE, *OOff);

	*IOff = (*IOff) + Octet + Off;
	*OOff = (*OOff) + sizeof(Q931ie_TransNetSel) + x - 1;
	pie->Size = (L3UCHAR)(sizeof(Q931ie_TransNetSel) + x - 1);

	return Q931E_NO_ERROR;
}

L3INT Q931Pie_TransNetSel(Q931_TrunkInfo_t *trunk, L3UCHAR *IBuf, L3UCHAR *OBuf, L3INT *Octet)
{
	Q931ie_TransNetSel *pIE = (Q931ie_TransNetSel*)IBuf;
	L3INT rc = Q931E_NO_ERROR;
	L3INT Beg = *Octet;
	L3INT li;
	L3INT x;
	L3INT l;

	OBuf[(*Octet)++] = Q931ie_TRANSIT_NETWORK_SELECTION;
	li = (*Octet)++;

	/* Octet 3 */
	OBuf[(*Octet)++] = 0x80 | (pIE->Type << 4) | pIE->NetIDPlan;

	/* Octet 4 */
	l = pIE->Size - sizeof(Q931ie_TransNetSel) + 1;
	for (x = 0; x < l; x++) {
		OBuf[(*Octet)++] = pIE->NetID[x];
	}

	OBuf[li] = (L3UCHAR)((*Octet) - Beg) - 2;
	return rc;
}

L3INT Q931Die_TransNetSel(Q931_TrunkInfo_t *trunk, Q931ie_Generic *ie, struct strstream *ostream)
{
	Q931ie_TransNetSel *pie = (Q931ie_TransNetSel *)ie;
	strstream_printf(ostream,
			"\tTransit Network Selection IE [%#02x]:\n"
			"\t\tType of Network ID:    %s\n"
			"\t\tNetwork ID plan:       %s\n"
			"\t\tNetwork ID:            %.*s\n",
			pie->IEId, q931_network_id_type(pie->Type), q931_network_id_plan(pie->NetIDPlan),
			pie->Size - sizeof(*pie), pie->NetID);
	return Q931E_NO_ERROR;
}


/********************************************************************************************************************************
 * User-User IE
 ********************************************************************************************************************************/

L3INT Q931Uie_UserUser(Q931_TrunkInfo_t *trunk, Q931mes_Generic *pMsg, L3UCHAR * IBuf, L3UCHAR * OBuf, L3INT *IOff, L3INT *OOff)
{
	Q931ie_UserUser *pie = (Q931ie_UserUser*)OBuf;
	ie *pIE = &pMsg->UserUser;
	L3INT Octet = 0;
	L3INT Off = 0;
	L3INT l;

	*pIE = 0;

	pie->IEId = IBuf[Octet++];

	/* Octet 2 */
	l = IBuf[Octet++] - 1;

	/* Octet 3 */
	pie->ProtDisc = IBuf[Octet++];

	for (Off = 0; Off < l; Off++) {
		pie->User[Off] = IBuf[Octet + Off];
	}

	Q931SetIE(*pIE, *OOff);

	*IOff = (*IOff) + Octet + Off;
	*OOff = (*OOff) + sizeof(Q931ie_UserUser) + Off - 1;
	pie->Size = (L3UCHAR)(sizeof(Q931ie_UserUser) + Off - 1);

	return Q931E_NO_ERROR;
}

L3INT Q931Pie_UserUser(Q931_TrunkInfo_t *trunk, L3UCHAR *IBuf, L3UCHAR *OBuf, L3INT *Octet)
{
	Q931ie_UserUser *pIE = (Q931ie_UserUser*)IBuf;
	L3INT rc = Q931E_NO_ERROR;
	L3INT Beg = *Octet;
	L3INT li;
	L3INT x;
	L3INT l;

	OBuf[(*Octet)++] = Q931ie_USER_USER;
	li = (*Octet)++;

	/* Octet 3 */
	OBuf[(*Octet)++] = pIE->ProtDisc;

	/* Octet 4 */
	l = pIE->Size - sizeof(Q931ie_UserUser) + 1;
	for (x = 0; x < l; x++) {
		OBuf[(*Octet)++] = pIE->User[x];
	}

	OBuf[li] = (L3UCHAR)((*Octet) - Beg) - 2;
	return rc;
}

L3INT Q931Die_UserUser(Q931_TrunkInfo_t *trunk, Q931ie_Generic *ie, struct strstream *ostream)
{
	Q931ie_UserUser *pie = (Q931ie_UserUser *)ie;
	strstream_printf(ostream,
			"\tUser-User IE [%#02x]:\n",
			pie->IEId);

	strstream_puts(ostream, "\t\tValue:    [ ");
	strstream_printhex(ostream, (char *)pie->User, pie->Size - sizeof(*pie));
	strstream_puts(ostream, " ]\n");
	return Q931E_NO_ERROR;
}


/********************************************************************************************************************************
 * Generic Digits IE
 ********************************************************************************************************************************/

L3INT Q931Uie_GenericDigits(Q931_TrunkInfo_t *trunk, Q931mes_Generic *pMsg, L3UCHAR * IBuf, L3UCHAR * OBuf, L3INT *IOff, L3INT *OOff)
{
	Q931ie_GenericDigits *pie = (Q931ie_GenericDigits*)OBuf;
	ie *pIE = &pMsg->GenericDigits;
	L3INT Off = 0;
	L3INT Octet = 0;
	L3INT x;
	L3INT IESize;

	*pIE = 0;

	/* Octet 1 */
	pie->IEId = IBuf[Octet++];

	/* Octet 2 */
	IESize = IBuf[Octet++];

	/* Octet 3 */
	pie->Type     = (IBuf[Octet]) & 0x1F;
	pie->Encoding = (IBuf[Octet] >> 5) & 0x07;
	Octet++;

	/* Octet 4*/
	if (pie->Encoding == 0) { /* BCD Even */
		x = 0;
		do {
			pie->Digit[x++] =  IBuf[Octet + Off] & 0x0f;
			pie->Digit[x++] = (IBuf[Octet + Off] >> 4) & 0x0f;
			Off++;
		} while (Q931MoreIE());
	} else if (pie->Encoding == 1) { /* BCD Odd */
		x = 0;
		do {
			pie->Digit[x++] = IBuf[Octet + Off] & 0x0f;
			if (Q931MoreIE()) {
				pie->Digit[x] = (IBuf[Octet + Off] >> 4) & 0x0f;
			}
			x++;
			Off++;
		} while (Q931MoreIE());
	} else if (pie->Encoding == 2) { /* IA5 */
		x = 0;
		do {
			pie->Digit[x++] = IBuf[Octet + Off] & 0x7f;
			Off++;
		} while (Q931MoreIE());
	} else {
		/* Binary encoding type unkown */
		Q931SetError(trunk, Q931E_GENERIC_DIGITS, Octet, Off);
		return Q931E_GENERIC_DIGITS;
	}

	Q931IESizeTest(Q931E_GENERIC_DIGITS);
	Q931SetIE(*pIE, *OOff);

	*IOff = (*IOff) + Octet + Off;
	*OOff = (*OOff) + sizeof(Q931ie_CallingSub) + x - 1;
	pie->Size = (L3UCHAR)(sizeof(Q931ie_CallingSub) + x - 1);

	return Q931E_NO_ERROR;
}

L3INT Q931Pie_GenericDigits(Q931_TrunkInfo_t *trunk, L3UCHAR *IBuf, L3UCHAR *OBuf, L3INT *Octet)
{
	OBuf[(*Octet)++] = (Q931ie_GENERIC_DIGITS & 0xFF);
	OBuf[(*Octet)++] = 0;

	return Q931E_NO_ERROR;
}

L3INT Q931Die_GenericDigits(Q931_TrunkInfo_t *trunk, Q931ie_Generic *ie, struct strstream *ostream)
{
	Q931ie_GenericDigits *pie = (Q931ie_GenericDigits *)ie;
	strstream_printf(ostream,
			"\tGeneric Digits IE [%#02x]:\n",
			pie->IEId);
	return Q931E_NO_ERROR;
}


/********************************************************************************************************************************
 * Change Status IE
 ********************************************************************************************************************************/

L3INT Q931Uie_ChangeStatus(Q931_TrunkInfo_t *trunk, Q931mes_Generic *pMsg, L3UCHAR * IBuf, L3UCHAR * OBuf, L3INT *IOff, L3INT *OOff)
{
	Q931ie_ChangeStatus *pie = (Q931ie_ChangeStatus*)OBuf;
	ie *pIE = &pMsg->ChangeStatus;
	L3INT Off = 0;
	L3INT Octet = 0;
	L3INT IESize;

	*pIE = 0;

	pie->IEId = IBuf[Octet++];

	/* Octet 2 */
	IESize = IBuf[Octet++];

	/* Octet 3 */
	pie->Preference = (IBuf[Octet + Off] >> 6) & 0x01;
	pie->Spare      =  IBuf[Octet + Off] & 0x38;
	pie->NewStatus  =  IBuf[Octet + Off] & 0x07;
	Octet++;

	Q931SetIE(*pIE, *OOff);

	*IOff = (*IOff) + Octet + Off;
	*OOff = (*OOff) + sizeof(Q931ie_ChangeStatus);
	pie->Size = sizeof(Q931ie_ChangeStatus);

	return Q931E_NO_ERROR;
}

L3INT Q931Pie_ChangeStatus(Q931_TrunkInfo_t *trunk, L3UCHAR *IBuf, L3UCHAR *OBuf, L3INT *Octet)
{
	Q931ie_ChangeStatus *pIE = (Q931ie_ChangeStatus*)IBuf;
	L3INT rc = Q931E_NO_ERROR;
	L3INT Beg = *Octet;
	L3INT li;

	OBuf[(*Octet)++] = Q931ie_CHANGE_STATUS;
	li = (*Octet)++;

	/* Octet 3 */
	OBuf[(*Octet)++] = 0x80 | pIE->NewStatus | ((pIE->Preference & 0x01) << 6);

	OBuf[li] = (L3UCHAR)((*Octet) - Beg) - 2;
	return rc;
}

L3INT Q931Die_ChangeStatus(Q931_TrunkInfo_t *trunk, Q931ie_Generic *ie, struct strstream *ostream)
{
	Q931ie_ChangeStatus *pie = (Q931ie_ChangeStatus *)ie;
	strstream_printf(ostream,
			"\tChange Status IE [%#02x]:\n"
			"\t\tNew Status: %#02x\n"
			"\t\tPreference: %s\n",
			pie->IEId, pie->NewStatus, pie->Preference ? "yes" : "no");
	return Q931E_NO_ERROR;
}


/********************************************************************************************************************************
 * Generic IE handling
 ********************************************************************************************************************************/

L3INT Q931Uie_Generic(Q931_TrunkInfo_t *trunk, Q931mes_Generic *pMsg, L3UCHAR *IBuf, L3UCHAR *OBuf, L3INT *IOff, L3INT *OOff)
{
	L3INT Octet = 0;
	L3UCHAR id = 0;

	/* id */
	id = IBuf[Octet++];

	/* Length */
	if (!(id & 0x80)) {
		Octet += IBuf[Octet];
		Octet++;

		Q931Log(trunk, Q931_LOG_DEBUG, "Discarding IE %#02x with length %d\n", id, Octet - 2);
	} else {
		Q931Log(trunk, Q931_LOG_DEBUG, "Discarding single Octet IE %#02x\n", id);
	}

	*IOff += Octet;
	return Q931E_NO_ERROR;
}

L3INT Q931Pie_Generic(Q931_TrunkInfo_t *trunk, L3UCHAR *IBuf, L3UCHAR *OBuf, L3INT *Octet)
{
	/* do nothing */
	return Q931E_NO_ERROR;
}

L3INT Q931Die_Generic(Q931_TrunkInfo_t *trunk, Q931ie_Generic *ie, struct strstream *ostream)
{
	strstream_printf(ostream, "\tUnhandled IE [%#02x]\n", ie->IEId);
	return Q931E_NO_ERROR;
}
