/*****************************************************************************

  FileName:	Q931mes.c

  Contents:	Pack/Unpack functions. These functions will unpack a Q931
		message from the bit packed original format into structs
		that contains variables sized by the user. It will also pack
		the struct back into a Q.931 message as required.

		See q931.h for description.

  License/Copyright:

  Copyright (c) 2007, Jan Vidar Berger, Case Labs, Ltd. All rights reserved.
  email:janvb@caselaboratories.com

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are
  met:

	* Redistributions of source code must retain the above copyright notice,
	  this list of conditions and the following disclaimer.
	* Redistributions in binary form must reproduce the above copyright notice,
	  this list of conditions and the following disclaimer in the documentation
	  and/or other materials provided with the distribution.
	* Neither the name of the Case Labs, Ltd nor the names of its contributors
	  may be used to endorse or promote products derived from this software
	  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.

*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <limits.h>

#include "Q931.h"
#include "Q931priv.h"
#include "Q932.h"

/*
 * TODO: Convert this into something like:
 *
 * { Q931mes_ALERTING,
 *			 { Q931ie_BEARER_CAPABILITY, 4, 12, Q931_IEF_TO_BOTH },
 *			 { Q931ie_CHANNEL_ID,        2,  0, Q931_IEF_TO_BOTH },
 *      		 ...
 * },
 */

struct Q931MessageIE Q931MessageIEs[] = {
	/*
	 * ITU-T Q.931 IE table
	 */

	/* ALERTING */
	{ Q931mes_ALERTING, 8, "ALERTING", {
			{ Q931ie_BEARER_CAPABILITY,        4, 12, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q931ie_CHANNEL_IDENTIFICATION,   2,  0, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q931ie_PROGRESS_INDICATOR,       2,  4, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q931ie_DISPLAY,                  2, 34, Q931_IE_CODESET_0, Q931_IEF_TO_USER },
			{ Q931ie_SIGNAL,                   2,  3, Q931_IE_CODESET_0, Q931_IEF_TO_USER },
			{ Q931ie_HIGH_LAYER_COMPATIBILITY, 2,  5, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },

			{ Q932ie_FACILITY,                 2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q932ie_EXTENDED_FACILITY,        2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },

/* Q.952 */
			{ Q931ie_REDIRECTION_NUMBER,       2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_USER },

			{ 0, 0, 0, 0 },
	}},

	/* CALL PROCEEDING */
	{ Q931mes_CALL_PROCEEDING, 8, "CALL PROCEEDING", {
			{ Q931ie_BEARER_CAPABILITY,        4, 12, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q931ie_CHANNEL_IDENTIFICATION,   2,  0, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q931ie_PROGRESS_INDICATOR,       2,  4, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q931ie_DISPLAY,                  2, 34, Q931_IE_CODESET_0, Q931_IEF_TO_USER },
			{ Q931ie_HIGH_LAYER_COMPATIBILITY, 2,  5, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },

			{ Q932ie_FACILITY,                 2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q932ie_EXTENDED_FACILITY,        2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },

			{ 0, 0, 0, 0 },
	}},

	/* CONNECT */
	{ Q931mes_CONNECT, 8, "CONNECT", {
			{ Q931ie_BEARER_CAPABILITY,        4, 12, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q931ie_CHANNEL_IDENTIFICATION,   2,  0, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q931ie_PROGRESS_INDICATOR,       2,  4, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q931ie_DISPLAY,                  2, 34, Q931_IE_CODESET_0, Q931_IEF_TO_USER },
			{ Q931ie_DATETIME,                 5,  8, Q931_IE_CODESET_0, Q931_IEF_TO_USER },
			{ Q931ie_SIGNAL,                   2,  3, Q931_IE_CODESET_0, Q931_IEF_TO_USER },
			{ Q931ie_LOW_LAYER_COMPATIBILITY,  2, 18, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q931ie_HIGH_LAYER_COMPATIBILITY, 2,  5, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },

			{ Q931ie_CONNECTED_NUMBER,         4, 24, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
/* Q.951.5 */		{ Q931ie_CONNECTED_SUBADDRESS,     2, 23, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },

			{ Q932ie_FACILITY,                 2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q932ie_EXTENDED_FACILITY,        2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },

/* Q.952 */
			{ Q931ie_REDIRECTION_NUMBER,       2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_USER },

			{ 0, 0, 0, 0 },
	}},

	/* CONNECT ACKNOWLEDGE */
	{ Q931mes_CONNECT_ACKNOWLEDGE, 8, "CONNECT ACKNOWLEDGE", {
			{ Q931ie_DISPLAY,           2, 34, Q931_IE_CODESET_0, Q931_IEF_TO_USER },
			{ Q931ie_SIGNAL,            2,  3, Q931_IE_CODESET_0, Q931_IEF_TO_USER },

			{ Q932ie_FACILITY,          2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q932ie_EXTENDED_FACILITY, 2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },

			{ 0, 0, 0, 0 },
	}},

	/* DISCONNECT */
	{ Q931mes_DISCONNECT, 8, "DISCONNECT", {
			{ Q931ie_CAUSE,              4, 32, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH | Q931_IEF_MANDATORY },
			{ Q931ie_PROGRESS_INDICATOR, 2,  4, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q931ie_DISPLAY,            2, 34, Q931_IE_CODESET_0, Q931_IEF_TO_USER },
			{ Q931ie_SIGNAL,             2,  3, Q931_IE_CODESET_0, Q931_IEF_TO_USER },

			{ Q932ie_FACILITY,           2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q932ie_EXTENDED_FACILITY,  2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },

			{ 0, 0, 0, 0 },
	}},

	/* INFORMATION */
	{ Q931mes_INFORMATION, 8, "INFORMATION", {
			{ Q931ie_SENDING_COMPLETE,    1,  1, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q931ie_DISPLAY,             2, 34, Q931_IE_CODESET_0, Q931_IEF_TO_USER },
			{ Q931ie_KEYPAD_FACILITY,     2, 34, Q931_IE_CODESET_0, Q931_IEF_TO_NET },
			{ Q931ie_SIGNAL,              2,  3, Q931_IE_CODESET_0, Q931_IEF_TO_USER },
			{ Q931ie_CALLED_PARTY_NUMBER, 2, 34, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ 0, 0, 0, 0 },
	}},

	/* NOTIFY */
	{ Q931mes_NOTIFY, 8, "NOTIFY", {
			{ Q931ie_BEARER_CAPABILITY,      2, 12, Q931_IE_CODESET_0, Q931_IEF_TO_USER },
			{ Q931ie_NOTIFICATION_INDICATOR, 3,  3, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH | Q931_IEF_MANDATORY },
			{ Q931ie_DISPLAY,                2, 34, Q931_IE_CODESET_0, Q931_IEF_TO_USER },

/* Q.952 */
			{ Q931ie_REDIRECTION_NUMBER,       2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_USER },

			{ 0, 0, 0, 0 },
	}},

	/* PROGRESS */
	{ Q931mes_PROGRESS, 8, "PROGRESS", {
			{ Q931ie_BEARER_CAPABILITY,        4, 12, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q931ie_CAUSE,                    4, 32, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q931ie_PROGRESS_INDICATOR,       2,  4, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH | Q931_IEF_MANDATORY },
			{ Q931ie_DISPLAY,                  2, 34, Q931_IE_CODESET_0, Q931_IEF_TO_USER },
			{ Q931ie_HIGH_LAYER_COMPATIBILITY, 2,  5, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },

			{ Q932ie_FACILITY,                 2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q932ie_EXTENDED_FACILITY,        2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },

/* Q.952 */
			{ Q931ie_REDIRECTION_NUMBER,       2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_USER },

			{ 0, 0, 0, 0 },
	}},

	/* RELEASE */
	{ Q931mes_RELEASE, 8, "RELEASE", {
			{ Q931ie_CAUSE,             4, 32, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q931ie_DISPLAY,           2, 34, Q931_IE_CODESET_0, Q931_IEF_TO_USER },
			{ Q931ie_SIGNAL,            2,  3, Q931_IE_CODESET_0, Q931_IEF_TO_USER },

			{ Q932ie_FACILITY,          2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q932ie_EXTENDED_FACILITY, 2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },

			{ 0, 0, 0, 0 },
	}},

	/* RELEASE COMPLETE */
	{ Q931mes_RELEASE_COMPLETE, 8, "RELEASE COMPLETE", {
			{ Q931ie_CAUSE,             4, 32, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q931ie_DISPLAY,           2, 34, Q931_IE_CODESET_0, Q931_IEF_TO_USER },
			{ Q931ie_SIGNAL,            2,  3, Q931_IE_CODESET_0, Q931_IEF_TO_USER },

			{ Q932ie_FACILITY,          2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q932ie_EXTENDED_FACILITY, 2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },

			{ 0, 0, 0, 0 },
	}},

	/* RESUME */
	{ Q931mes_RESUME, 8, "RESUME", {
			{ Q931ie_CALL_IDENTITY, 2, 10, Q931_IE_CODESET_0, Q931_IEF_TO_NET },
			{ 0, 0, 0, 0 },
	}},

	/* RESUME ACKNOWLEDGE */
	{ Q931mes_RESUME_ACKNOWLEDGE, 8, "RESUME ACKNOWLEDGE", {
			{ Q931ie_CHANNEL_IDENTIFICATION, 3,  0, Q931_IE_CODESET_0, Q931_IEF_TO_USER | Q931_IEF_MANDATORY },
			{ Q931ie_DISPLAY,                2, 34, Q931_IE_CODESET_0, Q931_IEF_TO_USER },
			{ 0, 0, 0, 0 },
	}},

	/* RESUME REJECT */
	{ Q931mes_RESUME_REJECT, 8, "RESUME REJECT", {
			{ Q931ie_CAUSE,    4, 32, Q931_IE_CODESET_0, Q931_IEF_TO_USER | Q931_IEF_MANDATORY },
			{ Q931ie_DISPLAY,  4, 34, Q931_IE_CODESET_0, Q931_IEF_TO_USER },
			{ 0, 0, 0, 0 },
	}},

	/* SETUP */
	{ Q931mes_SETUP, 8, "SETUP", {
			{ Q931ie_SENDING_COMPLETE,            1,  1, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q931ie_REPEAT_INDICATOR,            1,  1, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },		/* Multiple occurrences possible, context! */
			{ Q931ie_BEARER_CAPABILITY,           4, 12, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH | Q931_IEF_MANDATORY },
			{ Q931ie_CHANNEL_IDENTIFICATION,      3,  0, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q931ie_PROGRESS_INDICATOR,          2,  4, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q931ie_NETWORK_SPECIFIC_FACILITIES, 2,  0, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q931ie_DISPLAY,                     2, 34, Q931_IE_CODESET_0, Q931_IEF_TO_USER },
			{ Q931ie_DATETIME,                    5,  8, Q931_IE_CODESET_0, Q931_IEF_TO_NET },
			{ Q931ie_KEYPAD_FACILITY,             2, 34, Q931_IE_CODESET_0, Q931_IEF_TO_NET },
			{ Q931ie_SIGNAL,                      2,  3, Q931_IE_CODESET_0, Q931_IEF_TO_USER },
			{ Q931ie_CALLING_PARTY_NUMBER,        2, 34, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q931ie_CALLING_PARTY_SUBADDRESS,    2, 23, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q931ie_CALLED_PARTY_NUMBER,         2, 34, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q931ie_CALLED_PARTY_SUBADDRESS,     2, 23, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q931ie_TRANSIT_NETWORK_SELECTION,   2, 34, Q931_IE_CODESET_0, Q931_IEF_TO_NET },
			{ Q931ie_LOW_LAYER_COMPATIBILITY,     2, 18, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q931ie_HIGH_LAYER_COMPATIBILITY,    2,  5, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },

			{ Q932ie_FACILITY,                    2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q932ie_EXTENDED_FACILITY,           2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },

/* Q.952 */
			{ Q931ie_REDIRECTING_NUMBER,          2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_USER },

			{ 0, 0, 0, 0 },
	}},

	/* SETUP ACKNOWLEDGE */
	{ Q931mes_SETUP_ACKNOWLEDGE, 8, "SETUP ACKNOWLEDGE", {
			{ Q931ie_CHANNEL_IDENTIFICATION, 3,  0, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q931ie_PROGRESS_INDICATOR,     2,  4, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q931ie_DISPLAY,                2, 34, Q931_IE_CODESET_0, Q931_IEF_TO_USER },
			{ Q931ie_SIGNAL,                 2,  3, Q931_IE_CODESET_0, Q931_IEF_TO_USER },

			{ Q932ie_FACILITY,               2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q932ie_EXTENDED_FACILITY,      2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },

			{ 0, 0, 0, 0 },
	}},

	/* STATUS */
	{ Q931mes_STATUS, 8, "STATUS", {
			{ Q931ie_CAUSE,      4, 32, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH | Q931_IEF_MANDATORY },
			{ Q931ie_CALL_STATE, 3,  3, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH | Q931_IEF_MANDATORY },
			{ Q931ie_DISPLAY,    2, 34, Q931_IE_CODESET_0, Q931_IEF_TO_USER },
			{ 0, 0, 0, 0 },
	}},

	/* STATUS ENQUIRY */
	{ Q931mes_STATUS_ENQUIRY, 8, "STATUS ENQUIRY", {
			{ Q931ie_DISPLAY, 2, 34, Q931_IE_CODESET_0, Q931_IEF_TO_USER },
			{ 0, 0, 0, 0 },
	}},

	/* SUSPEND */
	{ Q931mes_SUSPEND, 8, "SUSPEND", {
			{ Q931ie_CALL_IDENTITY, 2, 10, Q931_IE_CODESET_0, Q931_IEF_TO_NET },
			{ 0, 0, 0, 0 },
	}},

	/* SUSPEND ACKNOWLEDGE */
	{ Q931mes_SUSPEND_ACKNOWLEDGE, 8, "SUSPEND ACKNOWLEDGE", {
			{ Q931ie_DISPLAY, 2, 34, Q931_IE_CODESET_0, Q931_IEF_TO_USER },
			{ 0, 0, 0, 0 },
	}},

	/* SUSPEND REJECT */
	{ Q931mes_SUSPEND_REJECT, 8, "SUSPEND REJECT", {
			{ Q931ie_CAUSE,   4, 32, Q931_IE_CODESET_0, Q931_IEF_TO_USER | Q931_IEF_MANDATORY },
			{ Q931ie_DISPLAY, 4, 34, Q931_IE_CODESET_0, Q931_IEF_TO_USER },
			{ 0, 0, 0, 0 },
	}},

	/* RESTART */
	{ Q931mes_RESTART, 8, "RESTART", {
			{ Q931ie_RESTART_INDICATOR,      3,  3, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH | Q931_IEF_MANDATORY },
			{ Q931ie_CHANNEL_IDENTIFICATION, 3,  0, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q931ie_DISPLAY,                4, 34, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ 0, 0, 0, 0 },
	}},

	/* RESTART ACKNOWLEDGE */
	{ Q931mes_RESTART_ACKNOWLEDGE, 8, "RESTART ACKNOWLEDGE", {
			{ Q931ie_RESTART_INDICATOR,      3,  3, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH | Q931_IEF_MANDATORY },
			{ Q931ie_CHANNEL_IDENTIFICATION, 3,  0, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q931ie_DISPLAY,                4, 34, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ 0, 0, 0, 0 },
	}},


	/*
	 * ITU-T Q.932 IE table
	 */

	/* HOLD */
	{ Q932mes_HOLD, 8, "HOLD", {
			{ Q931ie_DISPLAY,           2, 34, Q931_IE_CODESET_0, Q931_IEF_TO_USER },

			{ Q932ie_FACILITY,          2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q932ie_EXTENDED_FACILITY, 2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ 0, 0, 0, 0 },
	}},

	/* HOLD ACKNOWLEDGE */
	{ Q932mes_HOLD_ACKNOWLEDGE, 8, "HOLD ACKNOWLEDGE", {
			{ Q931ie_DISPLAY,           2, 34, Q931_IE_CODESET_0, Q931_IEF_TO_USER },

			{ Q932ie_FACILITY,          2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q932ie_EXTENDED_FACILITY, 2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ 0, 0, 0, 0 },
	}},

	/* HOLD REJECT */
	{ Q932mes_HOLD_REJECT, 8, "HOLD REJECT", {
			{ Q931ie_CAUSE,             4, 32, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH | Q931_IEF_MANDATORY },
			{ Q931ie_DISPLAY,           2, 34, Q931_IE_CODESET_0, Q931_IEF_TO_USER },

			{ Q932ie_FACILITY,          2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q932ie_EXTENDED_FACILITY, 2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ 0, 0, 0, 0 },
	}},

	/* REGISTER */
	{ Q932mes_REGISTER, 8, "REGISTER", {
			{ Q932ie_FACILITY,          2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH | Q931_IEF_MANDATORY },
			{ Q932ie_EXTENDED_FACILITY, 2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q931ie_DISPLAY,           2, 34, Q931_IE_CODESET_0, Q931_IEF_TO_USER },

			{ 0, 0, 0, 0 },
	}},

	/* RETRIEVE */
	{ Q932mes_RETRIEVE, 8, "RETRIEVE", {
			{ Q931ie_CHANNEL_IDENTIFICATION, 3,  0, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q931ie_DISPLAY,                2, 34, Q931_IE_CODESET_0, Q931_IEF_TO_USER },

			{ Q932ie_FACILITY,               2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q932ie_EXTENDED_FACILITY,      2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ 0, 0, 0, 0 },
	}},

	/* RETRIEVE ACKNOWLEDGE */
	{ Q932mes_RETRIEVE_ACKNOWLEDGE, 8, "RETRIEVE ACKNOWLEDGE", {
			{ Q931ie_CHANNEL_IDENTIFICATION, 3,  0, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q931ie_DISPLAY,                2, 34, Q931_IE_CODESET_0, Q931_IEF_TO_USER },

			{ Q932ie_FACILITY,               2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q932ie_EXTENDED_FACILITY,      2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ 0, 0, 0, 0 },
	}},

	/* RETRIEVE REJECT */
	{ Q932mes_RETRIEVE_REJECT, 8, "RETRIEVE REJECT", {
			{ Q931ie_CAUSE,             4, 32, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH | Q931_IEF_MANDATORY },
			{ Q931ie_DISPLAY,           2, 34, Q931_IE_CODESET_0, Q931_IEF_TO_USER },

			{ Q932ie_FACILITY,          2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ Q932ie_EXTENDED_FACILITY, 2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ 0, 0, 0, 0 },
	}},

	/* FACILITY */
	{ Q932mes_FACILITY, 8, "FACILITY", {
			{ Q932ie_FACILITY,          8, 99, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH | Q931_IEF_MANDATORY },
			{ Q932ie_EXTENDED_FACILITY, 2, 99, Q931_IE_CODESET_0, Q931_IEF_TO_BOTH },
			{ 0, 0, 0, 0 },
	}},

	/* TODO... */
	{ 0, 0, 0, {{ 0, 0, 0, 0 }}},
};

/**
 * Q931MesgIEGetEntry
 * @brief	Search matching IE entry in the Message IE table
 * @param[in]	MesType		Message type
 * @param[in]	id		IE id
 * @param	index		Offset into message array to speed up search
 * @return	pointer to the table entry, NULL on error
 * @note	Use on (from L2) incoming message only
 */
static const struct Q931MessageIEEntry *Q931MesgIEsGetEntry(struct Q931Dialect *dialect, L3INT MesType, L3UCHAR ProtDisc, L3UCHAR id, L3UCHAR *index)
{
	const struct Q931MessageIE *mes = NULL;
	const struct Q931MessageIE *map = NULL;
	const struct Q931MessageIEEntry *ie;
	int i, offset = 0;

	assert(dialect);

	map = Q931DialectGetMesIEMap(dialect);
	if (!map) {
		return NULL;
	}

	if (index) {
		offset = *index;
	}

	for (i = offset; map[i].id; i++) {
		if(map[i].id == MesType && map[i].protdisc == ProtDisc) {
			mes = &map[i];
			break;
		}
	}
	if (!mes) {
		/* Message not found */
		return NULL;
	}

	if (index) {
		*index = i;
	}

	ie = mes->ies;
	while (ie->id) {
		if (ie->id == id) {
			return ie;
		}
		ie++;
	}
	return NULL;
}

/**
 * Q931MesgIEGetFirstEntry
 * @brief	Search matching IE entry in the Message IE table
 * @param[in]	MesType		Message type
 * @return	pointer to the table entry, NULL on error
 * @note	Use on (from L2) incoming message only
 */
static const struct Q931MessageIEEntry *Q931MesgIEsGetFirstEntry(struct Q931Dialect *dialect, L3INT MesType, L3UCHAR ProtDisc)
{
	struct Q931MessageIE *map = NULL;

	assert(dialect);

	map = Q931DialectGetMesIEMap(dialect);
	if (!map) {
		return NULL;
	}

	while (map->id) {
		if(map->id == MesType && map->protdisc == ProtDisc)
			break;
		map++;
	}
	if (!map->id) {
		/* Message not found */
		return NULL;
	}
	return map->ies;
}

/**
 * Q931MesgIEsGetMandatory
 * @brief	Get a list of mandatory IEs for this message
 * @param[in]	MesType		Message type
 * @param[out]	list		List of mandatory IEs
 * @param[in]	size		Size of the list
 * @return	Number of elements in the list, < 0 on error
 *
 * @note	This s*cks somehow (bitmap?)
 */
static L3INT Q931MesgIEsGetMandatory(struct Q931Dialect *dialect, L3INT MesType, L3UCHAR ProtDisc, L3UCHAR *list, L3INT size)
{
	const struct Q931MessageIE *map = NULL;
	const struct Q931MessageIEEntry *ie;
	int i = 0;

	assert(dialect);

	map = Q931DialectGetMesIEMap(dialect);
	if (!map) {
		return -1;
	}

	while (map->id) {
		if(map->id == MesType && map->protdisc == ProtDisc)
			break;
		map++;
	}
	if (!map->id) {
		/* Message not found */
		return -1;
	}

	ie = map->ies;
	while (ie->id) {
		if (ie->flags & Q931_IEF_MANDATORY) {
			if (i >= size) {
				return -1;
			}
			list[i++] = ie->id;
		}
		ie++;
	}
	return i;
}

/**
 * Q931MesgIEsMandatoryMissing
 * @brief	Check if mandatory IE is missing
 * @param[in]	list	List of mandatory IEs
 * @param[in]	size	Size of the list
 * @return	0 if none are missing, the number of missing IEs otherwise
 */
static L3INT Q931MesgIEsMandatoryMissing(L3UCHAR *list, L3INT size)
{
	int count = 0;
	int i;

	for (i = 0; i < size; i++) {
		if (list[i]) {
			count++;
		}
	}
	return count;
}

/**
 * Q931MesgIEsMandatoryClear
 * @brief	Remove mandatory IE from list
 * @param[in]	list	mandatory IE list
 * @param[in]	size	size of the list
 * @param[in]	id	id of the mandatory IE
 */
static void Q931MesgIEsMandatoryClear(L3UCHAR *list, L3INT size, L3UCHAR id)
{
	int i;

	for (i = 0; i < size; i++) {
		if (list[i] == id) {
			list[i] = 0;
			return;
		}
	}
}

/**
 * Q931MesgIEIsDirectionValid
 * @brief	Check if IE is valid in this direction
 * @param[in]	ie		MessageIE entry
 * @param[in]	mode		Q931_NT or Q931_TE
 * @param[in]	outgoing	Is packet outgoing or incoming
 * @return	L3TRUE if valid, L3FALSE otherwise
 */
static L3BOOL Q931MesgIEIsDirectionValid(const struct Q931MessageIEEntry *ie, const L3INT mode, const L3BOOL outgoing)
{
	L3INT flags = Q931_IEF_NONE;

	if (mode == Q931_NT) {
		flags |= (outgoing) ? Q931_IEF_TO_USER : Q931_IEF_TO_NET;
	} else {
		flags |= (outgoing) ? Q931_IEF_TO_NET : Q931_IEF_TO_USER;
	}

	if (!(ie->flags & flags)) {	/* IE not allowed in this direction */
		return L3FALSE;
	}
	return L3TRUE;
}

/**
 * Q931MesgIEIsCodesetValid
 * @brief	Check if IE is valid in this codeset
 * @param[in]	ie	MessageIE entry
 * @param[in]	codeset	Current active codeset
 * @return	L3TRUE if valid, L3FALSE otherwise
 */
static L3BOOL Q931MesgIEIsCodesetValid(const struct Q931MessageIEEntry *ie, const L3INT codeset)
{
	L3INT flags = Q931_IE_CODESET_0;

	/* only check if any codesets have been specified at all */
	if (ie->codeset & Q931_IE_CODESET_ALL) {
		flags <<= codeset;

		if (!(ie->codeset & flags)) {	/* IE not allowed in this codeset */
			return L3FALSE;
		}
	}
	return L3TRUE;
}

/**
 * Q931MesgIEGetCodeset
 * @brief	Return first valid codeset for IE
 * @param[in]	ie	MessageIE entry
 * @return	First valid codeset id
 */
static L3UCHAR Q931MesgIEGetCodeset(const struct Q931MessageIEEntry *ie)
{
	L3INT mask = Q931_IE_CODESET_0;
	L3UCHAR codeset = 0;

	while (mask <= Q931_IE_CODESET_7) {
		if (ie->codeset & mask) {
			return codeset;
		}

		mask <<= 1;
		codeset++;
	}
	return 0;
}

/**
 * Q931MesgIEIsSizeValid
 * @brief	Check if IE size is valid
 * @param[in]	ie	MessageIE entry
 * @param[in]	size	the size of the IE
 * @return	L3TRUE if valid, L3FALSE otherwise
 */
static L3BOOL Q931MesgIEIsSizeValid(const struct Q931MessageIEEntry *ie, L3UCHAR size)
{
	if (ie->id & 0x80) {	/* Fixed size IE */
		return (L3BOOL)(size == 1);
	}
	if (!ie->maxsize) {
		return (L3BOOL)(size >= ie->minsize);
	}
	return (L3BOOL)(size >= ie->minsize && size <= ie->maxsize);
}

/**
 * Q931MesgIEIsMandatory
 * @brief	Check if IE is mandatory
 * @param[in]	ie	MessageIE entry
 * @return	L3TRUE if mandatory, L3FALSE otherwise
 */
static L3BOOL Q931MesgIEIsMandatory(const struct Q931MessageIEEntry *ie)
{
	return (L3BOOL)((ie->flags & Q931_IEF_MANDATORY) != 0);
}

/**
 * Q931MesgGetName
 * @brief	Get Message Name by ID
 * @param[in]	dialect		Q931 Dialect
 * @param[in]	MesType		Message ID
 * @param[in]	ProtDisc	Protocol Discriminator
 * @return	Name string or NULL
 */
static const char *Q931MesgGetName(struct Q931Dialect *dialect, const L3INT MesType, const L3INT ProtDisc)
{
	const struct Q931MessageIE *map = NULL;

	assert(dialect);

	map = Q931DialectGetMesIEMap(dialect);
	if (!map) {
		return NULL;
	}

	for (int i = 0; map[i].id; i++) {
		if (map[i].id == MesType && map[i].protdisc == ProtDisc)
			return map[i].name;
	}
	return NULL;
}

/**
 * Q931MesgHeader
 * @brief	Create Q.931 Message header
 */
L3INT Q931MesgHeader(Q931_TrunkInfo_t *trunk, Q931mes_Generic *mes, L3UCHAR *OBuf, L3INT Size, L3INT *IOff)
{
	L3INT Octet = *IOff;

	Q931Log(trunk, Q931_LOG_DEBUG, "Creating Q.931 Message Header:\n    ProtDisc %d (%#x), CRV %d (%#x), CRVflag: %d (%#x), MesType: %d (%#x)\n",
			 mes->ProtDisc, mes->ProtDisc, mes->CRV, mes->CRV, mes->CRVFlag, mes->CRVFlag, mes->MesType, mes->MesType);

	OBuf[Octet++] = mes->ProtDisc;				/* Protocol discriminator */
	if (!Q931_IS_BRI(trunk)) {
		OBuf[Octet++] = 2;									/* length is 2 octets */
		OBuf[Octet++] = (L3UCHAR)((mes->CRV >> 8) & 0x7f) | ((mes->CRVFlag << 7) & 0x80);	/* msb */
		OBuf[Octet++] = (L3UCHAR) (mes->CRV & 0xff);						/* lsb */
	} else {
		OBuf[Octet++] = 1;									/* length is 1 octet */
		OBuf[Octet++] = (L3UCHAR) (mes->CRV & 0x7f) | ((mes->CRVFlag << 7) & 0x80);		/* CRV & flag */
	}
	OBuf[Octet++] = mes->MesType;				/* message header */

	*IOff = Octet;
	return 0;
}

/*****************************************************************************
 *
 * Muahahaha, a generic Q931Umes implementation!!!
 *
 *****************************************************************************/

/**
 * Q931Umes_Generic
 * @brief	A generic table driven codeset aware decode function for messages
 * @param[in]	trunk		Q.931 Trunk
 * @param[in]	IBuf		Input buffer
 * @param[out]	mes		Decoded Q.931 message
 * @param[in]	IOff		Offset of input buffer
 * @param[in]	Size		Number of bytes in input buffer
 * @param[out]	Errs		Message parser errors
 *
 * @note	Use this to rule them all...
 */
L3INT Q931Umes_Generic(Q931_TrunkInfo_t *trunk, L3UCHAR *IBuf, Q931mes_Generic *mes, L3INT IOff, L3INT Size, struct Q931MesgErrors *Errs)
{
	L3INT ir = 0;
	L3INT OOff = 0;
	L3INT rc = Q931E_NO_ERROR;
	L3UCHAR last_codeset = 0, codeset = 0;
	L3UCHAR shift_nolock = 1;
	L3UCHAR mandatoryIEs[Q931MAXIE];
	L3UCHAR msgidx = 0;
	L3INT   mandsize = 0, missing_count;

	if (!mes->MesType) {
		Q931Log(trunk, Q931_LOG_ERROR, "Nationally specified message types not supported\n");
		return Q931E_UNEXPECTED_MESSAGE;
	}

	memset(mandatoryIEs, 0, sizeof(mandatoryIEs));

	mandsize = Q931MesgIEsGetMandatory(trunk->Dialect, mes->MesType, mes->ProtDisc, mandatoryIEs, sizeof(mandatoryIEs));
	if (mandsize < 0) {
		Q931Log(trunk, Q931_LOG_DEBUG, "Unable to load mandatory IEs for message\n");
		return Q931E_UNKNOWN_MESSAGE;	/* this one ok?? */
	}

	/* TODO: handle codeset shifts (more) correctly */
	/* TODO: improve error handling... */

	while (IOff < Size) {
		const struct Q931MessageIEEntry *entry;
		L3UCHAR id, size;

		/* Codeset shift */
		if ((IBuf[IOff] & 0xf0) == Q931ie_SHIFT) {
			shift_nolock = (IBuf[IOff] & 0x08);
			if (shift_nolock) {
				last_codeset = codeset;
			}
			codeset = (IBuf[IOff++] & 0x07);
#ifdef __TODO__
			/* TODO: Check if target codeset is valid */
			if (!Q931MesgCodesetIsValid(trunk, codeset)) {
				Q931Log(trunk, Q931_LOG_ERROR, "Shift to invalid codeset %d\n", codeset);
				Q931MesgErrorsAddIE(Errs, Q931ie_SHIFT, 1, IOff, Q931_MSGE_INVALID_CODESET);
			}
#endif
			continue;
		}

		/* Grab IE */
		id = IBuf[IOff];

		if (!(id & 0x80)) {
			/* Variable size IE, get length */
			size = IBuf[IOff + 1] + 2;	/* header + body */
		} else {
			/* Single octet IE */
			size = 1;
		}

		/* IE supported? */
		if (Q931UieIsNull(trunk, id)) {
			/* congrats, no idea how to cope with this */
			Q931Log(trunk, Q931_LOG_DEBUG, "Unhandled IE %02u (%#02x)\n", id, id);
			Q931MesgErrorsAddIE(Errs, id, size, IOff, Q931_MSGE_UNKNOWN);

			if (Q931TrunkIsSetFlag(trunk, Q931_TFLAG_IGNORE_UNKNOWN_IE)) {
				IOff += size;	/* "consume" IE */
				goto skip;
			} else {
				return Q931E_UNKNOWN_IE;
			}
		}

		/* Lookup IE in table and perform checks */
		entry = Q931MesgIEsGetEntry(trunk->Dialect, mes->MesType, mes->ProtDisc, id, &msgidx);
		if (!entry) {
			/* aww, not found */
			Q931Log(trunk, Q931_LOG_DEBUG, "IE %02u (%#02x) not allowed for message type\n", id, id);
			Q931MesgErrorsAddIE(Errs, id, size, IOff, Q931_MSGE_NOT_ALLOWED);

			if (Q931TrunkIsSetFlag(trunk, Q931_TFLAG_IGNORE_ILLEGAL_IE)) {
				IOff += size;	/* "consume" IE */
				goto skip;
			} else {
				return Q931E_ILLEGAL_IE;
			}
		}

		if (!Q931MesgIEIsDirectionValid(entry, trunk->NetUser, 0)) {
			/* oops, we shouldn't be getting this ie from our peer in this mode of operation */
			Q931Log(trunk, Q931_LOG_DEBUG, "IE %02u (%#02x) in wrong direction (flags: 0x%x, mode: %d, incoming)\n", id, id, entry->flags, trunk->NetUser);
			Q931MesgErrorsAddIE(Errs, id, size, IOff, Q931_MSGE_INVALID_DIRECTION);
			return Q931E_ILLEGAL_IE;
		}

		if (size > (Size - IOff)) {
			/* Overflow, IE larger than bytes in buffer */
			Q931Log(trunk, Q931_LOG_DEBUG, "IE %02u (%#02x) not enough bytes left\n", id, id);
			Q931MesgErrorsAddIE(Errs, id, size, IOff, Q931_MSGE_OVERFLOW);
			return Q931E_ILLEGAL_IE;
		}

		if (!Q931MesgIEIsSizeValid(entry, size)) {
			/* oops, IE is too large or short */
			Q931Log(trunk, Q931_LOG_DEBUG, "IE %02u (%#02x) size error [%02u octets]\n", id, id, size);
			Q931MesgErrorsAddIE(Errs, id, size, IOff, Q931_MSGE_INVALID_SIZE);
			return Q931E_ILLEGAL_IE;	/* meh, s*cks */
		}

		if (!Q931MesgIEIsCodesetValid(entry, codeset)) {
			/* arrrr, IE is not allowed in this codest */
			Q931Log(trunk, Q931_LOG_DEBUG, "IE %02u (%#02x) not allowed in codeset %d\n", id, id, codeset);
			Q931MesgErrorsAddIE(Errs, id, size, IOff, Q931_MSGE_INVALID_CODESET);
			return Q931E_ILLEGAL_IE;
		}

		switch (id) {
		case Q931ie_REPEAT_INDICATOR:	/* TODO: handle this one better... */
			if (ir < 2) {
				rc = Q931Uie(trunk, id, mes, &IBuf[IOff], &mes->buf[OOff], &IOff, &OOff);
				ir++;
			} else {
				Q931MesgErrorsAddIE(Errs, id, size, IOff, Q931_MSGE_NOT_ALLOWED);
				return Q931E_ILLEGAL_IE;
			}
			break;
		default:
			rc = Q931Uie(trunk, id, mes, &IBuf[IOff], &mes->buf[OOff], &IOff, &OOff);
			if (rc != Q931E_NO_ERROR)
				return rc;
			break;
		}

		/* Remove mandatory IE from list, if we've found one */
		if (Q931MesgIEIsMandatory(entry)) {
			Q931MesgIEsMandatoryClear(mandatoryIEs, mandsize, id);
		}

		/* increate message's ie counter */
		mes->nr_ie++;
skip:
		/* drop back to old codeset!? */
		if (shift_nolock) {
			codeset = last_codeset;
		}
	}

	if ((missing_count = Q931MesgIEsMandatoryMissing(mandatoryIEs, mandsize)) > 0) {
		L3INT i = 0;

		/* tadaa, at least one of the mandatory IEs is missing */
		Q931Log(trunk, Q931_LOG_ERROR, "%d Mandatory IEs missing in this incoming message\n", missing_count);

		/* Add missing mandatory IEs to error report */
		for (i = 0; i < mandsize; i++) {
			if (mandatoryIEs[i]) {
				Q931MesgErrorsAddIE(Errs, mandatoryIEs[i], 0, 0, Q931_MSGE_MANDATORY_MISSING);
			}
		}
		return Q931E_MANDATORY_IE_MISSING;
	}

	mes->Size = sizeof(Q931mes_Generic) - 1 + OOff;

	/* */
	Q931Dmes_Generic(trunk, mes, Q931_MSG_INCOMING);

	return Q931E_NO_ERROR;
}

#include "utils/strstream.h"

/**
 * Q931Dmes_Generic
 * @brief	Print message debug information
 * @param[in]	trunk	Q.931 trunk
 * @param[in]	mes	Decoded message
 * @param[in]	direction	Incoming/Outgoing direction
 */
L3INT Q931Dmes_Generic(Q931_TrunkInfo_t *trunk, Q931mes_Generic *msg, q931_msg_direction_t direction)
{
	L3INT ret = Q931E_NO_ERROR;
	L3INT offset = 0;
	struct strstream stream;
	char buf[STRSTREAM_SIZE];
	const char *name = NULL;

	if (trunk->loglevel != Q931_LOG_DEBUG)
		return Q931E_NO_ERROR;

	strstream_init_static(&stream, buf, sizeof(buf));

	name = Q931MesgGetName(trunk->Dialect, msg->MesType, msg->ProtDisc);
	if (!name) {
		name = "UNKNOWN MESSAGE";
	}

	strstream_puts(&stream, "\n----\n");

	/* print message header */
	strstream_printf(&stream,
			"%s\n"
			"\tMessage-ID:  0x%02x\n"
			"\tDirection:   %s\n"
			"\tCall Ref:    0x%04x (%s)\n"
			"\tL2 Link ID:  %u\n",
			name, msg->MesType, ((direction == Q931_MSG_INCOMING) ? "Incoming" : "Outgoing"),
			msg->CRV, (msg->CRVFlag ? "Terminator" : "Originator"), msg->Tei);

	strstream_puts(&stream, "\t---------------------\n");

	/* print IE information */
	while (offset < msg->Size && ret == Q931E_NO_ERROR) {
		Q931ie_Generic *ie = (Q931ie_Generic *)&msg->buf[offset];

		if (!ie->IEId && !ie->Size)
			break;

		if (!(ie->IEId & 0x80) && ie->Size)
			offset += ie->Size;
		else
			offset += sizeof(*ie);

		/* IE supported? */
		if (Q931DieIsNull(trunk, ie->IEId))
			ret = Q931Die_Generic(trunk, ie, &stream);
		else
			ret = Q931Die(trunk, ie, &stream);
	}

	strstream_puts(&stream, "\n----\n");

	/* output final string */
	Q931LogRaw(trunk, Q931_LOG_DEBUG, strstream_get(&stream));
	return ret;
}


/*
 * BIG FAT TODO:
 *
 * Yeah, this _IS_ ugly, but it's needed until we change the way IEs refs are handled in the message
 * (or we're going to do a full (= slow) lookup in the buffer...)
 *
 */
static Q931ie_Generic * Q931MesgGetIE(Q931_TrunkInfo_t *trunk, Q931mes_Generic *mes, L3UCHAR id)
{
	ie ie = 0;

	switch (id) {
	case Q931ie_SENDING_COMPLETE:
		ie = mes->SendComplete;
		break;
	case Q931ie_REPEAT_INDICATOR:
		ie = mes->RepeatInd;
		break;
	case Q931ie_BEARER_CAPABILITY:
		ie = mes->BearerCap;
		break;
	case Q931ie_CHANNEL_IDENTIFICATION:
		ie = mes->ChanID;
		break;
	case Q931ie_NETWORK_SPECIFIC_FACILITIES:
		ie = mes->NetFac;
		break;
	case Q931ie_DISPLAY:
		ie = mes->Display;
		break;
	case Q931ie_DATETIME:
		ie = mes->DateTime;
		break;
	case Q931ie_KEYPAD_FACILITY:
		ie = mes->KeypadFac;
		break;
	case Q931ie_SIGNAL:
		ie = mes->Signal;
		break;
	case Q931ie_CALLING_PARTY_NUMBER:
		ie = mes->CallingNum;
		break;
	case Q931ie_CALLING_PARTY_SUBADDRESS:
		ie = mes->CallingSub;
		break;
	case Q931ie_CALLED_PARTY_NUMBER:
		ie = mes->CalledNum;
		break;
	case Q931ie_CALLED_PARTY_SUBADDRESS:
		ie = mes->CalledSub;
		break;
	case Q931ie_TRANSIT_NETWORK_SELECTION:
		ie = mes->TransNetSel;
		break;
	case Q931ie_LOW_LAYER_COMPATIBILITY:
		ie = mes->LLComp;
		break;
	case Q931ie_HIGH_LAYER_COMPATIBILITY:
		ie = mes->HLComp;
		break;
	case Q931ie_PROGRESS_INDICATOR:
		ie = mes->ProgInd;
		break;
	case Q931ie_CAUSE:
		ie = mes->Cause;
		break;
	case Q931ie_CHANGE_STATUS:
		ie = mes->ChangeStatus;
		break;
	case Q931ie_RESTART_INDICATOR:
		ie = mes->RestartInd;
		break;
	case Q931ie_CALL_STATE:
		ie = mes->CallState;
		break;
	default:
		Q931Log(trunk, Q931_LOG_DEBUG, "Unable to get reference for IE %02u (%#02x)\n", id, id);
		return NULL;
	}

	if(Q931IsIEPresent(ie)) {
		return Q931GetIEPtr(ie, mes->buf);
	}

	return NULL;
}

L3INT Q931Pmes_Generic(Q931_TrunkInfo_t *trunk, Q931mes_Generic *msg, L3INT ISize, L3UCHAR *OBuf, L3INT *OSize)
{
	const struct Q931MessageIEEntry *entry;
	L3UCHAR mandatoryIEs[Q931MAXIE];
	L3UCHAR codeset = 0;
	L3INT rc = Q931E_NO_ERROR;
	L3INT Octet = 0;
	L3INT mandsize = 0, missing_count;

	memset(mandatoryIEs, 0, sizeof(mandatoryIEs));

	/* */
	Q931Dmes_Generic(trunk, msg, Q931_MSG_OUTGOING);

	/* Q931 Message Header */
	Q931MesgHeader(trunk, msg, OBuf, *OSize, &Octet);

	/* Get mandatory IE list */
	mandsize = Q931MesgIEsGetMandatory(trunk->Dialect, msg->MesType, msg->ProtDisc, mandatoryIEs, sizeof(mandatoryIEs));
	if (mandsize < 0) {
		Q931Log(trunk, Q931_LOG_DEBUG, "Unable to load mandatory IEs for message\n");
		return Q931E_UNKNOWN_MESSAGE;	/* TODO: this one ok?? */
	}

	/* Walk list of allowed IEs */
	entry = Q931MesgIEsGetFirstEntry(trunk->Dialect, msg->MesType, msg->ProtDisc);

	while (entry && entry->id) {
		L3UCHAR ie_codeset = 0;
		Q931ie_Generic *pIE;

		pIE = Q931MesgGetIE(trunk, msg, entry->id);

		/* Check + encode IE */
		if (pIE) {
			L3INT ieOffset;

			/* Direction check */
			if (!Q931MesgIEIsDirectionValid(entry, trunk->NetUser, 1)) {
				/* oops, we shouldn't be getting this ie from our peer in this mode of operation */
				Q931Log(trunk, Q931_LOG_DEBUG, "IE %02u (%#02x) in wrong direction\n", entry->id, entry->id);
				return Q931E_ILLEGAL_IE;
			}

			/* IE supported? */
			if (Q931PieIsNull(trunk, entry->id)) {
				/* congrats, no idea how to cope with this */
				Q931Log(trunk, Q931_LOG_DEBUG, "Unhandled IE %u (%#02x)\n", entry->id, entry->id);
				return Q931E_UNKNOWN_IE;
			}

			/* Check IEs codeset and change codeset if neccessary (non-locking shift only!) */
			ie_codeset = Q931MesgIEGetCodeset(entry);
			if (ie_codeset != codeset) {
#ifdef __TODO__
				if (dialect->flags & Q931_DIALECT_LOCKING_SHIFT_ONLY) {
					/* emit locking shift to target codeset */
					codeset       = rc;
					OBuf[Octet++] = Q931ie_SHIFT | (codeset & 0x07);
				} else {
#endif
					/* emit non-locking shift to target codeset */
					OBuf[Octet++] = Q931ie_SHIFT | 0x08 | (rc & 0x07);
#ifdef __TODO__
				}
#endif
			}

			/* TODO: repeat indicator? */

			/* Store current offset in output buffer */
			ieOffset = Octet;

			/* Encode IE */
			if ((rc = Q931Pie(trunk, entry->id, (L3UCHAR *)pIE, OBuf, &Octet)) != Q931E_NO_ERROR) {
				return rc;
			}

			/* Check encoded(!) size */
			if (!Q931MesgIEIsSizeValid(entry, Octet - ieOffset)) {
				/* oops, IE is too large or short */
				Q931Log(trunk, Q931_LOG_DEBUG, "IE %02u (%#02x) size error [%d octets]\n", entry->id, entry->id, Octet - ieOffset);
				return Q931E_ILLEGAL_IE;	/* meh, s*cks */
			}

			/* Remove mandatory IE from list, if we've found one */
			if (Q931MesgIEIsMandatory(entry)) {
				Q931MesgIEsMandatoryClear(mandatoryIEs, mandsize, entry->id);
			}
		}

		entry++;
	}

	if ((missing_count = Q931MesgIEsMandatoryMissing(mandatoryIEs, mandsize)) > 0) {
		/* tadaa, at least one of the mandatory IEs is missing */
		Q931Log(trunk, Q931_LOG_ERROR, "%d Mandatory IEs missing in this outgoing message\n", missing_count);

		return Q931E_MANDATORY_IE_MISSING;
	}

	*OSize = Octet;
	return rc;
}

/**
 * Names of message errors
 */
const static struct Q931MesgErrorName {
	const char id;
	const char *name;
	const char *diag;
} Q931MesgErrorNames[] = {
	/* Order is important here! */
	{ Q931_MSGE_NONE,               "None",                      NULL },
	{ Q931_MSGE_UNKNOWN,            "Unknown/unhandled",         "Message element is unknown (or invalid) and not handled by parser" },
	{ Q931_MSGE_OVERFLOW,           "Not enough octets left",    "Message element is larger than octets left in buffer" },
	{ Q931_MSGE_NOT_ALLOWED,        "Not allowed here",          "Message element is not allowed in this state/message" },
	{ Q931_MSGE_INVALID_DIRECTION,  "Wrong direction",           "Message element is sent to or coming from the wrong direction" },
	{ Q931_MSGE_INVALID_SIZE,       "Size invalid",              "Message element is either too small or too large" },
	{ Q931_MSGE_INVALID_CODESET,    "Wrong codeset",             "Message element is in the wrong codeset (wrong dialect?)" },
	{ Q931_MSGE_MANDATORY_MISSING,  "Mandatory element missing", "A mandatory message element is missing" }
};

/**
 * Get name of message error
 * @param	id	Error id
 * @return	Pointer to string with error name
 */
static const char *Q931MesgErrorGetName(const int id)
{
	if (id < 0 || id > Q931_MSGE_COUNT)
		return "- Invalid error id -";

	return Q931MesgErrorNames[id].name;
}

/**
 * Get diagnostic information for error code
 * @param	id	Error id
 * @return	Pointer to string with error name
 */
static const char *Q931MesgErrorGetDiagnostic(const int id)
{
	if (id < 0 || id > Q931_MSGE_COUNT)
		return NULL;

	return Q931MesgErrorNames[id].diag;
}

/**
 * Initialize message error buffer and set message id
 * @param	merrs	Message error buffer
 * @param	id	Message id
 * @return	Q931E_NO_ERROR on success
 */
L3INT Q931MesgErrorsInit(struct Q931MesgErrors *merrs, const char id)
{
	if (!merrs) {
		return Q931E_NO_ERROR;
	}

	memset(merrs, 0, sizeof(struct Q931MesgErrors));
	merrs->id = id;
	return Q931E_NO_ERROR;
}

/**
 * Add IE error to message error buffer
 * @param	merrs	Message error buffer
 * @param	id	IE id
 * @param	size	IE size
 * @param	offset	IE offset
 * @param	error	IE error code
 * @return	Q931E_NO_ERROR on success
 */
L3INT Q931MesgErrorsAddIE(struct Q931MesgErrors *merrs, const char id, const char size, const int offset, const int error)
{
	if (!merrs || !id || !error) {
		return Q931E_NO_ERROR;
	}

	if (merrs->nr_errors == Q931_MSGE_ERR_MAX) {
		/* overflow, at least count them */
		merrs->nr_overflow++;
		return Q931E_NO_ERROR;
	}

	merrs->ie_errors[merrs->nr_errors].id     = id;
	merrs->ie_errors[merrs->nr_errors].size   = size;
	merrs->ie_errors[merrs->nr_errors].offset = (offset < UCHAR_MAX) ? offset : UCHAR_MAX;
	merrs->ie_errors[merrs->nr_errors].error  = error;
	merrs->nr_errors++;
	return Q931E_NO_ERROR;
}

/**
 * Get number of errors
 * @param	merrs	Message error buffer
 * @return	Q931E_NO_ERROR on success
 */
L3INT Q931MesgErrorsCount(struct Q931MesgErrors *merrs)
{
	return merrs->nr_errors;
}


#define TMP_BUFSIZE	4096
#define TMP_PUTS(fmt) \
	strncat(tmp + offset, fmt, sizeof(tmp) - offset - 1); \
	offset += strlen(tmp + offset);
#define TMP_APPEND(fmt, args...) \
	offset += snprintf(tmp + offset, sizeof(tmp) - offset, fmt, ##args);

#ifdef HAVE_ALLOCA_H
#include <alloca.h>
#elif defined(PLATFORM_WIN32) || defined(PLATFORM_WIN64)
#include <malloc.h>
#endif

#ifdef HAVE_ALLOCA
#define TMP_APPEND_HEX(fmt, buf, size)						\
	do {									\
		char *__tmp = NULL;						\
		static const char hex[16] = { '0', '1', '2', '3', '4', '5',	\
					'6', '7', '8', '9', 'a', 'b', 'c',	\
					'e', 'f' };				\
										\
		__tmp = alloca((size << 1) + size);				\
		if (__tmp) {							\
			int off = 0, x = 0;					\
										\
			for (x = 0; x < size; x++) {				\
				__tmp[off++] = hex[((buf)[x] & 0xf0) >> 4];	\
				__tmp[off++] = hex[((buf)[x] & 0x0f)];		\
				if (x + 1 < size) {				\
					__tmp[off++] = ' ';			\
				}						\
			}							\
			__tmp[off] = '\0';					\
			offset += snprintf(tmp + offset, sizeof(tmp) - offset, fmt, __tmp);	\
		}								\
	} while (0);

#endif /* HAVE_ALLOCA */

#define TMP_MIN(x, y)  \
	((x) < (y)) ? (x) : (y)

/**
 * Print a detailed message error report
 * @param	trunk	Q.931 Trunk
 * @param	merrs	Message error report
 * @param	IBuf	Message input buffer
 * @param	Size	Message size
 * @return	Q931E_NO_ERROR on success
 */
L3INT Q931MesgErrorsPrint(Q931_TrunkInfo_t *trunk, const struct Q931MesgErrors *merrs, const unsigned char *IBuf, const int Size)
{
	char tmp[TMP_BUFSIZE] = { 0 };
	char last_ie = 0;
	char errcnt  = 0;
	int  offset  = 0;
	int  i;

	if (!trunk || !merrs) {
		return Q931E_NO_ERROR;	/* TODO: which error code here? */
	}

	/* skip if not in debug loglevel */
	if (trunk->loglevel != Q931_LOG_DEBUG) {
		return Q931E_NO_ERROR;
	}

	if (merrs->nr_errors == 0) {
		Q931Log(trunk, Q931_LOG_DEBUG, "Message parser did not report any errors\n");
		return Q931E_NO_ERROR;
	}

	/* report header */
	TMP_PUTS("Message parser error(s)\n");
	TMP_APPEND("=================== Errors reported for message - %3u (%#02x) ===================\n", merrs->id, merrs->id);

	/* detailed report of ie errors */
	for (i = 0; i < merrs->nr_errors; i++) {
		const struct Q931MessageIEEntry *entry;
		const char *diag;

		if (last_ie != merrs->ie_errors[i].id) {
			/* new ie, print header, TODO: add codeset */
			TMP_APPEND("* IE id: %3u (%#02x), name: \"%s\", size: %3u, offset: %3u\n\n",
				merrs->ie_errors[i].id, merrs->ie_errors[i].id, "N/A",
				merrs->ie_errors[i].size,
				merrs->ie_errors[i].offset);

			/* new ie, update */
			last_ie = merrs->ie_errors[i].id;
			errcnt  = 1;
		}
		TMP_APPEND("  <%02d>  error message: %s\n", errcnt, Q931MesgErrorGetName(merrs->ie_errors[i].error));

		/* print diagnostic information */
		if ((diag = Q931MesgErrorGetDiagnostic(merrs->ie_errors[i].error))) {
			TMP_APPEND("\tdescription:   %s\n", diag);
		}

		/*
		 * extended diagnostic info
		 * NOTE: protocol discriminator hardcoded to "8"!
		 */
		TMP_PUTS("\tdetails:       ");
		switch (merrs->ie_errors[i].error) {
		case Q931_MSGE_INVALID_SIZE:
			if ((entry = Q931MesgIEsGetEntry(trunk->Dialect, merrs->id, 8, merrs->ie_errors[i].id, NULL)) != NULL) {
				if (merrs->id & 0x80) {
					TMP_PUTS("expected size: 1 octet [fixed size '1xxx xxxx']\n");
				}
				else if (entry->minsize == entry->maxsize) {
					TMP_APPEND("expected size: %d octet(s) [fixed size '0xxx xxxx']\n", entry->minsize);
				}
				else if (!entry->maxsize) {
					TMP_APPEND("expected size: at least %d octet(s)\n", entry->minsize);
				}
				else {
					TMP_APPEND("expected size: %d - %d octet(s)\n", entry->minsize, entry->maxsize);
				}
			}
			break;

		case Q931_MSGE_INVALID_CODESET:
			if ((entry = Q931MesgIEsGetEntry(trunk->Dialect, merrs->id, 8, merrs->ie_errors[i].id, NULL)) != NULL) {
				TMP_APPEND("expected codeset: %d\n", Q931MesgIEGetCodeset(entry));
			}
			break;

		default:
			TMP_PUTS("N/A\n");
			break;
		}

		/* last line of this ie? */
		if ((i + 1 == merrs->nr_errors) || (last_ie != merrs->ie_errors[i + 1].id)) {
#ifdef HAVE_ALLOCA
			/*
			 * print hex, omit if this is an overflow error
			 */
			if (merrs->ie_errors[i].offset) {
				int length = TMP_MIN((Size - merrs->ie_errors[i].offset), merrs->ie_errors[i].size);

				TMP_PUTS("\traw octets:\n");
				TMP_APPEND_HEX("\t\t[ %s ]\n", IBuf + merrs->ie_errors[i].offset, length);
			}
#endif
			/* IE separating footer? */
			TMP_PUTS("\n");
		}
		errcnt++;
	}

	if (merrs->nr_overflow > 0) {
		TMP_APPEND("* %d Additional errors have not been logged\n", merrs->nr_overflow);
	}

	/* print report footer */
	TMP_APPEND("---------------------------------------------------------------[ %2d error(s) ]--\n", merrs->nr_errors + merrs->nr_overflow);

	Q931Log(trunk, Q931_LOG_ERROR, "%s\n", tmp);
	return Q931E_NO_ERROR;
}

