/*
 * X.213 NSAP address handling (for Q.931 Calling/-ed Party Subaddress IE)
 *
 *
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <string.h>

#include "X213.h"
#include "ISO3166.h"

#include "utils/common.h"

/**********************************************************************
 * Constants
 **********************************************************************/

#define NSAP_IDI_X121_LEN		7	/* 7 octets: 14 digits */
#define NSAP_IDI_ISO_DCC_LEN		2	/* 2 octets:  3 digits + padding */
#define NSAP_IDI_F69_LEN		4	/* 4 octets:  8 digits */
#define NSAP_IDI_E163_LEN		6	/* 6 octets: 12 digits */
#define NSAP_IDI_E164_LEN		8	/* 8 octets: 15 digits + padding */
#define NSAP_IDI_ISO_6523_ICD_LEN	3	/* 3 octets:  6 digits */
#define NSAP_IDI_IANA_ICP_LEN		2	/* 2 octets:  4 digits */
#define NSAP_IDI_ITU_T_IND_LEN		3	/* 3 octets:  6 digits */
#define NSAP_IDI_LOCAL_LEN		0	/* no IDI */


/**********************************************************************
 * Tables
 **********************************************************************/

static const struct nsap_afi_info {
	const unsigned char afi;
	const unsigned char max_dsp_length;
	const nsap_idi_format_t idi_format;
	const nsap_dsp_syntax_t dsp_syntax;
} nsap_afi_info_map[] = {
	{ 0x36, NSAP_IDI_X121, NSAP_DSP_DECIMAL, 24 },
	{ 0x37, NSAP_IDI_X121, NSAP_DSP_BINARY,  12 },
	{ 0x52, NSAP_IDI_X121, NSAP_DSP_DECIMAL, 24 },
	{ 0x53, NSAP_IDI_X121, NSAP_DSP_BINARY,  12 },

	{ 0x38, NSAP_IDI_ISO_DCC, NSAP_DSP_DECIMAL, 35 },
	{ 0x39, NSAP_IDI_ISO_DCC, NSAP_DSP_BINARY,  17 },

	{ 0x40, NSAP_IDI_F69, NSAP_DSP_DECIMAL, 30 },
	{ 0x41, NSAP_IDI_F69, NSAP_DSP_BINARY,  15 },
	{ 0x54, NSAP_IDI_F69, NSAP_DSP_DECIMAL, 30 },
	{ 0x55, NSAP_IDI_F69, NSAP_DSP_BINARY,  15 },

	{ 0x42, NSAP_IDI_E163, NSAP_DSP_DECIMAL, 26 },
	{ 0x43, NSAP_IDI_E163, NSAP_DSP_BINARY,  13 },
	{ 0x56, NSAP_IDI_E163, NSAP_DSP_DECIMAL, 26 },
	{ 0x57, NSAP_IDI_E163, NSAP_DSP_BINARY,  13 },

	{ 0x44, NSAP_IDI_E164, NSAP_DSP_DECIMAL, 23 },
	{ 0x45, NSAP_IDI_E164, NSAP_DSP_BINARY,  11 },
	{ 0x58, NSAP_IDI_E164, NSAP_DSP_DECIMAL, 23 },
	{ 0x59, NSAP_IDI_E164, NSAP_DSP_BINARY,  11 },

	{ 0x46, NSAP_IDI_ISO_6523_ICD, NSAP_DSP_DECIMAL, 34 },
	{ 0x47, NSAP_IDI_ISO_6523_ICD, NSAP_DSP_BINARY,  17 },

	{ 0x34, NSAP_IDI_IANA_ICP, NSAP_DSP_DECIMAL, 34 },
	{ 0x35, NSAP_IDI_IANA_ICP, NSAP_DSP_BINARY,  17 },

	{ 0x76, NSAP_IDI_ITU_T_IND, NSAP_DSP_DECIMAL, 32 },
	{ 0x77, NSAP_IDI_ITU_T_IND, NSAP_DSP_BINARY,  16 },

	{ 0x48, NSAP_IDI_LOCAL, NSAP_DSP_DECIMAL,     38 },
	{ 0x49, NSAP_IDI_LOCAL, NSAP_DSP_BINARY,      19 },
	{ 0x50, NSAP_IDI_LOCAL, NSAP_DSP_ISO_IEC_646, 19 },
	{ 0x51, NSAP_IDI_LOCAL, NSAP_DSP_NATIONAL,     9 },
};

static const struct nsap_afi_info *nsap_afi(unsigned char afi)
{
	for (int i = 0; i < ARRAY_SIZE(nsap_afi_info_map); i++) {
		if (nsap_afi_info_map[i].afi == afi)
			return &nsap_afi_info_map[i];
	}
	return NULL;
}

static const char *nsap_idi_type(nsap_idi_format_t idi)
{
	switch (idi) {
	case NSAP_IDI_X121:
		return "X.121";
	case NSAP_IDI_ISO_DCC:
		return "ISO DCC";
	case NSAP_IDI_F69:
		return "F.69";
	case NSAP_IDI_E163:
		return "E.163";
	case NSAP_IDI_E164:
		return "E.164";
	case NSAP_IDI_ISO_6523_ICD:
		return "ISO 6523-ICD";
	case NSAP_IDI_IANA_ICP:
		return "IANA ICP";
	case NSAP_IDI_ITU_T_IND:
		return "ITU-T IND";
	case NSAP_IDI_LOCAL:
		return "Local";
	default:
		return "unknown/invalid";
	}
}


static int nsap_idi_decode_x121(struct nsap_addr *nsap, char *buf, int size)
{
	return NSAPE_NO_ERROR;
}


static inline int nsap_bcd2binary32(const char *buf, const int bufsize)
{
	int value = buf[0] & 0x0f;
	char nibble = 0;

	for (int i = !0, offset = 0; offset < bufsize; i = !i) {
		nibble = ((i) ? (buf[offset] >> 4) : buf[offset]) & 0x0f;
		if (nibble == 0x0f)	/* padding */
			break;
		value = (value * 10) + nibble;
		if (!(i)) offset++;
	}
	return value;
}

static inline int nsap_bcd2ascii(const char *inb, const int inbsize, char *outb, const int outbsize)
{
	static const char c[16] = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'X', 'X', 'X', 'X', 'X', 'X' };
	int nibble = 0, i, j;

	for (i = 0, j = 0; i < inbsize && j < outbsize; j++) {
		nibble = ((j & 1) ? (inb[i] >> 4) : inb[i])  & 0x0f;
		if (nibble == 0x0f)	/* padding */
			break;
		outb[j] = c[nibble];
		if (j & 1) i++;
	}
	return j;
}

static int nsap_idi_decode_iso_dcc(struct nsap_addr *nsap, char *buf, int size)
{
	if (size < NSAP_IDI_ISO_DCC_LEN)
		return NSAPE_TOO_SHORT;

	nsap->nsap_a_iso_dcc.dcc = nsap_bcd2binary32(buf, NSAP_IDI_ISO_DCC_LEN);

	if (iso3166_by_id(nsap->nsap_a_iso_dcc.dcc, &nsap->nsap_a_iso_dcc.name, &nsap->nsap_a_iso_dcc.code) < 0)
		return NSAPE_GENERR;

	return NSAPE_NO_ERROR;
}


static int nsap_idi_decode_f69(struct nsap_addr *nsap, char *buf, int size)
{
	int offset = 0;

	if (size < NSAP_IDI_F69_LEN)
		return NSAPE_TOO_SHORT;

	nsap_bcd2ascii(buf, NSAP_IDI_F69_LEN, nsap->nsap_a_f69.telex, sizeof(nsap->nsap_a_f69.telex) - 1);
	offset += NSAP_IDI_F69_LEN;

	switch (nsap->dsp) {
	case NSAP_DSP_BINARY:
		memcpy(nsap->nsap_a_f69.value, &buf[NSAP_IDI_F69_LEN], MIN(size - NSAP_IDI_F69_LEN, sizeof(nsap->nsap_a_f69.value) - 1));
		break;
	case NSAP_DSP_DECIMAL:
		nsap_bcd2ascii(&buf[NSAP_IDI_F69_LEN], size - NSAP_IDI_F69_LEN, nsap->nsap_a_f69.value, sizeof(nsap->nsap_a_f69.value) - 1);
		break;
	default:
		return NSAPE_INVALID_FORMAT;
	}
	return NSAPE_NO_ERROR;
}


static int nsap_idi_decode_e163(struct nsap_addr *nsap, char *buf, int size)
{
	if (size < NSAP_IDI_E163_LEN)
		return NSAPE_TOO_SHORT;

	memset(&nsap->nsap_a_e163, 0, sizeof(nsap->nsap_a_e163));

	nsap_bcd2ascii(buf, NSAP_IDI_E163_LEN, nsap->nsap_a_e163.number, sizeof(nsap->nsap_a_e163.number) - 1);

	switch (nsap->dsp) {
	case NSAP_DSP_BINARY:
		memcpy(nsap->nsap_a_e163.value, &buf[NSAP_IDI_E163_LEN], MIN(size - NSAP_IDI_E163_LEN, sizeof(nsap->nsap_a_e163.value) - 1));
		break;
	case NSAP_DSP_DECIMAL:
		nsap_bcd2ascii(&buf[NSAP_IDI_E163_LEN], size - NSAP_IDI_E163_LEN, nsap->nsap_a_e163.value, sizeof(nsap->nsap_a_e163.value) - 1);
		break;
	default:
		return NSAPE_INVALID_FORMAT;
	}
	return NSAPE_NO_ERROR;
}


static int nsap_idi_decode_e164(struct nsap_addr *nsap, char *buf, int size)
{
	if (size < NSAP_IDI_E164_LEN)
		return NSAPE_TOO_SHORT;

	memset(&nsap->nsap_a_e164, 0, sizeof(nsap->nsap_a_e164));

	nsap_bcd2ascii(buf, NSAP_IDI_E164_LEN, nsap->nsap_a_e164.number, sizeof(nsap->nsap_a_e164.number) - 1);

	switch (nsap->dsp) {
	case NSAP_DSP_BINARY:
		memcpy(nsap->nsap_a_e164.value, &buf[NSAP_IDI_E164_LEN], MIN(size - NSAP_IDI_E164_LEN, sizeof(nsap->nsap_a_e164.value) - 1));
		break;
	case NSAP_DSP_DECIMAL:
		nsap_bcd2ascii(&buf[NSAP_IDI_E164_LEN], size - NSAP_IDI_E164_LEN, nsap->nsap_a_e164.value, sizeof(nsap->nsap_a_e164.value) - 1);
		break;
	default:
		return NSAPE_INVALID_FORMAT;
	}
	return NSAPE_NO_ERROR;
}


static int nsap_idi_decode_iso_6523_icd(struct nsap_addr *nsap, char *buf, int size)
{
	int offset = 0;

	if (size < NSAP_IDI_ISO_6523_ICD_LEN)
		return NSAPE_TOO_SHORT;

	memset(&nsap->nsap_a_iso_6523_icd, 0, sizeof(nsap->nsap_a_iso_6523_icd));

	nsap->nsap_a_iso_6523_icd.icd = nsap_bcd2binary32(&buf[offset], NSAP_IDI_ISO_6523_ICD_LEN);
	offset += NSAP_IDI_ISO_6523_ICD_LEN;

	switch (nsap->dsp) {
	case NSAP_DSP_BINARY:
		break;
	case NSAP_DSP_DECIMAL:
		break;
	default:
		return NSAPE_INVALID_FORMAT;
	}
	return NSAPE_NO_ERROR;
}


static int nsap_idi_decode_iana_icp(struct nsap_addr *nsap, char *buf, int size)
{
	unsigned short int icp = 0;
	int offset = 0;

	if (size < NSAP_IDI_IANA_ICP_LEN)
		return NSAPE_TOO_SHORT;

	memset(&nsap->nsap_a_iana_icp, 0, sizeof(nsap->nsap_a_iana_icp));

	memcpy(&icp, &buf[offset], 2);	/* copy 4 octets of ICP */
	offset += 2;

	switch (icp) {
	case 0x0000:	/* IPv6 */
		if (size < 18)
			return NSAPE_TOO_SHORT;
		nsap->nsap_a_iana_icp.family = AF_INET6;
		memcpy(&nsap->nsap_a_iana_icp.icp_a_ip6, &buf[offset], 16);
		break;
	case 0x0001:	/* IPv4 */
		if (size < 6)
			return NSAPE_TOO_SHORT;
		nsap->nsap_a_iana_icp.family = AF_INET;
		memcpy(&nsap->nsap_a_iana_icp.icp_a_ip4, &buf[offset], 4);
		break;
	default:
		nsap->nsap_a_iana_icp.family = AF_UNSPEC;
	}
	return NSAPE_NO_ERROR;
}


static int nsap_idi_decode_itu_t_ind(struct nsap_addr *nsap, char *buf, int size)
{
	int offset = 0;

	if (size < NSAP_IDI_ITU_T_IND_LEN)
		return NSAPE_TOO_SHORT;

	memset(&nsap->nsap_a_itu_t_ind, 0, sizeof(nsap->nsap_a_itu_t_ind));

	nsap->nsap_a_itu_t_ind.ind = nsap_bcd2binary32(&buf[offset], NSAP_IDI_ITU_T_IND_LEN);
	offset += NSAP_IDI_ITU_T_IND_LEN;

	switch (nsap->dsp) {
	case NSAP_DSP_BINARY:
		break;
	case NSAP_DSP_DECIMAL:
		break;
	default:
		return NSAPE_INVALID_FORMAT;
	}
	return NSAPE_NO_ERROR;
}


static int nsap_idi_decode_local(struct nsap_addr *nsap, char *buf, int size)
{
	switch (nsap->dsp) {
	case NSAP_DSP_BINARY:
		break;
	case NSAP_DSP_DECIMAL:
		break;
	case NSAP_DSP_ISO_IEC_646:
		break;
	case NSAP_DSP_NATIONAL:
		break;
	default:
		return NSAPE_INVALID_FORMAT;
	}
	return NSAPE_NO_ERROR;
}

/**
 * Check whether NSAP AFI can be decoded
 * @param[in]	afi	AFI to check
 * @return	true (1) or false (0)
 */
int nsap_afi_valid(const char afi)
{
	return !((afi >= 0x00 && afi <= 0x0f) || afi == 0xff);
}

int nsap_decode(struct nsap_addr *addr, char *buf, int size)
{
	int offset = 0;
	int ret = 0;

	const struct nsap_afi_info *afi = nsap_afi(buf[offset++]);
	if (!afi)
		return NSAPE_NO_ERROR;

	addr->idi = afi->idi_format;
	addr->dsp = afi->dsp_syntax;
	addr->max_dsp_length = afi->max_dsp_length;

	switch (addr->idi) {
	case NSAP_IDI_X121:
		ret = nsap_idi_decode_x121(addr, &buf[offset], size - offset);
		break;
	case NSAP_IDI_ISO_DCC:
		ret = nsap_idi_decode_iso_dcc(addr, &buf[offset], size - offset);
		break;
	case NSAP_IDI_F69:
		ret = nsap_idi_decode_f69(addr, &buf[offset], size - offset);
		break;
	case NSAP_IDI_E163:
		ret = nsap_idi_decode_e163(addr, &buf[offset], size - offset);
		break;
	case NSAP_IDI_E164:
		ret = nsap_idi_decode_e164(addr, &buf[offset], size - offset);
		break;
	case NSAP_IDI_ISO_6523_ICD:
		ret = nsap_idi_decode_iso_6523_icd(addr, &buf[offset], size - offset);
		break;
	case NSAP_IDI_IANA_ICP:
		ret = nsap_idi_decode_iana_icp(addr, &buf[offset], size - offset);
		break;
	case NSAP_IDI_ITU_T_IND:
		ret = nsap_idi_decode_itu_t_ind(addr, &buf[offset], size - offset);
		break;
	case NSAP_IDI_LOCAL:
		ret = nsap_idi_decode_local(addr, &buf[offset], size - offset);
		break;
	default:
		break;
	}

	if (ret <= 0)
		return ret;

	offset += ret;

	return offset;
}

/**
 * Create textual representation of NSAP address
 * @param[in]	addr	NSAP address
 * @param[out]	buf	Output buffer
 * @param[in]	size	Size of output buffer
 * @return	NSAPE_NO_ERROR on success, NSAPE_* otherwise
 */
int nsap_print(struct nsap_addr *addr, char *buf, int size)
{
	int offset = 0;

	offset += snprintf(buf, size, "%s ", nsap_idi_type(addr->idi));

	switch (addr->idi) {
	case NSAP_IDI_IANA_ICP:
		switch (addr->nsap_a_iana_icp.family) {
#if defined(HAVE_INET_NTOP)
		case AF_INET:
			strncat(&buf[offset], "IPv4: ", size - offset);
			offset = strlen(buf);
			inet_ntop(AF_INET, &addr->nsap_a_iana_icp.icp_a_ip4, &buf[offset], size - offset);
			break;
		case AF_INET6:
			strncat(&buf[offset], "IPv6: ", size - offset);
			offset = strlen(buf);
			inet_ntop(AF_INET6, &addr->nsap_a_iana_icp.icp_a_ip6, &buf[offset], size - offset);
			break;
#elif defined(HAVE_GETNAMEINFO)
		case AF_INET:
			{
				struct sockaddr_in tmp;

				memcpy(&tmp.sin_addr, &addr->nsap_a_iana_icp.icp_a_ip4, sizeof(addr->nsap_a_iana_icp.icp_a_ip4));
				tmp.sin_family = AF_INET;
				tmp.sin_port = 0;

				strncat(&buf[offset], "IPv4: ", size - offset);
				offset = strlen(buf);
				getnameinfo((struct sockaddr *)&tmp, sizeof(tmp), &buf[offset], size - offset, NULL, 0, NI_NUMERICHOST);
			}
			break;
		case AF_INET6:
			{
				struct sockaddr_in6 tmp;

				memcpy(&tmp.sin6_addr, &addr->nsap_a_iana_icp.icp_a_ip6, sizeof(addr->nsap_a_iana_icp.icp_a_ip6));
				tmp.sin6_family = AF_INET6;
				tmp.sin6_port = 0;

				strncat(&buf[offset], "IPv6: ", size - offset);
				offset = strlen(buf);
				getnameinfo((struct sockaddr *)&tmp, sizeof(tmp), &buf[offset], size - offset, NULL, 0, NI_NUMERICHOST);
			}
			break;
#else
		/*
		 * Poor man's (probably endianess-broken) replacement for missing getnameinfo() and inet_ntop()
		 * Only added to silence the WIN32 build which keeps failing to find its getnameinfo() function provided
		 * by winsock2
		 *
		 * @todo	Add custom getnameinfo() and inet_ntop() autoconf checks which includes ws2tcpip.h et.al.
		 *		(and then remove this stuff)
		 */
		case AF_INET:
			{
				char tmp[4];

				memcpy(tmp, &addr->nsap_a_iana_icp.icp_a_ip4.s_addr, 4);
				snprintf(&buf[offset], size - offset, "IPv4: %d.%d.%d.%d", tmp[0], tmp[1], tmp[2], tmp[3]);
				offset = strlen(buf);
			}
			break;
		case AF_INET6:
			{
				short tmp[8];

				memcpy(tmp, &addr->nsap_a_iana_icp.icp_a_ip6.s6_addr, 16);
				snprintf(&buf[offset], size - offset, "IPv6: %x:%x:%x:%x:%x:%x:%x:%x",
					tmp[0], tmp[1], tmp[2], tmp[3], tmp[4], tmp[5], tmp[6], tmp[7]);
				offset = strlen(buf);
			}
			break;
#endif
		case AF_UNSPEC:
			strncat(&buf[offset], "-N/A-", size - offset);
			break;
		}
		break;
	case NSAP_IDI_X121:
		break;
	case NSAP_IDI_F69:
		offset += snprintf(&buf[offset], size - offset, "Authority: %s, Value: ", addr->nsap_a_f69.telex);
		switch (addr->dsp) {
		case NSAP_DSP_DECIMAL:
			strncat(&buf[offset], addr->nsap_a_f69.value, size - offset);
			offset = strlen(buf);
			break;
		default:
			/* print hex */
			break;
		}
		break;
	case NSAP_IDI_E163:
		strncat(&buf[offset], "Number: ", size - offset);
		offset += strlen(buf);
		switch (addr->dsp) {
		case NSAP_DSP_DECIMAL:
			strncat(&buf[offset], addr->nsap_a_e163.value, size - offset);
			offset = strlen(buf);
			break;
		default:
			/* print hex */
			break;
		}
		break;
	case NSAP_IDI_E164:
		strncat(&buf[offset], "Number: ", size - offset);
		offset += strlen(buf);
		switch (addr->dsp) {
		case NSAP_DSP_DECIMAL:
			strncat(&buf[offset], addr->nsap_a_e164.value, size - offset);
			offset = strlen(buf);
			break;
		default:
			/* print hex */
			break;
		}
		break;
	case NSAP_IDI_ITU_T_IND:
		offset += snprintf(&buf[offset], size - offset, "IND: %6d, Value: ", addr->nsap_a_itu_t_ind.ind);
		switch (addr->dsp) {
		case NSAP_DSP_DECIMAL:
			strncat(&buf[offset], addr->nsap_a_itu_t_ind.value, size - offset);
			offset = strlen(buf);
			break;
		default:
			/* print hex */
			break;
		}
		break;
	case NSAP_IDI_ISO_DCC:
		offset += snprintf(&buf[offset], size - offset, "ISO Code: %2s, Value: ", addr->nsap_a_iso_dcc.code);
		switch (addr->dsp) {
		case NSAP_DSP_DECIMAL:
			strncat(&buf[offset], addr->nsap_a_iso_dcc.value, size - offset);
			offset = strlen(buf);
			break;
		default:
			/* print hex */
			break;
		}
		break;
	case NSAP_IDI_ISO_6523_ICD:
	case NSAP_IDI_LOCAL:
		strncat(buf, "decoding not implemented", size);
		break;
	default:
		strncat(buf, "-N/A-", size);
	}
	return NSAPE_NO_ERROR;
}
