/*****************************************************************************

  FileName:     q921.h

  Description:  Contains headers of a Q.921 protocol.

  Note:         This header file is the only include file that should be
                acessed by users of the Q.921 stack.

  Interface:    The Q.921 stack contains 2 layers.

                -   One interface layer.
                -   One driver layer.

                The interface layer contains the interface functions required
                for a layer 2 stack to be able to send and receive messages.

                The driver layer will simply feed bytes into the ship as
                required and queue messages received out from the ship.

                Q921TimeTick        The Q.921 like any other blackbox
                                    modules contains no thread by it's own
                                    and must therefore be called regularly
                                    by an external 'thread' to do maintenance
                                    etc.

                Q921Rx32            Receive message from layer 3. Called by
                                    the layer 3 stack to send a message.


				NOTE: The following are not yet implemented

                OnQ921Error         Function called every if an error is
                                    detected.

                OnQ921Log           Function called if logging is active.


                <TODO> Maintenance/Configuration interface
				<TODO> Logging
				<TODO> DL_ message passing to layer 3
				<TODO> Timers
				<TODO> Api commands to tell 921 to stop and start for a trunk

  Created:      27.dec.2000/JVB

  License/Copyright:

  Copyright (c) 2007, Jan Vidar Berger, Case Labs, Ltd. All rights reserved.
  email:janvb@caselaboratories.com

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are
  met:

    * Redistributions of source code must retain the above copyright notice,
	  this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
	  this list of conditions and the following disclaimer in the documentation
	  and/or other materials provided with the distribution.
    * Neither the name of the Case Labs, Ltd nor the names of its contributors
	  may be used to endorse or promote products derived from this software
	  without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.

*****************************************************************************/

/****************************************************************************
 * Changes:
 *
 * - June,July 2008: Stefan Knoblich <s.knoblich@axsentis.de>:
 *     Add PTMP TEI management
 *     Add timers
 *     Add retransmit counters
 *     Add logging
 *     Various cleanups
 *
 * - April 2009: Stefan Knoblich <s.knoblich@axsentis.de>:
 *     More cleanups
 *     Doxygen documentation
 *
 ****************************************************************************/

#ifndef _Q921_H_
#define _Q921_H_

#include "Q9XXplatform.h"

#define	Q921_STATISTICS
#define	Q921_STATISTICS_EXTENDED

#define Q921MAXHDLCSPACE 3000
#define L2UCHAR		unsigned char		/* Min 8 bit			*/
#define L2USHORT	unsigned short		/* 16 bit			*/
#define L2INT		int			/* Min 16 bit signed		*/
#define L2ULONG		unsigned long		/* Min 32 bit			*/
#define L2TRUNK		Q921Data_t *

#define Q921_TEI_BCAST		127
#define Q921_TEI_MAX		Q921_TEI_BCAST

#define Q921_TEI_DYN_MIN	64
#define Q921_TEI_DYN_MAX	126


/**
 * Mode
 */
typedef enum
{
	Q921_TE = 0,		/*!< User Mode			*/
	Q921_NT = 1		/*!< Network Mode		*/
} Q921NetUser_t;

/**
 * Link type
 */
typedef enum
{
	Q921_PTP  = 0,		/*!< Point-To-Point		*/
	Q921_PTMP = 1		/*!< Point-To-Multipoint	*/
} Q921NetType_t;

/**
 * Log levels
 */
typedef enum
{
	Q921_LOG_NONE  = -1,
	Q921_LOG_EMERG =  0,
	Q921_LOG_ALERT,
	Q921_LOG_CRIT,
	Q921_LOG_ERROR,
	Q921_LOG_WARNING,
	Q921_LOG_NOTICE,
	Q921_LOG_INFO,
	Q921_LOG_DEBUG
} Q921LogLevel_t;

/*
 * Messages for L2 <-> L3 communication
 */
typedef enum {
	Q921_DL_ESTABLISH = 0,
	Q921_DL_ESTABLISH_CONFIRM,
	Q921_DL_RELEASE,
	Q921_DL_RELEASE_CONFIRM,
	Q921_DL_DATA,
	Q921_DL_UNIT_DATA
} Q921DLMsg_t;

typedef int (*Q921Tx21CB_t) (void *, L2UCHAR *, L2INT);
typedef int (*Q921Tx23CB_t) (void *, Q921DLMsg_t ind, L2UCHAR tei, L2UCHAR *, L2INT);
typedef int (*Q921LogCB_t)  (void *, Q921LogLevel_t, char *, L2INT);

struct Q921_Link;

#ifdef Q921_STATISTICS
enum {
	Q921_STATS_NONE = 0,

	/* state changes */
	Q921_STATS_ST01,
	Q921_STATS_ST02,
	Q921_STATS_ST03,
	Q921_STATS_ST04,
	Q921_STATS_ST05,
	Q921_STATS_ST06,
	Q921_STATS_ST07,
	Q921_STATS_ST08,

	/* timer events */
	Q921_STATS_T200,
	Q921_STATS_T201,
	Q921_STATS_T202,
	Q921_STATS_T203,

	Q921_STATS_TM01,

	/* limit / counter overruns */
	Q921_STATS_N200,
	Q921_STATS_N201,
	Q921_STATS_N202,

#ifdef Q921_STATISTICS_EXTENDED
	/* packet send counter */
	Q921_STATS_SEND_S,
	Q921_STATS_SEND_U,
	Q921_STATS_SEND_I,

	Q921_STATS_SEND_DISC,
	Q921_STATS_SEND_DM,
	Q921_STATS_SEND_REJ,
	Q921_STATS_SEND_RR,
	Q921_STATS_SEND_RNR,
	Q921_STATS_SEND_SABME,
	Q921_STATS_SEND_UA,
	Q921_STATS_SEND_UN,

	/* packet receive counter */
	Q921_STATS_RECV_S,
	Q921_STATS_RECV_U,
	Q921_STATS_RECV_I,

	Q921_STATS_RECV_DISC,
	Q921_STATS_RECV_DM,
	Q921_STATS_RECV_REJ,
	Q921_STATS_RECV_RNR,
	Q921_STATS_RECV_RR,
	Q921_STATS_RECV_SABME,
	Q921_STATS_RECV_UA,
	Q921_STATS_RECV_UN,

	Q921_STATS_RECV_INVALID_S,
	Q921_STATS_RECV_INVALID_U,
	Q921_STATS_RECV_INVALID_I,

	/* TEI-management counter */
	Q921_STATS_TEI_SEND_AR,		/*!< Assign   Request  sent */
	Q921_STATS_TEI_SEND_AS,		/*!< Assigned Response sent */
	Q921_STATS_TEI_SEND_CR,		/*!< Check    Request  sent */
	Q921_STATS_TEI_SEND_CS,		/*!< Check    Response sent */
	Q921_STATS_TEI_SEND_DS,		/*!< Deny     Response sent */
	Q921_STATS_TEI_SEND_RR,		/*!< Remove   Request  sent */
	Q921_STATS_TEI_SEND_VR,		/*!< Verify   Request  sent */

	Q921_STATS_TEI_RECV_AR,		/*!< Assign Request  received */
	Q921_STATS_TEI_RECV_AS,		/*!< Assign Response received */
	Q921_STATS_TEI_RECV_CR,		/*!< Check  Request  received */
	Q921_STATS_TEI_RECV_CS,		/*!< Check  Response received */
	Q921_STATS_TEI_RECV_RR,		/*!< Remove Request  received */
	Q921_STATS_TEI_RECV_VR,		/*!< Verify Request  received */
#endif
	/* don't touch */
	Q921_STATS_MAX
};
#endif

/**
 * Q921Data_t
 * @brief	Q.921 Trunk context
 */
typedef struct Q921Data
{
	L2INT initialized;

	L2UCHAR sapi;			/*!< User assigned SAPI */
	L2UCHAR tei;			/*!< User assigned TEI value */

	L2INT Q921HeaderSpace;
	Q921NetUser_t NetUser;
	Q921NetType_t NetType;

	struct Q921_Link *context;	/*!< per-TEI / link context space */

	/* timers */
	L2ULONG T202;			/*!< PTMP TE mode TEI retransmit timer */
	L2ULONG T200Timeout;
	L2ULONG T201Timeout;
	L2ULONG T202Timeout;
	L2ULONG T203Timeout;

	L2ULONG TM01Timeout;

	/* counters */
	L2ULONG N200Limit;		/*!< max retransmit */

	L2ULONG N202;			/*!< PTMP TE mode retransmit counter */
	L2ULONG N202Limit;		/*!< PTMP TE mode max retransmit */

	L2ULONG N201Limit;		/*!< max number of octets */
	L2ULONG k;			/*!< max number of unacknowledged I frames */

	/* callbacks and callback data pointers */
	Q921Tx21CB_t Q921Tx21Proc;
	Q921Tx23CB_t Q921Tx23Proc;
	void *PrivateData21;
	void *PrivateData23;

	/* logging */
	Q921LogLevel_t	loglevel;	/*!< trunk loglevel */
	Q921LogCB_t	Q921LogProc;	/*!< log callback procedure */
	void *PrivateDataLog;		/*!< private data pointer for log proc */

	/* tei mgmt */
	L2UCHAR tei_map[Q921_TEI_MAX];	/*!< */

	L2UCHAR HDLCInQueue[Q921MAXHDLCSPACE];	/*!< HDLC input queue */
} Q921Data_t;

/**
 * Public functions
 */

/**
 * Initialize Q.921 Trunk
 * @param[in]	trunk	Q.921 trunk
 * @param[in]	sapi	SAPI
 * @param[in]	tei	TEI (ignored in PTMP mode)
 * @param[in]	NetUser Mode (Network / User)
 * @param[in]	NetType Type (PTP / PTMP)
 * @param[in]	hsize
 * @param[in]	cb21	Layer1 TX callback, called to send Q.921 to Layer 1
 * @param[in]	cb23	Layer3 RX callback, called to submit frames to Layer 3
 * @param[in]	priv21	Custom data for Layer1 TX callback
 * @param[in]	priv23	Custom data for Layer3 RX callback
 */
Q921_API int Q921_InitTrunk(L2TRUNK trunk,
					L2UCHAR sapi,
					L2UCHAR tei,
					Q921NetUser_t NetUser,
					Q921NetType_t NetType,
					L2INT hsize,
					Q921Tx21CB_t cb21,
					Q921Tx23CB_t cb23,
					void *priv21,
					void *priv23);
/**
 * Start Q.921 Trunk
 * @param[in]	trunk	Q.921 trunk
 */
Q921_API int Q921Start(L2TRUNK trunk);

/**
 * Stop Q.921 Trunk
 * @param[in]	trunk	Q.921 trunk
 */
Q921_API int Q921Stop(L2TRUNK trunk);

/**
 * Set logging callback and private data
 * @param[in]	trunk	Q.921 trunk
 * @param[in]	func	Callback function
 * @param[in]	priv	Custom data passed to callback
 */
Q921_API void Q921SetLogCB(L2TRUNK trunk, Q921LogCB_t func, void *priv);

/**
 * Set log level
 * @param[in]	trunk	Q.921 trunk
 * @param[in]	level	Log level
 */
Q921_API void Q921SetLogLevel(L2TRUNK trunk, Q921LogLevel_t level);

/**
 * Get log level
 * @param[in]	trunk	Q.921 trunk
 * @return	Log level
 */
Q921_API Q921LogLevel_t Q921GetLogLevel(L2TRUNK trunk);

/**
 * Get log level name
 * @param[in]	trunk	Q.921 trunk
 * @return	Name of current log level
 */
Q921_API const char * Q921GetLogLevelName(L2TRUNK trunk);

/**
 * Receive incoming data from Layer 1
 * @brief	Process Q.921 frame that has been enqueued using Q921QueueHDLCFrame()
 * @param[in]	trunk	Q.921 trunk
 */
Q921_API int Q921Rx12(L2TRUNK trunk);

/**
 * Receive data from Layer 3
 * @param[in]	trunk	Q.921 trunk
 * @param[in]	ind	Type of message
 * @param[in]	tei	TEI of this message
 * @param[in]	Mes	Message
 * @param[in]	Size	Size of message (octets)
 */
Q921_API int Q921Rx32(L2TRUNK trunk, Q921DLMsg_t ind, L2UCHAR tei, L2UCHAR * Mes, L2INT Size);

/**
 * Enqueue new incoming Q.921 frame from Layer 1
 * @param[in]	trunk	Q.921 trunk
 * @param[in]	b	Buffer to incoming data
 * @param[in]	size	Size of incoming data
 */
Q921_API int Q921QueueHDLCFrame(L2TRUNK trunk, L2UCHAR *b, L2INT size);

/**
 * Set custom callback to get current time in ms
 * @param[in]	callback	Callback function
 */
Q921_API void Q921SetGetTimeCB(L2ULONG (*callback)(void));

/**
 * Update internal timers
 * @param[in]	trunk	Q.921 trunk
 */
Q921_API void Q921TimerTick(L2TRUNK trunk);

/**
 * Establish Q.921 link
 * @param[in]	trunk	Q.921 trunk
 * @param[in]	tei	TEI of link to establish
 */
Q921_API int Q921Establish(L2TRUNK trunk, L2UCHAR tei);

/**
 * Release Q.921 link
 * @param[in]	trunk	Q.921 trunk
 * @param[in]	tei	TEI of link to release
 */
Q921_API int Q921Release(L2TRUNK trunk, L2UCHAR tei);

/**
 * Is Q.921 link ready to transmit data?
 * @param[in]	trunk	Q.921 trunk
 * @param[in]	tei	TEI of link to check
 */
Q921_API int Q921IsEstablished(L2TRUNK trunk, L2UCHAR tei);

/**
 * Q.921 Statistic functions
 * @defgroup	q921_stats	Q.921 Statistics
 * @{
 */
#ifdef Q921_STATISTICS
/**
 * Q.921 Statistic counter
 * @note	Used as an iterator
 */
struct Q921StatsCounter
{
	const int  id;
	const char *name;
	const char *desc;
};

/**
 * Q921StatsCounterResetAll
 * @brief	Reset all counter to zero
 * @param[in]	trunk	pointer to Q.921 data struct
 * @param[in]	tei	Q921_TEI_MAX = all, otherwise reset only counters of given tei
 *
 * @author	Stefan Knoblich <s.knoblich@axsentis.de>
 */
Q921_API void Q921StatsCounterResetAll(L2TRUNK trunk, const int tei);

/**
 * Q921StatsCounterReset
 * @brief	Reset all counter to zero
 * @param[in]	trunk	pointer to Q.921 data struct
 * @param[in]	id	id of counter to reset
 * @param[in]	tei	Q921_TEI_MAX = all, otherwise reset only counters of given tei
 *
 * @author	Stefan Knoblich <s.knoblich@axsentis.de>
 */
Q921_API void Q921StatsCounterReset(L2TRUNK trunk, const int id, const int tei);

/**
 * Counter iteration API
 * @defgroup q921_stats_iterator	Q.921 Statistics, iterator-based
 * @{
 */

Q921_API const struct Q921StatsCounter *Q921StatsCounterFirst(void);
Q921_API const struct Q921StatsCounter *Q921StatsCounterNext(const struct Q921StatsCounter *cur);

/**
 * Q921StatsCounterValue
 * @brief	Retrieve current value of specified statistic event counter of one TEI, or the sum of the counters of all TEIs
 * @param[in]	trunk	pointer to Q.921 data struct
 * @param[in]	counter	Counter
 * @param[in]	tei	TEI of link: 0 for trunk, Q921_TEI_BCAST for sum of all counters, TEI of link otherwise
 * @return	counter value or 0 on error
 *
 * @author	Stefan Knoblich <s.knoblich@axsentis.de>
 */
Q921_API unsigned int Q921StatsCounterValue(const L2TRUNK trunk, const struct Q921StatsCounter *counter, const int tei);

/**
 * Get ID of current counter
 * @param[in]	counter	Counter
 * @return	id on success Q931_STATS_NONE on error
 *
 * @author	Stefan Knoblich <s.knoblich@axsentis.de>
 */
Q921_API int Q921StatsCounterID(const struct Q921StatsCounter *counter);

/**
 * Get Name of current counter
 * @param[in]	counter	Counter
 * @return	name on success, NULL on error
 *
 * @author	Stefan Knoblich <s.knoblich@axsentis.de>
 */
Q921_API const char *Q921StatsCounterName(const struct Q921StatsCounter *counter);

/**
 * Get Description of current counter
 * @param[in]	counter	Counter
 * @return	description on success, NULL on error
 *
 * @author	Stefan Knoblich <s.knoblich@axsentis.de>
 */
Q921_API const char *Q921StatsCounterDescription(const struct Q921StatsCounter *counter);

/**
 * Is current counter shared between multiple TEIs (= global)
 * @param[in]	counter	Counter
 * @return	0 on false or error, 1 on true
 *
 * @author	Stefan Knoblich <s.knoblich@axsentis.de>
 */
Q921_API int Q921StatsCounterIsGlobal(const struct Q921StatsCounter *counter);

/**
 * Does current counter indicate an error condition
 * @param[in]	counter	Counter
 * @return	0 on false or error, 1 on true
 *
 * @author	Stefan Knoblich <s.knoblich@axsentis.de>
 */
Q921_API int Q921StatsCounterIsError(const struct Q921StatsCounter *counter);

/**
 * Is current counter available (in this mode)
 * @param[in]	trunk	Q.921 trunk, needed for configuration details
 * @param[in]	counter	Counter
 * @return	0 on false or error, 1 on true
 *
 * @author	Stefan Knoblich <s.knoblich@axsentis.de>
 */
Q921_API int Q921StatsCounterIsAvailable(const L2TRUNK trunk, const struct Q921StatsCounter *counter);

/**
 * Iterate over all counter
 * @param[in]	name	Name of local counter variable
 *
 * @code
 * Q921StatsCounterForeach(cur) {
 * 	printf("Counter '%s' on tei %d has value %u\n",
 * 		Q921StatsCounterName(cur),
 * 		tei,
 * 		Q921StatsCounterValue(q921_trunk, cur, tei));
 * }
 * @endcode
 *
 * @author	Stefan Knoblich <s.knoblich@axsentis.de>
 */
#define Q921StatsCounterForeach(name)	\
	for (const struct Q921StatsCounter * name = Q921StatsCounterFirst(); name; name = Q921StatsCounterNext(name))

/** @} */

/**
 * Counter direct access API
 * @defgroup	q921_stats_direct	Q.921 Statistics, access by id
 * @{
 */

Q921_API const char * Q921StatsCounterGetDescription(const int id);
Q921_API const char * Q921StatsCounterGetName(const int id);
Q921_API int Q921StatsCounterGetIsGlobal(const int id);
Q921_API unsigned int Q921StatsCounterGetValue(const L2TRUNK trunk, const int id, const int tei);

/** @} */

#endif
/** @} */

#endif /* _Q921_H_ */
