/*****************************************************************************

  FileName:		Q931call.h

  Contents:		call related api functions.

  License/Copyright:

  Copyright (c) 2008, Stefan Knoblich, axsentis GmbH. All rights reserved.
  email: s.knoblich@axsentis.de

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are
  met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.
    * Neither the name of the axsentis GmbH nor the names of its contributors
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.
*****************************************************************************/

#ifndef _Q931CALL_H_
#define _Q931CALL_H_

#define Q931_CALL_MAX_EVENTS	3	/*!< Should be enough, maybe even too much... */

/**
 * Events
 */
enum {
	Q931_EVENT_NONE = 0,			/*!< NOP, sort of */

	/* Indications (L3 -> L4) */
	Q931_EVENT_SETUP_INDICATION,		/*!< Incoming SETUP */
	Q931_EVENT_SETUP_CONFIRM,		/*!< */
	Q931_EVENT_SETUP_COMPLETE_INDICATION,	/*!< */
	Q931_EVENT_MORE_INFO_INDICATION,	/*!< */
	Q931_EVENT_REJECT_INDICATION,		/*!< */
	Q931_EVENT_PROCEEDING_INDICATION,	/*!< */
	Q931_EVENT_ALERTING_INDICATION,		/*!< */
	Q931_EVENT_PROGRESS_INDICATION,		/*!< */
	Q931_EVENT_NOTIFY_INDICATION,		/*!< */
	Q931_EVENT_RELEASE_INDICATION,		/*!< */
	Q931_EVENT_RELEASE_CONFIRM,		/*!< */
	Q931_EVENT_SUSPEND_CONFIRM,		/*!< */
	Q931_EVENT_RESUME_CONFIRM,		/*!< */
	Q931_EVENT_DISCONNECT_INDICATION,	/*!< */
	Q931_EVENT_INFORMATION_INDICATION,	/*!< */
	Q931_EVENT_STATUS_INDICATION,		/*!< */
	Q931_EVENT_RESTART_CONFIRM,		/*!< */

	/* Requests (L4 -> L3) (?) */
	Q931_EVENT_SETUP_RESPONSE,		/*!< */
	Q931_EVENT_REJECT_REQUEST,		/*!< */
	Q931_EVENT_INFORMATION_REQUEST,		/*!< */
	Q931_EVENT_ALERTING_REQUEST,		/*!< */
	Q931_EVENT_PROGRESS_REQUEST,		/*!< */
	Q931_EVENT_PROCEEDING_REQUEST,		/*!< */
	Q931_EVENT_NOTIFY_REQUEST,		/*!< */
	Q931_EVENT_SUSPEND_REQUEST,		/*!< */
	Q931_EVENT_DISCONNECT_REQUEST,		/*!< */
	Q931_EVENT_RESTART_REQUEST,		/*!< */

	/* Misc events (L2/L3 -> L4) */
	Q931_EVENT_TIMEOUT_INDICATION,		/*!< Overlap dial(?) timer timeout */
	Q931_EVENT_DL_FAILURE_INDICATION,	/*!< Failure on Layer 2 */

	/* CRV events */
	Q931_EVENT_RELEASE_CRV,			/*!< Call/CRV will be released */
};

/**
 * @brief	Possible event types
 */
enum {
	Q931_EVENT_TYPE_NONE = 0,
	Q931_EVENT_TYPE_CRV,
	Q931_EVENT_TYPE_TIMER,
	Q931_EVENT_TYPE_MESSAGE,
};

/**
 * @brief	Additional data for Timer event
 */
struct Q931_CallTimerEvent
{
	L3INT id;
};

/**
 * @brief	Additional data for Message event
 */
struct Q931_CallMessageEvent
{
	L3INT		type;
	Q931mes_Generic	*data;
};

/**
 * @brief	Q.931 Call event
 */
struct Q931_CallEvent
{
	L3INT	id;		/*!< */
	L3INT	type;		/*!< */
	L3BOOL	error;		/*!< Error event */

	/* Additional event data */
	union {
		struct Q931_CallTimerEvent	timer;
		struct Q931_CallMessageEvent	message;
	} data;
};

/**
 * @brief	Q.931 ITU-T Call states
 */
enum {
	/* Q.931 User mode states */
	Q931_U0  = 0,			/*!< Null */
	Q931_U1,			/*!< Call initiated */
	Q931_U2,			/*!< Overlap sending */
	Q931_U3,			/*!< Outgoing call proceeding */
	Q931_U4,			/*!< Call delivered */
	Q931_U6  = 6,			/*!< Call present */
	Q931_U7,			/*!< Call received */
	Q931_U8,			/*!< Connect request */
	Q931_U9,			/*!< Incoming call proceeding */
	Q931_U10,			/*!< Active */
	Q931_U11,			/*!< Disconnect request */
	Q931_U12,			/*!< Disconnect indication */
	Q931_U15 = 15,			/*!< Suspend request */
	Q931_U17 = 17,			/*!< Resume request */
	Q931_U19 = 19,			/*!< Release request */
	Q931_U25 = 25,			/*!< Overlap receiving */

	Q931_UANY = 0xff,		/*!< Any U state */

	/* Q.931 Network mode states */
	Q931_N0  = (0x0100 | 0),	/*!< Null */
	Q931_N1,			/*!< Call initiation */
	Q931_N2,			/*!< Overlap sending */
	Q931_N3,			/*!< Outgoing call proceeding */
	Q931_N4,			/*!< Call delivered */
	Q931_N6  = (0x0100 | 6),	/*!< Call present */
	Q931_N7,			/*!< Call received */
	Q931_N8,			/*!< Connect request */
	Q931_N9,			/*!< Incoming call proceeding */
	Q931_N10,			/*!< Active */
	Q931_N11,			/*!< Disconnect request */
	Q931_N12,			/*!< Disconnect indication */
	Q931_N15 = (0x0100 | 15),	/*!< Suspend request */
	Q931_N17 = (0x0100 | 17),	/*!< Resume request */
	Q931_N19 = (0x0100 | 19),	/*!< Release request */
	Q931_N22 = (0x0100 | 22),	/*!< ?? */
	Q931_N25 = (0x0100 | 25),	/*!< Overlap receiving */

	Q931_NANY = 0x1ff		/*!< Any N state */
};


/**
 * @brief	Call direction
 */
enum {
	Q931_DIRECTION_INBOUND = 0,
	Q931_DIRECTION_OUTBOUND
};

/**
 * @brief	Call context
 */
struct Q931_Call
{
	L3UCHAR InUse;			/* Indicate if entry is in use.         */
					/*  0 = Not in Use                      */
					/*  1 = Active Call.                    */

	L3UCHAR Tei;			/* Associated TEI 			*/

	L3UCHAR BChan;			/* Associated B Channel.                */
					/* 0 - 31 valid B chan			*/
					/* 255 = Not allocated			*/

	L3UCHAR	Direction;		/* Direction of call                    */
	L3INT   CRV;			/* Associated CRV.                      */

	L3UINT  State;			/* Call State.                          */
					/*  0 is Idle, but other values are	*/
					/*  defined per dialect.		*/
					/*  Default usage is 1-99 for TE and    */
					/*  101 - 199 for NT.			*/

	L3ULONG Timer;			/* Timer in ms. The TimeTick will check	*/
					/* if this has exceeded the timeout, and*/
					/* if so call the timers timeout proc.	*/

	L3USHORT TimerID;		/* Timer Identification/State           */
					/* actual values defined by dialect	*/
					/*  0 : No timer running                */
					/*  ITU-T Q.931:301 - 322 Timer running */

	L3UCHAR TimerCnt;		/* Counter to track the number of expire events */


	Q931_TrunkInfo_t *Trunk;	/* Short-cut pointer to access the	*/
					/* associated trunk 			*/

	void *pvt;			/* L4 Private data */

	/*
	 * Per-call event queue
	 */
	L3INT			numevents;			/*!< Number of queued events */
	struct Q931_CallEvent	events[Q931_CALL_MAX_EVENTS];	/*!< Event queue */
};

typedef struct Q931_Call Q931_Call_t;

typedef void (*Q931CallEventCB_t) (struct Q931_Call *call, struct Q931_CallEvent *event, void *priv);


/*
 * Helper functions
 */

void Q931CallSetState(struct Q931_Call *call, const L3INT state);	/* TODO: private? */
Q931_API L3INT Q931CallGetState(const struct Q931_Call *call);
Q931_API const char * Q931CallGetStateName(const struct Q931_Call *call);
Q931_API const char * Q931CallGetEventName(const L3INT event);
Q931_API const char * Q931CallGetTimerName(const L3USHORT timer);

Q931_API void   Q931CallSetPrivate(struct Q931_Call *call, const void *priv);
Q931_API void * Q931CallGetPrivate(const struct Q931_Call *call);

Q931_API Q931_TrunkInfo_t * Q931CallGetTrunk(const struct Q931_Call *call);

Q931_API void Q931SetCallEventCB(Q931_TrunkInfo_t *trunk, const Q931CallEventCB_t cb, const void *priv);

/*
 * Call control
 */

/* *********************************************** */

struct Q931_Call *Q931CallNew(Q931_TrunkInfo_t *trunk);
L3INT Q931CallRelease(struct Q931_Call *call);

Q931_API L3INT Q931CallGetCRV(const struct Q931_Call *call);
Q931_API L3INT Q931CallGetDirection(const struct Q931_Call *call);
Q931_API L3BOOL Q931CallIsOutgoing(const struct Q931_Call *call);
Q931_API L3BOOL Q931CallIsGlobal(const struct Q931_Call *call);

Q931_API L3INT Q931CallEventGetType(const struct Q931_CallEvent *event);
Q931_API L3INT Q931CallEventGetId(const struct Q931_CallEvent *event);
Q931_API void * Q931CallEventGetData(const struct Q931_CallEvent *event);

Q931_API struct Q931_Call * Q931GetCallByCRV(const Q931_TrunkInfo_t *trunk, const L3INT crv);
Q931_API struct Q931_Call * Q931GetGlobalCall(const Q931_TrunkInfo_t *trunk);

/*
 * Debug
 */
Q931_API L3INT Q931CallDump(const struct Q931_Call *call);
Q931_API L3INT Q931DumpAllCalls(const Q931_TrunkInfo_t *trunk);

/*
 * Q.931 Internal
 */
#if defined(Q931_PRIVATE)

void Q931CallInitEvent(struct Q931_CallEvent *event);
void Q931CallSendEvent(const struct Q931_Call *call, const struct Q931_CallEvent *event);

/* V2 part of the call event api */
struct Q931_CallEvent *Q931CallNewEvent(struct Q931_Call *call);
void Q931CallSendQueuedEvents(struct Q931_Call *call);
void Q931CallQueueEvent(struct Q931_Call *call, struct Q931_CallEvent *event);
void Q931CallFlushEvents(struct Q931_Call *call);


struct Q931_Call *Q931CallNewIncoming(Q931_TrunkInfo_t *trunk, const L3INT crv);
struct Q931_Call *Q931GetCallByIndex(const Q931_TrunkInfo_t *trunk, const L3INT callindex);

void Q931CallStartTimer(struct Q931_Call *call, L3USHORT timer);
void Q931CallRestartTimer(struct Q931_Call *call, L3USHORT timer);
void Q931CallStopTimer(struct Q931_Call *call, L3USHORT timer);
void Q931CallStopAllTimers(struct Q931_Call *call);
Q931_API L3USHORT Q931CallGetTimerID(const struct Q931_Call *call);
Q931_API L3UCHAR Q931CallGetTimerExpireCount(const struct Q931_Call *call);
void Q931CallIncrementTimerExpireCount(struct Q931_Call *call);

#endif /* Q931_PRIVATE */

#endif
