/*
 *
 *
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "common.h"
#include "bitmap.h"

#define BITMAP_GET(m, b)        (!!((m)[BITMAP_OFFSET(b)] & (1 << BITMAP_BIT(b))))
#define BITMAP_SET(m, b)        ((m)[BITMAP_OFFSET(b)] |= (1 << BITMAP_BIT(b)))
#define BITMAP_UNSET(m, b)      ((m)[BITMAP_OFFSET(b)] &= ~(1 << BITMAP_BIT(b)))


//#define bitmap_debug(x, ...)		printf(x, __VA_ARGS__)
#define bitmap_debug(x, ...)

int bitmap_create(bitmap_t **bm, const int nbits)
{
	int bytes = BITMAP_SIZE_BYTES(nbits);

	if (bm == NULL || nbits <= 0)
		return -1;
	if ((*bm = calloc(1, bytes)) == NULL)
		return -1;
	return 0;
}

int bitmap_destroy(bitmap_t *bm)
{
	if (!bm)
		return -1;
	free(bm);
	return 0;
}

int bitmap_init(bitmap_t *bm, const int nbits)
{
	if (bm == NULL || nbits <= 0)
		return -1;
	memset(bm, 0, BITMAP_SIZE_BYTES(nbits));
	return 0;
}

int bitmap_is_set(const bitmap_t *bm, const int nbits, const int pos)
{
	if (!bm || nbits <= 0 || pos < 0 || pos >= nbits)
		return -1;
	return BITMAP_GET(bm, pos);
}

int bitmap_set(bitmap_t *bm, const int nbits, const int pos)
{
	if (!bm || nbits <= 0 || pos < 0 || pos >= nbits)
		return -1;
	BITMAP_SET(bm, pos);
	return 0;
}

int bitmap_unset(bitmap_t *bm, const int nbits, const int pos)
{
	if (!bm || nbits <= 0 || pos < 0 || pos >= nbits)
		return -1;
	BITMAP_UNSET(bm, pos);
	return 0;
}

int bitmap_clear(bitmap_t *bm, const int nbits)
{
	if (!bm || nbits <= 0)
		return -1;
	memset(bm, 0, BITMAP_SIZE_BYTES(nbits));
	return 0;
}

int bitmap_is_empty(const bitmap_t *bm, const int nbits)
{
	bitmap_t *p = (bitmap_t *)bm;

	for (int i = 0; i < BITMAP_SIZE_BYTES(nbits); i++)
		if (*(p++)) return 0;
	return 1;
}

int bitmap_next_bit(const bitmap_t *bm, const int nbits, const int last)
{
	bitmap_t *ptr = NULL;
	bitmap_t mask = 0;

	if (!bm || !nbits)
		return -1;
	if (!(last >= -1 || last < nbits))
		return -1;

	ptr  = (bitmap_t *)&bm[BITMAP_OFFSET(last + 1)];
	mask = (~0ul) << BITMAP_BIT(last + 1);

	bitmap_debug("starting new search at last %d, mask %#lx (row %d)\n", last, mask, BITMAP_OFFSET(last + 1));

	for (int i = last + 1; i < nbits;) {
		bitmap_debug("%d/%d: last %d, mask %#lx, bm[%d] = %#lx\n", i, nbits, last, mask, BITMAP_OFFSET(i), bm[BITMAP_OFFSET(i)]);
		/* nothing in this row left? skip */
		if ((*ptr & mask) == 0) {
			i += BITS_PER_LONG - BITMAP_BIT(i);
			mask = ~0ul;
			ptr++;
			bitmap_debug("remaining row empty, fast forwarding to %d/%d\n", i, nbits);
			continue;
		}
		if ((*ptr & (1ul << BITMAP_BIT(i))) != 0) {
			bitmap_debug("%d/%d found bit %d\n", i, nbits, i);
			return i;
		}
		if (BITMAP_BIT(++i) == 0)
			ptr++;
	}
	return -1;
}

int bitmap_first_bit(const bitmap_t *bm, const int nbits)
{
	return bitmap_next_bit(bm, nbits, -1);
}

int bitmap_count(const bitmap_t *bm, const int nbits)
{
	int count = 0;

	for (int i = bitmap_first_bit(bm, nbits); i >= 0; i = bitmap_next_bit(bm, nbits, i)) count++;

	return count;
}

int bitmap_dump(const bitmap_t *bm, const int nbits)
{
	if (!bm || nbits <= 0)
		return -1;

	for (int i = 0; i < BITMAP_SIZE(nbits); i++)
		printf("map[%05d]: %#lx\n", i, bm[i]);

	return 0;
}
