#ifndef _BITMAP_H
#define _BITMAP_H

#include <limits.h>
#include "common.h"

#ifndef LONG_MAX
#error "LONG_MAX not defined"
#endif

#if (LONG_MAX == 2147483647)
#define BITS_PER_LONG		32
#define BITMAP_SIZE(x)		(((x) + 31) >> 5)
#define BITMAP_SIZE_BYTES(x)	(BITMAP_SIZE(x) << 2)
#define BITMAP_BIT(x)		((x) & 0x1f)
#define BITMAP_OFFSET(x)	((x) >> 5)
#else
#define BITS_PER_LONG		64
#define BITMAP_SIZE(x)		(((x) + 63) >> 6)
#define BITMAP_SIZE_BYTES(x)	(BITMAP_SIZE(x) << 3)
#define BITMAP_BIT(x)		((x) & 0x3f)
#define BITMAP_OFFSET(x)	((x) >> 6)
#endif

typedef unsigned long bitmap_t;

#define DECLARE_BITMAP(_name, _nbits) \
	bitmap_t _name[BITMAP_SIZE(_nbits)]

/**
 * Allocate new bitmap with size of <nbits> bits ([0:nbits-1])
 * @param	bm	Bitmap
 * @param[in]	nbits	Size in bits
 * @return	0 on success, -1 on error
 */
int bitmap_create(bitmap_t **bm, const int nbits);

/**
 * Destroy dynamically allocated bitmap
 * @param	bm	Bitmap
 * @return	0 on success, -1 on error
 */
int bitmap_destroy(bitmap_t *bm);

/**
 * Initialize a static bitmap with size <nbits> bits ([0:nbits-1])
 * @param[in]	bm	Bitmap
 * @param[in]	nbits	Size of bitmap in bits
 * @return	0 on success, -1 on error
 */
int bitmap_init(bitmap_t *bm, const int nbits);

/**
 * Check whether a bit in the map is set
 * @param[in]	bm	Bitmap
 * @param[in]	nbits	Size of bitmap in bits
 * @param[in]	pos	Bit to check
 * @return	1 if the bit is set, 0 if it's not, -1 on error
 */
int bitmap_is_set(const bitmap_t *bm, const int nbits, const int pos);

/**
 * Set a bit in the map
 * @param[in]	bm	Bitmap
 * @param[in]	pos	Bit to set
 * @return	0 on success, -1 on error
 */
int bitmap_set(bitmap_t *bm, const int nbits, const int pos);

/**
 * Unset a bit
 * @param[in]	bm	Bitmap
 * @param[in]	nbits	Size of bitmap in bits
 * @param[in]	pos	Bit to clear
 * @return	0 on success, -1 on error
 */
int bitmap_unset(bitmap_t *bm, const int nbits, const int pos);

/**
 * Clear all bits
 * @param[in]	bm	Bitmap
 * @param[in]	nbits	Size of bitmap in bits
 * @reurn	0 on success, -1 on error
 */
int bitmap_clear(bitmap_t *bm, const int nbits);

/**
 * Check if map has no entries
 * @param[in]	bm	Bitmap
 * @param[in]	nbits	Size of bitmap in bits
 * @return	0 if map has entries, 1 otherwise
 */
int bitmap_is_empty(const bitmap_t *bm, const int nbits);


/**
 * Return first bit set in bitmap
 * @param[in]	bm	Bitmap
 * @param[in]	nbits	Size of bitmap in bits
 * @return	-1 if none found or error, >= 0 pos of first bit
 */
int bitmap_first_bit(const bitmap_t *bm, const int nbits);

/**
 * Return next bit set in bitmap
 * @param[in]	bm	Bitmap
 * @param[in]	nbits	Size of bitmap in bits
 * @param[in]	last	Position of last set bit (output of bitmap_next_bit())
 * @return	-1 if none found or error, >= 0 pos of first bit
 */
int bitmap_next_bit(const bitmap_t *bm, const int nbits, const int last);

/**
 * Count set bits in bitmap
 * @param[in]	bm	Bitmap
 * @param[in]	nbits	Size of bitmap (number of bits)
 * @return	Number of '1' bits in bitmap
 */
int bitmap_count(const bitmap_t *bm, const int nbits);

#endif
