/*
 *
 *
 */
#ifndef _DLIST_H_
#define _DLIST_H_

#include "common.h"

/*
 * Double-linked list implementation
 */
struct dlist_head;

/**
 * Double-linked list structure
 */
struct dlist_head {
	struct dlist_head *head;	/*!< pointer to list head */
	struct dlist_head *next;	/*!< next element (first) */
	struct dlist_head *prev;	/*!< previous element (last) */
};

#define DLIST_HEAD(_name)	\
	struct dlist_head _name = { &(_name), NULL, NULL }

/**
 * dlist_init_head
 * @brief	Initialize linked list
 * @param[in]	h	pointer to linked list head
 */
void dlist_init_head(struct dlist_head *h);

void dlist_init(struct dlist_head *e);

/**
 * dlist_entry
 * @brief Get current element
 */
#define dlist_entry(ptr, type, member) \
	container_of(ptr, type, member)

/**
 * dlist_get_head
 * @brief	Get head element of list
 */
#define dlist_get_head(ptr)	\
	(ptr)->head

/**
 * dlist_insert_after
 * @brief	Insert new element after given one
 * @param[in]	elem	Element to insert after
 * @param[in]	new	New element
 */
void dlist_insert_after(struct dlist_head *elem, struct dlist_head *enew);


/**
 * dlist_insert_head
 * @brief	Insert new element at first position
 * @param[in]	list	The list
 * @param[in]	new	New element
 */
#define dlist_insert_head(list, enew) \
	dlist_insert_after((list)->head, enew)

#define dlist_push_head(list, enew)	\
	dlist_insert_after((list)->head, enew)


/**
 * dlist_insert_before
 * @brief	Insert new element before given one
 * @param[in]	elem	Element to insert before
 * @param[in]	new	New element
 */
void dlist_insert_before(struct dlist_head *elem, struct dlist_head *enew);


/**
 * dlist_insert_tail
 * @brief	Insert new element at the end of the list
 * @param[in]	list	The list
 * @param[in]	new	New element
 */
void dlist_insert_tail(struct dlist_head *list, struct dlist_head *enew);

#define dlist_push_tail(list, enew)	\
	dlist_insert_tail(list, enew)


/**
 * dlist_remove
 * @brief	Remove element from the list
 * @param[in]	elem	Element to remove
 */
void dlist_remove(struct dlist_head *elem);


/**
 * dlist_first
 * @brief	Get first element in the list
 * @param[in]	elem	Any list element
 * @return	First element
 */
struct dlist_head *dlist_first(const struct dlist_head *elem);


/**
 * dlist_last
 * @brief	Get last element in the list
 * @param[in]	elem	Any list element
 * @return	Last element
 */
struct dlist_head *dlist_last(const struct dlist_head *elem);


/**
 * dlist_next
 * @brief	Get next element after given one
 * @param[in]	elem	Any list element
 * @return	Next element
 */
struct dlist_head *dlist_next(const struct dlist_head *elem);


/**
 * dlist_prev
 * @brief	Get element before given one
 * @param[in]	elem	Any list element
 * @return	Previous element
 */
struct dlist_head *dlist_prev(const struct dlist_head *elem);


/**
 * dlist_is_head
 * @brief	Check if the element passed is the list head
 * @param[in]	elem	Any list element
 * @return	0 if the element passed is not the head, 1 otherwise
 */
int dlist_is_head(const struct dlist_head *elem);


/**
 * dlist_is_first
 * @brief	Check if the element passed is the first one
 * @param[in]	elem	Any list element
 * @return	0 if the element passed is not the first, 1 otherwise
 */
int dlist_is_first(const struct dlist_head *elem);


/**
 * dlist_is_last
 * @brief	Check if the element passed is the last one
 * @param[in]	elem	Any list element
 * @return	0 if the element passed is not the last, 1 otherwise
 */
int dlist_is_last(const struct dlist_head *elem);


/**
 * dlist_replace
 * @brief	Replace an element in the list with a new one
 * @param[in]	old	Element to replace
 * @param[in]	new	Replacement
 */
void dlist_replace(struct dlist_head *old, struct dlist_head *enew);


/**
 * Remove and return first element of list
 * @param[in]	elem	List
 * @return	First element of list or NULL if list is empty
 */
struct dlist_head *dlist_pop_head(const struct dlist_head *elem);


/**
 * Remove and return last element of list
 * @param[in]	elem	List
 * @return	Last element of list or NULL if list is empty
 */
struct dlist_head *dlist_pop_tail(const struct dlist_head *elem);


/**
 * dlist_for_each
 * @brief	Walk the list of elements in a for like loop
 * @param[in]	head	The list
 * @param[in]	p	Pointer to store the current element
 * @note	Not dlist_remove() safe!
 */
#define dlist_for_each(head, p)		\
	for (p = (typeof(*p) *)(head)->next;	\
	p != (typeof(*p) *)(head) && p;		\
	p = (typeof(*p) *)(p)->next)


/**
 * dlist_for_each_safe
 * @brief	Walk the list of elements in a for like loop
 * @param[in]	head	The list
 * @param[in]	n	Temporary storage
 * @param[in]	p	Pointer to store the current element
 * @note	dlist_remove() safe version
 */
#define dlist_for_each_safe(head, n, p)	\
	for (p = (typeof(*p) *)(head)->next, n = (typeof(*p) *)(p)->next; \
	p != (typeof(*p) *)(head) && p;		\
	p = (typeof(*p) *)(n), n = (typeof(*p) *)(p)->next)


/**
 * dlist_for_each_reverse
 * @brief	Reverse-walk the list of elements in a for like loop
 * @param[in]	head	The list
 * @param[in]	p	Pointer to store the current element
 * @note	Not dlist_remove() safe!
 */
#define dlist_for_each_reverse(head, p)	\
	for (p = (typeof(*p) *)(head)->prev;	\
	p != (typeof(*p) *)(head) && p;		\
	p = (typeof(*p) *)(p)->prev)


/**
 * dlist_is_empty
 * @brief	Check if the list has no elements
 * @param[in]	head	The list
 * @return	0 if the list is not empty, 1 otherwise
 */
#define dlist_is_empty(head) ({		\
		((head)->next == head);	\
	})


/**
 * dlist_num_entries
 * @brief	Get the number of elements in the list
 * @param[in]	head	The list
 * @return	number of entries
 */
#define dlist_num_entries(head) ({		\
		struct dlist_head *p = (head);	\
		int _x = 0;			\
		for (; p && p->next && p->next != (head); _x++)	\
			p = p->next;		\
		_x;				\
	})

#endif /* _DLIST_H_ */

