/*
 *
 *
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "msgb.h"

enum {
	MSGBF_NONE = 0,
	MSGBF_BUFFER_OWNED = (1 << 0),
};

/*******************************************************************************
 * Buffer initialization / destruction
 *******************************************************************************/

static inline int _msgb_init(struct msgb *msgb, const uint8_t *buf, const int size, const int flags)
{
	if (!msgb || size <= 0)
		return -1;

	if (flags & MSGBF_BUFFER_OWNED) {
		memset(msgb, 0, sizeof(*msgb) + size);
		msgb->buf = (uint8_t *)((uintptr_t)msgb + sizeof(*msgb));
	} else if (buf) {
		memset(msgb, 0, sizeof(*msgb));
		msgb->buf = (uint8_t *)buf;
	} else {
		return -1;
	}

	msgb->flags = flags;
	msgb->size  = size;
	msgb->next  = NULL;
	return 0;
}

int msgb_init(struct msgb *msgb, const uint8_t *buf, const int size)
{
	return _msgb_init(msgb, buf, size, MSGBF_NONE);
}

struct msgb *msgb_alloc(const int size)
{
	struct msgb *msgb = NULL;

	if (size <= 0)
		return NULL;

	msgb = malloc(sizeof(*msgb) + size);
	if (!msgb)
		return NULL;

	if (_msgb_init(msgb, NULL, size, MSGBF_BUFFER_OWNED) < 0) {
		free(msgb);
		return NULL;
	}

	return msgb;
}

struct msgb *msgb_alloc_reserve(const int size, const int reserve)
{
	struct msgb *msgb = NULL;

	if (size <= 0 || reserve <= 0 || size <= reserve)
		return NULL;

	msgb = msgb_alloc(size);
	if (!msgb)
		return NULL;

	if (msgb_reserve(msgb, reserve) < 0) {
		free(msgb);
		return NULL;
	}

	return msgb;
}

int msgb_destroy(struct msgb *msgb)
{
	if (!msgb)
		return -1;
	for (struct msgb *cur = msgb, *next = NULL; cur; cur = next) {
		next = cur->next;
		/* clear content on exit */
		memset(cur->buf, 0, cur->size);
		free(cur);
	}
	return 0;
}

/*******************************************************************************
 * Buffer status
 *******************************************************************************/

int msgb_size(const struct msgb *msgb)
{
	return (msgb && msgb->buf) ? msgb->size : -1;
}

int msgb_length(const struct msgb *msgb)
{
	if (!msgb || !msgb->size || !msgb->buf)
		return -1;

	return (msgb->end - msgb->head);
}

int msgb_space(const struct msgb *msgb)
{
	if (!msgb || !msgb->size || !msgb->buf)
		return -1;

	return (msgb->size - msgb->tail);
}

int msgb_reserved_space(const struct msgb *msgb)
{
	if (!msgb || !msgb->size || !msgb->buf)
		return -1;

	return msgb->head;
}

int msgb_empty(const struct msgb *msgb)
{
	if (!msgb || !msgb->size || !msgb->buf)
		return 0;

	return msgb->end == msgb->head;
}

int msgb_full(const struct msgb *msgb)
{
	if (!msgb || !msgb->size || !msgb->buf)
		return 1;

	return msgb->end == msgb->size;
}

int msgb_left(const struct msgb *msgb)
{
	if (!msgb || !msgb->size || !msgb->buf)
		return -1;

	return (msgb->end - msgb->tail);
}

int msgb_tell(const struct msgb *msgb)
{
	if (!msgb || !msgb->size || !msgb->buf)
		return -1;

	return (msgb->tail - msgb->head);
}

uint8_t *msgb_head_ptr(const struct msgb *msgb)
{
	if (!msgb || !msgb->size || !msgb->buf)
		return NULL;

	return &msgb->buf[msgb->head];
}

uint8_t *msgb_tail_ptr(const struct msgb *msgb)
{
	if (!msgb || !msgb->size || !msgb->buf)
		return NULL;

	return &msgb->buf[msgb->tail];
}


/*******************************************************************************
 * Buffer manipulation
 *******************************************************************************/

int msgb_reserve(struct msgb *msgb, const int reserve)
{
	if (!msgb || msgb->size <= 0 || reserve <= 0)
		return -1;
	if (msgb->head + reserve >= msgb->size)
		return -1;
	/* move head */
	msgb->head += reserve;
	/* check if tail is still after head (update if it's not) */
	if (msgb->tail < msgb->head)
		msgb->tail = msgb->head;
	/* check if data end is still after head (update if it's not) */
	if (msgb->end < msgb->head)
		msgb->end = msgb->head;
	return 0;
}

int msgb_unreserve(struct msgb *msgb, const int size)
{
	int old_pos;

	if (!msgb || msgb->size <= 0)
		return -1;

	old_pos = msgb->head;

	if (size <= 0) {
		msgb->head = 0;
	} else if (msgb->head >= size) {
		msgb->head -= size;
	} else {
		/* invalid number of bytes */
		return -1;
	}

	/*
	 * Move tail and end pointer only
	 * if they pointed to the old head position
	 * (i.e. nothing has been read or written yet)
	 */
	if (msgb->tail == old_pos)
		msgb->tail = msgb->head;
	if (msgb->end == old_pos)
		msgb->end = msgb->head;
	return 0;
}

int msgb_rewind(struct msgb *msgb)
{
	if (!msgb || !msgb->buf)
		return -1;
	msgb->tail = msgb->head;
	return 0;
}

int msgb_seek(struct msgb *msgb, const msgb_seek_t type, const int offset)
{
	unsigned int pos = 0;

	if (!msgb || !msgb->buf)
		return -1;

	switch (type) {
	case MSGB_SEEK_CUR:
		pos = msgb->tail + offset;
		break;
	case MSGB_SEEK_SET:
		pos = msgb->head + offset;
		break;
	case MSGB_SEEK_END:
		pos = msgb->end - offset;
		break;
	default:
		return -1;
	}

	if (pos < msgb->head || pos > msgb->end)
		return -1;

	msgb->tail = pos;
	return 0;
}

int msgb_clear(struct msgb *msgb)
{
	if (!msgb || !msgb->buf)
		return -1;
	memset(msgb->buf, 0, msgb->size);
	msgb->end = msgb->tail = msgb->head;
	return 0;
}

int msgb_reset(struct msgb *msgb)
{
	if (!msgb || !msgb->buf)
		return -1;
	memset(msgb->buf, 0, msgb->size);
	msgb->end = msgb->tail = msgb->head = 0;
	return 0;
}

/*******************************************************************************
 * Buffer accessors
 *******************************************************************************/

uint8_t *msgb_put(struct msgb *msgb, const int size)
{
	uint8_t *tmp = NULL;

	if (!msgb || !msgb->size || !msgb->buf || size < 0)
		return NULL;
	if ((msgb->tail + size) > msgb->size)
		return NULL;

	tmp = &msgb->buf[msgb->tail];
	msgb->tail += size;
	msgb->end   = MAX(msgb->end, msgb->tail);
	return tmp;
}

uint8_t *msgb_get(struct msgb *msgb, const int size)
{
	uint8_t *tmp = NULL;

	if (!msgb || !msgb->size || !msgb->buf || size < 0)
		return NULL;
	if ((msgb->tail + size) > msgb->end)
		return NULL;

	tmp = &msgb->buf[msgb->tail];
	msgb->tail += size;
	return tmp;
}

int msgb_write(struct msgb *msgb, const void *buf, const int size)
{
	uint8_t *tmp = NULL;
	int bytes = size, ret;

	if (!msgb || !buf || size <= 0)
		return -1;
	if ((ret = msgb_space(msgb)) < size)
		bytes = ret;
	if (bytes <= 0)
		return -1;
	if ((tmp = msgb_put(msgb, bytes)) == NULL)
		return -1;

	memcpy(tmp, buf, bytes);
	return bytes;
}

int msgb_read(struct msgb *msgb, void *buf, const int size)
{
	uint8_t *tmp = NULL;
	int bytes = size, ret;

	if (!msgb || !buf || size <= 0)
		return -1;
	if ((ret = msgb_left(msgb)) < size)
		bytes = ret;
	if ((tmp = msgb_get(msgb, bytes)) == NULL)
		return -1;

	memcpy(buf, tmp, bytes);
	return bytes;
}

int msgb_fill(struct msgb *msgb, const int c, const int size)
{
	uint8_t *tmp = NULL;
	int bytes = size, ret;

	if (!msgb || size <= 0)
		return -1;
	if ((ret = msgb_space(msgb)) < size)
		bytes = ret;
	if (bytes <= 0)
		return -1;
	if ((tmp = msgb_put(msgb, bytes)) == NULL)
		return -1;
	memset(tmp, c, bytes);
	return bytes;
}

/*******************************************************************************
 * Buffer list accessors
 *******************************************************************************/


/*******************************************************************************
 * Buffer chaining
 *******************************************************************************/

int msgb_chain_append(struct msgb *chain, struct msgb *msgb)
{
	if (!chain || !msgb)
		return -1;
	for (struct msgb *cur = chain; cur; cur = cur->next) {
		if (!cur->next) {
			cur->next = msgb;
			return 0;
		}
	}
	return -1;
}

struct msgb *msgb_chain_split(struct msgb *chain, struct msgb *where)
{
	if (!chain || !where)
		return NULL;
	if (chain == where)
		return chain;
	/* find point to split */
	for (struct msgb *cur = chain; cur; cur = cur->next) {
		if (cur->next == where) {
			cur->next = NULL;
			return where;
		}
	}
	return NULL;
}

struct msgb *msgb_chain_split_at(struct msgb *chain, const int pos)
{
	int idx = 0;
	if (!chain || pos < 0)
		return NULL;
	if (pos == 0)
		return chain;
	/* find point to split */
	for (struct msgb *cur = chain; cur; cur = cur->next) {
		if (++idx == pos) {
			struct msgb *tmp = cur->next;
			cur->next = NULL;
			return tmp;
		}
	}
	return NULL;
}


/* */
static int _msgb_chain_length(const struct msgb *msgb)
{
	int bytes = 0, ret;

	if (!msgb)
		return -1;
	for (const struct msgb *cur = msgb; cur; cur = cur->next) {
		if ((ret = msgb_length(cur)) < 0)
			return -1;
		bytes += ret;
	}
	return bytes;
}

/* */
static int _msgb_chain_size(const struct msgb *msgb)
{
	int bytes = 0, ret;

	if (!msgb)
		return -1;
	for (const struct msgb *cur = msgb; cur; cur = cur->next) {
		if ((ret = msgb_size(cur)) < 0)
			return -1;
		bytes += ret;
	}
	return bytes;
}

static struct msgb* _msgb_linearize(struct msgb *msgb, const int reserve)
{
	struct msgb *tmp = NULL;
	int size;

	if (!msgb || reserve < 0)
		return NULL;
	if ((size = _msgb_chain_length(msgb)) <= 0)
		return NULL;

	if ((tmp = (reserve) ? msgb_alloc_reserve(size + reserve, reserve) : msgb_alloc(size)) == NULL)
		return NULL;

	for (struct msgb *cur = msgb; cur; cur = cur->next) {
		if ((size = msgb_length(cur)) <= 0)
			goto error;
		if (msgb_write(tmp, msgb_head_ptr(cur), size) < 0)
			goto error;
	}
	return tmp;
error:
	msgb_destroy(tmp);
	return NULL;
}

struct msgb *msgb_linearize(struct msgb *msgb)
{
	return _msgb_linearize(msgb, 0);
}

struct msgb *msgb_linearize_reserve(struct msgb *msgb, const int reserve)
{
	return _msgb_linearize(msgb, reserve);
}

int msgb_merge(struct msgb *msgb, struct msgb *outb)
{
	int size;

	if (!msgb || !outb)
		return -1;

	if ((size = _msgb_chain_length(msgb)) <= 0)
		return -1;
	if (size > msgb_left(outb))
		return -1;

	for (struct msgb *cur = msgb; cur; cur = cur->next) {
		if ((size = msgb_length(cur)) <= 0)
			return -1;
		if (msgb_write(outb, msgb_head_ptr(cur), size) < 0)
			return -1;
	}
	return 0;
}

static struct msgb *_msgb_split(struct msgb *msgb, const int size, const int reserve)
{
	struct msgb *chain = NULL;
	int bytes;

	if (!msgb || size <= 0 || reserve < 0)
		return NULL;
	if (msgb_left(msgb) == 0)
		return NULL;
	/* can't split chained buffers */
	if (msgb->next)
		return NULL;

	while ((bytes = msgb_left(msgb)) > 0) {
		struct msgb *tmp = NULL;

		if ((tmp = (reserve) ? msgb_alloc_reserve(size + reserve, reserve) : msgb_alloc(size)) == NULL)
			goto error;

		bytes = MIN(size, bytes);

		if (msgb_write(tmp, msgb_get(msgb, bytes), bytes) <= 0)
			goto error;

		if (!chain)
			chain = tmp;
		else if (msgb_chain_append(chain, tmp) < 0)	/* optimize later */
			goto error;
	}
	return chain;
error:
	msgb_destroy(chain);
	return NULL;
}

struct msgb *msgb_split(struct msgb *msgb, const int size)
{
	return _msgb_split(msgb, size, 0);
}

struct msgb *msgb_split_reserve(struct msgb *msgb, const int size, const int reserve)
{
	return _msgb_split(msgb, size, reserve);
}

int msgb_chain_count(const struct msgb *chain)
{
	int count = 0;

	for (const struct msgb *cur = chain; cur; cur = cur->next)
		count++;
	return count;
}

int msgb_chain_length(const struct msgb *chain)
{
	return _msgb_chain_length(chain);
}

int msgb_chain_size(const struct msgb *chain)
{
	return _msgb_chain_size(chain);
}
