/*
 *
 *
 *
 */
#ifndef _MSGB_H_
#define _MSGB_H_

#include <stdint.h>

#include "dlist.h"

struct msgb {
	struct dlist_head list;		/*!< List of message buffers */
	struct msgb *next;		/*!< Next buffer in chain */

	unsigned int flags;
	unsigned int size;

	unsigned int head;		/*!< Position of buffer head */
	unsigned int tail;		/*!< Current write position  */
	unsigned int end;		/*!< Pointer of last byte in buffer */

	uint8_t *buf;
};

#define MSGB_STATIC_SIZE(_size)	\
	(sizeof(struct msgb) + _size)

#define MSGB_INIT_STATIC(_msgb, _size)	\
	msgb_init((struct msgb *)(_msgb), (uint8_t *)(_msgb) + sizeof(struct msgb), (_size) - sizeof(struct msgb))

/**
 * Initialize custom message buffer
 * @note	Buffer content is _not_ zeroed, use msgb_clear()
 * @param	msgb	Message buffer
 * @param	buf	External buffer space
 * @param	size	Size of buffer space
 * @return	0 on success, <0 on error
 */
int msgb_init(struct msgb *msgb, const uint8_t *buf, const int size);

/**
 * Allocate message buffer of given size
 * @param	size	Size of buffer to allocate
 * @return	Newly allocated message buffer on success, NULL otherwise
 */
struct msgb *msgb_alloc(const int size);

/**
 * Allocate message buffer of given size and reserve space at the beginning
 * @note	Same as manually calling msgb_alloc() + msgb_reserve()
 * @param	size	Size of buffer to allocate
 * @param	reserve	Number of bytes to reserve
 * @return	Newly allocated message buffer on success, NULL otherwise
 */
struct msgb *msgb_alloc_reserve(const int size, const int reserve);

/**
 * Destroy message buffer
 * @note	Buffer content is zeroed
 * @param	msgb	Message buffer
 * @return	0 on success, <0 on error
 */
int msgb_destroy(struct msgb *msgb);

/**
 * Reserve space at the beginning of the buffer
 * @note	Updates head and tail pointer, existing data at the beginning may become inaccessible
 * @param	msgb	Message buffer
 * @param	reserve	Number of bytes to reserve
 * @return	0 on success, <0 on error
 */
int msgb_reserve(struct msgb *msgb, const int reserve);

/**
 * Make reserved space at the beginning available
 * @note	Moves buffer read and/or tail pointer to new head position if either pointed to the head before
 * @param	msgb	Message buffer
 * @param	size	Number of bytes to unreserve, <= 0 for everything
 * @return	0 on success, <0 on error
 */
int msgb_unreserve(struct msgb *msgb, const int size);

/**
 * Get allocated size of message buffer
 * @param	msgb	Message buffer
 * @return	Size of buffer on success, < 0 on error
 */
int msgb_size(const struct msgb *msgb);

/**
 * Get length of data in message buffer
 * @param	msgb	Message buffer
 * @return	Size of data on success, < 0 on error
 */
int msgb_length(const struct msgb *msgb);

/**
 * Get remaining free space in message buffer for writing (= free space after tail pointer)
 * @param	msgb	Message buffer
 * @return	Size of buffer on success, < 0 on error
 */
int msgb_space(const struct msgb *msgb);

/**
 * Get reserved space of message buffer (= space before head pointer)
 * @param	msgb	Message buffer
 * @return	Size of reserved buffer space on success, < 0 on error
 */
int msgb_reserved_space(const struct msgb *msgb);

/**
 * Is message buffer empty?
 * @param	msgb	Message buffer
 * @return	1 if buffer is empty, 0 otherwise
 */
int msgb_empty(const struct msgb *msgb);

/**
 * Is message buffer full?
 * @param	msgb	Message buffer
 * @return	1 if buffer is full, 0 otherwise
 */
int msgb_full(const struct msgb *msgb);

/**
 * Get number of remaining bytes to read in message buffer
 * @param	msgb	Message buffer
 * @return	Number of remaining bytes on success, < 0 on error
 */
int msgb_left(const struct msgb *msgb);


/**
 * Get pointer to head of buffer
 * @param	msgb	Message buffer
 * @return	Pointer to head of buffer on success, NULL otherwise
 */
uint8_t *msgb_head_ptr(const struct msgb *msgb);

/**
 * Get pointer to tail of buffer
 * @param	msgb	Message buffer
 * @return	Pointer to tail of buffer on success, NULL otherwise
 */
uint8_t *msgb_tail_ptr(const struct msgb *msgb);


/**
 * Move tail pointer
 * @param	msgb	Message buffer
 * @param	size	Number of bytes to move tail pointer
 * @return	Pointer to old position of tail on success, NULL otherwise
 */
uint8_t *msgb_put(struct msgb *msgb, const int size);

static inline int msgb_put_u8(struct msgb *msgb, const uint8_t val)
{
	uint8_t *tmp = msgb_put(msgb, 1);
	if (!tmp)
		return -1;
	tmp[0] = val;
	return 0;
}

static inline int msgb_put_s8(struct msgb *msgb, const int8_t val)
{
	uint8_t *tmp = msgb_put(msgb, 1);
	if (!tmp)
		return -1;
	tmp[0] = val;
	return 0;
}

static inline int msgb_put_u16(struct msgb *msgb, const uint16_t val)
{
	uint8_t *tmp = msgb_put(msgb, 2);
	if (!tmp)
		return -1;
	tmp[0] = (val >> 8) & 0xff;
	tmp[1] =  val       & 0xff;
	return 0;
}

static inline int msgb_put_s16(struct msgb *msgb, const int16_t val)
{
	uint8_t *tmp = msgb_put(msgb, 2);
	if (!tmp)
		return -1;
	tmp[0] = (val >> 8) & 0xff;
	tmp[1] =  val       & 0xff;
	return 0;
}

static inline int msgb_put_u32(struct msgb *msgb, const uint32_t val)
{
	uint8_t *tmp = msgb_put(msgb, 4);
	if (!tmp)
		return -1;
	tmp[0] = (val >> 24) & 0xff;
	tmp[1] = (val >> 16) & 0xff;
	tmp[2] = (val >>  8) & 0xff;
	tmp[3] =  val        & 0xff;
	return 0;
}

static inline int msgb_put_s32(struct msgb *msgb, const int32_t val)
{
	uint8_t *tmp = msgb_put(msgb, 4);
	if (!tmp)
		return -1;
	tmp[0] = (val >> 24) & 0xff;
	tmp[1] = (val >> 16) & 0xff;
	tmp[2] = (val >>  8) & 0xff;
	tmp[3] =  val        & 0xff;
	return 0;
}

static inline int msgb_put_u64(struct msgb *msgb, const uint64_t val)
{
	uint8_t *tmp = msgb_put(msgb, 8);
	if (!tmp)
		return -1;
	tmp[0] = (val >> 56) & 0xff;
	tmp[1] = (val >> 48) & 0xff;
	tmp[2] = (val >> 40) & 0xff;
	tmp[3] = (val >> 32) & 0xff;
	tmp[4] = (val >> 24) & 0xff;
	tmp[5] = (val >> 16) & 0xff;
	tmp[6] = (val >>  8) & 0xff;
	tmp[7] =  val        & 0xff;
	return 0;
}

static inline int msgb_put_s64(struct msgb *msgb, const int64_t val)
{
	uint8_t *tmp = msgb_put(msgb, 8);
	if (!tmp)
		return -1;
	tmp[0] = (val >> 56) & 0xff;
	tmp[1] = (val >> 48) & 0xff;
	tmp[2] = (val >> 40) & 0xff;
	tmp[3] = (val >> 32) & 0xff;
	tmp[4] = (val >> 24) & 0xff;
	tmp[5] = (val >> 16) & 0xff;
	tmp[6] = (val >>  8) & 0xff;
	tmp[7] =  val        & 0xff;
	return 0;
}

/**
 * Copy data into buffer
 * @note	No space left is also an error condition (this emulates write(2) behaviour)
 * @param	msgb	Message buffer
 * @param	buf	Buffer to copy from
 * @param	size	Number of bytes to copy into message buffer
 * @return	0 on success, < 0 otherwise
 */
int msgb_write(struct msgb *msgb, const void *buf, const int size);

/**
 * Fill buffer with character, starting at current write position
 * @note	Updates tail pointer
 * @param[in]	msgb	Buffer to fill
 * @param[in]	c	Character
 * @param[in]	size	Number of characters to write
 * @return	Number of bytes filled, < 0 on error
 */
int msgb_fill(struct msgb *msgb, const int c, const int size);


/**
 * Move read pointer
 * @param	msgb	Message buffer
 * @param	size	Number of bytes to move read pointer
 * @return	Pointer to old position of read pointer on success, NULL otherwise
 */
uint8_t *msgb_get(struct msgb *msgb, const int size);

static inline uint8_t msgb_get_u8(struct msgb *msgb)
{
	uint8_t *tmp = msgb_get(msgb, 1);
	if (!tmp)
		return 0;
	return tmp[0];
}

static inline int8_t msgb_get_s8(struct msgb *msgb)
{
	uint8_t *tmp = msgb_get(msgb, 1);
	if (!tmp)
		return 0;
	return tmp[0];
}

static inline uint16_t msgb_get_u16(struct msgb *msgb)
{
	uint8_t *tmp = msgb_get(msgb, 2);
	if (!tmp)
		return 0;
	return tmp[0] << 8 | tmp[1];
}

static inline int16_t msgb_get_s16(struct msgb *msgb)
{
	uint8_t *tmp = msgb_get(msgb, 2);
	if (!tmp)
		return 0;
	return tmp[0] << 8 | tmp[1];
}

static inline uint32_t msgb_get_u32(struct msgb *msgb)
{
	uint8_t *tmp = msgb_get(msgb, 4);
	if (!tmp)
		return 0;
	return tmp[0] << 24 | tmp[1] << 16 | tmp[2] << 8 | tmp[3];
}

static inline int32_t msgb_get_s32(struct msgb *msgb)
{
	uint8_t *tmp = msgb_get(msgb, 4);
	if (!tmp)
		return 0;
	return tmp[0] << 24 | tmp[1] << 16 | tmp[2] << 8 | tmp[3];
}

static inline uint64_t msgb_get_u64(struct msgb *msgb)
{
	uint8_t *tmp = msgb_get(msgb, 8);
	uint64_t val = 0;
	if (!tmp)
		return 0;
	val |= tmp[0]; val <<= 8;
	val |= tmp[1]; val <<= 8;
	val |= tmp[2]; val <<= 8;
	val |= tmp[3]; val <<= 8;
	val |= tmp[4]; val <<= 8;
	val |= tmp[5]; val <<= 8;
	val |= tmp[6]; val <<= 8;
	val |= tmp[7];
	return val;
}

static inline int64_t msgb_get_s64(struct msgb *msgb)
{
	uint8_t *tmp = msgb_get(msgb, 8);
	uint64_t val = 0;
	if (!tmp)
		return 0;
	val |= tmp[0]; val <<= 8;
	val |= tmp[1]; val <<= 8;
	val |= tmp[2]; val <<= 8;
	val |= tmp[3]; val <<= 8;
	val |= tmp[4]; val <<= 8;
	val |= tmp[5]; val <<= 8;
	val |= tmp[6]; val <<= 8;
	val |= tmp[7];
	return val;
}

/**
 * Copy data from buffer
 * @note	Returns 0 if no data is left (this emulates read(2) behaviour)
 * @param	msgb	Message buffer
 * @param	buf	Buffer to copy into
 * @param	size	Number of bytes to copy from message buffer
 * @return	0 on success, < 0 otherwise
 */
int msgb_read(struct msgb *msgb, void *buf, const int size);

/**
 * Set read and tail pointer to beginning of data
 * @param	msgb	Message buffer
 * @return	0 on success, < 0 on error
 */
int msgb_rewind(struct msgb *msgb);

/**
 * Type of seek operation
 */
typedef enum {
	MSGB_SEEK_CUR = 0,	/*!< Seek from current location (= relative) */
	MSGB_SEEK_SET,		/*!< Seek starting at head pointer (= absolute position) */
	MSGB_SEEK_END,		/*!< Reverse seek from end pointer (= reverse absolute position) */
} msgb_seek_t;

/**
 * Move read and tail pointer to desired location
 * @param	msgb	Message buffer
 * @param	type	Type of seek (starting point and direction)
 * @param	offset	Number of bytes to seek
 * @return	0 on success, < 0 on error
 */
int msgb_seek(struct msgb *msgb, const msgb_seek_t type, const int offset);

/**
 * Get current tail pointer position
 * @param[in]	msgb	Message buffer
 * @return	>= 0 on success, < 0 on error
 * @note	Returns the position relative to head pointer
 */
int msgb_tell(const struct msgb *msgb);

/**
 * Clear data, move read, tail and end pointer to position of head
 * @note	Does not release reserved space
 * @param	msgb	Message buffer
 * @return	0 on success, < 0 on error
 */
int msgb_clear(struct msgb *msgb);

/**
 * Completely reset buffer
 * @note	Clears data and releases reserved space
 * @param	msgb	Message buffer
 * @return	0 on success, < 0 on error
 */
int msgb_reset(struct msgb *msgb);

/**
 * Allocate new msgb and merge content of input (chain) into new buffer
 * @param[in]	msgb	Input (chain of msg) buffer(s)
 * @return	Liniearized output buffer on success, NULL on error
 */
struct msgb *msgb_linearize(struct msgb *msgb);

/**
 * Allocate new msgb with reserved headroom and merge content of input (chain) into new buffer
 * @note	Output buffer will be <size of input> + <reserve> bytes large
 * @param[in]	msgb	Input (chain of msg) buffer(s)
 * @param[in]	reserve	Space to reserve in bytes
 * @return	Liniearized output buffer on success, NULL on error
 */
struct msgb *msgb_linearize_reserve(struct msgb *msgb, const int reserve);

/**
 * Merge chain of input buffers into output buffer
 * @note	Output buffer should have enough free space or this operation will fail
 * @note	Read pointer of input buffers is reset
 * @param[in]	msgb	Input (chain of) msg buffer(s)
 * @param[in]	outb	Output msg buffer
 * @return	0 on success, -1 on error (e.g. not enough space in output buffer)
 */
int msgb_merge(struct msgb *msgb, struct msgb *outb);

/**
 * Split msgb into chain of <size> each
 * @param[in]	msgb	Input msg buffer
 * @param[in]	size	Size of fragments to split input buffer into
 * @return	Chain of msg buffers on succes, NULL on error
 */
struct msgb *msgb_split(struct msgb *msgb, const int size);

/**
 * Split msgb into chain of <size> each with additional headroom of <reserve> bytes
 * @note	Output buffers will be <size> + <reserve> bytes large
 * @param[in]	msgb	Input msg buffer
 * @param[in]	msgb	Input msg buffer
 * @param[in]	size	Size of fragments to split input buffer into
 * @param[in]	reserve	Size of headroom to reserve
 * @return	Chain of msg buffers on succes, NULL on error
 */
struct msgb *msgb_split_reserve(struct msgb *msgb, const int size, const int reserve);

/**
 * Append msgb to chain
 * @param[in]	chain	Chain of msg buffers
 * @param[in]	msgb	Buffer to append (can be a chain too)
 * @return	0 on success, -1 on error
 */
int msgb_chain_append(struct msgb *chain, struct msgb *msgb);

/**
 * Split chain at indicated element
 * @param[in]	chain	Chain of msg buffers
 * @param[in]	msgb	Buffer to append (can be a chain too)
 * @return	Remaining part of the chain or NULL on error
 */
struct msgb *msgb_chain_split(struct msgb *chain, struct msgb *where);

/**
 * Split chain at indicated element (index)
 * @param[in]	chain	Chain of msg buffers
 * @param[in]	msgb	Buffer to append (can be a chain too)
 * @return	Remaining part of the chain or NULL on error
 */
struct msgb *msgb_chain_split_at(struct msgb *chain, const int pos);

/**
 * Number of msgbs in chain
 * @param[in]	chain	msgb chain
 * @return	number of buffers in chain
 */
int msgb_chain_count(const struct msgb *chain);

/**
 * Total length of data in chain
 * @param[in]	chain	msgb chain
 * @return	Sum of msgb_length() of all msgbs on success, -1 on error
 */
int msgb_chain_length(const struct msgb *chain);

/**
 * Total size of msgbs in chain
 * @param[in]	chain	msgb chain
 * @return	Sum of msgb_size() of all msgbs on success, -1 on error
 */
int msgb_chain_size(const struct msgb *chain);

#endif /* _MSGB_H_ */
